import copy
import logging
import pathlib
import rapidjson
import freqtrade.vendor.qtpylib.indicators as qtpylib
import numpy as np
import talib.abstract as ta
import pandas as pd
import pandas_ta as pta
from freqtrade.strategy.interface import IStrategy
from freqtrade.strategy import merge_informative_pair
from freqtrade.strategy import DecimalParameter, CategoricalParameter
from pandas import DataFrame, Series
from functools import reduce
from freqtrade.persistence import Trade
from datetime import datetime, timedelta
import time
from typing import Optional, Tuple
import warnings
from enum import Enum


from collections import deque

log = logging.getLogger(__name__)

warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)










































class NostalgiaForInfinityX4_L(IStrategy):
  INTERFACE_VERSION = 3

  def version(self) -> str:
    return "v14.1.686"

  stoploss = -0.99

  trailing_stop = False
  trailing_only_offset_is_reached = True
  trailing_stop_positive = 0.01
  trailing_stop_positive_offset = 0.03

  use_custom_stoploss = False

  timeframe = "5m"
  info_timeframes = ["15m", "1h", "4h", "1d"]

  btc_info_timeframes = ["5m", "15m", "1h", "4h", "1d"]

  has_bt_agefilter = False
  bt_min_age_days = 3

  has_downtime_protection = False

  hold_support_enabled = True

  process_only_new_candles = True

  use_exit_signal = True
  exit_profit_only = False
  ignore_roi_if_entry_signal = True

  startup_candle_count: int = 800

  long_normal_mode_tags = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12", "13"]

  long_pump_mode_tags = ["21", "22", "23", "24", "25", "26"]

  long_quick_mode_tags = ["41", "42", "43", "44", "45", "46", "47", "48", "49", "50", "51"]

  long_rebuy_mode_tags = ["61", "62"]

  long_mode_tags = ["80", "81", "82"]

  long_rapid_mode_tags = ["101", "102", "103", "104", "105", "106", "107", "108", "109", "110"]

  long_grind_mode_tags = ["120"]

  long_normal_mode_name = "long_normal"
  long_pump_mode_name = "long_pump"
  long_quick_mode_name = "long_quick"
  long_rebuy_mode_name = "long_rebuy"
  long_high_profit_mode_name = "long_hp"
  long_rapid_mode_name = "long_rapid"
  long_grind_mode_name = "long_grind"


  short_normal_mode_tags = ["500", "501"]

  short_pump_mode_tags = ["521", "522", "523", "524", "525", "526"]

  short_quick_mode_tags = ["541", "542", "543", "544", "545", "546", "547", "548", "549", "550"]

  short_rebuy_mode_tags = ["561"]

  short_mode_tags = ["580", "581", "582"]

  short_rapid_mode_tags = ["601", "602", "603", "604", "605", "606", "607", "608", "609", "610"]

  short_grind_mode_tags = ["620"]

  short_normal_mode_name = "short_normal"
  short_pump_mode_name = "short_pump"
  short_quick_mode_name = "short_quick"
  short_rebuy_mode_name = "short_rebuy"
  short_high_profit_mode_name = "short_hp"
  short_rapid_mode_name = "short_rapid"

  is_futures_mode = False
  futures_mode_leverage = 5.0
  futures_mode_leverage_rebuy_mode = 5.0
  futures_mode_leverage_grind_mode = 5.0



  stop_thresholds = [-0.2, -0.2, -0.025, -0.025, 720, 720, 0.016, 0.016, 24.0, 24.0, False, False, True, True]

  stop_threshold = 4.0
  stop_threshold_futures = 12.0
  stop_threshold_futures_rapid = 12.0
  stop_threshold_spot_rapid = 4.0
  stop_threshold_spot_rebuy = 2.0
  stop_threshold_futures_rebuy = 4.0

  rebuy_mode_min_free_slots = 2

  position_adjustment_enable = True

  grinding_enable = True

  grind_derisk_spot = -0.40
  grind_derisk_futures = -0.30

  grind_1_stop_grinds_spot = -0.50
  grind_1_profit_threshold_spot = 0.018
  grind_1_stakes_spot = [0.20, 0.20, 0.20]
  grind_1_sub_thresholds_spot = [-0.10, -0.11, -0.12]

  grind_1_stop_grinds_futures = -0.50
  grind_1_profit_threshold_futures = 0.018
  grind_1_stakes_futures = [0.20, 0.20, 0.20]
  grind_1_sub_thresholds_futures = [-0.10, -0.11, -0.12]

  grind_2_stop_grinds_spot = -0.50
  grind_2_profit_threshold_spot = 0.018
  grind_2_stakes_spot = [0.15, 0.25, 0.30]
  grind_2_sub_thresholds_spot = [-0.10, -0.11, -0.12]

  grind_2_stop_grinds_futures = -0.50
  grind_2_profit_threshold_futures = 0.018
  grind_2_stakes_futures = [0.15, 0.25, 0.30]
  grind_2_sub_thresholds_futures = [-0.10, -0.11, -0.12]

  grind_3_stop_grinds_spot = -0.50
  grind_3_profit_threshold_spot = 0.018
  grind_3_stakes_spot = [0.10, 0.10, 0.10]
  grind_3_sub_thresholds_spot = [-0.10, -0.11, -0.12]

  grind_3_stop_grinds_futures = -0.50
  grind_3_profit_threshold_futures = 0.018
  grind_3_stakes_futures = [0.10, 0.10, 0.10]
  grind_3_sub_thresholds_futures = [-0.10, -0.11, -0.12]

  grind_4_stop_grinds_spot = -0.50
  grind_4_profit_threshold_spot = 0.018
  grind_4_stakes_spot = [0.10, 0.10, 0.10]
  grind_4_sub_thresholds_spot = [-0.10, -0.11, -0.12]

  grind_4_stop_grinds_futures = -0.50
  grind_4_profit_threshold_futures = 0.018
  grind_4_stakes_futures = [0.10, 0.10, 0.10]
  grind_4_sub_thresholds_futures = [-0.10, -0.11, -0.12]

  grind_5_stop_grinds_spot = -0.50
  grind_5_profit_threshold_spot = 0.048
  grind_5_stakes_spot = [0.10, 0.10, 0.10]
  grind_5_sub_thresholds_spot = [-0.10, -0.11, -0.12]

  grind_5_stop_grinds_futures = -0.50
  grind_5_profit_threshold_futures = 0.048
  grind_5_stakes_futures = [0.10, 0.10, 0.10]
  grind_5_sub_thresholds_futures = [-0.10, -0.11, -0.12]

  grind_6_stop_grinds_spot = -0.50
  grind_6_profit_threshold_spot = 0.018
  grind_6_stakes_spot = [0.05, 0.06, 0.07, 0.08, 0.09]
  grind_6_sub_thresholds_spot = [-0.05, -0.06, -0.07, -0.08, -0.09]

  grind_6_stop_grinds_futures = -0.50
  grind_6_profit_threshold_futures = 0.018
  grind_6_stakes_futures = [0.05, 0.06, 0.07, 0.08, 0.09]
  grind_6_sub_thresholds_futures = [-0.05, -0.06, -0.07, -0.08, -0.09]

  grind_1_derisk_1_stop_grinds_spot = -0.10
  grind_1_derisk_1_profit_threshold_spot = 0.018
  grind_1_derisk_1_stakes_spot = [0.50]
  grind_1_derisk_1_sub_thresholds_spot = [-0.10]

  grind_1_derisk_1_stop_grinds_futures = -0.10
  grind_1_derisk_1_profit_threshold_futures = 0.018
  grind_1_derisk_1_stakes_futures = [0.50]
  grind_1_derisk_1_sub_thresholds_futures = [-0.10]

  grind_2_derisk_1_stop_grinds_spot = -0.50
  grind_2_derisk_1_profit_threshold_spot = 0.018
  grind_2_derisk_1_stakes_spot = [0.15, 0.20, 0.25]
  grind_2_derisk_1_sub_thresholds_spot = [-0.10, -0.11, -0.12]

  grind_2_derisk_1_stop_grinds_futures = -0.50
  grind_2_derisk_1_profit_threshold_futures = 0.018
  grind_2_derisk_1_stakes_futures = [0.15, 0.20, 0.25]
  grind_2_derisk_1_sub_thresholds_futures = [-0.10, -0.11, -0.12]

  regular_mode_stake_multiplier_spot = [1.0]
  regular_mode_stake_multiplier_futures = [1.0]
  regular_mode_use_grind_stops = False

  regular_mode_rebuy_stakes_spot = [0.10, 0.10, 0.10]
  regular_mode_rebuy_thresholds_spot = [-0.12, -0.14, -0.16]
  regular_mode_grind_1_stakes_spot = [0.20, 0.20, 0.20]
  regular_mode_grind_1_thresholds_spot = [-0.06, -0.12, -0.14]
  regular_mode_grind_1_stop_grinds_spot = -0.20
  regular_mode_grind_1_profit_threshold_spot = 0.018
  regular_mode_grind_2_stakes_spot = [0.10, 0.15, 0.20]
  regular_mode_grind_2_thresholds_spot = [-0.04, -0.10, -0.12]
  regular_mode_grind_2_stop_grinds_spot = -0.20
  regular_mode_grind_2_profit_threshold_spot = 0.018
  regular_mode_grind_3_stakes_spot = [0.12, 0.13, 0.14]
  regular_mode_grind_3_thresholds_spot = [-0.03, -0.10, -0.12]
  regular_mode_grind_3_stop_grinds_spot = -0.20
  regular_mode_grind_3_profit_threshold_spot = 0.018
  regular_mode_grind_4_stakes_spot = [0.12, 0.13, 0.14]
  regular_mode_grind_4_thresholds_spot = [-0.03, -0.10, -0.12]
  regular_mode_grind_4_stop_grinds_spot = -0.20
  regular_mode_grind_4_profit_threshold_spot = 0.018
  regular_mode_grind_5_stakes_spot = [0.12, 0.13, 0.14]
  regular_mode_grind_5_thresholds_spot = [-0.03, -0.10, -0.12]
  regular_mode_grind_5_stop_grinds_spot = -0.20
  regular_mode_grind_5_profit_threshold_spot = 0.048
  regular_mode_derisk_1_spot = -0.80
  regular_mode_derisk_1_spot_old = -0.16
  regular_mode_derisk_1_reentry_spot = -0.08
  regular_mode_derisk_spot = -1.60
  regular_mode_derisk_spot_old = -0.40

  regular_mode_rebuy_stakes_futures = [0.10, 0.10, 0.10]
  regular_mode_rebuy_thresholds_futures = [-0.12, -0.14, -0.16]
  regular_mode_grind_1_stakes_futures = [0.20, 0.20, 0.20]
  regular_mode_grind_1_thresholds_futures = [-0.06, -0.12, -0.14]
  regular_mode_grind_1_stop_grinds_futures = -0.20
  regular_mode_grind_1_profit_threshold_futures = 0.018
  regular_mode_grind_2_stakes_futures = [0.10, 0.15, 0.20]
  regular_mode_grind_2_thresholds_futures = [-0.04, -0.10, -0.12]
  regular_mode_grind_2_stop_grinds_futures = -0.20
  regular_mode_grind_2_profit_threshold_futures = 0.018
  regular_mode_grind_3_stakes_futures = [0.12, 0.13, 0.14]
  regular_mode_grind_3_thresholds_futures = [-0.03, -0.10, -0.12]
  regular_mode_grind_3_stop_grinds_futures = -0.20
  regular_mode_grind_3_profit_threshold_futures = 0.018
  regular_mode_grind_4_stakes_futures = [0.12, 0.13, 0.14]
  regular_mode_grind_4_thresholds_futures = [-0.03, -0.10, -0.12]
  regular_mode_grind_4_stop_grinds_futures = -0.20
  regular_mode_grind_4_profit_threshold_futures = 0.018
  regular_mode_grind_5_stakes_futures = [0.12, 0.13, 0.14]
  regular_mode_grind_5_thresholds_futures = [-0.03, -0.10, -0.12]
  regular_mode_grind_5_stop_grinds_futures = -0.20
  regular_mode_grind_5_profit_threshold_futures = 0.048
  regular_mode_derisk_1_futures = -2.40
  regular_mode_derisk_1_futures_old = -0.48
  regular_mode_derisk_1_reentry_futures = -0.08  # without leverage
  regular_mode_derisk_futures = -3.20
  regular_mode_derisk_futures_old = -0.80

  rebuy_mode_stake_multiplier = 0.2
  rebuy_mode_stake_multiplier_alt = 0.3
  rebuy_mode_max = 3
  rebuy_mode_derisk_spot = -0.9
  rebuy_mode_derisk_futures = -2.0
  rebuy_mode_stakes_spot = [1.0, 2.0, 4.0]
  rebuy_mode_stakes_futures = [1.0, 2.0, 4.0]
  rebuy_mode_thresholds_spot = [-0.08, -0.10, -0.12]
  rebuy_mode_thresholds_futures = [-0.08, -0.10, -0.12]

  grind_mode_stake_multiplier_spot = [0.20, 0.30, 0.40, 0.50, 0.60, 0.70]
  grind_mode_stake_multiplier_futures = [0.20, 0.30, 0.40, 0.50]
  grind_mode_first_entry_profit_threshold_spot = 0.018
  grind_mode_first_entry_profit_threshold_futures = 0.018
  grind_mode_first_entry_stop_threshold_spot = -0.20
  grind_mode_first_entry_stop_threshold_futures = -0.20
  grind_mode_max_slots = 2
  grind_mode_coins = [
    "MATIC",
    "ADA",
    "ARB",
    "DOT",
    "XLM",
    "ALGO",
    "ETH",
    "RNDR",
    "XMR",
    "AVAX",
    "NEAR",
    "DOGE",
    "BCH",
    "ETC",
    "FTM",
    "KAS",
    "HBAR",
    "SUI",
    "TON",
    "XRP",
    "UNI",
    "LTC",
    "FIL",
    "ATOM",
    "GRT",
    "LINK",
    "VET",
    "THETA",
    "EOS",
    "LRC",
    "QTUM",
    "CELR",
  ]

  short_mode_max_slots = 1

  profit_max_thresholds = [0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.05, 0.05]

  max_slippage = 0.012

  btc_stakes = ["BTC", "ETH"]



  long_entry_signal_params = {


    "long_entry_condition_1_enable": True,
    "long_entry_condition_2_enable": True,
    "long_entry_condition_3_enable": True,
    "long_entry_condition_4_enable": True,
    "long_entry_condition_5_enable": True,
    "long_entry_condition_6_enable": True,
    "long_entry_condition_7_enable": True,
    "long_entry_condition_8_enable": True,
    "long_entry_condition_9_enable": True,
    "long_entry_condition_10_enable": True,
    "long_entry_condition_11_enable": True,
    "long_entry_condition_12_enable": True,
    "long_entry_condition_13_enable": True,
    "long_entry_condition_21_enable": True,
    "long_entry_condition_22_enable": True,
    "long_entry_condition_23_enable": True,
    "long_entry_condition_24_enable": True,
    "long_entry_condition_25_enable": True,
    "long_entry_condition_26_enable": True,
    "long_entry_condition_41_enable": True,
    "long_entry_condition_42_enable": True,
    "long_entry_condition_43_enable": True,
    "long_entry_condition_44_enable": True,
    "long_entry_condition_45_enable": True,
    "long_entry_condition_46_enable": True,
    "long_entry_condition_47_enable": True,
    "long_entry_condition_48_enable": True,
    "long_entry_condition_49_enable": True,
    "long_entry_condition_50_enable": True,
    "long_entry_condition_51_enable": True,
    "long_entry_condition_61_enable": True,
    "long_entry_condition_62_enable": True,
    "long_entry_condition_80_enable": False,
    "long_entry_condition_81_enable": True,
    "long_entry_condition_82_enable": True,
    "long_entry_condition_101_enable": True,
    "long_entry_condition_102_enable": True,
    "long_entry_condition_103_enable": True,
    "long_entry_condition_104_enable": True,
    "long_entry_condition_105_enable": True,
    "long_entry_condition_106_enable": True,
    "long_entry_condition_107_enable": True,
    "long_entry_condition_108_enable": True,
    "long_entry_condition_109_enable": True,
    "long_entry_condition_110_enable": True,
    "long_entry_condition_120_enable": False,
  }

  short_entry_signal_params = {


    "short_entry_condition_500_enable": True,
    "short_entry_condition_501_enable": False,
    "short_entry_condition_580_enable": True,
  }

  buy_protection_params = {}


  entry_10_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_10_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_10_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_10_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_10_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_10_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_10_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.80, decimals=2, space="buy", optimize=False)
  entry_10_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.90, decimals=2, space="buy", optimize=False)
  entry_10_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=0.95, decimals=2, space="buy", optimize=False)
  entry_10_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.00, decimals=2, space="buy", optimize=False)
  entry_10_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_10_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_10_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_10_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_10_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_10_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_10_ema_200_not_dec_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_10_ema_200_not_dec_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_10_ema_200_not_dec_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_10_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_10_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_10_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_10_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_10_rsi_3_min = DecimalParameter(00.0, 30.0, default=4.0, decimals=0, space="buy", optimize=False)
  entry_10_rsi_3_max = DecimalParameter(30.0, 70.0, default=46.0, decimals=0, space="buy", optimize=False)
  entry_10_rsi_3_15m_min = DecimalParameter(00.0, 36.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_10_rsi_3_1h_min = DecimalParameter(00.0, 36.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_10_rsi_3_4h_min = DecimalParameter(00.0, 36.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_10_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_10_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.90, decimals=2, space="buy", optimize=False)
  entry_10_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_10_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.90, decimals=2, space="buy", optimize=False)
  entry_10_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_10_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_10_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_10_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_10_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_10_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_10_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_10_rsi_14_max = DecimalParameter(20.0, 60.0, default=30.0, decimals=0, space="buy", optimize=False)
  entry_10_ema_offset = DecimalParameter(0.940, 0.972, default=0.952, decimals=3, space="buy", optimize=False)
  entry_10_ema_open_offset = DecimalParameter(0.0100, 0.0400, default=0.0200, decimals=4, space="buy", optimize=False)

  entry_11_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_11_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_11_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_11_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_11_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_11_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_11_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.80, decimals=2, space="buy", optimize=False)
  entry_11_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.90, decimals=2, space="buy", optimize=False)
  entry_11_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=0.95, decimals=2, space="buy", optimize=False)
  entry_11_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.00, decimals=2, space="buy", optimize=False)
  entry_11_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_11_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_11_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_11_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_11_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_11_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_11_ema_200_not_dec_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_11_ema_200_not_dec_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_11_ema_200_not_dec_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_11_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_11_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_11_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_11_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_11_ema_100_over_ema_200_enabled = CategoricalParameter(
    [True, False], default=True, space="buy", optimize=False
  )
  entry_11_ema_12_1h_over_ema_200_1h_enabled = CategoricalParameter(
    [True, False], default=True, space="buy", optimize=False
  )
  entry_11_rsi_3_min = DecimalParameter(00.0, 30.0, default=2.0, decimals=0, space="buy", optimize=False)
  entry_11_rsi_3_max = DecimalParameter(30.0, 70.0, default=46.0, decimals=0, space="buy", optimize=False)
  entry_11_rsi_3_15m_min = DecimalParameter(00.0, 36.0, default=16.0, decimals=0, space="buy", optimize=False)
  entry_11_rsi_3_1h_min = DecimalParameter(00.0, 36.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_11_rsi_3_4h_min = DecimalParameter(00.0, 36.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_11_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_11_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_11_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=85.0, decimals=0, space="buy", optimize=False)
  entry_11_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_11_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=85.0, decimals=0, space="buy", optimize=False)
  entry_11_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_11_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=85.0, decimals=0, space="buy", optimize=False)
  entry_11_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_11_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_11_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_11_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_11_rsi_14_max = DecimalParameter(20.0, 60.0, default=45.0, decimals=0, space="buy", optimize=False)
  entry_11_cti_20_max = DecimalParameter(-0.99, -0.60, default=-0.50, decimals=2, space="buy", optimize=False)
  entry_11_ema_open_offset = DecimalParameter(0.0200, 0.0400, default=0.0260, decimals=4, space="buy", optimize=False)
  entry_11_sma_offset = DecimalParameter(0.940, 0.988, default=0.978, decimals=3, space="buy", optimize=False)

  entry_12_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_12_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_12_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_12_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_12_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_12_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_12_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.80, decimals=2, space="buy", optimize=False)
  entry_12_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.90, decimals=2, space="buy", optimize=False)
  entry_12_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=1.10, decimals=2, space="buy", optimize=False)
  entry_12_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.20, decimals=2, space="buy", optimize=False)
  entry_12_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_12_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_12_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_12_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_12_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_12_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_12_ema_200_not_dec_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_12_ema_200_not_dec_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_12_ema_200_not_dec_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_12_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_12_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_12_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_12_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_12_ema_50_1h_over_ema_200_1h_enabled = CategoricalParameter(
    [True, False], default=True, space="buy", optimize=False
  )
  entry_12_rsi_3_min = DecimalParameter(00.0, 30.0, default=2.0, decimals=0, space="buy", optimize=False)
  entry_12_rsi_3_max = DecimalParameter(30.0, 70.0, default=46.0, decimals=0, space="buy", optimize=False)
  entry_12_rsi_3_15m_min = DecimalParameter(00.0, 36.0, default=12.0, decimals=0, space="buy", optimize=False)
  entry_12_rsi_3_1h_min = DecimalParameter(00.0, 36.0, default=20.0, decimals=0, space="buy", optimize=False)
  entry_12_rsi_3_4h_min = DecimalParameter(00.0, 36.0, default=20.0, decimals=0, space="buy", optimize=False)
  entry_12_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=20.0, decimals=0, space="buy", optimize=False)

  entry_12_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_12_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_12_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_12_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_12_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_12_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_12_r_14_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_12_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_12_r_14_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_12_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_12_r_480_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_12_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_12_r_480_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_12_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_12_r_14_max = DecimalParameter(-100.0, 80.0, default=-88.0, decimals=0, space="buy", optimize=False)
  entry_12_bb_offset = DecimalParameter(0.970, 0.999, default=0.984, decimals=3, space="buy", optimize=False)
  entry_12_sma_offset = DecimalParameter(0.930, 0.960, default=0.940, decimals=3, space="buy", optimize=False)

  entry_24_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=True)
  entry_24_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=True)
  entry_24_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=True)
  entry_24_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=True)
  entry_24_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=True)
  entry_24_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=True)
  entry_24_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.80, decimals=2, space="buy", optimize=True)
  entry_24_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.90, decimals=2, space="buy", optimize=True)
  entry_24_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=1.10, decimals=2, space="buy", optimize=True)
  entry_24_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.20, decimals=2, space="buy", optimize=True)
  entry_24_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=True)
  entry_24_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=True)
  entry_24_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=True)
  entry_24_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=True)
  entry_24_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=True)
  entry_24_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=True)
  entry_24_ema_200_not_dec_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=True)
  entry_24_ema_200_not_dec_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=True)
  entry_24_ema_200_not_dec_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=True)
  entry_24_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=True)
  entry_24_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=True)
  entry_24_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=True)
  entry_24_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=True)
  entry_24_ema_50_over_ema_200_enabled = CategoricalParameter([True, False], default=True, space="buy", optimize=True)
  entry_24_ema_12_1h_over_ema_200_1h_enabled = CategoricalParameter(
    [True, False], default=True, space="buy", optimize=True
  )
  entry_24_rsi_3_min = DecimalParameter(00.0, 30.0, default=2.0, decimals=0, space="buy", optimize=True)
  entry_24_rsi_3_max = DecimalParameter(30.0, 70.0, default=46.0, decimals=0, space="buy", optimize=True)
  entry_24_rsi_3_15m_min = DecimalParameter(00.0, 36.0, default=2.0, decimals=0, space="buy", optimize=True)
  entry_24_rsi_3_1h_min = DecimalParameter(00.0, 36.0, default=8.0, decimals=0, space="buy", optimize=True)
  entry_24_rsi_3_4h_min = DecimalParameter(00.0, 36.0, default=8.0, decimals=0, space="buy", optimize=True)
  entry_24_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=8.0, decimals=0, space="buy", optimize=True)
  entry_24_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=True)
  entry_24_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=True)
  entry_24_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=True)
  entry_24_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=True)
  entry_24_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=True)
  entry_24_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=True)
  entry_24_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=True)
  entry_24_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=True)
  entry_24_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=True)
  entry_24_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=True)
  entry_24_rsi_14_min = DecimalParameter(20.0, 60.0, default=26.0, decimals=0, space="buy", optimize=True)
  entry_24_rsi_14_max = DecimalParameter(20.0, 60.0, default=46.0, decimals=0, space="buy", optimize=True)
  entry_24_cti_20_max = DecimalParameter(-0.99, -0.60, default=-0.75, decimals=2, space="buy", optimize=True)
  entry_24_r_14_max = DecimalParameter(-100.0, 80.0, default=-97.0, decimals=0, space="buy", optimize=True)
  entry_24_ewo_50_200_min = DecimalParameter(2.0, 10.0, default=7.0, decimals=1, space="buy", optimize=True)
  entry_24_ewo_50_200_max = DecimalParameter(10.0, 30.0, default=24.0, decimals=1, space="buy", optimize=True)
  entry_24_sma_offset = DecimalParameter(0.960, 0.999, default=0.978, decimals=3, space="buy", optimize=True)

  entry_25_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_25_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_25_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_25_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_25_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_25_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_25_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.5, decimals=2, space="buy", optimize=False)
  entry_25_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.75, decimals=2, space="buy", optimize=False)
  entry_25_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=0.90, decimals=2, space="buy", optimize=False)
  entry_25_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.00, decimals=2, space="buy", optimize=False)
  entry_25_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_25_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_25_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_25_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_25_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_25_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_25_ema_200_not_dec_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_25_ema_200_not_dec_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_25_ema_200_not_dec_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_25_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_25_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_25_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_25_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_25_rsi_3_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_25_rsi_3_max = DecimalParameter(30.0, 60.0, default=60.0, decimals=0, space="buy", optimize=False)
  entry_25_rsi_3_15m_min = DecimalParameter(00.0, 30.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_25_rsi_3_1h_min = DecimalParameter(00.0, 30.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_25_rsi_3_4h_min = DecimalParameter(00.0, 30.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_25_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_25_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_25_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=75.0, decimals=0, space="buy", optimize=False)
  entry_25_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_25_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=75.0, decimals=0, space="buy", optimize=False)
  entry_25_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_25_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_25_r_14_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_25_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_25_r_14_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_25_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_25_r_480_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_25_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_25_r_480_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_25_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-20.0, decimals=0, space="buy", optimize=False)
  entry_25_rsi_14_max = DecimalParameter(20.0, 46.0, default=46.0, decimals=0, space="buy", optimize=False)
  entry_25_cti_20_max = DecimalParameter(-0.9, 0.0, default=-0.9, decimals=1, space="buy", optimize=False)
  entry_25_ewo_50_200_min = DecimalParameter(1.0, 8.0, default=2.6, decimals=1, space="buy", optimize=False)
  entry_25_sma_offset = DecimalParameter(0.920, 0.950, default=0.942, decimals=3, space="buy", optimize=False)

  entry_26_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_26_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_26_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_26_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_26_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_26_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_26_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.80, decimals=2, space="buy", optimize=False)
  entry_26_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.90, decimals=2, space="buy", optimize=False)
  entry_26_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=1.10, decimals=2, space="buy", optimize=False)
  entry_26_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.20, decimals=2, space="buy", optimize=False)
  entry_26_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_26_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_26_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_26_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_26_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_26_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_26_ema_200_not_dec_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_26_ema_200_not_dec_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_26_ema_200_not_dec_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_26_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_26_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_26_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_26_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_26_rsi_3_min = DecimalParameter(00.0, 30.0, default=2.0, decimals=0, space="buy", optimize=False)
  entry_26_rsi_3_max = DecimalParameter(30.0, 70.0, default=46.0, decimals=0, space="buy", optimize=False)
  entry_26_rsi_3_15m_min = DecimalParameter(00.0, 36.0, default=4.0, decimals=0, space="buy", optimize=False)
  entry_26_rsi_3_1h_min = DecimalParameter(00.0, 36.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_26_rsi_3_4h_min = DecimalParameter(00.0, 36.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_26_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_26_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_26_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_26_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_26_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_26_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_26_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_26_r_14_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_26_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_26_r_14_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_26_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_26_r_480_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_26_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_26_r_480_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_26_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_26_bb_offset = DecimalParameter(0.978, 0.999, default=0.992, decimals=3, space="buy", optimize=False)
  entry_26_ema_open_offset = DecimalParameter(0.0100, 0.0400, default=0.018, decimals=3, space="buy", optimize=False)
  entry_26_ewo_50_200_1h_min = DecimalParameter(1.0, 8.0, default=1.2, decimals=1, space="buy", optimize=False)

  entry_45_close_max_12 = DecimalParameter(00.50, 0.95, default=0.88, decimals=2, space="buy", optimize=False)
  entry_45_close_max_24 = DecimalParameter(00.50, 0.95, default=0.84, decimals=2, space="buy", optimize=False)
  entry_45_close_max_48 = DecimalParameter(00.50, 0.95, default=0.8, decimals=2, space="buy", optimize=False)
  entry_45_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.7, decimals=2, space="buy", optimize=False)
  entry_45_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.66, decimals=2, space="buy", optimize=False)
  entry_45_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.6, decimals=2, space="buy", optimize=False)
  entry_45_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.56, decimals=2, space="buy", optimize=False)
  entry_45_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.86, decimals=2, space="buy", optimize=False)
  entry_45_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=0.9, decimals=2, space="buy", optimize=False)
  entry_45_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.36, decimals=2, space="buy", optimize=False)
  entry_45_sup_level_1h_enabled = CategoricalParameter([True, False], default=True, space="buy", optimize=False)
  entry_45_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_45_sup_level_4h_enabled = CategoricalParameter([True, False], default=True, space="buy", optimize=False)
  entry_45_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_45_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_45_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_45_rsi_3_min = DecimalParameter(00.0, 30.0, default=5.0, decimals=0, space="buy", optimize=False)
  entry_45_rsi_3_max = DecimalParameter(30.0, 60.0, default=46.0, decimals=0, space="buy", optimize=False)
  entry_45_rsi_3_15m_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_45_rsi_3_1h_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_45_rsi_3_4h_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_45_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_45_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.92, decimals=2, space="buy", optimize=False)
  entry_45_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_45_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.92, decimals=2, space="buy", optimize=False)
  entry_45_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=75.0, decimals=0, space="buy", optimize=False)
  entry_45_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.92, decimals=2, space="buy", optimize=False)
  entry_45_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=75.0, decimals=0, space="buy", optimize=False)
  entry_45_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-6.0, decimals=0, space="buy", optimize=False)
  entry_45_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_45_rsi_14_min = DecimalParameter(10.0, 40.0, default=26.0, decimals=0, space="buy", optimize=False)
  entry_45_rsi_14_max = DecimalParameter(20.0, 60.0, default=40.0, decimals=0, space="buy", optimize=False)
  entry_45_cti_20_max = DecimalParameter(-0.99, -0.50, default=-0.54, decimals=2, space="buy", optimize=False)
  entry_45_sma_offset = DecimalParameter(0.940, 0.984, default=0.954, decimals=3, space="buy", optimize=False)

  entry_46_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_46_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_46_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_46_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_46_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_46_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_46_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.5, decimals=2, space="buy", optimize=False)
  entry_46_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.75, decimals=2, space="buy", optimize=False)
  entry_46_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=0.90, decimals=2, space="buy", optimize=False)
  entry_46_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.00, decimals=2, space="buy", optimize=False)
  entry_46_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_46_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_46_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_46_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_46_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_46_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_46_ema_200_not_dec_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_46_ema_200_not_dec_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_46_ema_200_not_dec_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_46_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_46_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_46_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_46_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=True, space="buy", optimize=False)
  entry_46_rsi_3_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_46_rsi_3_max = DecimalParameter(30.0, 60.0, default=60.0, decimals=0, space="buy", optimize=False)
  entry_46_rsi_3_15m_min = DecimalParameter(00.0, 30.0, default=20.0, decimals=0, space="buy", optimize=False)
  entry_46_rsi_3_1h_min = DecimalParameter(00.0, 30.0, default=20.0, decimals=0, space="buy", optimize=False)
  entry_46_rsi_3_4h_min = DecimalParameter(00.0, 30.0, default=20.0, decimals=0, space="buy", optimize=False)
  entry_46_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=20.0, decimals=0, space="buy", optimize=False)
  entry_46_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_46_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_46_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_46_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_46_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_46_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_46_r_14_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_46_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_46_r_14_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_46_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_46_r_480_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_46_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_46_r_480_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_46_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_46_rsi_14_max = DecimalParameter(26.0, 60.0, default=60.0, decimals=0, space="buy", optimize=False)

  entry_47_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_47_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_47_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_47_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_47_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_47_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_47_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.5, decimals=2, space="buy", optimize=False)
  entry_47_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.75, decimals=2, space="buy", optimize=False)
  entry_47_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=0.90, decimals=2, space="buy", optimize=False)
  entry_47_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.00, decimals=2, space="buy", optimize=False)
  entry_47_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_47_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_47_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_47_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_47_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_47_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_47_ema_200_not_dec_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_47_ema_200_not_dec_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_47_ema_200_not_dec_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_47_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=True, space="buy", optimize=False)
  entry_47_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_47_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_47_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_47_rsi_3_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_47_rsi_3_max = DecimalParameter(30.0, 60.0, default=60.0, decimals=0, space="buy", optimize=False)
  entry_47_rsi_3_15m_min = DecimalParameter(00.0, 30.0, default=16.0, decimals=0, space="buy", optimize=False)
  entry_47_rsi_3_1h_min = DecimalParameter(00.0, 30.0, default=20.0, decimals=0, space="buy", optimize=False)
  entry_47_rsi_3_4h_min = DecimalParameter(00.0, 30.0, default=20.0, decimals=0, space="buy", optimize=False)
  entry_47_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=20.0, decimals=0, space="buy", optimize=False)
  entry_47_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.85, decimals=2, space="buy", optimize=False)
  entry_47_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_47_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.85, decimals=2, space="buy", optimize=False)
  entry_47_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_47_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_47_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_47_r_14_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_47_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_47_r_14_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_47_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_47_r_480_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_47_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_47_r_480_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_47_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_47_rsi_14_min = DecimalParameter(20.0, 40.0, default=24.0, decimals=0, space="buy", optimize=False)
  entry_47_rsi_14_max = DecimalParameter(26.0, 60.0, default=60.0, decimals=0, space="buy", optimize=False)
  entry_47_rsi_20_min = DecimalParameter(20.0, 40.0, default=24.0, decimals=0, space="buy", optimize=False)
  entry_47_rsi_20_max = DecimalParameter(26.0, 60.0, default=60.0, decimals=0, space="buy", optimize=False)
  entry_47_cti_20_max = DecimalParameter(-0.8, 0.8, default=-0.7, decimals=1, space="buy", optimize=False)
  entry_47_ema_offset = DecimalParameter(0.980, 0.999, default=0.994, decimals=3, space="buy", optimize=False)
  entry_47_high_max_12_1h_max = DecimalParameter(00.70, 0.95, default=0.88, decimals=2, space="buy", optimize=False)

  entry_48_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_48_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_48_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_48_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_48_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_48_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_48_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.5, decimals=2, space="buy", optimize=False)
  entry_48_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.75, decimals=2, space="buy", optimize=False)
  entry_48_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=0.90, decimals=2, space="buy", optimize=False)
  entry_48_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.00, decimals=2, space="buy", optimize=False)
  entry_48_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_48_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_48_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_48_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_48_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_48_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_48_ema_200_not_dec_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_48_ema_200_not_dec_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_48_ema_200_not_dec_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_48_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_48_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_48_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_48_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_48_ema_50_1h_over_ema_200_1h_enabled = CategoricalParameter(
    [True, False], default=True, space="buy", optimize=True
  )
  entry_48_rsi_3_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_48_rsi_3_max = DecimalParameter(30.0, 60.0, default=60.0, decimals=0, space="buy", optimize=False)
  entry_48_rsi_3_15m_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_48_rsi_3_1h_min = DecimalParameter(00.0, 30.0, default=10.0, decimals=0, space="buy", optimize=False)
  entry_48_rsi_3_4h_min = DecimalParameter(00.0, 30.0, default=10.0, decimals=0, space="buy", optimize=False)
  entry_48_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=10.0, decimals=0, space="buy", optimize=False)
  entry_48_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_48_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_48_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_48_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_48_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_48_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_48_r_14_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_48_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_48_r_14_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_48_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_48_r_480_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_48_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_48_r_480_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_48_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_48_rsi_14_max = DecimalParameter(26.0, 50.0, default=38.0, decimals=0, space="buy", optimize=False)
  entry_48_cci_20_max = DecimalParameter(-180.0, -80.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_48_sma_offset = DecimalParameter(0.940, 0.978, default=0.962, decimals=3, space="buy", optimize=False)
  entry_48_inc_min = DecimalParameter(0.01, 0.04, default=0.022, decimals=3, space="buy", optimize=False)

  entry_49_close_max_12 = DecimalParameter(00.50, 0.95, default=0.88, decimals=2, space="buy", optimize=False)
  entry_49_close_max_24 = DecimalParameter(00.50, 0.95, default=0.86, decimals=2, space="buy", optimize=False)
  entry_49_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_49_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_49_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_49_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_49_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.5, decimals=2, space="buy", optimize=False)
  entry_49_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.75, decimals=2, space="buy", optimize=False)
  entry_49_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=0.90, decimals=2, space="buy", optimize=False)
  entry_49_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.00, decimals=2, space="buy", optimize=False)
  entry_49_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_49_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_49_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_49_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_49_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_49_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_49_ema_200_not_dec_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_49_ema_200_not_dec_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_49_ema_200_not_dec_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_49_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_49_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_49_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_49_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_49_ema_12_1h_over_ema_200_1h_enabled = CategoricalParameter(
    [True, False], default=True, space="buy", optimize=True
  )
  entry_49_rsi_3_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_49_rsi_3_max = DecimalParameter(30.0, 60.0, default=60.0, decimals=0, space="buy", optimize=False)
  entry_49_rsi_3_15m_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_49_rsi_3_1h_min = DecimalParameter(00.0, 30.0, default=10.0, decimals=0, space="buy", optimize=False)
  entry_49_rsi_3_4h_min = DecimalParameter(00.0, 30.0, default=10.0, decimals=0, space="buy", optimize=False)
  entry_49_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=10.0, decimals=0, space="buy", optimize=False)
  entry_49_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_49_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_49_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_49_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_49_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_49_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_49_r_14_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_49_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_49_r_14_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_49_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_49_r_480_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_49_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_49_r_480_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_49_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_49_rsi_14_max = DecimalParameter(26.0, 50.0, default=36.0, decimals=0, space="buy", optimize=False)
  entry_49_r_14_max = DecimalParameter(-99.0, -70.0, default=-50.0, decimals=0, space="buy", optimize=False)
  entry_49_inc_min = DecimalParameter(0.01, 0.04, default=0.028, decimals=3, space="buy", optimize=False)

  entry_50_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_50_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_50_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_50_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_50_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_50_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_50_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.80, decimals=2, space="buy", optimize=False)
  entry_50_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.90, decimals=2, space="buy", optimize=False)
  entry_50_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=1.10, decimals=2, space="buy", optimize=False)
  entry_50_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.20, decimals=2, space="buy", optimize=False)
  entry_50_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_50_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_50_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_50_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_50_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_50_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_50_ema_200_not_dec_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_50_ema_200_not_dec_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_50_ema_200_not_dec_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_50_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_50_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_50_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_50_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_50_ema_12_1h_over_ema_200_1h_enabled = CategoricalParameter(
    [True, False], default=True, space="buy", optimize=False
  )
  entry_50_rsi_3_min = DecimalParameter(00.0, 30.0, default=0.0, decimals=0, space="buy", optimize=False)
  entry_50_rsi_3_max = DecimalParameter(30.0, 70.0, default=46.0, decimals=0, space="buy", optimize=False)
  entry_50_rsi_3_15m_min = DecimalParameter(00.0, 36.0, default=0.0, decimals=0, space="buy", optimize=False)
  entry_50_rsi_3_1h_min = DecimalParameter(00.0, 36.0, default=0.0, decimals=0, space="buy", optimize=False)
  entry_50_rsi_3_4h_min = DecimalParameter(00.0, 36.0, default=0.0, decimals=0, space="buy", optimize=False)
  entry_50_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=0.0, decimals=0, space="buy", optimize=False)
  entry_50_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.80, decimals=2, space="buy", optimize=False)
  entry_50_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_50_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_50_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_50_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_50_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_50_r_14_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_50_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_50_r_14_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_50_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_50_r_480_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_50_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_50_r_480_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_50_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_50_bb_offset = DecimalParameter(0.970, 0.999, default=0.995, decimals=3, space="buy", optimize=False)
  entry_50_ema_open_offset = DecimalParameter(0.020, 0.040, default=0.020, decimals=3, space="buy", optimize=False)

  entry_102_close_max_12 = DecimalParameter(00.50, 0.95, default=0.92, decimals=2, space="buy", optimize=False)
  entry_102_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_102_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_102_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.65, decimals=2, space="buy", optimize=False)
  entry_102_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_102_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.55, decimals=2, space="buy", optimize=False)
  entry_102_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.5, decimals=2, space="buy", optimize=False)
  entry_102_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.75, decimals=2, space="buy", optimize=False)
  entry_102_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=0.90, decimals=2, space="buy", optimize=False)
  entry_102_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.00, decimals=2, space="buy", optimize=False)
  entry_102_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_102_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_102_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_102_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_102_sup_level_1d_enabled = CategoricalParameter([True, False], default=True, space="buy", optimize=False)
  entry_102_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_102_ema_200_not_dec_1h_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_102_ema_200_not_dec_4h_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_102_ema_200_not_dec_1d_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_102_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=True, space="buy", optimize=False)
  entry_102_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_102_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=True, space="buy", optimize=False)
  entry_102_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_102_rsi_3_min = DecimalParameter(00.0, 30.0, default=12.0, decimals=0, space="buy", optimize=False)
  entry_102_rsi_3_max = DecimalParameter(30.0, 70.0, default=30.0, decimals=0, space="buy", optimize=False)
  entry_102_rsi_3_15m_min = DecimalParameter(00.0, 30.0, default=16.0, decimals=0, space="buy", optimize=False)
  entry_102_rsi_3_1h_min = DecimalParameter(00.0, 30.0, default=12.0, decimals=0, space="buy", optimize=False)
  entry_102_rsi_3_4h_min = DecimalParameter(00.0, 30.0, default=12.0, decimals=0, space="buy", optimize=False)
  entry_102_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_102_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_102_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_102_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_102_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_102_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_102_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_102_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_102_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_102_rsi_14_max = DecimalParameter(20.0, 60.0, default=46.0, decimals=0, space="buy", optimize=False)
  entry_102_ema_offset = DecimalParameter(0.940, 0.984, default=0.966, decimals=3, space="buy", optimize=False)
  entry_102_bb_offset = DecimalParameter(0.970, 1.010, default=0.999, decimals=3, space="buy", optimize=False)

  entry_103_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_103_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_103_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_103_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_103_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_103_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_103_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.80, decimals=2, space="buy", optimize=False)
  entry_103_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.90, decimals=2, space="buy", optimize=False)
  entry_103_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=0.95, decimals=2, space="buy", optimize=False)
  entry_103_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.00, decimals=2, space="buy", optimize=False)
  entry_103_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_103_res_level_1h_enabled = CategoricalParameter([True, False], default=True, space="buy", optimize=False)
  entry_103_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_103_res_level_4h_enabled = CategoricalParameter([True, False], default=True, space="buy", optimize=False)
  entry_103_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_103_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_103_ema_200_not_dec_1h_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_103_ema_200_not_dec_4h_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_103_ema_200_not_dec_1d_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_103_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_103_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_103_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=True, space="buy", optimize=False)
  entry_103_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_103_rsi_3_min = DecimalParameter(00.0, 30.0, default=10.0, decimals=0, space="buy", optimize=False)
  entry_103_rsi_3_max = DecimalParameter(30.0, 70.0, default=65.0, decimals=0, space="buy", optimize=False)
  entry_103_rsi_3_15m_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_103_rsi_3_1h_min = DecimalParameter(00.0, 30.0, default=20.0, decimals=0, space="buy", optimize=False)
  entry_103_rsi_3_4h_min = DecimalParameter(00.0, 30.0, default=20.0, decimals=0, space="buy", optimize=False)
  entry_103_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=1.0, decimals=0, space="buy", optimize=False)
  entry_103_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.85, decimals=2, space="buy", optimize=False)
  entry_103_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=85.0, decimals=0, space="buy", optimize=False)
  entry_103_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_103_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=85.0, decimals=0, space="buy", optimize=False)
  entry_103_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_103_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=85.0, decimals=0, space="buy", optimize=False)
  entry_103_r_14_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_103_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_103_r_14_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_103_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-5.0, decimals=0, space="buy", optimize=False)
  entry_103_r_480_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_103_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-25.0, decimals=0, space="buy", optimize=False)
  entry_103_r_480_4h_min = DecimalParameter(-100.0, -70.0, default=-90.0, decimals=0, space="buy", optimize=False)
  entry_103_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-25.0, decimals=0, space="buy", optimize=False)
  entry_103_rsi_14_min = DecimalParameter(20.0, 60.0, default=24.0, decimals=0, space="buy", optimize=False)
  entry_103_sma_offset = DecimalParameter(0.930, 0.972, default=0.958, decimals=3, space="buy", optimize=False)
  entry_103_bb_offset = DecimalParameter(0.940, 1.010, default=0.948, decimals=3, space="buy", optimize=False)

  entry_104_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_104_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_104_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_104_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_104_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_104_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_104_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.80, decimals=2, space="buy", optimize=False)
  entry_104_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.90, decimals=2, space="buy", optimize=False)
  entry_104_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=0.95, decimals=2, space="buy", optimize=False)
  entry_104_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.00, decimals=2, space="buy", optimize=False)
  entry_104_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_104_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_104_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_104_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_104_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_104_res_level_1d_enabled = CategoricalParameter([True, False], default=True, space="buy", optimize=False)
  entry_104_ema_200_not_dec_1h_enabled = CategoricalParameter([True, False], default=True, space="buy", optimize=False)
  entry_104_ema_200_not_dec_4h_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_104_ema_200_not_dec_1d_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_104_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_104_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_104_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_104_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_104_rsi_3_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_104_rsi_3_max = DecimalParameter(30.0, 70.0, default=46.0, decimals=0, space="buy", optimize=False)
  entry_104_rsi_3_15m_min = DecimalParameter(00.0, 36.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_104_rsi_3_1h_min = DecimalParameter(00.0, 36.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_104_rsi_3_4h_min = DecimalParameter(00.0, 36.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_104_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_104_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.80, decimals=2, space="buy", optimize=False)
  entry_104_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=70.0, decimals=0, space="buy", optimize=False)
  entry_104_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_104_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=70.0, decimals=0, space="buy", optimize=False)
  entry_104_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_104_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=70.0, decimals=0, space="buy", optimize=False)
  entry_104_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_104_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_104_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_104_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_104_rsi_14_min = DecimalParameter(20.0, 60.0, default=30.0, decimals=0, space="buy", optimize=False)
  entry_104_rsi_14_max = DecimalParameter(20.0, 60.0, default=46.0, decimals=0, space="buy", optimize=False)
  entry_104_sma_offset = DecimalParameter(0.940, 0.984, default=0.956, decimals=3, space="buy", optimize=False)

  entry_106_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_106_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_106_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_106_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_106_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_106_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_106_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.5, decimals=2, space="buy", optimize=False)
  entry_106_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.75, decimals=2, space="buy", optimize=False)
  entry_106_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=0.90, decimals=2, space="buy", optimize=False)
  entry_106_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.00, decimals=2, space="buy", optimize=False)
  entry_106_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_106_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_106_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_106_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_106_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_106_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_106_ema_200_not_dec_1h_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_106_ema_200_not_dec_4h_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_106_ema_200_not_dec_1d_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_106_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_106_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_106_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_106_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_106_rsi_3_min = DecimalParameter(00.0, 30.0, default=2.0, decimals=0, space="buy", optimize=False)
  entry_106_rsi_3_max = DecimalParameter(30.0, 60.0, default=60.0, decimals=0, space="buy", optimize=False)
  entry_106_rsi_3_15m_min = DecimalParameter(00.0, 30.0, default=2.0, decimals=0, space="buy", optimize=False)
  entry_106_rsi_3_1h_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_106_rsi_3_4h_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_106_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=6.0, decimals=0, space="buy", optimize=False)
  entry_106_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_106_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_106_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_106_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_106_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_106_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_106_r_14_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_106_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_106_r_14_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_106_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_106_r_480_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_106_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_106_r_480_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_106_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_106_cti_20_max = DecimalParameter(-0.9, 0.0, default=-0.7, decimals=1, space="buy", optimize=False)
  entry_106_ewo_50_200_max = DecimalParameter(-10.0, -2.0, default=-8.0, decimals=1, space="buy", optimize=True)
  entry_106_sma_offset = DecimalParameter(0.980, 0.999, default=0.986, decimals=3, space="buy", optimize=True)

  entry_107_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_107_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_107_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_107_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_107_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_107_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_107_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.80, decimals=2, space="buy", optimize=False)
  entry_107_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.90, decimals=2, space="buy", optimize=False)
  entry_107_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=1.10, decimals=2, space="buy", optimize=False)
  entry_107_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.20, decimals=2, space="buy", optimize=False)
  entry_107_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_107_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_107_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_107_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_107_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_107_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_107_ema_200_not_dec_1h_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_107_ema_200_not_dec_4h_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_107_ema_200_not_dec_1d_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_107_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_107_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_107_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_107_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_107_rsi_3_min = DecimalParameter(00.0, 30.0, default=2.0, decimals=0, space="buy", optimize=False)
  entry_107_rsi_3_max = DecimalParameter(30.0, 70.0, default=46.0, decimals=0, space="buy", optimize=False)
  entry_107_rsi_3_15m_min = DecimalParameter(00.0, 36.0, default=2.0, decimals=0, space="buy", optimize=False)
  entry_107_rsi_3_1h_min = DecimalParameter(00.0, 36.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_107_rsi_3_4h_min = DecimalParameter(00.0, 36.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_107_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_107_cti_20_1h_min = DecimalParameter(-0.99, -0.50, default=-0.95, decimals=2, space="buy", optimize=False)
  entry_107_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.80, decimals=2, space="buy", optimize=False)
  entry_107_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_107_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.90, decimals=2, space="buy", optimize=False)
  entry_107_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_107_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_107_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_107_r_14_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_107_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_107_r_14_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_107_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_107_r_480_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_107_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_107_r_480_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_107_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_107_bb40_bbdelta_close = DecimalParameter(0.030, 0.060, default=0.040, decimals=3, space="buy", optimize=False)
  entry_107_bb40_closedelta_close = DecimalParameter(
    0.010, 0.040, default=0.020, decimals=3, space="buy", optimize=False
  )
  entry_107_bb40_tail_bbdelta = DecimalParameter(0.10, 0.60, default=0.40, decimals=2, space="buy", optimize=False)
  entry_107_cti_20_max = DecimalParameter(-0.90, -0.50, default=-0.75, decimals=2, space="buy", optimize=False)
  entry_107_r_480_min = DecimalParameter(-100.0, -80.0, default=-94.0, decimals=0, space="buy", optimize=False)

  entry_108_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_108_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_108_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_108_high_max_24_1h = DecimalParameter(00.40, 0.95, default=0.60, decimals=2, space="buy", optimize=False)
  entry_108_high_max_24_4h = DecimalParameter(00.40, 0.95, default=0.50, decimals=2, space="buy", optimize=False)
  entry_108_high_max_6_1d = DecimalParameter(00.30, 0.95, default=0.45, decimals=2, space="buy", optimize=False)
  entry_108_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.80, decimals=2, space="buy", optimize=False)
  entry_108_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.90, decimals=2, space="buy", optimize=False)
  entry_108_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=1.10, decimals=2, space="buy", optimize=False)
  entry_108_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.20, decimals=2, space="buy", optimize=False)
  entry_108_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_108_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_108_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_108_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_108_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_108_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_108_ema_200_not_dec_1h_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_108_ema_200_not_dec_4h_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_108_ema_200_not_dec_1d_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_108_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_108_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_108_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_108_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_108_rsi_3_min = DecimalParameter(00.0, 30.0, default=0.0, decimals=0, space="buy", optimize=False)
  entry_108_rsi_3_max = DecimalParameter(30.0, 70.0, default=46.0, decimals=0, space="buy", optimize=False)
  entry_108_rsi_3_15m_min = DecimalParameter(00.0, 36.0, default=2.0, decimals=0, space="buy", optimize=False)
  entry_108_rsi_3_1h_min = DecimalParameter(00.0, 36.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_108_rsi_3_4h_min = DecimalParameter(00.0, 36.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_108_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_108_cti_20_1h_min = DecimalParameter(-0.99, -0.50, default=-0.95, decimals=2, space="buy", optimize=False)
  entry_108_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.90, decimals=2, space="buy", optimize=False)
  entry_108_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_108_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.90, decimals=2, space="buy", optimize=False)
  entry_108_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_108_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.95, decimals=2, space="buy", optimize=False)
  entry_108_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=90.0, decimals=0, space="buy", optimize=False)
  entry_108_r_14_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_108_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_108_r_14_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_108_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_108_r_480_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_108_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_108_r_480_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_108_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_108_rsi_14_min = DecimalParameter(10.0, 40.0, default=15.0, decimals=0, space="buy", optimize=False)
  entry_108_cti_20_max = DecimalParameter(-0.70, -0.40, default=-0.50, decimals=2, space="buy", optimize=False)
  entry_108_r_14_max = DecimalParameter(-100.0, 80.0, default=-90.0, decimals=0, space="buy", optimize=False)
  entry_108_bb_offset = DecimalParameter(0.970, 0.999, default=0.999, decimals=3, space="buy", optimize=False)
  entry_108_ema_open_offset = DecimalParameter(0.0100, 0.0400, default=0.0200, decimals=4, space="buy", optimize=False)

  entry_110_close_max_12 = DecimalParameter(00.50, 0.95, default=0.80, decimals=2, space="buy", optimize=False)
  entry_110_close_max_24 = DecimalParameter(00.50, 0.95, default=0.75, decimals=2, space="buy", optimize=False)
  entry_110_close_max_48 = DecimalParameter(00.50, 0.95, default=0.70, decimals=2, space="buy", optimize=False)
  entry_110_high_max_24_1h = DecimalParameter(00.40, 0.90, default=0.65, decimals=2, space="buy", optimize=False)
  entry_110_high_max_24_4h = DecimalParameter(00.40, 0.85, default=0.60, decimals=2, space="buy", optimize=False)
  entry_110_high_max_6_1d = DecimalParameter(00.30, 0.80, default=0.55, decimals=2, space="buy", optimize=False)
  entry_110_hl_pct_change_6_1h = DecimalParameter(00.30, 0.90, default=0.5, decimals=2, space="buy", optimize=False)
  entry_110_hl_pct_change_12_1h = DecimalParameter(00.40, 1.00, default=0.75, decimals=2, space="buy", optimize=False)
  entry_110_hl_pct_change_24_1h = DecimalParameter(00.50, 1.20, default=0.90, decimals=2, space="buy", optimize=False)
  entry_110_hl_pct_change_48_1h = DecimalParameter(00.60, 1.60, default=1.00, decimals=2, space="buy", optimize=False)
  entry_110_sup_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_110_res_level_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_110_sup_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_110_res_level_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_110_sup_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_110_res_level_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_110_ema_200_not_dec_1h_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_110_ema_200_not_dec_4h_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_110_ema_200_not_dec_1d_enabled = CategoricalParameter(
    [True, False], default=False, space="buy", optimize=False
  )
  entry_110_not_downtrend_15m_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_110_not_downtrend_1h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_110_not_downtrend_4h_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_110_not_downtrend_1d_enabled = CategoricalParameter([True, False], default=False, space="buy", optimize=False)
  entry_110_rsi_3_min = DecimalParameter(00.0, 30.0, default=2.0, decimals=0, space="buy", optimize=False)
  entry_110_rsi_3_max = DecimalParameter(30.0, 60.0, default=60.0, decimals=0, space="buy", optimize=False)
  entry_110_rsi_3_15m_min = DecimalParameter(00.0, 30.0, default=8.0, decimals=0, space="buy", optimize=False)
  entry_110_rsi_3_1h_min = DecimalParameter(00.0, 30.0, default=16.0, decimals=0, space="buy", optimize=False)
  entry_110_rsi_3_4h_min = DecimalParameter(00.0, 30.0, default=10.0, decimals=0, space="buy", optimize=False)
  entry_110_rsi_3_1d_min = DecimalParameter(00.0, 30.0, default=10.0, decimals=0, space="buy", optimize=False)
  entry_110_cti_20_1h_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_110_rsi_14_1h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_110_cti_20_4h_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_110_rsi_14_4h_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_110_cti_20_1d_max = DecimalParameter(0.0, 0.99, default=0.9, decimals=2, space="buy", optimize=False)
  entry_110_rsi_14_1d_max = DecimalParameter(50.0, 90.0, default=80.0, decimals=0, space="buy", optimize=False)
  entry_110_r_14_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_110_r_14_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_110_r_14_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_110_r_14_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_110_r_480_1h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_110_r_480_1h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_110_r_480_4h_min = DecimalParameter(-100.0, -70.0, default=-100.0, decimals=0, space="buy", optimize=False)
  entry_110_r_480_4h_max = DecimalParameter(-40.0, -0.0, default=-0.0, decimals=0, space="buy", optimize=False)
  entry_110_cti_20_max = DecimalParameter(-0.99, -0.50, default=-0.95, decimals=1, space="buy", optimize=False)
  entry_110_ewo_50_200_max = DecimalParameter(-10.0, -2.0, default=-3.0, decimals=1, space="buy", optimize=True)
  entry_110_ema_offset = DecimalParameter(0.980, 0.999, default=0.994, decimals=3, space="buy", optimize=True)



  hold_trades_cache = None
  target_profit_cache = None




























  def __init__(self, config: dict) -> None:
    if "ccxt_config" not in config["exchange"]:
      config["exchange"]["ccxt_config"] = {}
    if "ccxt_async_config" not in config["exchange"]:
      config["exchange"]["ccxt_async_config"] = {}

    options = {
      "brokerId": None,
      "broker": {"spot": None, "margin": None, "future": None, "delivery": None},
      "partner": {
        "spot": {"id": None, "key": None},
        "future": {"id": None, "key": None},
        "id": None,
        "key": None,
      },
    }

    config["exchange"]["ccxt_config"]["options"] = options
    config["exchange"]["ccxt_async_config"]["options"] = options
    super().__init__(config)
    if ("exit_profit_only" in self.config and self.config["exit_profit_only"]) or (
      "sell_profit_only" in self.config and self.config["sell_profit_only"]
    ):
      self.exit_profit_only = True
    if "stop_threshold_spot" in self.config:
      self.stop_threshold = self.config["stop_threshold_spot"]
    if "stop_threshold_futures" in self.config:
      self.stop_threshold_futures = self.config["stop_threshold_futures"]
    if "regular_mode_stake_multiplier_spot" in self.config:
      self.regular_mode_stake_multiplier_spot = self.config["regular_mode_stake_multiplier_spot"]
    if "regular_mode_stake_multiplier_futures" in self.config:
      self.regular_mode_stake_multiplier_futures = self.config["regular_mode_stake_multiplier_futures"]
    if "regular_mode_derisk_1_spot" in self.config:
      self.regular_mode_derisk_1_spot = self.config["regular_mode_derisk_1_spot"]
    if "regular_mode_derisk_1_spot_old" in self.config:
      self.regular_mode_derisk_1_spot_old = self.config["regular_mode_derisk_1_spot_old"]
    if "regular_mode_derisk_spot" in self.config:
      self.regular_mode_derisk_spot = self.config["regular_mode_derisk_spot"]
    if "regular_mode_derisk_spot_old" in self.config:
      self.regular_mode_derisk_spot_old = self.config["regular_mode_derisk_spot_old"]
    if "regular_mode_derisk_1_futures" in self.config:
      self.regular_mode_derisk_1_futures = self.config["regular_mode_derisk_1_futures"]
    if "regular_mode_derisk_1_futures_old" in self.config:
      self.regular_mode_derisk_1_futures_old = self.config["regular_mode_derisk_1_futures_old"]
    if "regular_mode_derisk_futures" in self.config:
      self.regular_mode_derisk_futures = self.config["regular_mode_derisk_futures"]
    if "regular_mode_derisk_futures_old" in self.config:
      self.regular_mode_derisk_futures_old = self.config["regular_mode_derisk_futures_old"]
    if "grind_mode_max_slots" in self.config:
      self.grind_mode_max_slots = self.config["grind_mode_max_slots"]
    if "grind_mode_coins" in self.config:
      self.grind_mode_coins = self.config["grind_mode_coins"]
    if "max_slippage" in self.config:
      self.max_slippage = self.config["max_slippage"]
    if self.target_profit_cache is None:
      bot_name = ""
      if "bot_name" in self.config:
        bot_name = self.config["bot_name"] + "-"
      self.target_profit_cache = Cache(
        self.config["user_data_dir"]
        / (
          "nfix4-profit_max-"
          + bot_name
          + self.config["exchange"]["name"]
          + "-"
          + self.config["stake_currency"]
          + ("-(backtest)" if (self.config["runmode"].value == "backtest") else "")
          + ("-(hyperopt)" if (self.config["runmode"].value == "hyperopt") else "")
          + ".json"
        )
      )

    if self.config["exchange"]["name"] in ["okx", "okex"]:
      self.startup_candle_count = 480
    elif self.config["exchange"]["name"] in ["kraken"]:
      self.startup_candle_count = 710
    elif self.config["exchange"]["name"] in ["bybit"]:
      self.startup_candle_count = 199
    elif self.config["exchange"]["name"] in ["bitget"]:
      self.startup_candle_count = 499
    elif self.config["exchange"]["name"] in ["bingx"]:
      self.startup_candle_count = 499

    if ("trading_mode" in self.config) and (self.config["trading_mode"] in ["futures", "margin"]):
      self.is_futures_mode = True
      self.can_short = True

    self.target_profit_cache.save()


  def get_ticker_indicator(self):
    return int(self.timeframe[:-1])


  def mark_profit_target(
    self,
    mode_name: str,
    pair: str,
    sell: bool,
    signal_name: str,
    trade: Trade,
    current_time: datetime,
    current_rate: float,
    current_profit: float,
    last_candle,
    previous_candle_1,
  ) -> tuple:
    if sell and (signal_name is not None):
      return pair, signal_name

    return None, None


  def exit_profit_target(
    self,
    mode_name: str,
    pair: str,
    trade: Trade,
    current_time: datetime,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    last_candle,
    previous_candle_1,
    previous_rate,
    previous_profit,
    previous_sell_reason,
    previous_time_profit_reached,
    enter_tags,
  ) -> tuple:
    if previous_sell_reason in [f"exit_{mode_name}_stoploss_doom", f"exit_{mode_name}_stoploss"]:
      if profit_ratio > 0.04:

        self._remove_profit_target(pair)
        return False, None
      if profit_ratio < -0.18:
        if profit_ratio < (previous_profit - 0.04):
          return True, previous_sell_reason
      elif profit_ratio < -0.1:
        if profit_ratio < (previous_profit - 0.04):
          return True, previous_sell_reason
      elif profit_ratio < -0.04:
        if profit_ratio < (previous_profit - 0.04):
          return True, previous_sell_reason
      else:
        if profit_ratio < (previous_profit - 0.04):
          return True, previous_sell_reason
    elif previous_sell_reason in [f"exit_{mode_name}_stoploss_u_e"]:
      if profit_current_stake_ratio > 0.04:

        self._remove_profit_target(pair)
        return False, None
      if profit_ratio < (previous_profit - (0.20 if trade.realized_profit == 0.0 else 0.26)):
        return True, previous_sell_reason
    elif previous_sell_reason in [f"exit_profit_{mode_name}_max"]:
      if profit_init_ratio < -0.08:

        self._remove_profit_target(pair)
        return False, None
      if self.is_futures_mode:
        if 0.01 <= profit_init_ratio < 0.02:
          if profit_init_ratio < (previous_profit * 0.5):
            return True, previous_sell_reason
        elif 0.02 <= profit_init_ratio < 0.03:
          if profit_init_ratio < (previous_profit * 0.6):
            return True, previous_sell_reason
        elif 0.03 <= profit_init_ratio < 0.04:
          if profit_init_ratio < (previous_profit * 0.7):
            return True, previous_sell_reason
        elif 0.04 <= profit_init_ratio < 0.08:
          if profit_init_ratio < (previous_profit * 0.8):
            return True, previous_sell_reason
        elif 0.08 <= profit_init_ratio < 0.16:
          if profit_init_ratio < (previous_profit * 0.9):
            return True, previous_sell_reason
        elif 0.16 <= profit_init_ratio:
          if profit_init_ratio < (previous_profit * 0.95):
            return True, previous_sell_reason
      else:
        if 0.01 <= profit_init_ratio < 0.03:
          if profit_init_ratio < (previous_profit * 0.6):
            return True, previous_sell_reason
        elif 0.03 <= profit_init_ratio < 0.08:
          if profit_init_ratio < (previous_profit * 0.65):
            return True, previous_sell_reason
        elif 0.08 <= profit_init_ratio < 0.16:
          if profit_init_ratio < (previous_profit * 0.7):
            return True, previous_sell_reason
        elif 0.16 <= profit_init_ratio:
          if profit_init_ratio < (previous_profit * 0.75):
            return True, previous_sell_reason
    else:
      return False, None

    return False, None


  def calc_total_profit(
    self, trade: "Trade", filled_entries: "Orders", filled_exits: "Orders", exit_rate: float
  ) -> tuple:
    """
    Calculates the absolute profit for open trades.

    :param trade: trade object.
    :param filled_entries: Filled entries list.
    :param filled_exits: Filled exits list.
    :param exit_rate: The exit rate.
    :return tuple: The total profit in stake, ratio, ratio based on current stake, and ratio based on the first entry stake.
    """
    total_amount = 0.0
    total_stake = 0.0
    total_profit = 0.0
    current_stake = 0.0
    for entry_order in filled_entries:
      if trade.is_short:
        entry_stake = entry_order.safe_filled * entry_order.safe_price * (1 - trade.fee_open)
        total_amount += entry_order.safe_filled
        total_stake += entry_stake
        total_profit += entry_stake
      else:
        entry_stake = entry_order.safe_filled * entry_order.safe_price * (1 + trade.fee_open)
        total_amount += entry_order.safe_filled
        total_stake += entry_stake
        total_profit -= entry_stake
    for exit_order in filled_exits:
      if trade.is_short:
        exit_stake = exit_order.safe_filled * exit_order.safe_price * (1 + trade.fee_close)
        total_amount -= exit_order.safe_filled
        total_profit -= exit_stake
      else:
        exit_stake = exit_order.safe_filled * exit_order.safe_price * (1 - trade.fee_close)
        total_amount -= exit_order.safe_filled
        total_profit += exit_stake
    if trade.is_short:
      current_stake = total_amount * exit_rate * (1 + trade.fee_close)
      total_profit -= current_stake
    else:
      current_stake = total_amount * exit_rate * (1 - trade.fee_close)
      total_profit += current_stake
    if self.is_futures_mode:
      total_profit += trade.funding_fees
    total_profit_ratio = total_profit / total_stake
    current_profit_ratio = total_profit / current_stake
    init_profit_ratio = total_profit / filled_entries[0].cost
    return total_profit, total_profit_ratio, current_profit_ratio, init_profit_ratio


  def custom_exit(
    self, pair: str, trade: "Trade", current_time: "datetime", current_rate: float, current_profit: float, **kwargs
  ):
    df, _ = self.dp.get_analyzed_dataframe(pair, self.timeframe)
    last_candle = df.iloc[-1].squeeze()
    previous_candle_1 = df.iloc[-2].squeeze()
    previous_candle_2 = df.iloc[-3].squeeze()
    previous_candle_3 = df.iloc[-4].squeeze()
    previous_candle_4 = df.iloc[-5].squeeze()
    previous_candle_5 = df.iloc[-6].squeeze()

    enter_tag = "empty"
    if hasattr(trade, "enter_tag") and trade.enter_tag is not None:
      enter_tag = trade.enter_tag
    enter_tags = enter_tag.split()

    filled_entries = trade.select_filled_orders(trade.entry_side)
    filled_exits = trade.select_filled_orders(trade.exit_side)

    profit_stake = 0.0
    profit_ratio = 0.0
    profit_current_stake_ratio = 0.0
    profit_init_ratio = 0.0
    profit_stake, profit_ratio, profit_current_stake_ratio, profit_init_ratio = self.calc_total_profit(
      trade, filled_entries, filled_exits, current_rate
    )

    max_profit = (trade.max_rate - trade.open_rate) / trade.open_rate
    max_loss = (trade.open_rate - trade.min_rate) / trade.min_rate

    count_of_entries = len(filled_entries)
    if count_of_entries > 1:
      initial_entry = filled_entries[0]
      if initial_entry is not None and initial_entry.average is not None:
        max_profit = (trade.max_rate - initial_entry.average) / initial_entry.average
        max_loss = (initial_entry.average - trade.min_rate) / trade.min_rate

    if any(c in self.long_normal_mode_tags for c in enter_tags):
      sell, signal_name = self.long_exit_normal(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if any(c in self.long_pump_mode_tags for c in enter_tags):
      sell, signal_name = self.long_exit_pump(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if any(c in self.long_quick_mode_tags for c in enter_tags):
      sell, signal_name = self.long_exit_quick(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if all(c in self.long_rebuy_mode_tags for c in enter_tags):
      sell, signal_name = self.long_exit_rebuy(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if any(c in self.long_mode_tags for c in enter_tags):
      sell, signal_name = self.long_exit_high_profit(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if any(c in self.long_rapid_mode_tags for c in enter_tags):
      sell, signal_name = self.long_exit_rapid(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if all(c in self.long_grind_mode_tags for c in enter_tags):
      sell, signal_name = self.long_exit_grind(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if any(c in self.short_normal_mode_tags for c in enter_tags):
      sell, signal_name = self.short_exit_normal(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if any(c in self.short_pump_mode_tags for c in enter_tags):
      sell, signal_name = self.short_exit_pump(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if any(c in self.short_quick_mode_tags for c in enter_tags):
      sell, signal_name = self.short_exit_quick(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if all(c in self.short_rebuy_mode_tags for c in enter_tags):
      sell, signal_name = self.short_exit_rebuy(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if any(c in self.short_mode_tags for c in enter_tags):
      sell, signal_name = self.short_exit_high_profit(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if any(c in self.short_rapid_mode_tags for c in enter_tags):
      sell, signal_name = self.short_exit_rapid(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if not trade.is_short and (
      not any(
        c
        in (
          self.long_normal_mode_tags
          + self.long_pump_mode_tags
          + self.long_quick_mode_tags
          + self.long_rebuy_mode_tags
          + self.long_mode_tags
          + self.long_rapid_mode_tags
          + self.long_grind_mode_tags
        )
        for c in enter_tags
      )
    ):

      sell, signal_name = self.long_exit_normal(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    if trade.is_short and (
      not any(
        c
        in (
          self.short_normal_mode_tags
          + self.short_pump_mode_tags
          + self.short_quick_mode_tags
          + self.short_rebuy_mode_tags
          + self.short_mode_tags
          + self.short_rapid_mode_tags
          + self.short_grind_mode_tags
        )
        for c in enter_tags
      )
    ):

      sell, signal_name = self.short_exit_normal(
        pair,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )
      if sell and (signal_name is not None):
        return f"{signal_name} ( {enter_tag})"

    return None


  def custom_stake_amount(
    self,
    pair: str,
    current_time: datetime,
    current_rate: float,
    proposed_stake: float,
    min_stake: Optional[float],
    max_stake: float,
    leverage: float,
    entry_tag: Optional[str],
    side: str,
    **kwargs,
  ) -> float:
    enter_tags = entry_tag.split()
    if side == "long":

      if all(c in self.long_rebuy_mode_tags for c in enter_tags) or (
        any(c in self.long_rebuy_mode_tags for c in enter_tags)
        and all(c in (self.long_rebuy_mode_tags + self.long_grind_mode_tags) for c in enter_tags)
      ):
        stake_multiplier = self.rebuy_mode_stake_multiplier

        if (proposed_stake * self.rebuy_mode_stake_multiplier) < min_stake:
          stake_multiplier = self.rebuy_mode_stake_multiplier_alt
        return proposed_stake * stake_multiplier

      elif all(c in self.long_grind_mode_tags for c in enter_tags):
        for _, item in enumerate(
          self.grind_mode_stake_multiplier_futures if self.is_futures_mode else self.grind_mode_stake_multiplier_spot
        ):
          if (proposed_stake * item) > min_stake:
            stake_multiplier = item
            return proposed_stake * stake_multiplier
      else:
        stake_multiplier = (
          self.regular_mode_stake_multiplier_futures[0]
          if self.is_futures_mode
          else self.regular_mode_stake_multiplier_spot[0]
        )
        if (proposed_stake * stake_multiplier) > min_stake:
          return proposed_stake * stake_multiplier
        else:
          return min_stake
    else:

      if all(c in self.short_rebuy_mode_tags for c in enter_tags) or (
        any(c in self.short_rebuy_mode_tags for c in enter_tags)
        and all(c in (self.short_rebuy_mode_tags + self.short_grind_mode_tags) for c in enter_tags)
      ):
        stake_multiplier = self.rebuy_mode_stake_multiplier

        if (proposed_stake * self.rebuy_mode_stake_multiplier) < min_stake:
          stake_multiplier = self.rebuy_mode_stake_multiplier_alt
        return proposed_stake * stake_multiplier

      elif all(c in self.short_grind_mode_tags for c in enter_tags):
        for _, item in enumerate(
          self.grind_mode_stake_multiplier_futures if self.is_futures_mode else self.grind_mode_stake_multiplier_spot
        ):
          if (proposed_stake * item) > min_stake:
            stake_multiplier = item
            return proposed_stake * stake_multiplier
      else:
        stake_multiplier = (
          self.regular_mode_stake_multiplier_futures[0]
          if self.is_futures_mode
          else self.regular_mode_stake_multiplier_spot[0]
        )
        if (proposed_stake * stake_multiplier) > min_stake:
          return proposed_stake * stake_multiplier
        else:
          return min_stake

    return proposed_stake


  def adjust_trade_position(
    self,
    trade: Trade,
    current_time: datetime,
    current_rate: float,
    current_profit: float,
    min_stake: Optional[float],
    max_stake: float,
    current_entry_rate: float,
    current_exit_rate: float,
    current_entry_profit: float,
    current_exit_profit: float,
    **kwargs,
  ):
    if self.position_adjustment_enable == False:
      return None

    enter_tag = "empty"
    if hasattr(trade, "enter_tag") and trade.enter_tag is not None:
      enter_tag = trade.enter_tag
    enter_tags = enter_tag.split()

    if not trade.is_short and (
      all(c in self.long_rebuy_mode_tags for c in enter_tags)
      or (
        any(c in self.long_rebuy_mode_tags for c in enter_tags)
        and all(c in (self.long_rebuy_mode_tags + self.long_grind_mode_tags) for c in enter_tags)
      )
    ):
      return self.long_rebuy_adjust_trade_position(
        trade,
        enter_tags,
        current_time,
        current_rate,
        current_profit,
        min_stake,
        max_stake,
        current_entry_rate,
        current_exit_rate,
        current_entry_profit,
        current_exit_profit,
      )

    elif not trade.is_short and (
      any(
        c
        in (
          self.long_normal_mode_tags
          + self.long_pump_mode_tags
          + self.long_quick_mode_tags
          + self.long_mode_tags
          + self.long_rapid_mode_tags
          + self.long_grind_mode_tags
        )
        for c in enter_tags
      )
      or not any(
        c
        in (
          self.long_normal_mode_tags
          + self.long_pump_mode_tags
          + self.long_quick_mode_tags
          + self.long_rebuy_mode_tags
          + self.long_mode_tags
          + self.long_rapid_mode_tags
          + self.long_grind_mode_tags
        )
        for c in enter_tags
      )
    ):
      return self.long_grind_adjust_trade_position(
        trade,
        enter_tags,
        current_time,
        current_rate,
        current_profit,
        min_stake,
        max_stake,
        current_entry_rate,
        current_exit_rate,
        current_entry_profit,
        current_exit_profit,
      )

    elif trade.is_short and (
      any(
        c
        in (
          self.short_normal_mode_tags
          + self.short_pump_mode_tags
          + self.short_quick_mode_tags
          + self.short_mode_tags
          + self.short_rapid_mode_tags
          + self.short_grind_mode_tags
        )
        for c in enter_tags
      )
      or not any(
        c
        in (
          self.short_normal_mode_tags
          + self.short_pump_mode_tags
          + self.short_quick_mode_tags
          + self.short_rebuy_mode_tags
          + self.short_mode_tags
          + self.short_rapid_mode_tags
          + self.short_grind_mode_tags
        )
        for c in enter_tags
      )
    ):
      return self.short_grind_adjust_trade_position(
        trade,
        enter_tags,
        current_time,
        current_rate,
        current_profit,
        min_stake,
        max_stake,
        current_entry_rate,
        current_exit_rate,
        current_entry_profit,
        current_exit_profit,
      )

    return None


  def informative_pairs(self):

    pairs = self.dp.current_whitelist()

    informative_pairs = []
    for info_timeframe in self.info_timeframes:
      informative_pairs.extend([(pair, info_timeframe) for pair in pairs])

    if self.config["stake_currency"] in [
      "USDT",
      "BUSD",
      "USDC",
      "DAI",
      "TUSD",
      "FDUSD",
      "PAX",
      "USD",
      "EUR",
      "GBP",
      "TRY",
    ]:
      if ("trading_mode" in self.config) and (self.config["trading_mode"] in ["futures", "margin"]):
        btc_info_pair = f"BTC/{self.config['stake_currency']}:{self.config['stake_currency']}"
      else:
        btc_info_pair = f"BTC/{self.config['stake_currency']}"
    else:
      if ("trading_mode" in self.config) and (self.config["trading_mode"] in ["futures", "margin"]):
        btc_info_pair = "BTC/USDT:USDT"
      else:
        btc_info_pair = "BTC/USDT"

    informative_pairs.extend([(btc_info_pair, btc_info_timeframe) for btc_info_timeframe in self.btc_info_timeframes])

    return informative_pairs


  def informative_1d_indicators(self, metadata: dict, info_timeframe) -> DataFrame:
    tik = time.perf_counter()
    assert self.dp, "DataProvider is required for multiple timeframes."

    informative_1d = self.dp.get_pair_dataframe(pair=metadata["pair"], timeframe=info_timeframe)



    informative_1d["rsi_3"] = pta.rsi(informative_1d["close"], length=3, fillna=0.0)
    informative_1d["rsi_14"] = pta.rsi(informative_1d["close"], length=14, fillna=0.0)

    informative_1d["rsi_14_max_6"] = informative_1d["rsi_14"].rolling(6).max()

    informative_1d["ema_200"] = ta.EMA(informative_1d, timeperiod=200)

    informative_1d["ema_200_dec_4"] = (informative_1d["ema_200"].isnull()) | (
      informative_1d["ema_200"] <= informative_1d["ema_200"].shift(4)
    )

    informative_1d["cti_20"] = pta.cti(informative_1d["close"], length=20)

    informative_1d["cti_20_dec_3"] = informative_1d["cti_20"] < informative_1d["cti_20"].shift(3)

    (
      informative_1d["pivot"],
      informative_1d["res1"],
      informative_1d["res2"],
      informative_1d["res3"],
      informative_1d["sup1"],
      informative_1d["sup2"],
      informative_1d["sup3"],
    ) = pivot_points(informative_1d, mode="fibonacci")

    res_series = (
      informative_1d["high"].rolling(window=5, center=True).apply(lambda row: is_resistance(row), raw=True).shift(2)
    )
    sup_series = (
      informative_1d["low"].rolling(window=5, center=True).apply(lambda row: is_support(row), raw=True).shift(2)
    )
    informative_1d["res_level"] = Series(
      np.where(
        res_series,
        np.where(informative_1d["close"] > informative_1d["open"], informative_1d["close"], informative_1d["open"]),
        float("NaN"),
      )
    ).ffill()
    informative_1d["res_hlevel"] = Series(np.where(res_series, informative_1d["high"], float("NaN"))).ffill()
    informative_1d["sup_level"] = Series(
      np.where(
        sup_series,
        np.where(informative_1d["close"] < informative_1d["open"], informative_1d["close"], informative_1d["open"]),
        float("NaN"),
      )
    ).ffill()

    informative_1d["not_downtrend"] = (informative_1d["close"] > informative_1d["close"].shift(2)) | (
      informative_1d["rsi_14"] > 50.0
    )

    informative_1d["is_downtrend_3"] = (
      (informative_1d["close"] < informative_1d["open"])
      & (informative_1d["close"].shift(1) < informative_1d["open"].shift(1))
      & (informative_1d["close"].shift(2) < informative_1d["open"].shift(2))
    )

    informative_1d["top_wick_pct"] = (
      informative_1d["high"] - np.maximum(informative_1d["open"], informative_1d["close"])
    ) / np.maximum(informative_1d["open"], informative_1d["close"])
    informative_1d["bot_wick_pct"] = abs(
      (informative_1d["low"] - np.minimum(informative_1d["open"], informative_1d["close"]))
      / np.minimum(informative_1d["open"], informative_1d["close"])
    )

    informative_1d["change_pct"] = (informative_1d["close"] - informative_1d["open"]) / informative_1d["open"]

    informative_1d["hl_pct_change_3"] = range_percent_change(self, informative_1d, "HL", 3)
    informative_1d["hl_pct_change_6"] = range_percent_change(self, informative_1d, "HL", 6)

    informative_1d["high_max_6"] = informative_1d["high"].rolling(6).max()
    informative_1d["high_max_12"] = informative_1d["high"].rolling(12).max()

    informative_1d["low_min_6"] = informative_1d["low"].rolling(6).min()
    informative_1d["low_min_12"] = informative_1d["low"].rolling(12).min()


    tok = time.perf_counter()
    log.debug(f"[{metadata['pair']}] informative_1d_indicators took: {tok - tik:0.4f} seconds.")

    return informative_1d


  def informative_4h_indicators(self, metadata: dict, info_timeframe) -> DataFrame:
    tik = time.perf_counter()
    assert self.dp, "DataProvider is required for multiple timeframes."

    informative_4h = self.dp.get_pair_dataframe(pair=metadata["pair"], timeframe=info_timeframe)



    informative_4h["rsi_3"] = pta.rsi(informative_4h["close"], length=3, fillna=0.0)
    informative_4h["rsi_14"] = pta.rsi(informative_4h["close"], length=14, fillna=0.0)

    informative_4h["rsi_14_max_3"] = informative_4h["rsi_14"].rolling(3).max()
    informative_4h["rsi_14_max_6"] = informative_4h["rsi_14"].rolling(6).max()

    informative_4h["ema_12"] = ta.EMA(informative_4h, timeperiod=12)
    informative_4h["ema_26"] = ta.EMA(informative_4h, timeperiod=26)
    informative_4h["ema_50"] = ta.EMA(informative_4h, timeperiod=50)
    informative_4h["ema_200"] = ta.EMA(informative_4h, timeperiod=200)

    informative_4h["ema_200_dec_24"] = (informative_4h["ema_200"].isnull()) | (
      informative_4h["ema_200"] <= informative_4h["ema_200"].shift(24)
    )

    informative_4h["sma_50"] = pta.sma(informative_4h["close"], length=50, fillna=0.0)
    informative_4h["sma_200"] = pta.sma(informative_4h["close"], length=200, fillna=0.0)

    informative_4h["sma_200_dec_48"] = (
      (informative_4h["sma_200"].isnull())
      | (informative_4h["sma_200"].eq(0.0))
      | (informative_4h["sma_200"] <= informative_4h["sma_200"].shift(48))
    )

    informative_4h["zlma_50"] = pta.zlma(informative_4h["close"], length=50, mamode="ema", offset=0, fillna=0.0)

    informative_4h["zlma_50_dec"] = (
      (informative_4h["zlma_50"].isnull())
      | (informative_4h["zlma_50"].eq(0.0))
      | (informative_4h["zlma_50"] <= informative_4h["zlma_50"].shift(1))
    )

    informative_4h["r_14"] = pta.willr(
      informative_4h["high"], informative_4h["low"], informative_4h["close"], length=14
    )
    informative_4h["r_480"] = pta.willr(
      informative_4h["high"], informative_4h["low"], informative_4h["close"], length=480
    )

    informative_4h["cti_20"] = pta.cti(informative_4h["close"], length=20)

    res_series = (
      informative_4h["high"].rolling(window=5, center=True).apply(lambda row: is_resistance(row), raw=True).shift(2)
    )
    sup_series = (
      informative_4h["low"].rolling(window=5, center=True).apply(lambda row: is_support(row), raw=True).shift(2)
    )
    informative_4h["res_level"] = Series(
      np.where(
        res_series,
        np.where(informative_4h["close"] > informative_4h["open"], informative_4h["close"], informative_4h["open"]),
        float("NaN"),
      )
    ).ffill()
    informative_4h["res_hlevel"] = Series(np.where(res_series, informative_4h["high"], float("NaN"))).ffill()
    informative_4h["sup_level"] = Series(
      np.where(
        sup_series,
        np.where(informative_4h["close"] < informative_4h["open"], informative_4h["close"], informative_4h["open"]),
        float("NaN"),
      )
    ).ffill()

    informative_4h["not_downtrend"] = (informative_4h["close"] > informative_4h["close"].shift(2)) | (
      informative_4h["rsi_14"] > 50.0
    )

    informative_4h["is_downtrend_3"] = (
      (informative_4h["close"] < informative_4h["open"])
      & (informative_4h["close"].shift(1) < informative_4h["open"].shift(1))
      & (informative_4h["close"].shift(2) < informative_4h["open"].shift(2))
    )

    informative_4h["top_wick_pct"] = (
      informative_4h["high"] - np.maximum(informative_4h["open"], informative_4h["close"])
    ) / np.maximum(informative_4h["open"], informative_4h["close"])
    informative_4h["bot_wick_pct"] = abs(
      (informative_4h["low"] - np.minimum(informative_4h["open"], informative_4h["close"]))
      / np.minimum(informative_4h["open"], informative_4h["close"])
    )

    informative_4h["change_pct"] = (informative_4h["close"] - informative_4h["open"]) / informative_4h["open"]

    informative_4h["change_pct_min_3"] = informative_4h["change_pct"].rolling(3).min()

    informative_4h["change_pct_max_3"] = informative_4h["change_pct"].rolling(3).max()

    informative_4h["high_max_3"] = informative_4h["high"].rolling(3).max()
    informative_4h["high_max_12"] = informative_4h["high"].rolling(12).max()
    informative_4h["high_max_24"] = informative_4h["high"].rolling(24).max()
    informative_4h["high_max_36"] = informative_4h["high"].rolling(36).max()
    informative_4h["high_max_48"] = informative_4h["high"].rolling(48).max()

    informative_4h["volume_mean_factor_6"] = informative_4h["volume"] / informative_4h["volume"].rolling(6).mean()


    tok = time.perf_counter()
    log.debug(f"[{metadata['pair']}] informative_1d_indicators took: {tok - tik:0.4f} seconds.")

    return informative_4h


  def informative_1h_indicators(self, metadata: dict, info_timeframe) -> DataFrame:
    tik = time.perf_counter()
    assert self.dp, "DataProvider is required for multiple timeframes."

    informative_1h = self.dp.get_pair_dataframe(pair=metadata["pair"], timeframe=info_timeframe)



    informative_1h["rsi_3"] = pta.rsi(informative_1h["close"], length=3, fillna=0.0)
    informative_1h["rsi_14"] = pta.rsi(informative_1h["close"], length=14, fillna=0.0)

    informative_1h["rsi_14_max_3"] = informative_1h["rsi_14"].rolling(3).max()
    informative_1h["rsi_14_max_6"] = informative_1h["rsi_14"].rolling(6).max()

    informative_1h["ema_12"] = ta.EMA(informative_1h, timeperiod=12)
    informative_1h["ema_26"] = ta.EMA(informative_1h, timeperiod=26)
    informative_1h["ema_50"] = ta.EMA(informative_1h, timeperiod=50)
    informative_1h["ema_200"] = ta.EMA(informative_1h, timeperiod=200)

    informative_1h["ema_200_dec_48"] = (informative_1h["ema_200"].isnull()) | (
      informative_1h["ema_200"] <= informative_1h["ema_200"].shift(48)
    )

    informative_1h["sma_50"] = pta.sma(informative_1h["close"], length=50, fillna=0.0)
    informative_1h["sma_200"] = pta.sma(informative_1h["close"], length=200, fillna=0.0)

    informative_1h["sma_200_dec_48"] = (
      (informative_1h["sma_200"].isnull())
      | (informative_1h["sma_200"].eq(0.0))
      | (informative_1h["sma_200"] <= informative_1h["sma_200"].shift(48))
    )

    informative_1h["zlma_50"] = pta.zlma(informative_1h["close"], length=50, mamode="ema", offset=0, fillna=0.0)

    informative_1h["zlma_50_dec"] = (
      (informative_1h["zlma_50"].isnull())
      | (informative_1h["zlma_50"].eq(0.0))
      | (informative_1h["zlma_50"] <= informative_1h["zlma_50"].shift(1))
    )

    bollinger = qtpylib.bollinger_bands(qtpylib.typical_price(informative_1h), window=20, stds=2)
    informative_1h["bb20_2_low"] = bollinger["lower"]
    informative_1h["bb20_2_mid"] = bollinger["mid"]
    informative_1h["bb20_2_upp"] = bollinger["upper"]

    informative_1h["bb20_2_width"] = (informative_1h["bb20_2_upp"] - informative_1h["bb20_2_low"]) / informative_1h[
      "bb20_2_mid"
    ]

    informative_1h["r_14"] = pta.willr(
      informative_1h["high"], informative_1h["low"], informative_1h["close"], length=14
    )
    informative_1h["r_480"] = pta.willr(
      informative_1h["high"], informative_1h["low"], informative_1h["close"], length=480
    )

    informative_1h["cti_20"] = pta.cti(informative_1h["close"], length=20)
    informative_1h["cti_40"] = pta.cti(informative_1h["close"], length=40)

    informative_1h["cti_20_dec_3"] = informative_1h["cti_20"] < informative_1h["cti_20"].shift(3)

    informative_1h["ewo_50_200"] = ewo(informative_1h, 50, 200)

    high = informative_1h["high"]
    low = informative_1h["low"]
    close = informative_1h["close"]
    chandelier_atr = ta.ATR(high, low, close, 22) * 3.0
    long_stop = (high.rolling(22).max() if True else high.rolling(22).apply(lambda x: x[:-1].max())) - chandelier_atr
    long_stop_prev = long_stop.shift(1).fillna(long_stop)
    long_stop = close.shift(1).where(close.shift(1) > long_stop_prev, long_stop)
    short_stop = (low.rolling(22).min() if True else low.rolling(22).apply(lambda x: x[:-1].min())) + chandelier_atr
    short_stop_prev = short_stop.shift(1).fillna(short_stop)
    short_stop = close.shift(1).where(close.shift(1) < short_stop_prev, short_stop)
    informative_1h["chandelier_dir"] = 1
    informative_1h.loc[informative_1h["close"] <= long_stop_prev, "chandelier_dir"] = -1
    informative_1h.loc[informative_1h["close"] > short_stop_prev, "chandelier_dir"] = 1

    res_series = (
      informative_1h["high"].rolling(window=5, center=True).apply(lambda row: is_resistance(row), raw=True).shift(2)
    )
    sup_series = (
      informative_1h["low"].rolling(window=5, center=True).apply(lambda row: is_support(row), raw=True).shift(2)
    )
    informative_1h["res_level"] = Series(
      np.where(
        res_series,
        np.where(informative_1h["close"] > informative_1h["open"], informative_1h["close"], informative_1h["open"]),
        float("NaN"),
      )
    ).ffill()
    informative_1h["res_hlevel"] = Series(np.where(res_series, informative_1h["high"], float("NaN"))).ffill()
    informative_1h["sup_level"] = Series(
      np.where(
        sup_series,
        np.where(informative_1h["close"] < informative_1h["open"], informative_1h["close"], informative_1h["open"]),
        float("NaN"),
      )
    ).ffill()

    informative_1h["hl_pct_change_48"] = range_percent_change(self, informative_1h, "HL", 48)
    informative_1h["hl_pct_change_24"] = range_percent_change(self, informative_1h, "HL", 24)
    informative_1h["hl_pct_change_12"] = range_percent_change(self, informative_1h, "HL", 12)
    informative_1h["hl_pct_change_6"] = range_percent_change(self, informative_1h, "HL", 6)

    informative_1h["not_downtrend"] = (informative_1h["close"] > informative_1h["close"].shift(2)) | (
      informative_1h["rsi_14"] > 50.0
    )

    informative_1h["is_downtrend_3"] = (
      (informative_1h["close"] < informative_1h["open"])
      & (informative_1h["close"].shift(1) < informative_1h["open"].shift(1))
      & (informative_1h["close"].shift(2) < informative_1h["open"].shift(2))
    )

    informative_1h["top_wick_pct"] = (
      informative_1h["high"] - np.maximum(informative_1h["open"], informative_1h["close"])
    ) / np.maximum(informative_1h["open"], informative_1h["close"])

    informative_1h["change_pct"] = (informative_1h["close"] - informative_1h["open"]) / informative_1h["open"]

    informative_1h["high_max_6"] = informative_1h["high"].rolling(6).max()
    informative_1h["high_max_12"] = informative_1h["high"].rolling(12).max()
    informative_1h["high_max_24"] = informative_1h["high"].rolling(24).max()
    informative_1h["high_max_48"] = informative_1h["high"].rolling(48).max()

    informative_1h["low_min_12"] = informative_1h["low"].rolling(12).min()
    informative_1h["low_min_24"] = informative_1h["low"].rolling(24).min()
    informative_1h["low_min_48"] = informative_1h["low"].rolling(48).min()


    tok = time.perf_counter()
    log.debug(f"[{metadata['pair']}] informative_1h_indicators took: {tok - tik:0.4f} seconds.")

    return informative_1h


  def informative_15m_indicators(self, metadata: dict, info_timeframe) -> DataFrame:
    tik = time.perf_counter()
    assert self.dp, "DataProvider is required for multiple timeframes."

    informative_15m = self.dp.get_pair_dataframe(pair=metadata["pair"], timeframe=info_timeframe)



    informative_15m["rsi_3"] = pta.rsi(informative_15m["close"], length=3)
    informative_15m["rsi_14"] = pta.rsi(informative_15m["close"], length=14)

    informative_15m["ema_12"] = ta.EMA(informative_15m, timeperiod=12)
    informative_15m["ema_26"] = ta.EMA(informative_15m, timeperiod=26)
    informative_15m["ema_200"] = ta.EMA(informative_15m, timeperiod=200)

    informative_15m["ema_200_dec_24"] = (informative_15m["ema_200"].isnull()) | (
      informative_15m["ema_200"] <= informative_15m["ema_200"].shift(24)
    )

    informative_15m["sma_200"] = ta.SMA(informative_15m, timeperiod=200)

    informative_15m["zlma_50"] = pta.zlma(informative_15m["close"], length=50, mamode="ema", offset=0)

    informative_15m["zlma_50_dec"] = (informative_15m["zlma_50"].isnull()) | (
      informative_15m["zlma_50"] <= informative_15m["zlma_50"].shift(1)
    )

    bollinger = qtpylib.bollinger_bands(qtpylib.typical_price(informative_15m), window=20, stds=2)
    informative_15m["bb20_2_low"] = bollinger["lower"]
    informative_15m["bb20_2_mid"] = bollinger["mid"]
    informative_15m["bb20_2_upp"] = bollinger["upper"]

    informative_15m["cti_20"] = pta.cti(informative_15m["close"], length=20)

    informative_15m["ewo_50_200"] = ewo(informative_15m, 50, 200)

    informative_15m["not_downtrend"] = (
      (informative_15m["close"] > informative_15m["open"])
      | (informative_15m["close"].shift(1) > informative_15m["open"].shift(1))
      | (informative_15m["close"].shift(2) > informative_15m["open"].shift(2))
      | (informative_15m["rsi_14"] > 50.0)
      | (informative_15m["rsi_3"] > 25.0)
    )


    tok = time.perf_counter()
    log.debug(f"[{metadata['pair']}] informative_15m_indicators took: {tok - tik:0.4f} seconds.")

    return informative_15m


  def base_tf_5m_indicators(self, metadata: dict, df: DataFrame) -> DataFrame:
    tik = time.perf_counter()











































    df["rsi_3"] = pta.rsi(df["close"], length=3)
    df["rsi_14"] = pta.rsi(df["close"], length=14)
    df["rsi_20"] = pta.rsi(df["close"], length=20)

    df["ema_12"] = ta.EMA(df, timeperiod=12)
    df["ema_16"] = ta.EMA(df, timeperiod=16)
    df["ema_20"] = ta.EMA(df, timeperiod=20)
    df["ema_26"] = ta.EMA(df, timeperiod=26)
    df["ema_50"] = ta.EMA(df, timeperiod=50)
    df["ema_100"] = ta.EMA(df, timeperiod=100)
    df["ema_200"] = ta.EMA(df, timeperiod=200)

    df["ema_200_dec_24"] = (df["ema_200"].isnull()) | (df["ema_200"] <= df["ema_200"].shift(24))

    df["ema_200_pct_change_144"] = (df["ema_200"] - df["ema_200"].shift(144)) / df["ema_200"].shift(144)
    df["ema_200_pct_change_288"] = (df["ema_200"] - df["ema_200"].shift(288)) / df["ema_200"].shift(288)

    df["sma_16"] = ta.SMA(df, timeperiod=16)
    df["sma_30"] = ta.SMA(df, timeperiod=30)
    df["sma_75"] = ta.SMA(df, timeperiod=75)
    df["sma_200"] = ta.SMA(df, timeperiod=200)

    bb_20_std2 = qtpylib.bollinger_bands(qtpylib.typical_price(df), window=20, stds=2)
    df["bb20_2_low"] = bb_20_std2["lower"]
    df["bb20_2_mid"] = bb_20_std2["mid"]
    df["bb20_2_upp"] = bb_20_std2["upper"]

    bb_40_std2 = qtpylib.bollinger_bands(df["close"], window=40, stds=2)
    df["bb40_2_low"] = bb_40_std2["lower"]
    df["bb40_2_mid"] = bb_40_std2["mid"]
    df["bb40_2_delta"] = (bb_40_std2["mid"] - df["bb40_2_low"]).abs()
    df["bb40_2_tail"] = (df["close"] - df["bb40_2_low"]).abs()

    df["r_14"] = pta.willr(df["high"], df["low"], df["close"], length=14)
    df["r_480"] = pta.willr(df["high"], df["low"], df["close"], length=480)

    df["cti_20"] = pta.cti(df["close"], length=20)

    df["cci_20"] = ta.CCI(df, source="hlc3", timeperiod=20)

    df["ewo_50_200"] = ewo(df, 50, 200)

    df["hma_55"] = pta.hma(df["close"], length=55)
    df["hma_70"] = pta.hma(df["close"], length=70)

    df["hma_55_buy"] = (df["hma_55"] > df["hma_55"].shift(1)) & (df["hma_55"].shift(1) < df["hma_55"].shift(2))
    df["hma_70_buy"] = (df["hma_70"] > df["hma_70"].shift(1)) & (df["hma_70"].shift(1) < df["hma_70"].shift(2))

    df["zlma_50"] = pta.zlma(df["close"], length=50, mamode="ema", offset=0)

    high = df["high"]
    low = df["low"]
    close = df["close"]
    chandelier_atr = ta.ATR(high, low, close, 22) * 3.0
    long_stop = (high.rolling(22).max() if True else high.rolling(22).apply(lambda x: x[:-1].max())) - chandelier_atr
    long_stop_prev = long_stop.shift(1).fillna(long_stop)
    long_stop = close.shift(1).where(close.shift(1) > long_stop_prev, long_stop)
    short_stop = (low.rolling(22).min() if True else low.rolling(22).apply(lambda x: x[:-1].min())) + chandelier_atr
    short_stop_prev = short_stop.shift(1).fillna(short_stop)
    short_stop = close.shift(1).where(close.shift(1) < short_stop_prev, short_stop)
    df["chandelier_dir"] = 1
    df.loc[df["close"] <= long_stop_prev, "chandelier_dir"] = -1
    df.loc[df["close"] > short_stop_prev, "chandelier_dir"] = 1

    heikinashi = qtpylib.heikinashi(df)
    df["ha_open"] = heikinashi["open"]
    df["ha_close"] = heikinashi["close"]
    df["ha_high"] = heikinashi["high"]
    df["ha_low"] = heikinashi["low"]

    df["tpct_change_0"] = top_percent_change(self, df, 0)
    df["tpct_change_2"] = top_percent_change(self, df, 2)

    df["change_pct"] = (df["close"] - df["open"]) / df["open"]

    df["close_max_12"] = df["close"].rolling(12).max()
    df["close_max_24"] = df["close"].rolling(24).max()
    df["close_max_48"] = df["close"].rolling(48).max()

    df["open_min_6"] = df["open"].rolling(6).min()
    df["open_min_12"] = df["open"].rolling(12).min()

    df["close_min_12"] = df["close"].rolling(12).min()
    df["close_min_24"] = df["close"].rolling(24).min()
    df["close_min_48"] = df["close"].rolling(48).min()

    df["close_delta"] = (df["close"] - df["close"].shift()).abs()

    df["num_empty_288"] = (df["volume"] <= 0).rolling(window=288, min_periods=288).sum()

    df["crossed_below_ema_12_26"] = qtpylib.crossed_below(df["ema_12"], df["ema_26"])
    df["crossed_above_ema_12_26"] = qtpylib.crossed_above(df["ema_12"], df["ema_26"])



    df["rsi"] = ta.RSI(df)

    df["stoch"] = ta.STOCH(df)["slowk"]

    df["roc"] = ta.ROC(df)

    df["uo"] = ta.ULTOSC(df)

    df["ao"] = qtpylib.awesome_oscillator(df)

    df["macd"] = ta.MACD(df)["macd"]

    df["cci"] = ta.CCI(df)

    df["cmf"] = chaikin_money_flow(df, 20)

    df["obv"] = ta.OBV(df)

    df["mfi"] = ta.MFI(df)

    df["adx"] = ta.ADX(df)

    df["atr"] = qtpylib.atr(df, window=14, exp=False)

    keltner = qtpylib.keltner_channel(df, window=20, atrs=1)
    keltner = emaKeltner(df)
    df["kc_upperband"] = keltner["upper"]
    df["kc_middleband"] = keltner["mid"]
    df["kc_lowerband"] = keltner["lower"]

    bollinger = qtpylib.bollinger_bands(qtpylib.typical_price(df), window=20, stds=2)
    df["bollinger_upperband"] = bollinger["upper"]
    df["bollinger_middleband"] = bollinger["mid"]
    df["bollinger_lowerband"] = bollinger["lower"]

    df["ema9"] = ta.EMA(df, timeperiod=9)
    df["ema20"] = ta.EMA(df, timeperiod=20)
    df["ema50"] = ta.EMA(df, timeperiod=50)
    df["ema200"] = ta.EMA(df, timeperiod=200)

    df["tema"] = ta.TEMA(df, timeperiod=9)

    df["sar"] = ta.SAR(df)

    pivots = pivot_points_S(df)
    df["pivot_lows"] = pivots["pivot_lows"]
    df["pivot_highs"] = pivots["pivot_highs"]

    initialize_divergences_lists(df)
    add_divergences(df, "rsi")
    add_divergences(df, "stoch")
    add_divergences(df, "roc")
    add_divergences(df, "uo")
    add_divergences(df, "ao")
    add_divergences(df, "macd")
    add_divergences(df, "cci")
    add_divergences(df, "cmf")
    add_divergences(df, "obv")
    add_divergences(df, "mfi")
    add_divergences(df, "adx")


    if not self.config["runmode"].value in ("live", "dry_run"):

      df["bt_agefilter_ok"] = False
      df.loc[df.index > (12 * 24 * self.bt_min_age_days), "bt_agefilter_ok"] = True
    else:

      df["live_data_ok"] = df["volume"].rolling(window=72, min_periods=72).min() > 0


    tok = time.perf_counter()
    log.debug(f"[{metadata['pair']}] base_tf_5m_indicators took: {tok - tik:0.4f} seconds.")

    return df


  def info_switcher(self, metadata: dict, info_timeframe) -> DataFrame:
    if info_timeframe == "1d":
      return self.informative_1d_indicators(metadata, info_timeframe)
    elif info_timeframe == "4h":
      return self.informative_4h_indicators(metadata, info_timeframe)
    elif info_timeframe == "1h":
      return self.informative_1h_indicators(metadata, info_timeframe)
    elif info_timeframe == "15m":
      return self.informative_15m_indicators(metadata, info_timeframe)
    else:
      raise RuntimeError(f"{info_timeframe} not supported as informative timeframe for BTC pair.")


  def btc_info_1d_indicators(self, btc_info_pair, btc_info_timeframe, metadata: dict) -> DataFrame:
    tik = time.perf_counter()
    btc_info_1d = self.dp.get_pair_dataframe(btc_info_pair, btc_info_timeframe)


    btc_info_1d["rsi_14"] = pta.rsi(btc_info_1d["close"], length=14)



    ignore_columns = ["date"]
    btc_info_1d.rename(columns=lambda s: f"btc_{s}" if s not in ignore_columns else s, inplace=True)

    tok = time.perf_counter()
    log.debug(f"[{metadata['pair']}] btc_info_1d_indicators took: {tok - tik:0.4f} seconds.")

    return btc_info_1d


  def btc_info_4h_indicators(self, btc_info_pair, btc_info_timeframe, metadata: dict) -> DataFrame:
    tik = time.perf_counter()
    btc_info_4h = self.dp.get_pair_dataframe(btc_info_pair, btc_info_timeframe)



    btc_info_4h["rsi_14"] = pta.rsi(btc_info_4h["close"], length=14)

    btc_info_4h["sma_200"] = ta.SMA(btc_info_4h, timeperiod=200)

    btc_info_4h["is_bull"] = btc_info_4h["close"] > btc_info_4h["sma_200"]


    ignore_columns = ["date"]
    btc_info_4h.rename(columns=lambda s: f"btc_{s}" if s not in ignore_columns else s, inplace=True)

    tok = time.perf_counter()
    log.debug(f"[{metadata['pair']}] btc_info_4h_indicators took: {tok - tik:0.4f} seconds.")

    return btc_info_4h


  def btc_info_1h_indicators(self, btc_info_pair, btc_info_timeframe, metadata: dict) -> DataFrame:
    tik = time.perf_counter()
    btc_info_1h = self.dp.get_pair_dataframe(btc_info_pair, btc_info_timeframe)



    btc_info_1h["rsi_14"] = pta.rsi(btc_info_1h["close"], length=14)

    btc_info_1h["not_downtrend"] = (btc_info_1h["close"] > btc_info_1h["close"].shift(2)) | (
      btc_info_1h["rsi_14"] > 50.0
    )


    ignore_columns = ["date"]
    btc_info_1h.rename(columns=lambda s: f"btc_{s}" if s not in ignore_columns else s, inplace=True)

    tok = time.perf_counter()
    log.debug(f"[{metadata['pair']}] btc_info_1h_indicators took: {tok - tik:0.4f} seconds.")

    return btc_info_1h


  def btc_info_15m_indicators(self, btc_info_pair, btc_info_timeframe, metadata: dict) -> DataFrame:
    tik = time.perf_counter()
    btc_info_15m = self.dp.get_pair_dataframe(btc_info_pair, btc_info_timeframe)


    btc_info_15m["rsi_14"] = pta.rsi(btc_info_15m["close"], length=14)


    ignore_columns = ["date"]
    btc_info_15m.rename(columns=lambda s: f"btc_{s}" if s not in ignore_columns else s, inplace=True)

    tok = time.perf_counter()
    log.debug(f"[{metadata['pair']}] btc_info_15m_indicators took: {tok - tik:0.4f} seconds.")

    return btc_info_15m


  def btc_info_5m_indicators(self, btc_info_pair, btc_info_timeframe, metadata: dict) -> DataFrame:
    tik = time.perf_counter()
    btc_info_5m = self.dp.get_pair_dataframe(btc_info_pair, btc_info_timeframe)



    btc_info_5m["rsi_14"] = pta.rsi(btc_info_5m["close"], length=14)

    btc_info_5m["close_max_24"] = btc_info_5m["close"].rolling(24).max()
    btc_info_5m["close_max_72"] = btc_info_5m["close"].rolling(72).max()

    btc_info_5m["pct_close_max_24"] = (btc_info_5m["close_max_24"] - btc_info_5m["close"]) / btc_info_5m["close"]
    btc_info_5m["pct_close_max_72"] = (btc_info_5m["close_max_72"] - btc_info_5m["close"]) / btc_info_5m["close"]

    btc_info_5m["close_min_24"] = btc_info_5m["close"].rolling(24).min()
    btc_info_5m["close_min_72"] = btc_info_5m["close"].rolling(72).min()

    btc_info_5m["pct_close_min_24"] = (btc_info_5m["close_min_24"] - btc_info_5m["close"]) / btc_info_5m["close"]
    btc_info_5m["pct_close_min_72"] = (btc_info_5m["close_min_72"] - btc_info_5m["close"]) / btc_info_5m["close"]


    ignore_columns = ["date"]
    btc_info_5m.rename(columns=lambda s: f"btc_{s}" if s not in ignore_columns else s, inplace=True)

    tok = time.perf_counter()
    log.debug(f"[{metadata['pair']}] btc_info_5m_indicators took: {tok - tik:0.4f} seconds.")

    return btc_info_5m


  def btc_info_switcher(self, btc_info_pair, btc_info_timeframe, metadata: dict) -> DataFrame:
    if btc_info_timeframe == "1d":
      return self.btc_info_1d_indicators(btc_info_pair, btc_info_timeframe, metadata)
    elif btc_info_timeframe == "4h":
      return self.btc_info_4h_indicators(btc_info_pair, btc_info_timeframe, metadata)
    elif btc_info_timeframe == "1h":
      return self.btc_info_1h_indicators(btc_info_pair, btc_info_timeframe, metadata)
    elif btc_info_timeframe == "15m":
      return self.btc_info_15m_indicators(btc_info_pair, btc_info_timeframe, metadata)
    elif btc_info_timeframe == "5m":
      return self.btc_info_5m_indicators(btc_info_pair, btc_info_timeframe, metadata)
    else:
      raise RuntimeError(f"{btc_info_timeframe} not supported as informative timeframe for BTC pair.")


  def populate_indicators(self, df: DataFrame, metadata: dict) -> DataFrame:
    tik = time.perf_counter()
    """
        --> BTC informative indicators
        ___________________________________________________________________________________________
        """
    if self.config["stake_currency"] in [
      "USDT",
      "BUSD",
      "USDC",
      "DAI",
      "TUSD",
      "FDUSD",
      "PAX",
      "USD",
      "EUR",
      "GBP",
      "TRY",
    ]:
      if ("trading_mode" in self.config) and (self.config["trading_mode"] in ["futures", "margin"]):
        btc_info_pair = f"BTC/{self.config['stake_currency']}:{self.config['stake_currency']}"
      else:
        btc_info_pair = f"BTC/{self.config['stake_currency']}"
    else:
      if ("trading_mode" in self.config) and (self.config["trading_mode"] in ["futures", "margin"]):
        btc_info_pair = "BTC/USDT:USDT"
      else:
        btc_info_pair = "BTC/USDT"

    for btc_info_timeframe in self.btc_info_timeframes:
      btc_informative = self.btc_info_switcher(btc_info_pair, btc_info_timeframe, metadata)
      df = merge_informative_pair(df, btc_informative, self.timeframe, btc_info_timeframe, ffill=True)


      drop_columns = {
        "1d": [f"btc_{s}_{btc_info_timeframe}" for s in ["date", "open", "high", "low", "close", "volume"]],
        "4h": [f"btc_{s}_{btc_info_timeframe}" for s in ["date", "open", "high", "low", "close", "volume"]],
        "1h": [f"btc_{s}_{btc_info_timeframe}" for s in ["date", "open", "high", "low", "close", "volume"]],
        "15m": [f"btc_{s}_{btc_info_timeframe}" for s in ["date", "open", "high", "low", "close", "volume"]],
        "5m": [f"btc_{s}_{btc_info_timeframe}" for s in ["date", "open", "high", "low", "close", "volume"]],
      }.get(
        btc_info_timeframe,
        [f"{s}_{btc_info_timeframe}" for s in ["date", "open", "high", "low", "close", "volume"]],
      )
      drop_columns.append(f"date_{btc_info_timeframe}")
      df.drop(columns=df.columns.intersection(drop_columns), inplace=True)

    """
        --> Indicators on informative timeframes
        ___________________________________________________________________________________________
        """
    for info_timeframe in self.info_timeframes:
      info_indicators = self.info_switcher(metadata, info_timeframe)
      df = merge_informative_pair(df, info_indicators, self.timeframe, info_timeframe, ffill=True)


      drop_columns = {
        "1d": [f"{s}_{info_timeframe}" for s in ["date", "open", "high", "low", "close", "volume"]],
        "4h": [f"{s}_{info_timeframe}" for s in ["date", "open", "high", "low", "close", "volume"]],
        "1h": [f"{s}_{info_timeframe}" for s in ["date", "open", "high", "low", "close", "volume"]],
        "15m": [f"{s}_{info_timeframe}" for s in ["date", "high", "low", "volume"]],
      }.get(info_timeframe, [f"{s}_{info_timeframe}" for s in ["date", "open", "high", "low", "close", "volume"]])
      df.drop(columns=df.columns.intersection(drop_columns), inplace=True)

    """
        --> The indicators for the base timeframe  (5m)
        ___________________________________________________________________________________________
        """
    df = self.base_tf_5m_indicators(metadata, df)

    df["zlma_50_1h"] = df["zlma_50_1h"].astype(np.float64).replace(to_replace=[np.nan, None], value=(0.0))
    df["cti_20_1d"] = df["cti_20_1d"].astype(np.float64).replace(to_replace=[np.nan, None], value=(0.0))
    df["r_480_1h"] = df["r_480_1h"].astype(np.float64).replace(to_replace=[np.nan, None], value=(-50.0))
    df["r_480_4h"] = df["r_480_4h"].astype(np.float64).replace(to_replace=[np.nan, None], value=(-50.0))
    df["rsi_14_1d"] = df["rsi_14_1d"].astype(np.float64).replace(to_replace=[np.nan, None], value=(50.0))

    tok = time.perf_counter()
    log.debug(f"[{metadata['pair']}] Populate indicators took a total of: {tok - tik:0.4f} seconds.")

    return df


  def confirm_trade_entry(
    self,
    pair: str,
    order_type: str,
    amount: float,
    rate: float,
    time_in_force: str,
    current_time: datetime,
    entry_tag: Optional[str],
    **kwargs,
  ) -> bool:

    if entry_tag == "force_entry":
      return True

    df, _ = self.dp.get_analyzed_dataframe(pair, self.timeframe)

    if len(df) < 1:
      return False

    df = df.iloc[-1].squeeze()

    entry_tags = entry_tag.split()
    if all(c in self.long_grind_mode_tags for c in entry_tags):
      is_pair_grind_mode = pair.split("/")[0] in self.grind_mode_coins
      if is_pair_grind_mode:
        num_open_grind_mode = 0
        open_trades = Trade.get_trades_proxy(is_open=True)
        for open_trade in open_trades:
          enter_tag = open_trade.enter_tag
          enter_tags = enter_tag.split()
          if all(c in self.long_grind_mode_tags for c in enter_tags):
            num_open_grind_mode += 1
        if num_open_grind_mode >= self.grind_mode_max_slots:

          return False
      else:

        return False

    if rate > df["close"]:
      slippage = (rate / df["close"]) - 1.0

      if slippage < self.max_slippage:
        return True
      else:
        log.warning(f"Cancelling buy for {pair} due to slippage {(slippage * 100.0):.2f}%")
        return False

    return True


  def confirm_trade_exit(
    self,
    pair: str,
    trade: Trade,
    order_type: str,
    amount: float,
    rate: float,
    time_in_force: str,
    exit_reason: str,
    current_time: datetime,
    **kwargs,
  ) -> bool:

    if exit_reason != "force_exit":
      if self._should_hold_trade(trade, rate, exit_reason):
        return False
      if exit_reason == "stop_loss":
        return False
      if self.exit_profit_only:
        if self.exit_profit_only:
          profit = 0.0
          if trade.realized_profit != 0.0:
            profit = ((rate - trade.open_rate) / trade.open_rate) * trade.stake_amount * (1 - trade.fee_close)
            profit = profit + trade.realized_profit
            profit = profit / trade.stake_amount
          else:
            profit = trade.calc_profit_ratio(rate)
          if profit < self.exit_profit_offset:
            return False

    self._remove_profit_target(pair)
    return True


  def bot_loop_start(self, current_time: datetime, **kwargs) -> None:
    if self.config["runmode"].value not in ("live", "dry_run"):
      return super().bot_loop_start(datetime, **kwargs)

    if self.hold_support_enabled:
      self.load_hold_trades_config()

    return super().bot_loop_start(current_time, **kwargs)


  def leverage(
    self,
    pair: str,
    current_time: datetime,
    current_rate: float,
    proposed_leverage: float,
    max_leverage: float,
    entry_tag: Optional[str],
    side: str,
    **kwargs,
  ) -> float:
    enter_tags = entry_tag.split()
    if all(c in self.long_rebuy_mode_tags for c in enter_tags):
      return self.futures_mode_leverage_rebuy_mode
    elif all(c in self.long_grind_mode_tags for c in enter_tags):
      return self.futures_mode_leverage_grind_mode
    return self.futures_mode_leverage


  def _set_profit_target(
    self, pair: str, sell_reason: str, rate: float, current_profit: float, current_time: datetime
  ):
    self.target_profit_cache.data[pair] = {
      "rate": rate,
      "profit": current_profit,
      "sell_reason": sell_reason,
      "time_profit_reached": current_time.isoformat(),
    }
    self.target_profit_cache.save()


  def _remove_profit_target(self, pair: str):
    if self.target_profit_cache is not None:
      self.target_profit_cache.data.pop(pair, None)
      self.target_profit_cache.save()


  def get_hold_trades_config_file(self):
    proper_holds_file_path = self.config["user_data_dir"].resolve() / "nfi-hold-trades.json"
    if proper_holds_file_path.is_file():
      return proper_holds_file_path

    strat_file_path = pathlib.Path(__file__)
    hold_trades_config_file_resolve = strat_file_path.resolve().parent / "hold-trades.json"
    if hold_trades_config_file_resolve.is_file():
      log.warning(
        "Please move %s to %s which is now the expected path for the holds file",
        hold_trades_config_file_resolve,
        proper_holds_file_path,
      )
      return hold_trades_config_file_resolve

    hold_trades_config_file_absolute = strat_file_path.absolute().parent / "hold-trades.json"
    if hold_trades_config_file_absolute.is_file():
      log.warning(
        "Please move %s to %s which is now the expected path for the holds file",
        hold_trades_config_file_absolute,
        proper_holds_file_path,
      )
      return hold_trades_config_file_absolute


  def load_hold_trades_config(self):
    if self.hold_trades_cache is None:
      hold_trades_config_file = self.get_hold_trades_config_file()
      if hold_trades_config_file:
        log.warning("Loading hold support data from %s", hold_trades_config_file)
        self.hold_trades_cache = HoldsCache(hold_trades_config_file)

    if self.hold_trades_cache:
      self.hold_trades_cache.load()


  def _should_hold_trade(self, trade: "Trade", rate: float, sell_reason: str) -> bool:
    if self.config["runmode"].value not in ("live", "dry_run"):
      return False

    if not self.hold_support_enabled:
      return False

    self.load_hold_trades_config()

    if not self.hold_trades_cache:

      return False

    if not self.hold_trades_cache.data:

      return False

    hold_trade = False

    trade_ids: dict = self.hold_trades_cache.data.get("trade_ids")
    if trade_ids and trade.id in trade_ids:
      trade_profit_ratio = trade_ids[trade.id]
      profit = 0.0
      if trade.realized_profit != 0.0:
        profit = ((rate - trade.open_rate) / trade.open_rate) * trade.stake_amount * (1 - trade.fee_close)
        profit = profit + trade.realized_profit
        profit = profit / trade.stake_amount
      else:
        profit = trade.calc_profit_ratio(rate)
      current_profit_ratio = profit
      if sell_reason == "force_sell":
        formatted_profit_ratio = f"{trade_profit_ratio * 100}%"
        formatted_current_profit_ratio = f"{current_profit_ratio * 100}%"
        log.warning(
          "Force selling %s even though the current profit of %s < %s",
          trade,
          formatted_current_profit_ratio,
          formatted_profit_ratio,
        )
        return False
      elif current_profit_ratio >= trade_profit_ratio:

        formatted_profit_ratio = f"{trade_profit_ratio * 100}%"
        formatted_current_profit_ratio = f"{current_profit_ratio * 100}%"
        log.warning(
          "Selling %s because the current profit of %s >= %s",
          trade,
          formatted_current_profit_ratio,
          formatted_profit_ratio,
        )
        return False

      hold_trade = True

    trade_pairs: dict = self.hold_trades_cache.data.get("trade_pairs")
    if trade_pairs and trade.pair in trade_pairs:
      trade_profit_ratio = trade_pairs[trade.pair]
      profit = 0.0
      if trade.realized_profit != 0.0:
        profit = ((rate - trade.open_rate) / trade.open_rate) * trade.stake_amount * (1 - trade.fee_close)
        profit = profit + trade.realized_profit
        profit = profit / trade.stake_amount
      else:
        profit = trade.calc_profit_ratio(rate)
      current_profit_ratio = profit
      if sell_reason == "force_sell":
        formatted_profit_ratio = f"{trade_profit_ratio * 100}%"
        formatted_current_profit_ratio = f"{current_profit_ratio * 100}%"
        log.warning(
          "Force selling %s even though the current profit of %s < %s",
          trade,
          formatted_current_profit_ratio,
          formatted_profit_ratio,
        )
        return False
      elif current_profit_ratio >= trade_profit_ratio:

        formatted_profit_ratio = f"{trade_profit_ratio * 100}%"
        formatted_current_profit_ratio = f"{current_profit_ratio * 100}%"
        log.warning(
          "Selling %s because the current profit of %s >= %s",
          trade,
          formatted_current_profit_ratio,
          formatted_profit_ratio,
        )
        return False

      hold_trade = True

    return hold_trade


  def populate_exit_trend(self, df: DataFrame, metadata: dict) -> DataFrame:
    df.loc[:, "exit_long"] = 0
    df.loc[:, "exit_short"] = 0

    return df






















  def populate_entry_trend(self, df: DataFrame, metadata: dict) -> DataFrame:
    long_entry_conditions = []
    short_entry_conditions = []

    df.loc[:, "enter_tag"] = ""

    is_backtest = self.dp.runmode.value in ["backtest", "hyperopt", "plot"]

    current_free_slots = self.config["max_open_trades"]
    if not is_backtest:
      current_free_slots = self.config["max_open_trades"] - Trade.get_open_trade_count()

    num_open_grind_mode = 0
    num_open_short_mode = 0
    is_pair_grind_mode = metadata["pair"].split("/")[0] in self.grind_mode_coins
    if not is_backtest:
      open_trades = Trade.get_trades_proxy(is_open=True)
      for open_trade in open_trades:
        enter_tag = open_trade.enter_tag
        if enter_tag is not None:
          enter_tags = enter_tag.split()
          if all(c in self.long_grind_mode_tags for c in enter_tags):
            num_open_grind_mode += 1
          elif all(c in self.short_normal_mode_tags for c in enter_tags):
            num_open_short_mode += 1

    is_btc_stake = self.config["stake_currency"] in self.btc_stakes
    allowed_empty_candles = 144 if is_btc_stake else 60











    for enabled_long_entry_signal in self.long_entry_signal_params:
      index = int(enabled_long_entry_signal.split("_")[3])
      item_buy_protection_list = [True]
      if self.long_entry_signal_params[f"{enabled_long_entry_signal}"]:


        long_entry_logic = []
        long_entry_logic.append(reduce(lambda x, y: x & y, item_buy_protection_list))

        if index == 1:

          long_entry_logic.append(df["ema_26"] > df["ema_12"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * 0.02))
          long_entry_logic.append((df["ema_26"].shift() - df["ema_12"].shift()) > (df["open"] / 100))
          long_entry_logic.append(df["close"] < (df["bb20_2_low"] * 0.999))

        if index == 2:

          long_entry_logic.append(df["bb40_2_delta"].gt(df["close"] * 0.06))
          long_entry_logic.append(df["close_delta"].gt(df["close"] * 0.02))
          long_entry_logic.append(df["bb40_2_tail"].lt(df["bb40_2_delta"] * 0.2))
          long_entry_logic.append(df["close"].lt(df["bb40_2_low"].shift()))
          long_entry_logic.append(df["close"].le(df["close"].shift()))

        if index == 3:

          long_entry_logic.append(df["rsi_14"] < 36.0)
          long_entry_logic.append(df["ha_close"] > df["ha_open"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * 0.020))

        if index == 4:

          long_entry_logic.append(df["ema_26"] > df["ema_12"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * 0.018))
          long_entry_logic.append((df["ema_26"].shift() - df["ema_12"].shift()) > (df["open"] / 100))
          long_entry_logic.append(df["close"] < (df["bb20_2_low"] * 0.996))

        if index == 5:

          long_entry_logic.append(df["ema_26"] > df["ema_12"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * 0.03))
          long_entry_logic.append((df["ema_26"].shift() - df["ema_12"].shift()) > (df["open"] / 100))
          long_entry_logic.append(df["rsi_14"] < 36.0)

        if index == 6:

          long_entry_logic.append(df["close"] < (df["ema_26"] * 0.94))
          long_entry_logic.append(df["close"] < (df["bb20_2_low"] * 0.996))

        if index == 7:

          long_entry_logic.append(df["close"] < (df["ema_16"] * 0.974))
          long_entry_logic.append(df["ewo_50_200"] > 2.0)
          long_entry_logic.append(df["rsi_14"] < 30.0)

        if index == 8:

          long_entry_logic.append(df["close"] < (df["ema_16"] * 0.944))
          long_entry_logic.append(df["ewo_50_200"] < -4.0)
          long_entry_logic.append(df["rsi_14"] < 30.0)

        if index == 9:

          long_entry_logic.append(df["ema_26_15m"] > df["ema_12_15m"])
          long_entry_logic.append((df["ema_26_15m"] - df["ema_12_15m"]) > (df["open_15m"] * 0.020))
          long_entry_logic.append((df["ema_26_15m"].shift(3) - df["ema_12_15m"].shift(3)) > (df["open_15m"] / 100.0))
          long_entry_logic.append(df["close_15m"] < (df["bb20_2_low_15m"] * 0.99))

        if index == 10:

          long_entry_logic.append(df["rsi_14"] < self.entry_10_rsi_14_max.value)
          long_entry_logic.append(df["ema_26"] > df["ema_12"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * self.entry_10_ema_open_offset.value))
          long_entry_logic.append((df["ema_26"].shift() - df["ema_12"].shift()) > (df["open"] / 100.0))
          long_entry_logic.append(df["close"] < (df["ema_20"] * self.entry_10_ema_offset.value))

        if index == 11:

          long_entry_logic.append(df["rsi_14"] < self.entry_11_rsi_14_max.value)
          long_entry_logic.append(df["cti_20"] < self.entry_11_cti_20_max.value)
          long_entry_logic.append(df["ema_26"] > df["ema_12"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * self.entry_11_ema_open_offset.value))
          long_entry_logic.append((df["ema_26"].shift() - df["ema_12"].shift()) > (df["open"] / 100.0))
          long_entry_logic.append(df["close"] < (df["sma_30"] * self.entry_11_sma_offset.value))

        if index == 12:




          long_entry_logic.append(df["r_14"] < self.entry_12_r_14_max.value)
          long_entry_logic.append(df["close"] < (df["bb20_2_low"] * self.entry_12_bb_offset.value))
          long_entry_logic.append(df["close"] < (df["sma_30"] * self.entry_12_sma_offset.value))

        if index == 13:

          long_entry_logic.append(df["ema_26"] > df["ema_12"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * 0.0150))
          long_entry_logic.append((df["ema_26"].shift() - df["ema_12"].shift()) > (df["open"] / 100))
          long_entry_logic.append(df["close"] < (df["bb20_2_low"] * 0.994))

        if index == 21:

          long_entry_logic.append(df["ema_26"] > df["ema_12"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * 0.016))
          long_entry_logic.append((df["ema_26"].shift() - df["ema_12"].shift()) > (df["open"] / 100))
          long_entry_logic.append(df["rsi_14"] < 36.0)

        if index == 22:

          long_entry_logic.append(df["close"] < (df["ema_16"] * 0.968))
          long_entry_logic.append(df["cti_20"] < -0.9)
          long_entry_logic.append(df["rsi_14"] < 50.0)

        if index == 23:

          long_entry_logic.append(df["ewo_50_200_15m"] > 4.2)
          long_entry_logic.append(df["rsi_14_15m"].shift(1) < 30.0)
          long_entry_logic.append(df["rsi_14_15m"] < 30.0)
          long_entry_logic.append(df["rsi_14"] < 35.0)
          long_entry_logic.append(df["cti_20"] < -0.8)
          long_entry_logic.append(df["close"] < (df["ema_26_15m"] * 0.958))

        if index == 24:

          long_entry_logic.append(df["rsi_14"] > self.entry_24_rsi_14_min.value)
          long_entry_logic.append(df["rsi_14"] < self.entry_24_rsi_14_max.value)
          long_entry_logic.append(df["cti_20"] < self.entry_24_cti_20_max.value)
          long_entry_logic.append(df["r_14"] < self.entry_24_r_14_max.value)
          long_entry_logic.append(df["ewo_50_200"] > self.entry_24_ewo_50_200_min.value)
          long_entry_logic.append(df["ewo_50_200"] < self.entry_24_ewo_50_200_max.value)
          long_entry_logic.append(df["close"] < (df["sma_75"] * self.entry_24_sma_offset.value))

        if index == 25:

          long_entry_logic.append(df["rsi_14"] < self.entry_25_rsi_14_max.value)
          long_entry_logic.append(df["cti_20"] < self.entry_25_cti_20_max.value)
          long_entry_logic.append(df["ewo_50_200"] > self.entry_25_ewo_50_200_min.value)
          long_entry_logic.append(df["close"] < (df["sma_30"] * self.entry_25_sma_offset.value))

        if index == 26:

          long_entry_logic.append(df["close"] < (df["bb20_2_low"] * self.entry_26_bb_offset.value))
          long_entry_logic.append(df["ewo_50_200_1h"] > self.entry_26_ewo_50_200_1h_min.value)
          long_entry_logic.append(df["ema_26"] > df["ema_12"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * self.entry_26_ema_open_offset.value))
          long_entry_logic.append((df["ema_26"].shift() - df["ema_12"].shift()) > (df["open"] / 100.0))

        if index == 41:

          long_entry_logic.append(df["bb40_2_delta"].gt(df["close"] * 0.036))
          long_entry_logic.append(df["close_delta"].gt(df["close"] * 0.02))
          long_entry_logic.append(df["bb40_2_tail"].lt(df["bb40_2_delta"] * 0.4))
          long_entry_logic.append(df["close"].lt(df["bb40_2_low"].shift()))
          long_entry_logic.append(df["close"].le(df["close"].shift()))
          long_entry_logic.append(df["rsi_14"] < 36.0)

        if index == 42:

          long_entry_logic.append(df["ema_26"] > df["ema_12"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * 0.018))
          long_entry_logic.append((df["ema_26"].shift() - df["ema_12"].shift()) > (df["open"] / 100))
          long_entry_logic.append(df["close"] < (df["bb20_2_low"] * 0.996))
          long_entry_logic.append(df["rsi_14"] < 40.0)

        if index == 43:

          long_entry_logic.append(df["close"] < (df["ema_26"] * 0.938))
          long_entry_logic.append(df["cti_20"] < -0.75)
          long_entry_logic.append(df["r_14"] < -94.0)

        if index == 44:

          long_entry_logic.append(df["btc_pct_close_max_24_5m"] < 0.03)
          long_entry_logic.append(df["btc_pct_close_max_72_5m"] < 0.03)
          long_entry_logic.append(df["close_max_12"] < (df["close"] * 1.2))
          long_entry_logic.append(df["close_max_24"] < (df["close"] * 1.24))
          long_entry_logic.append(df["close_max_48"] < (df["close"] * 1.3))
          long_entry_logic.append(df["high_max_24_1h"] < (df["close"] * 1.5))
          long_entry_logic.append(df["high_max_24_4h"] < (df["close"] * 1.75))
          long_entry_logic.append(df["hl_pct_change_6_1h"] < 0.4)
          long_entry_logic.append(df["hl_pct_change_12_1h"] < 0.5)
          long_entry_logic.append(df["hl_pct_change_24_1h"] < 0.75)
          long_entry_logic.append(df["hl_pct_change_48_1h"] < 0.9)
          long_entry_logic.append(df["num_empty_288"] < allowed_empty_candles)

          long_entry_logic.append(df["cti_20_1h"] < 0.8)
          long_entry_logic.append(df["rsi_14_1h"] < 80.0)
          long_entry_logic.append(df["cti_20_4h"] < 0.8)
          long_entry_logic.append(df["rsi_14_4h"] < 80.0)

          long_entry_logic.append(df["close_max_48"] > (df["close"] * 1.1))

          long_entry_logic.append(df["bb20_2_width_1h"] > 0.132)
          long_entry_logic.append(df["cti_20"] < -0.8)
          long_entry_logic.append(df["r_14"] < -90.0)

        if index == 45:

          long_entry_logic.append(df["rsi_14"] > self.entry_45_rsi_14_min.value)
          long_entry_logic.append(df["rsi_14"] < self.entry_45_rsi_14_max.value)
          long_entry_logic.append(df["rsi_20"] < df["rsi_20"].shift(1))
          long_entry_logic.append(df["cti_20"] < self.entry_45_cti_20_max.value)
          long_entry_logic.append(df["close"] < (df["sma_16"] * self.entry_45_sma_offset.value))

        if index == 46:

          long_entry_logic.append(df["rsi_14"] < self.entry_46_rsi_14_max.value)
          long_entry_logic.append(df["chandelier_dir_1h"].shift(1) < -0)
          long_entry_logic.append(df["chandelier_dir_1h"] > 0)
          long_entry_logic.append(df["close"] > df["zlma_50_1h"])
          long_entry_logic.append(df["ema_12"] < df["ema_26"])

        if index == 47:

          long_entry_logic.append(df["rsi_14"] > self.entry_47_rsi_14_min.value)
          long_entry_logic.append(df["rsi_14"] < self.entry_47_rsi_14_max.value)
          long_entry_logic.append(df["rsi_20"] > self.entry_47_rsi_20_min.value)
          long_entry_logic.append(df["rsi_20"] < self.entry_47_rsi_20_max.value)
          long_entry_logic.append(df["cti_20"] < self.entry_47_cti_20_max.value)
          long_entry_logic.append(df["chandelier_dir"].shift(1) < -0)
          long_entry_logic.append(df["chandelier_dir"] > 0)
          long_entry_logic.append(df["ema_12"] < (df["ema_26"] * self.entry_47_ema_offset.value))
          long_entry_logic.append(df["close"] < (df["high_max_12_1h"] * self.entry_47_high_max_12_1h_max.value))

        if index == 48:

          long_entry_logic.append(df["rsi_14"] < self.entry_48_rsi_14_max.value)
          long_entry_logic.append(df["cci_20"] < self.entry_48_cci_20_max.value)
          long_entry_logic.append(df["close"] < (df["sma_30"] * self.entry_48_sma_offset.value))
          long_entry_logic.append(((df["close"] - df["open_min_6"]) / df["open_min_6"]) > self.entry_48_inc_min.value)

        if index == 49:

          long_entry_logic.append(df["rsi_14"] < self.entry_49_rsi_14_max.value)



          long_entry_logic.append(
            ((df["close"] - df["open_min_12"]) / df["open_min_12"]) > self.entry_49_inc_min.value
          )

        if index == 50:

          long_entry_logic.append(df["close"] < (df["bb20_2_low"] * self.entry_50_bb_offset.value))
          long_entry_logic.append(df["ema_26"] > df["ema_12"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * self.entry_50_ema_open_offset.value))
          long_entry_logic.append((df["ema_26"].shift() - df["ema_12"].shift()) > (df["open"] / 100.0))

        if index == 51:

          long_entry_logic.append(df["rsi_14"] < 32.0)
          long_entry_logic.append(df["close"] < (df["ema_26"] * 0.972))

        if index == 61:

          long_entry_logic.append(df["rsi_14"] < 40.0)
          long_entry_logic.append(df["bb40_2_delta"].gt(df["close"] * 0.03))
          long_entry_logic.append(df["close_delta"].gt(df["close"] * 0.018))
          long_entry_logic.append(df["bb40_2_tail"].lt(df["bb40_2_delta"] * 0.4))
          long_entry_logic.append(df["close"].lt(df["bb40_2_low"].shift()))
          long_entry_logic.append(df["close"].le(df["close"].shift()))

        if index == 62:

          long_entry_logic.append(df["rsi_14"] < 36.0)
          long_entry_logic.append(df["close"] < (df["ema_26"] * 0.992))

        if index == 81:

          long_entry_logic.append(df["rsi_14"] < 32.0)
          long_entry_logic.append(df["bb40_2_delta"].gt(df["close"] * 0.036))
          long_entry_logic.append(df["close_delta"].gt(df["close"] * 0.012))
          long_entry_logic.append(df["bb40_2_tail"].lt(df["bb40_2_delta"] * 0.4))
          long_entry_logic.append(df["close"].lt(df["bb40_2_low"].shift()))
          long_entry_logic.append(df["close"].le(df["close"].shift()))

        if index == 82:

          long_entry_logic.append(df["ema_26"] > df["ema_12"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * 0.024))
          long_entry_logic.append((df["ema_26"].shift() - df["ema_12"].shift()) > (df["open"] / 100))

        if index == 101:

          long_entry_logic.append(df["rsi_14"] < 36.0)
          long_entry_logic.append(df["rsi_14"] < df["rsi_14"].shift(1))
          long_entry_logic.append(df["close"] < (df["sma_16"] * 0.956))
          long_entry_logic.append(df["cti_20_15m"] < -0.5)

        if index == 102:

          long_entry_logic.append(df["rsi_14"] < self.entry_102_rsi_14_max.value)
          long_entry_logic.append(df["close"] < (df["ema_16"] * self.entry_102_ema_offset.value))
          long_entry_logic.append(df["close"] < (df["bb20_2_low"] * self.entry_102_bb_offset.value))

        if index == 103:

          long_entry_logic.append(df["rsi_14"] > self.entry_103_rsi_14_min.value)
          long_entry_logic.append(df["close"] < (df["sma_16"] * self.entry_103_sma_offset.value))
          long_entry_logic.append(df["close"] < (df["bb20_2_mid"] * self.entry_103_bb_offset.value))

        if index == 104:

          long_entry_logic.append(df["rsi_14"] > self.entry_104_rsi_14_min.value)
          long_entry_logic.append(df["rsi_14"] < self.entry_104_rsi_14_max.value)
          long_entry_logic.append(df["close"] < (df["sma_16"] * self.entry_104_sma_offset.value))

        if index == 105:

          long_entry_logic.append(df["rsi_3"] < 60.0)
          long_entry_logic.append(df["rsi_14"] < 46.0)
          long_entry_logic.append(df["ema_26"] > df["ema_12"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * 0.018))
          long_entry_logic.append((df["ema_26"].shift() - df["ema_12"].shift()) > (df["open"] / 100))

        if index == 106:

          long_entry_logic.append(df["cti_20"] < self.entry_106_cti_20_max.value)
          long_entry_logic.append(df["ewo_50_200"] < self.entry_106_ewo_50_200_max.value)
          long_entry_logic.append(df["close"] < (df["sma_30"] * self.entry_106_sma_offset.value))

        if index == 107:

          long_entry_logic.append(df["bb40_2_low"].shift().gt(0.0))
          long_entry_logic.append(df["bb40_2_delta"].gt(df["close"] * self.entry_107_bb40_bbdelta_close.value))
          long_entry_logic.append(df["close_delta"].gt(df["close"] * self.entry_107_bb40_closedelta_close.value))
          long_entry_logic.append(df["bb40_2_tail"].lt(df["bb40_2_delta"] * self.entry_107_bb40_tail_bbdelta.value))
          long_entry_logic.append(df["close"].lt(df["bb40_2_low"].shift()))
          long_entry_logic.append(df["close"].le(df["close"].shift()))
          long_entry_logic.append(df["cti_20"] < self.entry_107_cti_20_max.value)
          long_entry_logic.append(df["r_480"] > self.entry_107_r_480_min.value)

        if index == 108:

          long_entry_logic.append(df["rsi_14"] > self.entry_108_rsi_14_min.value)
          long_entry_logic.append(df["cti_20"] < self.entry_108_cti_20_max.value)
          long_entry_logic.append(df["r_14"] < self.entry_108_r_14_max.value)
          long_entry_logic.append(df["r_14"].shift(1) < self.entry_108_r_14_max.value)
          long_entry_logic.append(df["close"] < (df["bb20_2_low"] * self.entry_108_bb_offset.value))
          long_entry_logic.append(df["ema_26"] > df["ema_12"])
          long_entry_logic.append((df["ema_26"] - df["ema_12"]) > (df["open"] * self.entry_108_ema_open_offset.value))
          long_entry_logic.append((df["ema_26"].shift() - df["ema_12"].shift()) > (df["open"] / 100.0))

        if index == 109:

          long_entry_logic.append(df["cti_20"] < -0.75)
          long_entry_logic.append(df["r_14"] < -90.0)
          long_entry_logic.append(df["close"] < (df["bb20_2_low"] * 0.999))
          long_entry_logic.append(df["close"] < (df["ema_20"] * 0.960))

        if index == 110:

          long_entry_logic.append(df["cti_20"] < self.entry_110_cti_20_max.value)
          long_entry_logic.append(df["ewo_50_200"] < self.entry_110_ewo_50_200_max.value)
          long_entry_logic.append(df["close"] < (df["ema_20"] * self.entry_110_ema_offset.value))

        if index == 120:

          long_entry_logic.append(num_open_grind_mode < self.grind_mode_max_slots)

          long_entry_logic.append(df["btc_pct_close_max_24_5m"] < 0.03)
          long_entry_logic.append(df["btc_pct_close_max_72_5m"] < 0.06)
          long_entry_logic.append(df["close"] > (df["close_max_12"] * 0.94))
          long_entry_logic.append(df["close"] > (df["close_max_24"] * 0.92))
          long_entry_logic.append(df["close"] > (df["close_max_48"] * 0.90))
          long_entry_logic.append(df["close"] > (df["high_max_12_1h"] * 0.88))
          long_entry_logic.append(df["close"] > (df["high_max_24_1h"] * 0.86))
          long_entry_logic.append(df["close"] > (df["high_max_6_1d"] * 0.84))
          long_entry_logic.append(df["close"] > (df["high_max_12_1d"] * 0.80))
          long_entry_logic.append(df["hl_pct_change_6_1h"] < 0.30)
          long_entry_logic.append(df["hl_pct_change_12_1h"] < 0.40)
          long_entry_logic.append(df["hl_pct_change_24_1h"] < 0.50)
          long_entry_logic.append(df["hl_pct_change_48_1h"] < 0.60)
          long_entry_logic.append(df["num_empty_288"] < allowed_empty_candles)

          long_entry_logic.append(df["rsi_3"] >= 30.0)
          long_entry_logic.append(df["rsi_3"] <= 60.0)
          long_entry_logic.append(df["rsi_3_15m"] >= 30.0)
          long_entry_logic.append(df["rsi_3_1h"] >= 30.0)
          long_entry_logic.append(df["rsi_3_4h"] >= 30.0)
          long_entry_logic.append(df["rsi_3_1d"] >= 30.0)
          long_entry_logic.append(df["cti_20_1h"] <= 0.70)
          long_entry_logic.append(df["rsi_14_1h"] <= 60.0)
          long_entry_logic.append(df["cti_20_4h"] <= 0.70)
          long_entry_logic.append(df["rsi_14_4h"] <= 60.0)



          long_entry_logic.append(df["ema_200_dec_4_1d"] == False)

          long_entry_logic.append(df["rsi_14"] < 42.0)
          long_entry_logic.append(df["ema_12"] < df["ema_26"])
          long_entry_logic.append(df["ha_close"] > df["ha_open"])
          
        if index == 80:

          long_entry_logic.append(df["high"] > df["high"].shift())
          long_entry_logic.append(df["total_bullish_divergences"].shift() > 0)
          long_entry_logic.append(two_bands_check(df))


        long_entry_logic.append(df["volume"] > 0)
        item_long_entry = reduce(lambda x, y: x & y, long_entry_logic)
        df.loc[item_long_entry, "enter_tag"] += f"{index} "
        long_entry_conditions.append(item_long_entry)
        df.loc[:, "enter_long"] = item_long_entry

    if long_entry_conditions:
      df.loc[:, "enter_long"] = reduce(lambda x, y: x | y, long_entry_conditions)











    for enabled_short_entry_signal in self.short_entry_signal_params:
      short_index = int(enabled_short_entry_signal.split("_")[3])
      item_short_buy_protection_list = [True]
      if self.short_entry_signal_params[f"{enabled_short_entry_signal}"]:


        short_entry_logic = []
        short_entry_logic.append(reduce(lambda x, y: x & y, item_short_buy_protection_list))

        if short_index == 500:

          short_entry_logic.append(df["num_empty_288"] < allowed_empty_candles)

          short_entry_logic.append(df["rsi_3_15m"] <= 96.0)
          short_entry_logic.append(df["rsi_3_1h"] <= 96.0)

          short_entry_logic.append(df["ema_12"] > df["ema_26"])
          short_entry_logic.append((df["ema_12"] - df["ema_26"]) > (df["open"] * 0.028))
          short_entry_logic.append((df["ema_12"] - df["ema_26"]) > (df["open"] / 100.0))
          short_entry_logic.append(df["close"] > (df["bb20_2_upp"] * 1.012))
          short_entry_logic.append(df["rsi_14"] > 64.0)

        if short_index == 501:






          short_entry_logic.append(df["buy_short2"] > 0)
          
        if short_index == 580:


          short_entry_logic.append(df["low"] < df["low"].shift())
          short_entry_logic.append(df["total_bearish_divergences"].shift() > 0)
          short_entry_logic.append(two_bands_check(df))


        short_entry_logic.append(df["volume"] > 0)
        item_short_entry = reduce(lambda x, y: x & y, short_entry_logic)
        df.loc[item_short_entry, "enter_tag"] += f"{short_index} "
        short_entry_conditions.append(item_short_entry)
        df.loc[:, "enter_short"] = item_short_entry

    if short_entry_conditions:
      df.loc[:, "enter_short"] = reduce(lambda x, y: x | y, short_entry_conditions)

    return df




























  def long_exit_normal(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    sell = False

    sell, signal_name = self.long_exit_signals(
      self.long_normal_mode_name,
      profit_init_ratio,
      max_profit,
      max_loss,
      last_candle,
      previous_candle_1,
      previous_candle_2,
      previous_candle_3,
      previous_candle_4,
      previous_candle_5,
      trade,
      current_time,
      enter_tags,
    )

    if not sell:
      sell, signal_name = self.long_exit_main(
        self.long_normal_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_williams_r(
        self.long_normal_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_dec(
        self.long_normal_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_stoploss(
        self.long_normal_mode_name,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )


    if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
      previous_rate = self.target_profit_cache.data[pair]["rate"]
      previous_profit = self.target_profit_cache.data[pair]["profit"]
      previous_sell_reason = self.target_profit_cache.data[pair]["sell_reason"]
      previous_time_profit_reached = datetime.fromisoformat(self.target_profit_cache.data[pair]["time_profit_reached"])

      sell_max, signal_name_max = self.exit_profit_target(
        self.long_normal_mode_name,
        pair,
        trade,
        current_time,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        last_candle,
        previous_candle_1,
        previous_rate,
        previous_profit,
        previous_sell_reason,
        previous_time_profit_reached,
        enter_tags,
      )
      if sell_max and signal_name_max is not None:
        return True, f"{signal_name_max}_m"
      if previous_sell_reason in [f"exit_{self.long_normal_mode_name}_stoploss_u_e"]:
        if profit_ratio > (previous_profit + 0.005):
          mark_pair, mark_signal = self.mark_profit_target(
            self.long_normal_mode_name,
            pair,
            True,
            previous_sell_reason,
            trade,
            current_time,
            current_rate,
            profit_ratio,
            last_candle,
            previous_candle_1,
          )
          if mark_pair:
            self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
      elif (profit_init_ratio > (previous_profit + 0.001)) and (
        previous_sell_reason not in [f"exit_{self.long_normal_mode_name}_stoploss_doom"]
      ):

        mark_pair, mark_signal = self.mark_profit_target(
          self.long_normal_mode_name,
          pair,
          True,
          previous_sell_reason,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if sell and signal_name is not None:
      previous_profit = None
      if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
        previous_profit = self.target_profit_cache.data[pair]["profit"]
      if signal_name in [
        f"exit_{self.long_normal_mode_name}_stoploss_doom",
        f"exit_{self.long_normal_mode_name}_stoploss_u_e",
      ]:
        mark_pair, mark_signal = self.mark_profit_target(
          self.long_normal_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
        else:

          return True, f"{signal_name}"
      elif (previous_profit is None) or (previous_profit < profit_init_ratio):
        mark_pair, mark_signal = self.mark_profit_target(
          self.long_normal_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)
        else:

          return True, f"{signal_name}"
    else:
      if profit_init_ratio >= self.profit_max_thresholds[0]:
        previous_profit = None
        if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
          previous_profit = self.target_profit_cache.data[pair]["profit"]
        if (previous_profit is None) or (previous_profit < profit_init_ratio):
          mark_signal = f"exit_profit_{self.long_normal_mode_name}_max"
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if signal_name not in [
      f"exit_profit_{self.long_normal_mode_name}_max",
      f"exit_{self.long_normal_mode_name}_stoploss_doom",
      f"exit_{self.long_normal_mode_name}_stoploss_u_e",
    ]:
      if sell and (signal_name is not None):
        return True, f"{signal_name}"

    return False, None


  def long_exit_pump(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    sell = False

    sell, signal_name = self.long_exit_signals(
      self.long_pump_mode_name,
      profit_init_ratio,
      max_profit,
      max_loss,
      last_candle,
      previous_candle_1,
      previous_candle_2,
      previous_candle_3,
      previous_candle_4,
      previous_candle_5,
      trade,
      current_time,
      enter_tags,
    )

    if not sell:
      sell, signal_name = self.long_exit_main(
        self.long_pump_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_williams_r(
        self.long_pump_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_dec(
        self.long_pump_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_stoploss(
        self.long_pump_mode_name,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )


    if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
      previous_rate = self.target_profit_cache.data[pair]["rate"]
      previous_profit = self.target_profit_cache.data[pair]["profit"]
      previous_sell_reason = self.target_profit_cache.data[pair]["sell_reason"]
      previous_time_profit_reached = datetime.fromisoformat(self.target_profit_cache.data[pair]["time_profit_reached"])

      sell_max, signal_name_max = self.exit_profit_target(
        self.long_pump_mode_name,
        pair,
        trade,
        current_time,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        last_candle,
        previous_candle_1,
        previous_rate,
        previous_profit,
        previous_sell_reason,
        previous_time_profit_reached,
        enter_tags,
      )
      if sell_max and signal_name_max is not None:
        return True, f"{signal_name_max}_m"
      if previous_sell_reason in [f"exit_{self.long_pump_mode_name}_stoploss_u_e"]:
        if profit_ratio > (previous_profit + 0.005):
          mark_pair, mark_signal = self.mark_profit_target(
            self.long_pump_mode_name,
            pair,
            True,
            previous_sell_reason,
            trade,
            current_time,
            current_rate,
            profit_ratio,
            last_candle,
            previous_candle_1,
          )
          if mark_pair:
            self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
      elif (profit_init_ratio > (previous_profit + 0.001)) and (
        previous_sell_reason not in [f"exit_{self.long_pump_mode_name}_stoploss_doom"]
      ):

        mark_pair, mark_signal = self.mark_profit_target(
          self.long_pump_mode_name,
          pair,
          True,
          previous_sell_reason,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if sell and signal_name is not None:
      previous_profit = None
      if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
        previous_profit = self.target_profit_cache.data[pair]["profit"]
      if signal_name in [
        f"exit_{self.long_pump_mode_name}_stoploss_doom",
        f"exit_{self.long_pump_mode_name}_stoploss_u_e",
      ]:
        mark_pair, mark_signal = self.mark_profit_target(
          self.long_pump_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
        else:

          return True, f"{signal_name}"
      elif (previous_profit is None) or (previous_profit < profit_init_ratio):
        mark_pair, mark_signal = self.mark_profit_target(
          self.long_pump_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)
        else:

          return True, f"{signal_name}"
    else:
      if profit_init_ratio >= self.profit_max_thresholds[2]:
        previous_profit = None
        if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
          previous_profit = self.target_profit_cache.data[pair]["profit"]
        if (previous_profit is None) or (previous_profit < profit_init_ratio):
          mark_signal = f"exit_profit_{self.long_pump_mode_name}_max"
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if signal_name not in [
      f"exit_profit_{self.long_pump_mode_name}_max",
      f"exit_{self.long_pump_mode_name}_stoploss_doom",
      f"exit_{self.long_pump_mode_name}_stoploss_u_e",
    ]:
      if sell and (signal_name is not None):
        return True, f"{signal_name}"

    return False, None


  def long_exit_quick(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    sell = False

    sell, signal_name = self.long_exit_signals(
      self.long_quick_mode_name,
      profit_init_ratio,
      max_profit,
      max_loss,
      last_candle,
      previous_candle_1,
      previous_candle_2,
      previous_candle_3,
      previous_candle_4,
      previous_candle_5,
      trade,
      current_time,
      enter_tags,
    )

    if not sell:
      sell, signal_name = self.long_exit_main(
        self.long_quick_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_williams_r(
        self.long_quick_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_dec(
        self.long_quick_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_stoploss(
        self.long_quick_mode_name,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      if (0.09 >= profit_init_ratio > 0.02) and (last_candle["rsi_14"] > 78.0):
        sell, signal_name = True, f"exit_{self.long_quick_mode_name}_q_1"

      if (0.09 >= profit_init_ratio > 0.02) and (last_candle["cti_20"] > 0.95):
        sell, signal_name = True, f"exit_{self.long_quick_mode_name}_q_2"

      if (0.09 >= profit_init_ratio > 0.02) and (last_candle["r_14"] >= -0.1):
        sell, signal_name = True, f"exit_{self.long_quick_mode_name}_q_3"


    if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
      previous_rate = self.target_profit_cache.data[pair]["rate"]
      previous_profit = self.target_profit_cache.data[pair]["profit"]
      previous_sell_reason = self.target_profit_cache.data[pair]["sell_reason"]
      previous_time_profit_reached = datetime.fromisoformat(self.target_profit_cache.data[pair]["time_profit_reached"])

      sell_max, signal_name_max = self.exit_profit_target(
        self.long_quick_mode_name,
        pair,
        trade,
        current_time,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        last_candle,
        previous_candle_1,
        previous_rate,
        previous_profit,
        previous_sell_reason,
        previous_time_profit_reached,
        enter_tags,
      )
      if sell_max and signal_name_max is not None:
        return True, f"{signal_name_max}_m"
      if previous_sell_reason in [f"exit_{self.long_quick_mode_name}_stoploss_u_e"]:
        if profit_ratio > (previous_profit + 0.005):
          mark_pair, mark_signal = self.mark_profit_target(
            self.long_quick_mode_name,
            pair,
            True,
            previous_sell_reason,
            trade,
            current_time,
            current_rate,
            profit_ratio,
            last_candle,
            previous_candle_1,
          )
          if mark_pair:
            self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
      elif (profit_init_ratio > (previous_profit + 0.001)) and (
        previous_sell_reason not in [f"exit_{self.long_quick_mode_name}_stoploss_doom"]
      ):

        mark_pair, mark_signal = self.mark_profit_target(
          self.long_quick_mode_name,
          pair,
          True,
          previous_sell_reason,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if sell and signal_name is not None:
      previous_profit = None
      if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
        previous_profit = self.target_profit_cache.data[pair]["profit"]
      if signal_name in [
        f"exit_{self.long_quick_mode_name}_stoploss_doom",
        f"exit_{self.long_quick_mode_name}_stoploss_u_e",
      ]:
        mark_pair, mark_signal = self.mark_profit_target(
          self.long_quick_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
        else:

          return True, f"{signal_name}"
      elif (previous_profit is None) or (previous_profit < profit_init_ratio):
        mark_pair, mark_signal = self.mark_profit_target(
          self.long_quick_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)
        else:

          return True, f"{signal_name}"
    else:
      if profit_init_ratio >= self.profit_max_thresholds[4]:
        previous_profit = None
        if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
          previous_profit = self.target_profit_cache.data[pair]["profit"]
        if (previous_profit is None) or (previous_profit < profit_init_ratio):
          mark_signal = f"exit_profit_{self.long_quick_mode_name}_max"
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if signal_name not in [
      f"exit_profit_{self.long_quick_mode_name}_max",
      f"exit_{self.long_quick_mode_name}_stoploss_doom",
      f"exit_{self.long_quick_mode_name}_stoploss_u_e",
    ]:
      if sell and (signal_name is not None):
        return True, f"{signal_name}"

    return False, None


  def long_exit_rebuy(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    sell = False

    sell, signal_name = self.long_exit_signals(
      self.long_rebuy_mode_name,
      profit_init_ratio,
      max_profit,
      max_loss,
      last_candle,
      previous_candle_1,
      previous_candle_2,
      previous_candle_3,
      previous_candle_4,
      previous_candle_5,
      trade,
      current_time,
      enter_tags,
    )

    if not sell:
      sell, signal_name = self.long_exit_main(
        self.long_rebuy_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_williams_r(
        self.long_rebuy_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_dec(
        self.long_rebuy_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      if profit_stake < -(
        filled_entries[0].cost
        * (self.stop_threshold_futures_rebuy if self.is_futures_mode else self.stop_threshold_spot_rebuy)
        / (trade.leverage if self.is_futures_mode else 1.0)
      ):
        sell, signal_name = True, f"exit_{self.long_rebuy_mode_name}_stoploss_doom"


    if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
      previous_rate = self.target_profit_cache.data[pair]["rate"]
      previous_profit = self.target_profit_cache.data[pair]["profit"]
      previous_sell_reason = self.target_profit_cache.data[pair]["sell_reason"]
      previous_time_profit_reached = datetime.fromisoformat(self.target_profit_cache.data[pair]["time_profit_reached"])

      sell_max, signal_name_max = self.exit_profit_target(
        self.long_rebuy_mode_name,
        pair,
        trade,
        current_time,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        last_candle,
        previous_candle_1,
        previous_rate,
        previous_profit,
        previous_sell_reason,
        previous_time_profit_reached,
        enter_tags,
      )
      if sell_max and signal_name_max is not None:
        return True, f"{signal_name_max}_m"
      if previous_sell_reason in [f"exit_{self.long_rebuy_mode_name}_stoploss_u_e"]:
        if profit_ratio > (previous_profit + 0.005):
          mark_pair, mark_signal = self.mark_profit_target(
            self.long_rebuy_mode_name,
            pair,
            True,
            previous_sell_reason,
            trade,
            current_time,
            current_rate,
            profit_ratio,
            last_candle,
            previous_candle_1,
          )
          if mark_pair:
            self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
      elif (profit_init_ratio > (previous_profit + 0.001)) and (
        previous_sell_reason not in [f"exit_{self.long_rebuy_mode_name}_stoploss_doom"]
      ):

        mark_pair, mark_signal = self.mark_profit_target(
          self.long_rebuy_mode_name,
          pair,
          True,
          previous_sell_reason,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if sell and signal_name is not None:
      previous_profit = None
      if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
        previous_profit = self.target_profit_cache.data[pair]["profit"]
      if signal_name in [
        f"exit_{self.long_rebuy_mode_name}_stoploss_doom",
        f"exit_{self.long_rebuy_mode_name}_stoploss_u_e",
      ]:
        mark_pair, mark_signal = self.mark_profit_target(
          self.long_rebuy_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
        else:

          return True, f"{signal_name}"
      elif (previous_profit is None) or (previous_profit < profit_init_ratio):
        mark_pair, mark_signal = self.mark_profit_target(
          self.long_rebuy_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)
        else:

          return True, f"{signal_name}"
    else:
      if profit_init_ratio >= self.profit_max_thresholds[6]:
        previous_profit = None
        if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
          previous_profit = self.target_profit_cache.data[pair]["profit"]
        if (previous_profit is None) or (previous_profit < profit_init_ratio):
          mark_signal = f"exit_profit_{self.long_rebuy_mode_name}_max"
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if signal_name not in [f"exit_profit_{self.long_rebuy_mode_name}_max"]:
      if sell and (signal_name is not None):
        return True, f"{signal_name}"

    return False, None


  def long_exit_high_profit(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    sell = False

    sell, signal_name = self.long_exit_signals(
      self.long_high_profit_mode_name,
      profit_init_ratio,
      max_profit,
      max_loss,
      last_candle,
      previous_candle_1,
      previous_candle_2,
      previous_candle_3,
      previous_candle_4,
      previous_candle_5,
      trade,
      current_time,
      enter_tags,
    )

    if not sell:
      sell, signal_name = self.long_exit_main(
        self.long_high_profit_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_williams_r(
        self.long_high_profit_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_stoploss(
        self.long_high_profit_mode_name,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )


    if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
      previous_rate = self.target_profit_cache.data[pair]["rate"]
      previous_profit = self.target_profit_cache.data[pair]["profit"]
      previous_sell_reason = self.target_profit_cache.data[pair]["sell_reason"]
      previous_time_profit_reached = datetime.fromisoformat(self.target_profit_cache.data[pair]["time_profit_reached"])

      sell_max, signal_name_max = self.exit_profit_target(
        self.long_high_profit_mode_name,
        pair,
        trade,
        current_time,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        last_candle,
        previous_candle_1,
        previous_rate,
        previous_profit,
        previous_sell_reason,
        previous_time_profit_reached,
        enter_tags,
      )
      if sell_max and signal_name_max is not None:
        return True, f"{signal_name_max}_m"
      if previous_sell_reason in [f"exit_{self.long_high_profit_mode_name}_stoploss_u_e"]:
        if profit_ratio > (previous_profit + 0.005):
          mark_pair, mark_signal = self.mark_profit_target(
            self.long_high_profit_mode_name,
            pair,
            True,
            previous_sell_reason,
            trade,
            current_time,
            current_rate,
            profit_ratio,
            last_candle,
            previous_candle_1,
          )
          if mark_pair:
            self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
      elif (profit_init_ratio > (previous_profit + 0.001)) and (
        previous_sell_reason not in [f"exit_{self.long_high_profit_mode_name}_stoploss_doom"]
      ):

        mark_pair, mark_signal = self.mark_profit_target(
          self.long_high_profit_mode_name,
          pair,
          True,
          previous_sell_reason,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if sell and signal_name is not None:
      previous_profit = None
      if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
        previous_profit = self.target_profit_cache.data[pair]["profit"]
      if signal_name in [
        f"exit_{self.long_high_profit_mode_name}_stoploss_doom",
        f"exit_{self.long_high_profit_mode_name}_stoploss_u_e",
      ]:
        mark_pair, mark_signal = self.mark_profit_target(
          self.long_high_profit_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
        else:

          return True, f"{signal_name}"
      elif (previous_profit is None) or (previous_profit < profit_init_ratio):
        mark_pair, mark_signal = self.mark_profit_target(
          self.long_high_profit_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)
        else:

          return True, f"{signal_name}"
    else:
      if profit_init_ratio >= self.profit_max_thresholds[8]:
        previous_profit = None
        if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
          previous_profit = self.target_profit_cache.data[pair]["profit"]
        if (previous_profit is None) or (previous_profit < profit_init_ratio):
          mark_signal = f"exit_profit_{self.long_high_profit_mode_name}_max"
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if signal_name not in [
      f"exit_profit_{self.long_high_profit_mode_name}_max",
      f"exit_{self.long_high_profit_mode_name}_stoploss_doom",
      f"exit_{self.long_high_profit_mode_name}_stoploss_u_e",
    ]:
      if sell and (signal_name is not None):
        return True, f"{signal_name}"

    return False, None


  def long_exit_rapid(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    sell = False

    sell, signal_name = self.long_exit_signals(
      self.long_rapid_mode_name,
      profit_init_ratio,
      max_profit,
      max_loss,
      last_candle,
      previous_candle_1,
      previous_candle_2,
      previous_candle_3,
      previous_candle_4,
      previous_candle_5,
      trade,
      current_time,
      enter_tags,
    )

    if not sell:
      sell, signal_name = self.long_exit_main(
        self.long_rapid_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_williams_r(
        self.long_rapid_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_dec(
        self.long_rapid_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.long_exit_stoploss(
        self.long_rapid_mode_name,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      if (0.09 >= profit_init_ratio > 0.01) and (last_candle["rsi_14"] > 78.0):
        sell, signal_name = True, f"exit_{self.long_rapid_mode_name}_rpd_1"

      if (0.09 >= profit_init_ratio > 0.01) and (last_candle["cti_20"] > 0.95):
        sell, signal_name = True, f"exit_{self.long_rapid_mode_name}_rpd_2"

      if (0.09 >= profit_init_ratio > 0.01) and (last_candle["r_14"] >= -0.1):
        sell, signal_name = True, f"exit_{self.long_rapid_mode_name}_rpd_3"

      if profit_stake < -(
        filled_entries[0].cost
        * (self.stop_threshold_futures_rapid if self.is_futures_mode else self.stop_threshold_spot_rapid)
        / (trade.leverage if self.is_futures_mode else 1.0)
      ):
        sell, signal_name = True, f"exit_{self.long_rapid_mode_name}_stoploss_doom"


    if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
      previous_rate = self.target_profit_cache.data[pair]["rate"]
      previous_profit = self.target_profit_cache.data[pair]["profit"]
      previous_sell_reason = self.target_profit_cache.data[pair]["sell_reason"]
      previous_time_profit_reached = datetime.fromisoformat(self.target_profit_cache.data[pair]["time_profit_reached"])

      sell_max, signal_name_max = self.exit_profit_target(
        self.long_rapid_mode_name,
        pair,
        trade,
        current_time,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        last_candle,
        previous_candle_1,
        previous_rate,
        previous_profit,
        previous_sell_reason,
        previous_time_profit_reached,
        enter_tags,
      )
      if sell_max and signal_name_max is not None:
        return True, f"{signal_name_max}_m"
      if previous_sell_reason in [f"exit_{self.long_rapid_mode_name}_stoploss_u_e"]:
        if profit_ratio > (previous_profit + 0.005):
          mark_pair, mark_signal = self.mark_profit_target(
            self.long_rapid_mode_name,
            pair,
            True,
            previous_sell_reason,
            trade,
            current_time,
            current_rate,
            profit_ratio,
            last_candle,
            previous_candle_1,
          )
          if mark_pair:
            self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
      elif (profit_init_ratio > (previous_profit + 0.001)) and (
        previous_sell_reason not in [f"exit_{self.long_rapid_mode_name}_stoploss_doom"]
      ):

        mark_pair, mark_signal = self.mark_profit_target(
          self.long_rapid_mode_name,
          pair,
          True,
          previous_sell_reason,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if sell and signal_name is not None:
      previous_profit = None
      if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
        previous_profit = self.target_profit_cache.data[pair]["profit"]
      if signal_name in [
        f"exit_{self.long_rapid_mode_name}_stoploss_doom",
        f"exit_{self.long_rapid_mode_name}_stoploss_u_e",
      ]:
        mark_pair, mark_signal = self.mark_profit_target(
          self.long_rapid_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
        else:

          return True, f"{signal_name}"
      elif (previous_profit is None) or (previous_profit < profit_init_ratio):
        mark_pair, mark_signal = self.mark_profit_target(
          self.long_rapid_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)
        else:

          return True, f"{signal_name}"
    else:
      if profit_init_ratio >= 0.01:
        previous_profit = None
        if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
          previous_profit = self.target_profit_cache.data[pair]["profit"]
        if (previous_profit is None) or (previous_profit < profit_init_ratio):
          mark_signal = f"exit_profit_{self.long_rapid_mode_name}_max"
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if signal_name not in [f"exit_profit_{self.long_rapid_mode_name}_max"]:
      if sell and (signal_name is not None):
        return True, f"{signal_name}"

    return False, None


  def long_exit_grind(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    if len(filled_exits) > 30 and profit_init_ratio > 1.0:
      return True, f"exit_{self.long_grind_mode_name}_g"
    return False, None


  def long_exit_signals(
    self,
    mode_name: str,
    current_profit: float,
    max_profit: float,
    max_loss: float,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    buy_tag,
  ) -> tuple:

    if (
      (last_candle["rsi_14"] > 79.0)
      and (last_candle["close"] > last_candle["bb20_2_upp"])
      and (previous_candle_1["close"] > previous_candle_1["bb20_2_upp"])
      and (previous_candle_2["close"] > previous_candle_2["bb20_2_upp"])
      and (previous_candle_3["close"] > previous_candle_3["bb20_2_upp"])
      and (previous_candle_4["close"] > previous_candle_4["bb20_2_upp"])
    ):
      if last_candle["close"] > last_candle["ema_200"]:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_1_1_1"
      else:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_1_2_1"

    elif (
      (last_candle["rsi_14"] > 80.0)
      and (last_candle["close"] > last_candle["bb20_2_upp"])
      and (previous_candle_1["close"] > previous_candle_1["bb20_2_upp"])
      and (previous_candle_2["close"] > previous_candle_2["bb20_2_upp"])
    ):
      if last_candle["close"] > last_candle["ema_200"]:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_2_1_1"
      else:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_2_2_1"

    elif last_candle["rsi_14"] > 85.0:
      if last_candle["close"] > last_candle["ema_200"]:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_3_1_1"
      else:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_3_2_1"

    elif (last_candle["rsi_14"] > 80.0) and (last_candle["rsi_14_1h"] > 78.0):
      if last_candle["close"] > last_candle["ema_200"]:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_4_1_1"
      else:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_4_2_1"

    elif (
      (last_candle["close"] < last_candle["ema_200"])
      and (last_candle["close"] > last_candle["ema_50"])
      and (last_candle["rsi_14"] > 79.0)
    ):
      if current_profit > 0.01:
        return True, f"exit_{mode_name}_6_1"

    elif (last_candle["rsi_14_1h"] > 79.0) and (last_candle["crossed_below_ema_12_26"]):
      if last_candle["close"] > last_candle["ema_200"]:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_7_1_1"
      else:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_7_2_1"

    elif last_candle["close"] > last_candle["bb20_2_upp_1h"] * 1.08:
      if last_candle["close"] > last_candle["ema_200"]:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_8_1_1"
      else:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_8_2_1"

    return False, None


  def long_exit_main(
    self,
    mode_name: str,
    current_profit: float,
    max_profit: float,
    max_loss: float,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    buy_tag,
  ) -> tuple:
    if last_candle["close"] > last_candle["sma_200"]:
      if 0.02 > current_profit >= 0.01:
        if last_candle["rsi_14"] < 28.0:
          return True, f"exit_{mode_name}_o_1"
      elif 0.03 > current_profit >= 0.02:
        if last_candle["rsi_14"] < 30.0:
          return True, f"exit_{mode_name}_o_2"
      elif 0.04 > current_profit >= 0.03:
        if last_candle["rsi_14"] < 32.0:
          return True, f"exit_{mode_name}_o_3"
      elif 0.05 > current_profit >= 0.04:
        if last_candle["rsi_14"] < 34.0:
          return True, f"exit_{mode_name}_o_4"
      elif 0.06 > current_profit >= 0.05:
        if last_candle["rsi_14"] < 36.0:
          return True, f"exit_{mode_name}_o_5"
      elif 0.07 > current_profit >= 0.06:
        if last_candle["rsi_14"] < 38.0:
          return True, f"exit_{mode_name}_o_6"
      elif 0.08 > current_profit >= 0.07:
        if last_candle["rsi_14"] < 40.0:
          return True, f"exit_{mode_name}_o_7"
      elif 0.09 > current_profit >= 0.08:
        if last_candle["rsi_14"] < 42.0:
          return True, f"exit_{mode_name}_o_8"
      elif 0.1 > current_profit >= 0.09:
        if last_candle["rsi_14"] < 44.0:
          return True, f"exit_{mode_name}_o_9"
      elif 0.12 > current_profit >= 0.1:
        if last_candle["rsi_14"] < 46.0:
          return True, f"exit_{mode_name}_o_10"
      elif 0.2 > current_profit >= 0.12:
        if last_candle["rsi_14"] < 44.0:
          return True, f"exit_{mode_name}_o_11"
      elif current_profit >= 0.2:
        if last_candle["rsi_14"] < 42.0:
          return True, f"exit_{mode_name}_o_12"
    elif last_candle["close"] < last_candle["sma_200"]:
      if 0.02 > current_profit >= 0.01:
        if last_candle["rsi_14"] < 30.0:
          return True, f"exit_{mode_name}_u_1"
      elif 0.03 > current_profit >= 0.02:
        if last_candle["rsi_14"] < 32.0:
          return True, f"exit_{mode_name}_u_2"
      elif 0.04 > current_profit >= 0.03:
        if last_candle["rsi_14"] < 34.0:
          return True, f"exit_{mode_name}_u_3"
      elif 0.05 > current_profit >= 0.04:
        if last_candle["rsi_14"] < 36.0:
          return True, f"exit_{mode_name}_u_4"
      elif 0.06 > current_profit >= 0.05:
        if last_candle["rsi_14"] < 38.0:
          return True, f"exit_{mode_name}_u_5"
      elif 0.07 > current_profit >= 0.06:
        if last_candle["rsi_14"] < 40.0:
          return True, f"exit_{mode_name}_u_6"
      elif 0.08 > current_profit >= 0.07:
        if last_candle["rsi_14"] < 42.0:
          return True, f"exit_{mode_name}_u_7"
      elif 0.09 > current_profit >= 0.08:
        if last_candle["rsi_14"] < 44.0:
          return True, f"exit_{mode_name}_u_8"
      elif 0.1 > current_profit >= 0.09:
        if last_candle["rsi_14"] < 46.0:
          return True, f"exit_{mode_name}_u_9"
      elif 0.12 > current_profit >= 0.1:
        if last_candle["rsi_14"] < 48.0:
          return True, f"exit_{mode_name}_u_10"
      elif 0.2 > current_profit >= 0.12:
        if last_candle["rsi_14"] < 46.0:
          return True, f"exit_{mode_name}_u_11"
      elif current_profit >= 0.2:
        if last_candle["rsi_14"] < 44.0:
          return True, f"exit_{mode_name}_u_12"

    return False, None


  def long_exit_williams_r(
    self,
    mode_name: str,
    current_profit: float,
    max_profit: float,
    max_loss: float,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    buy_tag,
  ) -> tuple:
    if 0.02 > current_profit >= 0.01:
      if last_candle["r_480"] > -0.2:
        return True, f"exit_{mode_name}_w_1_1"
      elif (last_candle["r_14"] >= -1.0) and (last_candle["rsi_14"] > 78.0):
        return True, f"exit_{mode_name}_w_1_2"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["rsi_14"] < 46.0):
        return True, f"exit_{mode_name}_w_1_3"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["rsi_14"] > 74.0) and (last_candle["r_480_1h"] > -25.0):
        return True, f"exit_{mode_name}_w_1_4"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["cti_20"] > 0.95):
        return True, f"exit_{mode_name}_w_1_5"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] > 70.0)
        and (last_candle["r_480_1h"] > -10.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 60.0)
        and (last_candle["cti_20_1d"] > 0.80)
      ):
        return True, f"exit_{mode_name}_w_1_6"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_4h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 50.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_1_7"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["rsi_14_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.02)
      ):
        return True, f"exit_{mode_name}_w_1_8"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 76.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["cti_20_1d"] >= 0.70)
      ):
        return True, f"exit_{mode_name}_w_1_9"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_1_10"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["rsi_14_1h"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["r_480_1h"] > -30.0)
      ):
        return True, f"exit_{mode_name}_w_1_11"
      elif (
        (last_candle["r_14"] >= -24.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_1_12"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["rsi_14_4h"] >= 65.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_1_13"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1h"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
      ):
        return True, f"exit_{mode_name}_w_1_14"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["change_pct_1h"] < -0.00)
      ):
        return True, f"exit_{mode_name}_w_1_15"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["top_wick_pct_1d"] > 0.16)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.80))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_1_16"
      elif (
        (last_candle["r_14"] >= -16.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["r_480_1h"] > -30.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["top_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_1_17"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_1_18"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 80.0)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_1_19"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 65.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_1_20"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 65.0)
        and (last_candle["cti_20_dec_3_1h"] == True)
        and (last_candle["r_480_4h"] < -50.0)
      ):
        return True, f"exit_{mode_name}_w_1_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_1_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_1d"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["change_pct_1d"] < -0.01)
        and (last_candle["close"] < (last_candle["high_max_12_1d"] * 0.85))
      ):
        return True, f"exit_{mode_name}_w_1_23"
    elif 0.03 > current_profit >= 0.02:
      if last_candle["r_480"] > -0.3:
        return True, f"exit_{mode_name}_w_2_1"
      elif (last_candle["r_14"] >= -1.0) and (last_candle["rsi_14"] > 77.0):
        return True, f"exit_{mode_name}_w_2_2"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["rsi_14"] < 48.0):
        return True, f"exit_{mode_name}_w_2_3"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["rsi_14"] > 73.0) and (last_candle["r_480_1h"] > -25.0):
        return True, f"exit_{mode_name}_w_2_4"
      elif (last_candle["r_14"] >= -3.0) and (last_candle["cti_20"] > 0.95):
        return True, f"exit_{mode_name}_w_2_5"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -20.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 60.0)
        and (last_candle["cti_20_1d"] > 0.80)
      ):
        return True, f"exit_{mode_name}_w_2_6"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_4h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 50.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_2_7"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["rsi_14_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.02)
      ):
        return True, f"exit_{mode_name}_w_2_8"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["rsi_14_15m"] >= 68.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["cti_20_1d"] >= 0.70)
      ):
        return True, f"exit_{mode_name}_w_2_9"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_2_10"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["rsi_14_1h"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["r_480_1h"] > -30.0)
      ):
        return True, f"exit_{mode_name}_w_2_11"
      elif (
        (last_candle["r_14"] >= -24.0)
        and (last_candle["rsi_14"] >= 72.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_2_12"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["rsi_14_4h"] >= 65.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_2_13"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1h"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
      ):
        return True, f"exit_{mode_name}_w_2_14"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["change_pct_1h"] < -0.00)
      ):
        return True, f"exit_{mode_name}_w_2_15"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["top_wick_pct_1d"] > 0.16)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.80))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_2_16"
      elif (
        (last_candle["r_14"] >= -18.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["r_480_1h"] > -30.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["top_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_2_17"
      elif (
        (last_candle["r_14"] >= -5.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_2_18"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 80.0)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_2_19"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 65.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_2_20"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["rsi_14_15m"] >= 65.0)
        and (last_candle["cti_20_dec_3_1h"] == True)
        and (last_candle["r_480_4h"] < -50.0)
      ):
        return True, f"exit_{mode_name}_w_2_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_2_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 65.0)
        and (last_candle["rsi_14_1d"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["change_pct_1d"] < -0.01)
        and (last_candle["close"] < (last_candle["high_max_12_1d"] * 0.85))
      ):
        return True, f"exit_{mode_name}_w_2_23"
    elif 0.04 > current_profit >= 0.03:
      if last_candle["r_480"] > -0.4:
        return True, f"exit_{mode_name}_w_3_1"
      elif (last_candle["r_14"] >= -1.0) and (last_candle["rsi_14"] > 76.0):
        return True, f"exit_{mode_name}_w_3_2"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["rsi_14"] < 50.0):
        return True, f"exit_{mode_name}_w_3_3"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["rsi_14"] > 72.0) and (last_candle["r_480_1h"] > -25.0):
        return True, f"exit_{mode_name}_w_3_4"
      elif (last_candle["r_14"] >= -4.0) and (last_candle["cti_20"] > 0.95):
        return True, f"exit_{mode_name}_w_3_5"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -20.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 60.0)
        and (last_candle["cti_20_1d"] > 0.80)
      ):
        return True, f"exit_{mode_name}_w_3_6"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_4h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 50.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_3_7"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["rsi_14_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.02)
      ):
        return True, f"exit_{mode_name}_w_3_8"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 72.0)
        and (last_candle["rsi_14_15m"] >= 66.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["cti_20_1d"] >= 0.70)
      ):
        return True, f"exit_{mode_name}_w_3_9"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_3_10"
      elif (
        (last_candle["r_14"] >= -3.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["rsi_14_1h"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["r_480_1h"] > -30.0)
      ):
        return True, f"exit_{mode_name}_w_3_11"
      elif (
        (last_candle["r_14"] >= -24.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_3_12"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["rsi_14_4h"] >= 65.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_3_13"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1h"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
      ):
        return True, f"exit_{mode_name}_w_3_14"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["change_pct_1h"] < -0.00)
      ):
        return True, f"exit_{mode_name}_w_3_15"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["top_wick_pct_1d"] > 0.16)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.80))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_3_16"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["r_480_1h"] > -30.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["top_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_3_17"
      elif (
        (last_candle["r_14"] >= -35.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_3_18"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 80.0)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_3_19"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 65.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_3_20"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 66.0)
        and (last_candle["rsi_14_15m"] >= 65.0)
        and (last_candle["cti_20_dec_3_1h"] == True)
        and (last_candle["r_480_4h"] < -50.0)
      ):
        return True, f"exit_{mode_name}_w_3_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_3_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["change_pct_1d"] < -0.01)
        and (last_candle["close"] < (last_candle["high_max_12_1d"] * 0.85))
      ):
        return True, f"exit_{mode_name}_w_3_23"
    elif 0.05 > current_profit >= 0.04:
      if last_candle["r_480"] > -0.5:
        return True, f"exit_{mode_name}_w_4_1"
      elif (last_candle["r_14"] >= -1.0) and (last_candle["rsi_14"] > 75.0):
        return True, f"exit_{mode_name}_w_4_2"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["rsi_14"] < 52.0):
        return True, f"exit_{mode_name}_w_4_3"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["rsi_14"] > 71.0) and (last_candle["r_480_1h"] > -25.0):
        return True, f"exit_{mode_name}_w_4_4"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["cti_20"] > 0.95):
        return True, f"exit_{mode_name}_w_4_5"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -20.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 60.0)
        and (last_candle["cti_20_1d"] > 0.80)
      ):
        return True, f"exit_{mode_name}_w_4_6"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_4h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 50.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_4_7"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["rsi_14_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.02)
      ):
        return True, f"exit_{mode_name}_w_4_8"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 64.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["cti_20_1d"] >= 0.70)
      ):
        return True, f"exit_{mode_name}_w_4_9"
      elif (
        (last_candle["r_14"] >= -14.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_4_10"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["rsi_14_1h"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["r_480_1h"] > -30.0)
      ):
        return True, f"exit_{mode_name}_w_4_11"
      elif (
        (last_candle["r_14"] >= -24.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_4_12"
      elif (
        (last_candle["r_14"] >= -14.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["rsi_14_4h"] >= 65.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_4_13"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1h"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
      ):
        return True, f"exit_{mode_name}_w_4_14"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["change_pct_1h"] < -0.00)
      ):
        return True, f"exit_{mode_name}_w_4_15"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["top_wick_pct_1d"] > 0.16)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.80))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_4_16"
      elif (
        (last_candle["r_14"] >= -22.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["r_480_1h"] > -30.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["top_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_4_17"
      elif (
        (last_candle["r_14"] >= -40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_4_18"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 80.0)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_4_19"
      elif (
        (last_candle["r_14"] >= -14.0)
        and (last_candle["rsi_14"] >= 65.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_4_20"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 64.0)
        and (last_candle["rsi_14_15m"] >= 65.0)
        and (last_candle["cti_20_dec_3_1h"] == True)
        and (last_candle["r_480_4h"] < -50.0)
      ):
        return True, f"exit_{mode_name}_w_4_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_4_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["change_pct_1d"] < -0.01)
        and (last_candle["close"] < (last_candle["high_max_12_1d"] * 0.85))
      ):
        return True, f"exit_{mode_name}_w_4_23"
    elif 0.06 > current_profit >= 0.05:
      if last_candle["r_480"] > -0.6:
        return True, f"exit_{mode_name}_w_5_1"
      elif (last_candle["r_14"] >= -1.0) and (last_candle["rsi_14"] > 74.0):
        return True, f"exit_{mode_name}_w_5_2"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["rsi_14"] < 54.0):
        return True, f"exit_{mode_name}_w_5_3"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["rsi_14"] > 70.0) and (last_candle["r_480_1h"] > -25.0):
        return True, f"exit_{mode_name}_w_5_4"
      elif (last_candle["r_14"] >= -6.0) and (last_candle["cti_20"] > 0.95):
        return True, f"exit_{mode_name}_w_5_5"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -20.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 60.0)
        and (last_candle["cti_20_1d"] > 0.80)
      ):
        return True, f"exit_{mode_name}_w_5_6"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_4h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 50.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_5_7"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["rsi_14_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.02)
      ):
        return True, f"exit_{mode_name}_w_5_8"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["rsi_14_15m"] >= 62.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["cti_20_1d"] >= 0.70)
      ):
        return True, f"exit_{mode_name}_w_5_9"
      elif (
        (last_candle["r_14"] >= -15.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_5_10"
      elif (
        (last_candle["r_14"] >= -5.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["rsi_14_1h"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["r_480_1h"] > -30.0)
      ):
        return True, f"exit_{mode_name}_w_5_11"
      elif (
        (last_candle["r_14"] >= -24.0)
        and (last_candle["rsi_14"] >= 66.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_5_12"
      elif (
        (last_candle["r_14"] >= -16.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["rsi_14_4h"] >= 65.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_5_13"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1h"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
      ):
        return True, f"exit_{mode_name}_w_5_14"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["change_pct_1h"] < -0.00)
      ):
        return True, f"exit_{mode_name}_w_5_15"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["top_wick_pct_1d"] > 0.16)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.80))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_5_16"
      elif (
        (last_candle["r_14"] >= -24.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["r_480_1h"] > -30.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["top_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_5_17"
      elif (
        (last_candle["r_14"] >= -45.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_5_18"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 80.0)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_5_19"
      elif (
        (last_candle["r_14"] >= -16.0)
        and (last_candle["rsi_14"] >= 65.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_5_20"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 62.0)
        and (last_candle["rsi_14_15m"] >= 65.0)
        and (last_candle["cti_20_dec_3_1h"] == True)
        and (last_candle["r_480_4h"] < -50.0)
      ):
        return True, f"exit_{mode_name}_w_5_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_5_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["change_pct_1d"] < -0.01)
        and (last_candle["close"] < (last_candle["high_max_12_1d"] * 0.85))
      ):
        return True, f"exit_{mode_name}_w_5_23"
    elif 0.07 > current_profit >= 0.06:
      if last_candle["r_480"] > -0.7:
        return True, f"exit_{mode_name}_w_6_1"
      elif (last_candle["r_14"] >= -1.0) and (last_candle["rsi_14"] > 75.0):
        return True, f"exit_{mode_name}_w_6_2"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["rsi_14"] < 52.0):
        return True, f"exit_{mode_name}_w_6_3"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["rsi_14"] > 71.0) and (last_candle["r_480_1h"] > -25.0):
        return True, f"exit_{mode_name}_w_6_4"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["cti_20"] > 0.95):
        return True, f"exit_{mode_name}_w_6_5"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -20.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 60.0)
        and (last_candle["cti_20_1d"] > 0.80)
      ):
        return True, f"exit_{mode_name}_w_6_6"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_4h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 50.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_6_7"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["rsi_14_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.02)
      ):
        return True, f"exit_{mode_name}_w_6_8"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 64.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["cti_20_1d"] >= 0.70)
      ):
        return True, f"exit_{mode_name}_w_6_9"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_6_10"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["rsi_14_1h"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["r_480_1h"] > -30.0)
      ):
        return True, f"exit_{mode_name}_w_6_11"
      elif (
        (last_candle["r_14"] >= -14.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_6_12"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["rsi_14_4h"] >= 65.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_6_13"
      elif (
        (last_candle["r_14"] >= -18.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1h"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
      ):
        return True, f"exit_{mode_name}_w_6_14"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["change_pct_1h"] < -0.00)
      ):
        return True, f"exit_{mode_name}_w_6_15"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["top_wick_pct_1d"] > 0.16)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.80))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_6_16"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["r_480_1h"] > -30.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["top_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_6_17"
      elif (
        (last_candle["r_14"] >= -35.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_6_18"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 80.0)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_6_19"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 65.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_6_20"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 64.0)
        and (last_candle["rsi_14_15m"] >= 65.0)
        and (last_candle["cti_20_dec_3_1h"] == True)
        and (last_candle["r_480_4h"] < -50.0)
      ):
        return True, f"exit_{mode_name}_w_6_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_6_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_1d"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["change_pct_1d"] < -0.01)
        and (last_candle["close"] < (last_candle["high_max_12_1d"] * 0.85))
      ):
        return True, f"exit_{mode_name}_w_6_23"
    elif 0.08 > current_profit >= 0.07:
      if last_candle["r_480"] > -0.8:
        return True, f"exit_{mode_name}_w_7_1"
      elif (last_candle["r_14"] >= -1.0) and (last_candle["rsi_14"] > 76.0):
        return True, f"exit_{mode_name}_w_7_2"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["rsi_14"] < 50.0):
        return True, f"exit_{mode_name}_w_7_3"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["rsi_14"] > 72.0) and (last_candle["r_480_1h"] > -25.0):
        return True, f"exit_{mode_name}_w_7_4"
      elif (last_candle["r_14"] >= -4.0) and (last_candle["cti_20"] > 0.95):
        return True, f"exit_{mode_name}_w_7_5"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -15.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 60.0)
        and (last_candle["cti_20_1d"] > 0.80)
      ):
        return True, f"exit_{mode_name}_w_7_6"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_4h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 50.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_7_7"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["rsi_14_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.02)
      ):
        return True, f"exit_{mode_name}_w_7_8"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 72.0)
        and (last_candle["rsi_14_15m"] >= 66.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["cti_20_1d"] >= 0.70)
      ):
        return True, f"exit_{mode_name}_w_7_9"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 64.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_7_10"
      elif (
        (last_candle["r_14"] >= -3.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["rsi_14_1h"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["r_480_1h"] > -30.0)
      ):
        return True, f"exit_{mode_name}_w_7_11"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_7_12"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["rsi_14_4h"] >= 65.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_7_13"
      elif (
        (last_candle["r_14"] >= -16.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1h"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
      ):
        return True, f"exit_{mode_name}_w_7_14"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["change_pct_1h"] < -0.00)
      ):
        return True, f"exit_{mode_name}_w_7_15"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["top_wick_pct_1d"] > 0.16)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.80))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_7_16"
      elif (
        (last_candle["r_14"] >= -16.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["r_480_1h"] > -30.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["top_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_7_17"
      elif (
        (last_candle["r_14"] >= -25.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_7_18"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 80.0)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_7_19"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 65.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_7_20"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 66.0)
        and (last_candle["rsi_14_15m"] >= 65.0)
        and (last_candle["cti_20_dec_3_1h"] == True)
        and (last_candle["r_480_4h"] < -50.0)
      ):
        return True, f"exit_{mode_name}_w_7_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_7_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 65.0)
        and (last_candle["rsi_14_1d"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["change_pct_1d"] < -0.01)
        and (last_candle["close"] < (last_candle["high_max_12_1d"] * 0.85))
      ):
        return True, f"exit_{mode_name}_w_7_23"
    elif 0.09 > current_profit >= 0.08:
      if last_candle["r_480"] > -0.9:
        return True, f"exit_{mode_name}_w_8_1"
      elif (last_candle["r_14"] >= -1.0) and (last_candle["rsi_14"] > 77.0):
        return True, f"exit_{mode_name}_w_8_2"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["rsi_14"] < 48.0):
        return True, f"exit_{mode_name}_w_8_3"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["rsi_14"] > 73.0) and (last_candle["r_480_1h"] > -25.0):
        return True, f"exit_{mode_name}_w_8_4"
      elif (last_candle["r_14"] >= -3.0) and (last_candle["cti_20"] > 0.95):
        return True, f"exit_{mode_name}_w_8_5"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -15.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 60.0)
        and (last_candle["cti_20_1d"] > 0.80)
      ):
        return True, f"exit_{mode_name}_w_8_6"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_4h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 50.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_8_7"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["rsi_14_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.02)
      ):
        return True, f"exit_{mode_name}_w_8_8"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["rsi_14_15m"] >= 68.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["cti_20_1d"] >= 0.70)
      ):
        return True, f"exit_{mode_name}_w_8_9"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 66.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_8_10"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["rsi_14_1h"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["r_480_1h"] > -30.0)
      ):
        return True, f"exit_{mode_name}_w_8_11"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 72.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_8_12"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["rsi_14_4h"] >= 65.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_8_13"
      elif (
        (last_candle["r_14"] >= -14.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1h"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
      ):
        return True, f"exit_{mode_name}_w_8_14"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["change_pct_1h"] < -0.00)
      ):
        return True, f"exit_{mode_name}_w_8_15"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["top_wick_pct_1d"] > 0.16)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.80))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_8_16"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["r_480_1h"] > -30.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["top_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_8_17"
      elif (
        (last_candle["r_14"] >= -15.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_8_18"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 80.0)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_8_19"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 65.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_8_20"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["rsi_14_15m"] >= 65.0)
        and (last_candle["cti_20_dec_3_1h"] == True)
        and (last_candle["r_480_4h"] < -50.0)
      ):
        return True, f"exit_{mode_name}_w_8_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_8_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_1d"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["change_pct_1d"] < -0.01)
        and (last_candle["close"] < (last_candle["high_max_12_1d"] * 0.85))
      ):
        return True, f"exit_{mode_name}_w_8_23"
    elif 0.1 > current_profit >= 0.09:
      if last_candle["r_480"] > -1.0:
        return True, f"exit_{mode_name}_w_9_1"
      elif (last_candle["r_14"] >= -1.0) and (last_candle["rsi_14"] > 78.0):
        return True, f"exit_{mode_name}_w_9_2"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["rsi_14"] < 46.0):
        return True, f"exit_{mode_name}_w_9_3"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["rsi_14"] > 74.0) and (last_candle["r_480_1h"] > -25.0):
        return True, f"exit_{mode_name}_w_9_4"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["cti_20"] > 0.95):
        return True, f"exit_{mode_name}_w_9_5"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -15.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 60.0)
        and (last_candle["cti_20_1d"] > 0.80)
      ):
        return True, f"exit_{mode_name}_w_9_6"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] > 60.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_4h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 50.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_9_7"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["rsi_14_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.02)
      ):
        return True, f"exit_{mode_name}_w_9_8"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 76.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["cti_20_1d"] >= 0.70)
      ):
        return True, f"exit_{mode_name}_w_9_9"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_9_10"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["rsi_14_1h"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["r_480_1h"] > -30.0)
      ):
        return True, f"exit_{mode_name}_w_9_11"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_9_12"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["rsi_14_4h"] >= 65.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_9_13"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1h"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
      ):
        return True, f"exit_{mode_name}_w_9_14"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["change_pct_1h"] < -0.00)
      ):
        return True, f"exit_{mode_name}_w_9_15"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["top_wick_pct_1d"] > 0.16)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.80))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_9_16"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["r_480_1h"] > -30.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["top_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_9_17"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_9_18"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 80.0)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_9_19"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 65.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_9_20"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 65.0)
        and (last_candle["cti_20_dec_3_1h"] == True)
        and (last_candle["r_480_4h"] < -50.0)
      ):
        return True, f"exit_{mode_name}_w_9_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_9_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 72.0)
        and (last_candle["rsi_14_1d"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["change_pct_1d"] < -0.01)
        and (last_candle["close"] < (last_candle["high_max_12_1d"] * 0.85))
      ):
        return True, f"exit_{mode_name}_w_9_23"
    elif 0.12 > current_profit >= 0.1:
      if last_candle["r_480"] > -1.1:
        return True, f"exit_{mode_name}_w_10_1"
      elif (last_candle["r_14"] >= -1.0) and (last_candle["rsi_14"] > 79.0):
        return True, f"exit_{mode_name}_w_10_2"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["rsi_14"] < 44.0):
        return True, f"exit_{mode_name}_w_10_3"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["rsi_14"] > 75.0) and (last_candle["r_480_1h"] > -25.0):
        return True, f"exit_{mode_name}_w_10_4"
      elif (last_candle["r_14"] >= -1.0) and (last_candle["cti_20"] > 0.95):
        return True, f"exit_{mode_name}_w_10_5"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] > 65.0)
        and (last_candle["r_480_1h"] > -10.0)
        and (last_candle["r_480_4h"] > -5.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 60.0)
        and (last_candle["cti_20_1d"] > 0.80)
      ):
        return True, f"exit_{mode_name}_w_10_6"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] > 65.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_4h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 50.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_10_7"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 75.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["rsi_14_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.02)
      ):
        return True, f"exit_{mode_name}_w_10_8"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 78.0)
        and (last_candle["rsi_14_15m"] >= 72.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["cti_20_1d"] >= 0.70)
      ):
        return True, f"exit_{mode_name}_w_10_9"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_10_10"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 72.0)
        and (last_candle["rsi_14_15m"] >= 72.0)
        and (last_candle["rsi_14_1h"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["r_480_1h"] > -30.0)
      ):
        return True, f"exit_{mode_name}_w_10_11"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 76.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_10_12"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["rsi_14_4h"] >= 65.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_10_13"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1h"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
      ):
        return True, f"exit_{mode_name}_w_10_14"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["change_pct_1h"] < -0.00)
      ):
        return True, f"exit_{mode_name}_w_10_15"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["top_wick_pct_1d"] > 0.16)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.80))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_10_16"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["r_480_1h"] > -30.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["top_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_10_17"
      elif (
        (last_candle["r_14"] >= -5.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_10_18"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 80.0)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_10_19"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 65.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_10_20"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 72.0)
        and (last_candle["rsi_14_15m"] >= 65.0)
        and (last_candle["cti_20_dec_3_1h"] == True)
        and (last_candle["r_480_4h"] < -50.0)
      ):
        return True, f"exit_{mode_name}_w_10_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_10_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["rsi_14_1d"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["change_pct_1d"] < -0.01)
        and (last_candle["close"] < (last_candle["high_max_12_1d"] * 0.85))
      ):
        return True, f"exit_{mode_name}_w_10_23"
    elif 0.2 > current_profit >= 0.12:
      if last_candle["r_480"] > -0.4:
        return True, f"exit_{mode_name}_w_11_1"
      elif (last_candle["r_14"] >= -1.0) and (last_candle["rsi_14"] > 80.0):
        return True, f"exit_{mode_name}_w_11_2"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["rsi_14"] < 42.0):
        return True, f"exit_{mode_name}_w_11_3"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["rsi_14"] > 76.0) and (last_candle["r_480_1h"] > -25.0):
        return True, f"exit_{mode_name}_w_11_4"
      elif (last_candle["r_14"] >= -0.5) and (last_candle["cti_20"] > 0.95):
        return True, f"exit_{mode_name}_w_11_5"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] > 70.0)
        and (last_candle["r_480_1h"] > -10.0)
        and (last_candle["r_480_4h"] > -5.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 60.0)
        and (last_candle["cti_20_1d"] > 0.80)
      ):
        return True, f"exit_{mode_name}_w_11_6"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] > 70.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_4h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 50.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_11_7"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 76.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["rsi_14_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.02)
      ):
        return True, f"exit_{mode_name}_w_11_8"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 78.0)
        and (last_candle["rsi_14_15m"] >= 74.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["cti_20_1d"] >= 0.70)
      ):
        return True, f"exit_{mode_name}_w_11_9"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 72.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_11_10"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["rsi_14_15m"] >= 74.0)
        and (last_candle["rsi_14_1h"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["r_480_1h"] > -30.0)
      ):
        return True, f"exit_{mode_name}_w_11_11"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 78.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_11_12"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["rsi_14_4h"] >= 65.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_11_13"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1h"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
      ):
        return True, f"exit_{mode_name}_w_11_14"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["change_pct_1h"] < -0.00)
      ):
        return True, f"exit_{mode_name}_w_11_15"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["top_wick_pct_1d"] > 0.16)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.80))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_11_16"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["r_480_1h"] > -30.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["top_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_11_17"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_11_18"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 65.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 80.0)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_11_19"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_11_20"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["cti_20_dec_3_1h"] == True)
        and (last_candle["r_480_4h"] < -50.0)
      ):
        return True, f"exit_{mode_name}_w_11_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 75.0)
        and (last_candle["rsi_14_15m"] >= 65.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_11_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 78.0)
        and (last_candle["rsi_14_1d"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["change_pct_1d"] < -0.01)
        and (last_candle["close"] < (last_candle["high_max_12_1d"] * 0.85))
      ):
        return True, f"exit_{mode_name}_w_11_23"
    elif current_profit >= 0.2:
      if last_candle["r_480"] > -0.2:
        return True, f"exit_{mode_name}_w_12_1"
      elif (last_candle["r_14"] >= -1.0) and (last_candle["rsi_14"] > 81.0):
        return True, f"exit_{mode_name}_w_12_2"
      elif (last_candle["r_14"] >= -2.0) and (last_candle["rsi_14"] < 40.0):
        return True, f"exit_{mode_name}_w_12_3"
      elif (last_candle["r_14"] >= -5.0) and (last_candle["rsi_14"] > 77.0) and (last_candle["r_480_1h"] > -25.0):
        return True, f"exit_{mode_name}_w_12_4"
      elif (last_candle["r_14"] >= -0.1) and (last_candle["cti_20"] > 0.95):
        return True, f"exit_{mode_name}_w_12_5"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] > 75.0)
        and (last_candle["r_480_1h"] > -5.0)
        and (last_candle["r_480_4h"] > -5.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 60.0)
        and (last_candle["cti_20_1d"] > 0.80)
      ):
        return True, f"exit_{mode_name}_w_12_6"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] > 75.0)
        and (last_candle["r_480_1h"] > -25.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["rsi_14_1h"] > 50.0)
        and (last_candle["rsi_14_4h"] > 50.0)
        and (last_candle["rsi_14_1d"] > 50.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_12_7"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 78.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["rsi_14_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.02)
      ):
        return True, f"exit_{mode_name}_w_12_8"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 80.0)
        and (last_candle["rsi_14_15m"] >= 76.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["cti_20_1d"] >= 0.70)
      ):
        return True, f"exit_{mode_name}_w_12_9"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_12_10"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 76.0)
        and (last_candle["rsi_14_15m"] >= 76.0)
        and (last_candle["rsi_14_1h"] >= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["r_480_1h"] > -30.0)
      ):
        return True, f"exit_{mode_name}_w_12_11"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 80.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_12_12"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["rsi_14_4h"] >= 65.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.01)
      ):
        return True, f"exit_{mode_name}_w_12_13"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1h"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
      ):
        return True, f"exit_{mode_name}_w_12_14"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_4h"] >= 75.0)
        and (last_candle["r_480_4h"] > -15.0)
        and (last_candle["change_pct_1h"] < -0.00)
      ):
        return True, f"exit_{mode_name}_w_12_15"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["top_wick_pct_1d"] > 0.16)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.80))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_12_16"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["r_480_1h"] > -30.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["top_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_12_17"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_12_18"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 80.0)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_12_19"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 70.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_12_20"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 80.0)
        and (last_candle["rsi_14_15m"] >= 75.0)
        and (last_candle["cti_20_dec_3_1h"] == True)
        and (last_candle["r_480_4h"] < -50.0)
      ):
        return True, f"exit_{mode_name}_w_12_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 80.0)
        and (last_candle["rsi_14_15m"] >= 75.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_w_12_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 80.0)
        and (last_candle["rsi_14_1d"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["change_pct_1d"] < -0.01)
        and (last_candle["close"] < (last_candle["high_max_12_1d"] * 0.85))
      ):
        return True, f"exit_{mode_name}_w_12_23"

    return False, None


  def long_exit_dec(
    self,
    mode_name: str,
    current_profit: float,
    max_profit: float,
    max_loss: float,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    buy_tag,
  ) -> tuple:
    if 0.02 > current_profit >= 0.01:
      if (
        (last_candle["r_14"] > -10.0)
        and (last_candle["rsi_14"] > 66.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_1_1"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_1_2"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["cti_20_1d"] >= 0.50)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["change_pct_4h"] <= -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_3"
      elif (
        (last_candle["r_14"] >= -40.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1h"] < -0.03)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_4"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["change_pct_1d"] < -0.02)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_5"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_6"
      elif (
        (last_candle["rsi_3"] >= 60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_24"] == True)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_7"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_8"
      elif (
        (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_3_1h"] <= 10.0)
        and (last_candle["rsi_3_4h"] <= 6.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_9"
      elif (
        (last_candle["rsi_14"] <= 46.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_1h"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_1_10"
      elif (
        (last_candle["r_14"] >= -5.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1d"] < -0.10)
        and (last_candle["not_downtrend_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_11"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_12"
      elif (
        (last_candle["r_14"] >= -15.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["change_pct_4h"] < -0.03)
        and (last_candle["top_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_13"
      elif (
        (last_candle["rsi_14"] <= 46.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_3_1d"] <= 6.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_14"
      elif (
        (last_candle["r_14"] >= -5.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_15"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] >= 0.70)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_16"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_1_17"
      elif (
        (last_candle["r_14"] >= -60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_18"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_max_6_1d"] >= 85.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.06)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_1_19"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_20"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["close"] < (last_candle["high_max_48_1h"] * 0.75))
      ):
        return True, f"exit_{mode_name}_d_1_21"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.01)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.70))
      ):
        return True, f"exit_{mode_name}_d_1_22"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_1_23"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["change_pct_1h"] < -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_24"
      elif (
        (last_candle["r_14"] >= -25.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["rsi_3_1d"] <= 26.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_25"
    elif 0.03 > current_profit >= 0.02:
      if (
        (last_candle["r_14"] > -16.0)
        and (last_candle["rsi_14"] > 56.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_2_1"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] > 0.5)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_2_2"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["cti_20_1d"] >= 0.50)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["change_pct_4h"] <= -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_3"
      elif (
        (last_candle["r_14"] >= -40.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1h"] < -0.03)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_4"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["change_pct_1d"] < -0.02)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_5"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_3"] >= 70.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.7)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_6"
      elif (
        (last_candle["rsi_3"] >= 60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_24"] == True)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_7"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 66.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_8"
      elif (
        (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_3_1h"] <= 10.0)
        and (last_candle["rsi_3_4h"] <= 6.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_9"
      elif (
        (last_candle["rsi_14"] <= 48.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_1h"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_2_10"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1d"] < -0.10)
        and (last_candle["not_downtrend_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_11"
      elif (
        (last_candle["r_14"] >= -9.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_12"
      elif (
        (last_candle["r_14"] >= -15.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["change_pct_4h"] < -0.03)
        and (last_candle["top_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_13"
      elif (
        (last_candle["rsi_14"] <= 48.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_3_1d"] <= 6.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_14"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_15"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] >= 0.70)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_16"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_2_17"
      elif (
        (last_candle["r_14"] >= -60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_18"
      elif (
        (last_candle["r_14"] >= -40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_max_6_1d"] >= 85.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.06)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_2_19"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_20"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["close"] < (last_candle["high_max_48_1h"] * 0.75))
      ):
        return True, f"exit_{mode_name}_d_2_21"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.01)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.70))
      ):
        return True, f"exit_{mode_name}_d_2_22"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_2_23"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["change_pct_1h"] < -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_24"
      elif (
        (last_candle["r_14"] >= -25.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["rsi_3_1d"] <= 26.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_25"
    elif 0.04 > current_profit >= 0.03:
      if (
        (last_candle["r_14"] > -16.0)
        and (last_candle["rsi_14"] > 54.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_3_1"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_3_2"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["cti_20_1d"] >= 0.50)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["change_pct_4h"] <= -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_3"
      elif (
        (last_candle["r_14"] >= -40.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_1d"] > 0.5)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1h"] < -0.03)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_4"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["change_pct_1d"] < -0.02)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_5"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_3"] >= 70.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_6"
      elif (
        (last_candle["rsi_3"] >= 60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_24"] == True)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_7"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 64.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_8"
      elif (
        (last_candle["rsi_14"] <= 42.0)
        and (last_candle["rsi_3_1h"] <= 10.0)
        and (last_candle["rsi_3_4h"] <= 6.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_9"
      elif (
        (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_1h"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_3_10"
      elif (
        (last_candle["r_14"] >= -7.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1d"] < -0.10)
        and (last_candle["not_downtrend_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_11"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_12"
      elif (
        (last_candle["r_14"] >= -15.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["change_pct_4h"] < -0.03)
        and (last_candle["top_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_13"
      elif (
        (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_3_1d"] <= 6.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_14"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] >= 66.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_15"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] >= 0.70)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_16"
      elif (
        (last_candle["r_14"] >= -16.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_3_17"
      elif (
        (last_candle["r_14"] >= -60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_18"
      elif (
        (last_candle["r_14"] >= -40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_max_6_1d"] >= 85.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.06)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_3_19"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_20"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["close"] < (last_candle["high_max_48_1h"] * 0.75))
      ):
        return True, f"exit_{mode_name}_d_3_21"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.01)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.70))
      ):
        return True, f"exit_{mode_name}_d_3_22"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_3_23"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["change_pct_1h"] < -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_24"
      elif (
        (last_candle["r_14"] >= -25.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["rsi_3_1d"] <= 26.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_25"
    elif 0.05 > current_profit >= 0.04:
      if (
        (last_candle["r_14"] > -16.0)
        and (last_candle["rsi_14"] > 52.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_4_1"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_4_2"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["cti_20_1d"] >= 0.50)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["change_pct_4h"] <= -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_3"
      elif (
        (last_candle["r_14"] >= -40.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1h"] < -0.03)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_4"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["change_pct_1d"] < -0.02)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_5"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_3"] >= 70.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.7)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_6"
      elif (
        (last_candle["rsi_3"] >= 60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_24"] == True)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_7"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 62.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_8"
      elif (
        (last_candle["rsi_14"] <= 44.0)
        and (last_candle["rsi_3_1h"] <= 10.0)
        and (last_candle["rsi_3_4h"] <= 6.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_9"
      elif (
        (last_candle["rsi_14"] <= 52.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_1h"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_4_10"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1d"] < -0.10)
        and (last_candle["not_downtrend_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_11"
      elif (
        (last_candle["r_14"] >= -11.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_12"
      elif (
        (last_candle["r_14"] >= -15.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["change_pct_4h"] < -0.03)
        and (last_candle["top_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_13"
      elif (
        (last_candle["rsi_14"] <= 52.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_3_1d"] <= 6.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_14"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 64.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_15"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] >= 0.70)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_16"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_4_17"
      elif (
        (last_candle["r_14"] >= -60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_18"
      elif (
        (last_candle["r_14"] >= -40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_max_6_1d"] >= 85.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.06)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_4_19"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_20"
      elif (
        (last_candle["r_14"] >= -14.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["close"] < (last_candle["high_max_48_1h"] * 0.75))
      ):
        return True, f"exit_{mode_name}_d_4_21"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.01)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.70))
      ):
        return True, f"exit_{mode_name}_d_4_22"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_4_23"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["change_pct_1h"] < -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_24"
      elif (
        (last_candle["r_14"] >= -25.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["rsi_3_1d"] <= 26.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_25"
    elif 0.06 > current_profit >= 0.05:
      if (
        (last_candle["r_14"] > -16.0)
        and (last_candle["rsi_14"] > 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_5_1"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] > 0.5)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_5_2"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["cti_20_1d"] >= 0.50)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["change_pct_4h"] <= -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_3"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1h"] < -0.03)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_4"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["change_pct_1d"] < -0.02)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_5"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_3"] >= 70.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_6"
      elif (
        (last_candle["rsi_3"] >= 60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_24"] == True)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_7"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_8"
      elif (
        (last_candle["rsi_14"] <= 46.0)
        and (last_candle["rsi_3_1h"] <= 10.0)
        and (last_candle["rsi_3_4h"] <= 6.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_9"
      elif (
        (last_candle["rsi_14"] <= 54.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_1h"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_5_10"
      elif (
        (last_candle["r_14"] >= -9.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.8)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1d"] < -0.10)
        and (last_candle["not_downtrend_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_11"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.8)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_12"
      elif (
        (last_candle["r_14"] >= -15.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["change_pct_4h"] < -0.03)
        and (last_candle["top_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_13"
      elif (
        (last_candle["rsi_14"] <= 54.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_3_1d"] <= 6.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_14"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_15"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] >= 0.70)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_16"
      elif (
        (last_candle["r_14"] >= -24.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_5_17"
      elif (
        (last_candle["r_14"] >= -60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_18"
      elif (
        (last_candle["r_14"] >= -40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_max_6_1d"] >= 85.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.06)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_5_19"
      elif (
        (last_candle["r_14"] >= -30.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_20"
      elif (
        (last_candle["r_14"] >= -16.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["close"] < (last_candle["high_max_48_1h"] * 0.75))
      ):
        return True, f"exit_{mode_name}_d_5_21"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.01)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.70))
      ):
        return True, f"exit_{mode_name}_d_5_22"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_5_23"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["change_pct_1h"] < -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_24"
      elif (
        (last_candle["r_14"] >= -25.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["rsi_3_1d"] <= 26.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_25"
    elif 0.07 > current_profit >= 0.06:
      if (
        (last_candle["r_14"] > -16.0)
        and (last_candle["rsi_14"] > 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_6_1"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] > 0.5)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_6_2"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["cti_20_1d"] >= 0.50)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["change_pct_4h"] <= -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_3"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_1d"] > 0.5)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1h"] < -0.03)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_4"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_3"] >= 94.0)
        and (last_candle["rsi_14"] >= 64.0)
        and (last_candle["change_pct_1d"] < -0.02)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_5"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.7)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_6"
      elif (
        (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_24"] == True)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_7"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 62.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_8"
      elif (
        (last_candle["rsi_14"] <= 44.0)
        and (last_candle["rsi_3_1h"] <= 10.0)
        and (last_candle["rsi_3_4h"] <= 6.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_9"
      elif (
        (last_candle["rsi_14"] <= 52.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_1h"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_6_10"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.8)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1d"] < -0.10)
        and (last_candle["not_downtrend_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_11"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 66.0)
        and (last_candle["cti_20_1d"] >= 0.8)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_12"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["change_pct_4h"] < -0.03)
        and (last_candle["top_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_13"
      elif (
        (last_candle["rsi_14"] <= 52.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_3_1d"] <= 6.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_14"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] >= 64.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_15"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 66.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] >= 0.70)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_16"
      elif (
        (last_candle["r_14"] >= -22.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_6_17"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_18"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_max_6_1d"] >= 85.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.06)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_6_19"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_20"
      elif (
        (last_candle["r_14"] >= -14.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["close"] < (last_candle["high_max_48_1h"] * 0.75))
      ):
        return True, f"exit_{mode_name}_d_6_21"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.01)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.70))
      ):
        return True, f"exit_{mode_name}_d_6_22"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_6_23"
      elif (
        (last_candle["r_14"] >= -16.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["change_pct_1h"] < -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_24"
      elif (
        (last_candle["r_14"] >= -20.0)
        and (last_candle["rsi_14"] >= 62.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["rsi_3_1d"] <= 26.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_25"
    elif 0.08 > current_profit >= 0.07:
      if (
        (last_candle["r_14"] > -16.0)
        and (last_candle["rsi_14"] > 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_7_1"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_7_2"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["cti_20_1d"] >= 0.50)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["change_pct_4h"] <= -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_3"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1h"] < -0.03)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_4"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_3"] >= 96.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["change_pct_1d"] < -0.02)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_5"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.7)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_6"
      elif (
        (last_candle["rsi_3"] >= 95.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_24"] == True)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_7"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 64.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_8"
      elif (
        (last_candle["rsi_14"] <= 42.0)
        and (last_candle["rsi_3_1h"] <= 10.0)
        and (last_candle["rsi_3_4h"] <= 6.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_9"
      elif (
        (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_1h"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_7_10"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["cti_20_1d"] >= 0.8)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1d"] < -0.10)
        and (last_candle["not_downtrend_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_11"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_12"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["change_pct_4h"] < -0.03)
        and (last_candle["top_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_13"
      elif (
        (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_3_1d"] <= 6.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_14"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 66.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_15"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] >= 0.70)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_16"
      elif (
        (last_candle["r_14"] >= -14.0)
        and (last_candle["rsi_14"] >= 62.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_7_17"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_18"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_max_6_1d"] >= 85.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.06)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_7_19"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_20"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["close"] < (last_candle["high_max_48_1h"] * 0.75))
      ):
        return True, f"exit_{mode_name}_d_7_21"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.01)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.70))
      ):
        return True, f"exit_{mode_name}_d_7_22"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_7_23"
      elif (
        (last_candle["r_14"] >= -14.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["change_pct_1h"] < -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_24"
      elif (
        (last_candle["r_14"] >= -15.0)
        and (last_candle["rsi_14"] >= 64.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["rsi_3_1d"] <= 26.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_25"
    elif 0.09 > current_profit >= 0.08:
      if (
        (last_candle["r_14"] > -16.0)
        and (last_candle["rsi_14"] > 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_8_1"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_8_2"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["cti_20_1d"] >= 0.50)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["change_pct_4h"] <= -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_3"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1h"] < -0.03)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_4"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 98.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["change_pct_1d"] < -0.02)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_5"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_3"] >= 95.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.7)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_6"
      elif (
        (last_candle["rsi_3"] >= 98.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_24"] == True)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_7"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 66.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_8"
      elif (
        (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_3_1h"] <= 10.0)
        and (last_candle["rsi_3_4h"] <= 6.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_9"
      elif (
        (last_candle["rsi_14"] <= 48.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_1h"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_8_10"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 64.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1d"] < -0.10)
        and (last_candle["not_downtrend_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_11"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["cti_20_1d"] >= 0.8)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_12"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["change_pct_4h"] < -0.03)
        and (last_candle["top_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_13"
      elif (
        (last_candle["rsi_14"] <= 48.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_3_1d"] <= 6.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_14"
      elif (
        (last_candle["r_14"] >= -5.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_15"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] >= 0.70)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_16"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 64.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_8_17"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_18"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_max_6_1d"] >= 85.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.06)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_8_19"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_20"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["close"] < (last_candle["high_max_48_1h"] * 0.75))
      ):
        return True, f"exit_{mode_name}_d_8_21"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.01)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.70))
      ):
        return True, f"exit_{mode_name}_d_8_22"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_8_23"
      elif (
        (last_candle["r_14"] >= -12.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["change_pct_1h"] < -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_24"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_14"] >= 66.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["rsi_3_1d"] <= 26.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_25"
    elif 0.1 > current_profit >= 0.09:
      if (
        (last_candle["r_14"] > -16.0)
        and (last_candle["rsi_14"] > 52.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_9_1"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] > 0.5)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_9_2"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["cti_20_1d"] >= 0.50)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["change_pct_4h"] <= -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_3"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1h"] < -0.03)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_4"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 98.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["change_pct_1d"] < -0.02)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_5"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 98.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.70)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_6"
      elif (
        (last_candle["rsi_3"] >= 99.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_24"] == True)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_7"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_8"
      elif (
        (last_candle["rsi_14"] <= 38.0)
        and (last_candle["rsi_3_1h"] <= 10.0)
        and (last_candle["rsi_3_4h"] <= 6.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_9"
      elif (
        (last_candle["rsi_14"] <= 46.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_1h"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_9_10"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 64.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1d"] < -0.10)
        and (last_candle["not_downtrend_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_11"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_12"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["change_pct_4h"] < -0.03)
        and (last_candle["top_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_13"
      elif (
        (last_candle["rsi_14"] <= 46.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_3_1d"] <= 6.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_14"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_15"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 72.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] >= 0.70)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_16"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 66.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_9_17"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_18"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_max_6_1d"] >= 85.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.06)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_9_19"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_20"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["close"] < (last_candle["high_max_48_1h"] * 0.75))
      ):
        return True, f"exit_{mode_name}_d_9_21"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.01)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.70))
      ):
        return True, f"exit_{mode_name}_d_9_22"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_9_23"
      elif (
        (last_candle["r_14"] >= -10.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["change_pct_1h"] < -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_24"
      elif (
        (last_candle["r_14"] >= -8.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["rsi_3_1d"] <= 26.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_25"
    elif 0.12 > current_profit >= 0.1:
      if (
        (last_candle["r_14"] > -16.0)
        and (last_candle["rsi_14"] > 54.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_10_1"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_10_2"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["cti_20_1d"] >= 0.50)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["change_pct_4h"] <= -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_3"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_1d"] > 0.5)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1h"] < -0.03)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_4"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 98.0)
        and (last_candle["rsi_14"] >= 76.0)
        and (last_candle["change_pct_1d"] < -0.02)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_5"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 98.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_6"
      elif (
        (last_candle["rsi_3"] >= 99.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_24"] == True)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_7"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_8"
      elif (
        (last_candle["rsi_14"] <= 36.0)
        and (last_candle["rsi_3_1h"] <= 10.0)
        and (last_candle["rsi_3_4h"] <= 6.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_9"
      elif (
        (last_candle["rsi_14"] <= 44.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_1h"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_10_10"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["cti_20_1d"] >= 0.8)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1d"] < -0.10)
        and (last_candle["not_downtrend_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_11"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 76.0)
        and (last_candle["cti_20_1d"] >= 0.8)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_12"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["change_pct_4h"] < -0.03)
        and (last_candle["top_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_13"
      elif (
        (last_candle["rsi_14"] <= 44.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_3_1d"] <= 6.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_14"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 72.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_15"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] >= 0.70)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_16"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 68.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_10_17"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_18"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_max_6_1d"] >= 85.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.06)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_10_19"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_20"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["close"] < (last_candle["high_max_48_1h"] * 0.75))
      ):
        return True, f"exit_{mode_name}_d_10_21"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.01)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.70))
      ):
        return True, f"exit_{mode_name}_d_10_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_10_23"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["change_pct_1h"] < -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_24"
      elif (
        (last_candle["r_14"] >= -6.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["rsi_3_1d"] <= 26.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_25"
    elif 0.2 > current_profit >= 0.12:
      if (
        (last_candle["r_14"] > -16.0)
        and (last_candle["rsi_14"] > 56.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_11_1"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_11_2"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["cti_20_1d"] >= 0.50)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["change_pct_4h"] <= -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_3"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 60.0)
        and (last_candle["change_pct_1h"] < -0.03)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_4"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 98.0)
        and (last_candle["rsi_14"] >= 78.0)
        and (last_candle["change_pct_1d"] < -0.02)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_5"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 98.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_6"
      elif (
        (last_candle["rsi_3"] >= 99.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_24"] == True)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_7"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_8"
      elif (
        (last_candle["rsi_14"] <= 34.0)
        and (last_candle["rsi_3_1h"] <= 10.0)
        and (last_candle["rsi_3_4h"] <= 6.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_9"
      elif (
        (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_1h"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_11_10"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["cti_20_1d"] >= 0.80)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1d"] < -0.10)
        and (last_candle["not_downtrend_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_11"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 78.0)
        and (last_candle["cti_20_1d"] >= 0.8)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_12"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["change_pct_4h"] < -0.03)
        and (last_candle["top_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_13"
      elif (
        (last_candle["rsi_14"] <= 42.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_3_1d"] <= 6.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_14"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_15"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 76.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] >= 0.70)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_16"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_11_17"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_18"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_max_6_1d"] >= 85.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.06)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_11_19"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_20"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["close"] < (last_candle["high_max_48_1h"] * 0.75))
      ):
        return True, f"exit_{mode_name}_d_11_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.01)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.70))
      ):
        return True, f"exit_{mode_name}_d_11_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_11_23"
      elif (
        (last_candle["r_14"] >= -4.0)
        and (last_candle["rsi_3"] >= 80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["change_pct_1h"] < -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_24"
      elif (
        (last_candle["r_14"] >= -2.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["rsi_3_1d"] <= 26.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_25"
    elif current_profit >= 0.2:
      if (
        (last_candle["r_14"] > -10.0)
        and (last_candle["rsi_14"] > 66.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_12_1"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == True)
        and (last_candle["change_pct_4h"] < -0.03)
      ):
        return True, f"exit_{mode_name}_d_12_2"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_4h"] >= 50.0)
        and (last_candle["cti_20_1d"] >= 0.50)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["change_pct_4h"] <= -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_3"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_1d"] > 0.50)
        and (last_candle["rsi_14_1d"] >= 70.0)
        and (last_candle["change_pct_1h"] < -0.03)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_4"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 98.0)
        and (last_candle["rsi_14"] >= 80.0)
        and (last_candle["change_pct_1d"] < -0.02)
        and (last_candle["change_pct_4h"] < -0.02)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_5"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 98.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.80)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_6"
      elif (
        (last_candle["rsi_3"] >= 99.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_24"] == True)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_7"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 76.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_8"
      elif (
        (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_3_1h"] <= 10.0)
        and (last_candle["rsi_3_4h"] <= 6.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_9"
      elif (
        (last_candle["rsi_14"] <= 36.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["rsi_14_4h"] >= 60.0)
        and (last_candle["change_pct_1h"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_12_10"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["cti_20_1d"] >= 0.8)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["change_pct_1d"] < -0.10)
        and (last_candle["not_downtrend_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_11"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 80.0)
        and (last_candle["cti_20_1d"] >= 0.8)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_12"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 74.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["change_pct_4h"] < -0.03)
        and (last_candle["top_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_13"
      elif (
        (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_3_1d"] <= 6.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_14"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 76.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["r_480_4h"] >= -30.0)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_15"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 78.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["cti_20_1d"] >= 0.70)
        and (last_candle["change_pct_1d"] < -0.04)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_16"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 72.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_14_1h"] >= 60.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] < -0.04)
      ):
        return True, f"exit_{mode_name}_d_12_17"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_24_15m"] == True)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_24_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_18"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_max_6_1d"] >= 85.0)
        and (last_candle["change_pct_1h"] < -0.01)
        and (last_candle["change_pct_4h"] < -0.06)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_12_19"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_max_6_1d"] >= 70.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_20"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == True)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["close"] < (last_candle["high_max_48_1h"] * 0.75))
      ):
        return True, f"exit_{mode_name}_d_12_21"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_max_6_4h"] >= 70.0)
        and (last_candle["change_pct_4h"] < -0.01)
        and (last_candle["not_downtrend_4h"] == False)
        and (last_candle["close"] < (last_candle["high_max_24_1h"] * 0.70))
      ):
        return True, f"exit_{mode_name}_d_12_22"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 70.0)
        and (last_candle["not_downtrend_1h"] == False)
        and (last_candle["not_downtrend_1d"] == False)
        and (last_candle["close"] < (last_candle["high_max_6_1d"] * 0.80))
      ):
        return True, f"exit_{mode_name}_d_12_23"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_3"] >= 95.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_3_1h"] <= 20.0)
        and (last_candle["change_pct_1h"] < -0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_24"
      elif (
        (last_candle["r_14"] >= -1.0)
        and (last_candle["rsi_14"] >= 78.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_1d"] >= 50.0)
        and (last_candle["rsi_3_1d"] <= 26.0)
        and (last_candle["ema_200_dec_48_1h"] == True)
        and (last_candle["ema_200_dec_4_1d"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_25"

    return False, None


  def long_exit_stoploss(
    self,
    mode_name: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    buy_tag,
  ) -> tuple:
    is_backtest = self.dp.runmode.value in ["backtest", "hyperopt"]

    if (
      self.is_futures_mode is False
      and profit_stake
      < -(filled_entries[0].cost * self.stop_threshold / (trade.leverage if self.is_futures_mode else 1.0))

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 1) or is_backtest)
    ):
      return True, f"exit_{mode_name}_stoploss"

    if (
      self.is_futures_mode is True
      and profit_stake
      < -(filled_entries[0].cost * self.stop_threshold_futures / (trade.leverage if self.is_futures_mode else 1.0))

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2023, 10, 17) or is_backtest)
    ):
      return True, f"exit_{mode_name}_stoploss"

    return False, None















  def long_grind_adjust_trade_position(
    self,
    trade: Trade,
    enter_tags,
    current_time: datetime,
    current_rate: float,
    current_profit: float,
    min_stake: Optional[float],
    max_stake: float,
    current_entry_rate: float,
    current_exit_rate: float,
    current_entry_profit: float,
    current_exit_profit: float,
    **kwargs,
  ):
    is_backtest = self.dp.runmode.value in ["backtest", "hyperopt"]
    df, _ = self.dp.get_analyzed_dataframe(trade.pair, self.timeframe)
    if len(df) < 2:
      return None
    last_candle = df.iloc[-1].squeeze()
    previous_candle = df.iloc[-2].squeeze()

    filled_orders = trade.select_filled_orders()
    filled_entries = trade.select_filled_orders(trade.entry_side)
    filled_exits = trade.select_filled_orders(trade.exit_side)
    count_of_entries = trade.nr_of_successful_entries
    count_of_exits = trade.nr_of_successful_exits

    if count_of_entries == 0:
      return None

    if len(filled_orders) < 1:
      return None
    has_order_tags = False
    if hasattr(filled_orders[0], "ft_order_tag"):
      has_order_tags = True

    exit_rate = current_rate
    if self.dp.runmode.value in ("live", "dry_run"):
      ticker = self.dp.ticker(trade.pair)
      if ("bid" in ticker) and ("ask" in ticker):
        if trade.is_short:
          if self.config["exit_pricing"]["price_side"] in ["ask", "other"]:
            if ticker["ask"] is not None:
              exit_rate = ticker["ask"]
        else:
          if self.config["exit_pricing"]["price_side"] in ["bid", "other"]:
            if ticker["bid"] is not None:
              exit_rate = ticker["bid"]

    profit_stake, profit_ratio, profit_current_stake_ratio, profit_init_ratio = self.calc_total_profit(
      trade, filled_entries, filled_exits, exit_rate
    )

    slice_amount = filled_entries[0].cost
    slice_profit = (exit_rate - filled_orders[-1].safe_price) / filled_orders[-1].safe_price
    slice_profit_entry = (exit_rate - filled_entries[-1].safe_price) / filled_entries[-1].safe_price
    slice_profit_exit = (
      ((exit_rate - filled_exits[-1].safe_price) / filled_exits[-1].safe_price) if count_of_exits > 0 else 0.0
    )

    current_stake_amount = trade.amount * current_rate
    is_derisk = trade.amount < (filled_entries[0].safe_filled * 0.95)
    is_derisk_calc = False
    is_rebuy_mode = all(c in self.long_rebuy_mode_tags for c in enter_tags) or (
      any(c in self.long_rebuy_mode_tags for c in enter_tags)
      and all(c in (self.long_rebuy_mode_tags + self.long_grind_mode_tags) for c in enter_tags)
    )
    is_grind_mode = all(c in self.long_grind_mode_tags for c in enter_tags)

    if is_rebuy_mode:
      slice_amount /= self.rebuy_mode_stake_multiplier

    elif is_grind_mode:
      slice_amount /= (
        self.grind_mode_stake_multiplier_futures[0]
        if self.is_futures_mode
        else self.grind_mode_stake_multiplier_spot[0]
      )
    elif not is_derisk and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 2, 5) or is_backtest):
      rebuy_stake, order_tag, is_derisk_calc = self.long_adjust_trade_position_no_derisk(
        trade,
        enter_tags,
        current_time,
        current_rate,
        current_profit,
        min_stake,
        max_stake,
        current_entry_rate,
        current_exit_rate,
        current_entry_profit,
        current_exit_profit,
        last_candle,
        previous_candle,
        filled_orders,
        filled_entries,
        filled_exits,
        exit_rate,
        slice_amount,
        slice_profit_entry,
        slice_profit,
        profit_ratio,
        profit_stake,
        profit_init_ratio,
        current_stake_amount,
        has_order_tags,
      )
      if rebuy_stake is not None:
        if has_order_tags:
          return rebuy_stake, order_tag
        else:
          return rebuy_stake
      elif count_of_exits == 0:
        return None
      elif not is_derisk_calc:
        return None

    if not is_rebuy_mode and not is_grind_mode:

      if trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 16) or is_backtest:
        slice_amount /= (
          self.regular_mode_stake_multiplier_futures[0]
          if self.is_futures_mode
          else self.regular_mode_stake_multiplier_spot[0]
        )

    grind_derisk = self.grind_derisk_futures if self.is_futures_mode else self.grind_derisk_spot

    grind_1_max_sub_grinds = 0
    grind_1_stakes = self.grind_1_stakes_futures.copy() if self.is_futures_mode else self.grind_1_stakes_spot.copy()
    grind_1_sub_thresholds = (
      self.grind_1_sub_thresholds_futures if self.is_futures_mode else self.grind_1_sub_thresholds_spot
    )
    if (slice_amount * grind_1_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = min_stake / slice_amount / grind_1_stakes[0]
      for i, _ in enumerate(grind_1_stakes):
        grind_1_stakes[i] *= multi
    grind_1_max_sub_grinds = len(grind_1_stakes)
    grind_1_stop_grinds = self.grind_1_stop_grinds_futures if self.is_futures_mode else self.grind_1_stop_grinds_spot
    grind_1_profit_threshold = (
      self.grind_1_profit_threshold_futures if self.is_futures_mode else self.grind_1_profit_threshold_spot
    )

    grind_2_max_sub_grinds = 0
    grind_2_stakes = self.grind_2_stakes_futures.copy() if self.is_futures_mode else self.grind_2_stakes_spot.copy()
    grind_2_sub_thresholds = (
      self.grind_2_sub_thresholds_futures if self.is_futures_mode else self.grind_2_sub_thresholds_spot
    )
    if (slice_amount * grind_2_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = min_stake / slice_amount / grind_2_stakes[0]
      for i, _ in enumerate(grind_2_stakes):
        grind_2_stakes[i] *= multi
    grind_2_max_sub_grinds = len(grind_2_stakes)
    grind_2_stop_grinds = self.grind_2_stop_grinds_futures if self.is_futures_mode else self.grind_2_stop_grinds_spot
    grind_2_profit_threshold = (
      self.grind_2_profit_threshold_futures if self.is_futures_mode else self.grind_2_profit_threshold_spot
    )

    grind_3_max_sub_grinds = 0
    grind_3_stakes = self.grind_3_stakes_futures.copy() if self.is_futures_mode else self.grind_3_stakes_spot.copy()
    grind_3_sub_thresholds = (
      self.grind_3_sub_thresholds_futures if self.is_futures_mode else self.grind_3_sub_thresholds_spot
    )
    if (slice_amount * grind_3_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = min_stake / slice_amount / grind_3_stakes[0]
      for i, _ in enumerate(grind_3_stakes):
        grind_3_stakes[i] *= multi
    grind_3_max_sub_grinds = len(grind_3_stakes)
    grind_3_stop_grinds = self.grind_3_stop_grinds_futures if self.is_futures_mode else self.grind_3_stop_grinds_spot
    grind_3_profit_threshold = (
      self.grind_3_profit_threshold_futures if self.is_futures_mode else self.grind_3_profit_threshold_spot
    )

    grind_4_max_sub_grinds = 0
    grind_4_stakes = self.grind_4_stakes_futures.copy() if self.is_futures_mode else self.grind_4_stakes_spot.copy()
    grind_4_sub_thresholds = (
      self.grind_4_sub_thresholds_futures if self.is_futures_mode else self.grind_4_sub_thresholds_spot
    )
    if (slice_amount * grind_4_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = min_stake / slice_amount / grind_4_stakes[0]
      for i, _ in enumerate(grind_4_stakes):
        grind_4_stakes[i] *= multi
    grind_4_max_sub_grinds = len(grind_4_stakes)
    grind_4_stop_grinds = self.grind_4_stop_grinds_futures if self.is_futures_mode else self.grind_4_stop_grinds_spot
    grind_4_profit_threshold = (
      self.grind_4_profit_threshold_futures if self.is_futures_mode else self.grind_4_profit_threshold_spot
    )

    grind_5_max_sub_grinds = 0
    grind_5_stakes = self.grind_5_stakes_futures.copy() if self.is_futures_mode else self.grind_5_stakes_spot.copy()
    grind_5_sub_thresholds = (
      self.grind_5_sub_thresholds_futures if self.is_futures_mode else self.grind_5_sub_thresholds_spot
    )
    if (slice_amount * grind_5_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = min_stake / slice_amount / grind_5_stakes[0]
      for i, _ in enumerate(grind_5_stakes):
        grind_5_stakes[i] *= multi
    grind_5_max_sub_grinds = len(grind_5_stakes)
    grind_5_stop_grinds = self.grind_5_stop_grinds_futures if self.is_futures_mode else self.grind_5_stop_grinds_spot
    grind_5_profit_threshold = (
      self.grind_5_profit_threshold_futures if self.is_futures_mode else self.grind_5_profit_threshold_spot
    )

    grind_6_max_sub_grinds = 0
    grind_6_stakes = self.grind_6_stakes_futures.copy() if self.is_futures_mode else self.grind_6_stakes_spot.copy()
    grind_6_sub_thresholds = (
      self.grind_6_sub_thresholds_futures if self.is_futures_mode else self.grind_6_sub_thresholds_spot
    )
    if (slice_amount * grind_6_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = min_stake / slice_amount / grind_6_stakes[0]
      for i, _ in enumerate(grind_6_stakes):
        grind_6_stakes[i] *= multi
    grind_6_max_sub_grinds = len(grind_6_stakes)
    grind_6_stop_grinds = self.grind_6_stop_grinds_futures if self.is_futures_mode else self.grind_6_stop_grinds_spot
    grind_6_profit_threshold = (
      self.grind_6_profit_threshold_futures if self.is_futures_mode else self.grind_6_profit_threshold_spot
    )

    grind_1_derisk_1_max_sub_grinds = 0
    grind_1_derisk_1_stakes = (
      self.grind_1_derisk_1_stakes_futures.copy() if self.is_futures_mode else self.grind_1_derisk_1_stakes_spot.copy()
    )
    grind_1_derisk_1_sub_thresholds = (
      self.grind_1_derisk_1_sub_thresholds_futures
      if self.is_futures_mode
      else self.grind_1_derisk_1_sub_thresholds_spot
    )
    if (slice_amount * grind_1_derisk_1_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = min_stake / slice_amount / grind_1_derisk_1_stakes[0]
      for i, _ in enumerate(grind_1_derisk_1_stakes):
        grind_1_derisk_1_stakes[i] *= multi
    grind_1_derisk_1_max_sub_grinds = len(grind_1_derisk_1_stakes)
    grind_1_derisk_1_stop_grinds = (
      self.grind_1_derisk_1_stop_grinds_futures if self.is_futures_mode else self.grind_1_derisk_1_stop_grinds_spot
    )
    grind_1_derisk_1_profit_threshold = (
      self.grind_1_derisk_1_profit_threshold_futures
      if self.is_futures_mode
      else self.grind_1_derisk_1_profit_threshold_spot
    )

    grind_2_derisk_1_max_sub_grinds = 0
    grind_2_derisk_1_stakes = (
      self.grind_2_derisk_1_stakes_futures.copy() if self.is_futures_mode else self.grind_2_derisk_1_stakes_spot.copy()
    )
    grind_2_derisk_1_sub_thresholds = (
      self.grind_2_derisk_1_sub_thresholds_futures
      if self.is_futures_mode
      else self.grind_2_derisk_1_sub_thresholds_spot
    )
    if (slice_amount * grind_2_derisk_1_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = min_stake / slice_amount / grind_2_derisk_1_stakes[0]
      for i, _ in enumerate(grind_2_derisk_1_stakes):
        grind_2_derisk_1_stakes[i] *= multi
    grind_2_derisk_1_max_sub_grinds = len(grind_2_derisk_1_stakes)
    grind_2_derisk_1_stop_grinds = (
      self.grind_2_derisk_1_stop_grinds_futures if self.is_futures_mode else self.grind_2_derisk_1_stop_grinds_spot
    )
    grind_2_derisk_1_profit_threshold = (
      self.grind_2_derisk_1_profit_threshold_futures
      if self.is_futures_mode
      else self.grind_2_derisk_1_profit_threshold_spot
    )

    partial_sell = False
    is_derisk_found = False  # d de-risk
    is_derisk_1 = False
    is_derisk_1_found = False  # d1 de-risk exit
    derisk_1_order = None
    derisk_1_reentry_order = None
    derisk_1_sub_grind_count = 0
    derisk_1_total_amount = 0.0
    derisk_1_total_cost = 0.0
    derisk_1_current_open_rate = 0.0
    derisk_1_current_grind_stake = 0.0
    derisk_1_current_grind_stake_profit = 0.0
    derisk_1_is_sell_found = False
    derisk_1_reentry_found = False
    derisk_1_buy_orders = []
    derisk_1_distance_ratio = 0.0
    grind_1_sub_grind_count = 0
    grind_1_total_amount = 0.0
    grind_1_total_cost = 0.0
    grind_1_current_open_rate = 0.0
    grind_1_current_grind_stake = 0.0
    grind_1_current_grind_stake_profit = 0.0
    grind_1_is_sell_found = False
    grind_1_found = False
    grind_1_buy_orders = []
    grind_1_distance_ratio = 0.0
    grind_2_sub_grind_count = 0
    grind_2_total_amount = 0.0
    grind_2_total_cost = 0.0
    grind_2_current_open_rate = 0.0
    grind_2_current_grind_stake = 0.0
    grind_2_current_grind_stake_profit = 0.0
    grind_2_is_sell_found = False
    grind_2_found = False
    grind_2_buy_orders = []
    grind_2_distance_ratio = 0.0
    grind_3_sub_grind_count = 0
    grind_3_total_amount = 0.0
    grind_3_total_cost = 0.0
    grind_3_current_open_rate = 0.0
    grind_3_current_grind_stake = 0.0
    grind_3_current_grind_stake_profit = 0.0
    grind_3_is_sell_found = False
    grind_3_found = False
    grind_3_buy_orders = []
    grind_3_distance_ratio = 0.0
    grind_4_sub_grind_count = 0
    grind_4_total_amount = 0.0
    grind_4_total_cost = 0.0
    grind_4_current_open_rate = 0.0
    grind_4_current_grind_stake = 0.0
    grind_4_current_grind_stake_profit = 0.0
    grind_4_is_sell_found = False
    grind_4_found = False
    grind_4_buy_orders = []
    grind_4_distance_ratio = 0.0
    grind_5_sub_grind_count = 0
    grind_5_total_amount = 0.0
    grind_5_total_cost = 0.0
    grind_5_current_open_rate = 0.0
    grind_5_current_grind_stake = 0.0
    grind_5_current_grind_stake_profit = 0.0
    grind_5_is_sell_found = False
    grind_5_found = False
    grind_5_buy_orders = []
    grind_5_distance_ratio = 0.0
    grind_6_sub_grind_count = 0
    grind_6_total_amount = 0.0
    grind_6_total_cost = 0.0
    grind_6_current_open_rate = 0.0
    grind_6_current_grind_stake = 0.0
    grind_6_current_grind_stake_profit = 0.0
    grind_6_is_sell_found = False
    grind_6_found = False
    grind_6_buy_orders = []
    grind_6_distance_ratio = 0.0
    grind_1_derisk_1_sub_grind_count = 0
    grind_1_derisk_1_total_amount = 0.0
    grind_1_derisk_1_total_cost = 0.0
    grind_1_derisk_1_current_open_rate = 0.0
    grind_1_derisk_1_current_grind_stake = 0.0
    grind_1_derisk_1_current_grind_stake_profit = 0.0
    grind_1_derisk_1_is_sell_found = False
    grind_1_derisk_1_found = False
    grind_1_derisk_1_buy_orders = []
    grind_1_derisk_1_distance_ratio = 0.0
    grind_2_derisk_1_sub_grind_count = 0
    grind_2_derisk_1_total_amount = 0.0
    grind_2_derisk_1_total_cost = 0.0
    grind_2_derisk_1_current_open_rate = 0.0
    grind_2_derisk_1_current_grind_stake = 0.0
    grind_2_derisk_1_current_grind_stake_profit = 0.0
    grind_2_derisk_1_is_sell_found = False
    grind_2_derisk_1_found = False
    grind_2_derisk_1_buy_orders = []
    grind_2_derisk_1_distance_ratio = 0.0
    for order in reversed(filled_orders):
      if (order.ft_order_side == "buy") and (order is not filled_orders[0]):
        order_tag = ""
        if has_order_tags:
          if order.ft_order_tag is not None:
            order_tag = order.ft_order_tag
        if not is_derisk_1 and order_tag == "d1":
          derisk_1_sub_grind_count += 1
          derisk_1_total_amount += order.safe_filled
          derisk_1_total_cost += order.safe_filled * order.safe_price
          derisk_1_buy_orders.append(order.id)
          if not derisk_1_reentry_found and not is_derisk_1:
            derisk_1_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            derisk_1_reentry_found = True
            derisk_1_reentry_order = order
        elif not grind_1_derisk_1_is_sell_found and order_tag == "dl1":
          grind_1_derisk_1_sub_grind_count += 1
          grind_1_derisk_1_total_amount += order.safe_filled
          grind_1_derisk_1_total_cost += order.safe_filled * order.safe_price
          grind_1_derisk_1_buy_orders.append(order.id)
          if not grind_1_derisk_1_found:
            grind_1_derisk_1_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_1_derisk_1_found = True
        elif not grind_2_derisk_1_is_sell_found and order_tag == "dl2":
          grind_2_derisk_1_sub_grind_count += 1
          grind_2_derisk_1_total_amount += order.safe_filled
          grind_2_derisk_1_total_cost += order.safe_filled * order.safe_price
          grind_2_derisk_1_buy_orders.append(order.id)
          if not grind_2_derisk_1_found:
            grind_2_derisk_1_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_2_derisk_1_found = True
        elif not grind_6_is_sell_found and order_tag == "gd6":
          grind_6_sub_grind_count += 1
          grind_6_total_amount += order.safe_filled
          grind_6_total_cost += order.safe_filled * order.safe_price
          grind_6_buy_orders.append(order.id)
          if not grind_6_found:
            grind_6_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_6_found = True
        elif not grind_5_is_sell_found and order_tag == "gd5":
          grind_5_sub_grind_count += 1
          grind_5_total_amount += order.safe_filled
          grind_5_total_cost += order.safe_filled * order.safe_price
          grind_5_buy_orders.append(order.id)
          if not grind_5_found:
            grind_5_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_5_found = True
        elif not grind_4_is_sell_found and order_tag == "gd4":
          grind_4_sub_grind_count += 1
          grind_4_total_amount += order.safe_filled
          grind_4_total_cost += order.safe_filled * order.safe_price
          grind_4_buy_orders.append(order.id)
          if not grind_4_found:
            grind_4_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_4_found = True
        elif not grind_3_is_sell_found and order_tag == "gd3":
          grind_3_sub_grind_count += 1
          grind_3_total_amount += order.safe_filled
          grind_3_total_cost += order.safe_filled * order.safe_price
          grind_3_buy_orders.append(order.id)
          if not grind_3_found:
            grind_3_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_3_found = True
        elif not grind_2_is_sell_found and order_tag == "gd2":
          grind_2_sub_grind_count += 1
          grind_2_total_amount += order.safe_filled
          grind_2_total_cost += order.safe_filled * order.safe_price
          grind_2_buy_orders.append(order.id)
          if not grind_2_found:
            grind_2_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_2_found = True
        elif not grind_1_is_sell_found and order_tag not in [
          "r",
          "d1",
          "dl1",
          "dl2",
          "g1",
          "g2",
          "g3",
          "g4",
          "g5",
          "g6",
          "sg1",
          "sg2",
          "sg3",
          "sg4",
          "sg5",
          "sg6",
          "gd2",
          "gd3",
          "gd4",
          "gd5",
          "gd6",
          "gm0",
          "gmd0",
        ]:
          grind_1_sub_grind_count += 1
          grind_1_total_amount += order.safe_filled
          grind_1_total_cost += order.safe_filled * order.safe_price
          grind_1_buy_orders.append(order.id)
          if not grind_1_found:
            grind_1_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_1_found = True
      elif order.ft_order_side == "sell":
        if (
          order is filled_exits[-1]
          and (order.safe_remaining * exit_rate / (trade.leverage if self.is_futures_mode else 1.0)) > min_stake
        ):
          partial_sell = True
          break
        order_tag = ""
        if has_order_tags:
          if order.ft_order_tag is not None:
            sell_order_tag = order.ft_order_tag
            order_mode = sell_order_tag.split(" ", 1)
            if len(order_mode) > 0:
              order_tag = order_mode[0]
        if order_tag in ["dl1", "ddl1"]:
          grind_1_derisk_1_is_sell_found = True
        elif order_tag in ["dl2", "ddl2"]:
          grind_2_derisk_1_is_sell_found = True
        elif order_tag in ["gd6", "dd6"]:
          grind_6_is_sell_found = True
        elif order_tag in ["gd5", "dd5"]:
          grind_5_is_sell_found = True
        if order_tag in ["gd4", "dd4"]:
          grind_4_is_sell_found = True
        elif order_tag in ["gd3", "dd3"]:
          grind_3_is_sell_found = True
        elif order_tag in ["gd2", "dd2"]:
          grind_2_is_sell_found = True
        elif order_tag in ["d1"]:
          if not is_derisk_1_found:
            is_derisk_1_found = True
            is_derisk_1 = True
            derisk_1_order = order
        elif order_tag in ["p", "r", "d", "dd0", "partial_exit", ""]:
          if order_tag in ["d"]:
            is_derisk_found = True
            is_derisk = True
          grind_1_is_sell_found = True
          grind_2_is_sell_found = True
          grind_3_is_sell_found = True
          grind_4_is_sell_found = True
          grind_5_is_sell_found = True
          grind_6_is_sell_found = True
          grind_1_derisk_1_is_sell_found = True
          grind_2_derisk_1_is_sell_found = True
        elif order_tag not in [
          "dl1",
          "ddl1",
          "dl2",
          "ddl2",
          "g1",
          "g2",
          "g3",
          "g4",
          "g5",
          "g6",
          "sg1",
          "sg2",
          "sg3",
          "sg4",
          "sg5",
          "sg6",
          "gd2",
          "gd3",
          "gd4",
          "gd5",
          "gd6",
          "dd2",
          "dd3",
          "dd4",
          "dd5",
          "dd6",
          "gm0",
          "gmd0",
        ]:
          grind_1_is_sell_found = True

    if derisk_1_sub_grind_count > 0:
      derisk_1_current_open_rate = derisk_1_total_cost / derisk_1_total_amount
      derisk_1_current_grind_stake = derisk_1_total_amount * exit_rate * (1 - trade.fee_close)
      derisk_1_current_grind_stake_profit = derisk_1_current_grind_stake - derisk_1_total_cost
    if grind_1_sub_grind_count > 0:
      grind_1_current_open_rate = grind_1_total_cost / grind_1_total_amount
      grind_1_current_grind_stake = grind_1_total_amount * exit_rate * (1 - trade.fee_close)
      grind_1_current_grind_stake_profit = grind_1_current_grind_stake - grind_1_total_cost
    if grind_2_sub_grind_count > 0:
      grind_2_current_open_rate = grind_2_total_cost / grind_2_total_amount
      grind_2_current_grind_stake = grind_2_total_amount * exit_rate * (1 - trade.fee_close)
      grind_2_current_grind_stake_profit = grind_2_current_grind_stake - grind_2_total_cost
    if grind_3_sub_grind_count > 0:
      grind_3_current_open_rate = grind_3_total_cost / grind_3_total_amount
      grind_3_current_grind_stake = grind_3_total_amount * exit_rate * (1 - trade.fee_close)
      grind_3_current_grind_stake_profit = grind_3_current_grind_stake - grind_3_total_cost
    if grind_4_sub_grind_count > 0:
      grind_4_current_open_rate = grind_4_total_cost / grind_4_total_amount
      grind_4_current_grind_stake = grind_4_total_amount * exit_rate * (1 - trade.fee_close)
      grind_4_current_grind_stake_profit = grind_4_current_grind_stake - grind_4_total_cost
    if grind_5_sub_grind_count > 0:
      grind_5_current_open_rate = grind_5_total_cost / grind_5_total_amount
      grind_5_current_grind_stake = grind_5_total_amount * exit_rate * (1 - trade.fee_close)
      grind_5_current_grind_stake_profit = grind_5_current_grind_stake - grind_5_total_cost
    if grind_6_sub_grind_count > 0:
      grind_6_current_open_rate = grind_6_total_cost / grind_6_total_amount
      grind_6_current_grind_stake = grind_6_total_amount * exit_rate * (1 - trade.fee_close)
      grind_6_current_grind_stake_profit = grind_6_current_grind_stake - grind_6_total_cost
    if grind_1_derisk_1_sub_grind_count > 0:
      grind_1_derisk_1_current_open_rate = grind_1_derisk_1_total_cost / grind_1_derisk_1_total_amount
      grind_1_derisk_1_current_grind_stake = grind_1_derisk_1_total_amount * exit_rate * (1 - trade.fee_close)
      grind_1_derisk_1_current_grind_stake_profit = grind_1_derisk_1_current_grind_stake - grind_1_derisk_1_total_cost
    if grind_2_derisk_1_sub_grind_count > 0:
      grind_2_derisk_1_current_open_rate = grind_2_derisk_1_total_cost / grind_2_derisk_1_total_amount
      grind_2_derisk_1_current_grind_stake = grind_2_derisk_1_total_amount * exit_rate * (1 - trade.fee_close)
      grind_2_derisk_1_current_grind_stake_profit = grind_2_derisk_1_current_grind_stake - grind_2_derisk_1_total_cost

    num_open_grinds = (
      grind_1_sub_grind_count
      + grind_2_sub_grind_count
      + grind_3_sub_grind_count
      + grind_4_sub_grind_count
      + grind_5_sub_grind_count
      + grind_6_sub_grind_count
      + grind_1_derisk_1_sub_grind_count
      + grind_2_derisk_1_sub_grind_count
    )

    if partial_sell:
      order = filled_exits[-1]
      sell_amount = order.safe_remaining * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        self.dp.send_msg(
          f"Exit (remaining) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {order.safe_remaining} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        order_tag = "p"
        if has_order_tags:
          if order.ft_order_tag is not None:
            order_tag = order.ft_order_tag
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount

    if is_grind_mode and (
      (filled_entries[0].safe_filled * (trade.stake_amount / trade.amount) - (min_stake * 1.5)) > min_stake
    ):
      is_first_entry_exit_found = False
      for order in filled_orders:
        if order.ft_order_side == "sell":
          order_tag = ""
          if has_order_tags:
            if order.ft_order_tag is not None:
              sell_order_tag = order.ft_order_tag
              order_mode = sell_order_tag.split(" ", 1)
              if len(order_mode) > 0:
                order_tag = order_mode[0]
          else:

            is_first_entry_exit_found = True
          if order_tag in ["gm0", "gmd0"]:
            is_first_entry_exit_found = True
            break
      if not is_first_entry_exit_found:
        first_entry = filled_entries[0]
        first_entry_distance_ratio = (exit_rate - first_entry.safe_price) / first_entry.safe_price

        if first_entry_distance_ratio > (
          self.grind_mode_first_entry_profit_threshold_spot
          if self.is_futures_mode
          else self.grind_mode_first_entry_profit_threshold_spot
        ):
          sell_amount = first_entry.safe_filled * exit_rate / trade.leverage
          if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
            sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
          ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
          if sell_amount > min_stake and ft_sell_amount > min_stake:
            grind_profit = (exit_rate - first_entry.safe_price) / first_entry.safe_price
            coin_amount = sell_amount / exit_rate
            self.dp.send_msg(
              f"Grinding exit (gm0) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {coin_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
            )
            log.info(
              f"Grinding exit (gm0) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {coin_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
            )
            order_tag = "gm0"
            for grind_entry_id in grind_1_buy_orders:
              order_tag += " " + str(grind_entry_id)
            if has_order_tags:
              return -ft_sell_amount, order_tag
            else:
              return -ft_sell_amount

        if first_entry_distance_ratio < (
          self.grind_mode_first_entry_stop_threshold_spot
          if self.is_futures_mode
          else self.grind_mode_first_entry_stop_threshold_spot
        ):
          sell_amount = first_entry.safe_filled * exit_rate / trade.leverage
          if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
            sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
          ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
          if sell_amount > min_stake and ft_sell_amount > min_stake:
            grind_profit = (exit_rate - first_entry.safe_price) / first_entry.safe_price
            coin_amount = sell_amount / exit_rate
            self.dp.send_msg(
              f"Grinding de-risk (gmd0) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {coin_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
            )
            log.info(
              f"Grinding de-risk (gmd0) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {coin_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
            )
            order_tag = "gmd0"
            for grind_entry_id in grind_1_buy_orders:
              order_tag += " " + str(grind_entry_id)
            if has_order_tags:
              return -ft_sell_amount, order_tag
            else:
              return -ft_sell_amount

    is_long_grind_buy = self.long_grind_buy(last_candle, previous_candle, slice_profit)


    if (
      has_order_tags
      and is_derisk_1
      and not derisk_1_reentry_found
      and (not partial_sell)
      and (grind_1_derisk_1_sub_grind_count < grind_1_derisk_1_max_sub_grinds)
    ):
      if (
        (
          (
            (grind_1_derisk_1_sub_grind_count > 0)
            and grind_1_derisk_1_distance_ratio < grind_1_derisk_1_sub_thresholds[grind_1_derisk_1_sub_grind_count]
          )
          or ((is_derisk or is_derisk_calc) and grind_1_derisk_1_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )
        and ((num_open_grinds == 0) or (slice_profit < -0.02))
        and (
          (last_candle["close"] > (last_candle["close_max_12"] * 0.94))
          and (last_candle["close"] > (last_candle["close_max_24"] * 0.92))
          and (last_candle["close"] > (last_candle["close_max_48"] * 0.90))
          and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.88))
          and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.86))
          and (last_candle["close"] > (last_candle["high_max_6_1d"] * 0.84))
          and (last_candle["close"] > (last_candle["high_max_12_1d"] * 0.82))
        )
        and (
          (last_candle["zlma_50_dec_15m"] == False)
          and (last_candle["zlma_50_dec_1h"] == False)
          and (last_candle["zlma_50_dec_4h"] == False)
          and (last_candle["ema_200_dec_48_1h"] == False)
          and (last_candle["ema_200_dec_24_4h"] == False)
        )
        and (
          is_long_grind_buy
          or (
            (last_candle["rsi_3"] > 16.0)
            and (last_candle["rsi_3_15m"] > 16.0)
            and (last_candle["rsi_3_1h"] > 30.0)
            and (last_candle["rsi_3_4h"] > 30.0)
            and (last_candle["rsi_14"] < 36.0)
            and (last_candle["close"] < last_candle["res_hlevel_4h"])
            and (last_candle["close"] > last_candle["sup_level_4h"])
          )
        )
      ):
        buy_amount = (
          slice_amount
          * grind_1_derisk_1_stakes[grind_1_derisk_1_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_1_derisk_1_sub_grind_count > 0:
          grind_profit = (exit_rate - grind_1_derisk_1_current_open_rate) / grind_1_derisk_1_current_open_rate
          grind_profit_stake = grind_1_derisk_1_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (dl1) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_1_derisk_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (dl1) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_1_derisk_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "dl1"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_1_derisk_1_sub_grind_count > 0:
      grind_profit = (exit_rate - grind_1_derisk_1_current_open_rate) / grind_1_derisk_1_current_open_rate
      if grind_profit > grind_1_derisk_1_profit_threshold:
        sell_amount = grind_1_derisk_1_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (dl1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (dl1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "dl1"
          for grind_entry_id in grind_1_derisk_1_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (grind_1_derisk_1_distance_ratio < grind_1_derisk_1_stop_grinds) and (is_derisk or is_derisk_calc):
      sell_amount = grind_1_derisk_1_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_1_derisk_1_current_open_rate > 0.0:
          grind_profit = (
            ((exit_rate - grind_1_derisk_1_current_open_rate) / grind_1_derisk_1_current_open_rate)
            if grind_1_derisk_1_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (ddl1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (ddl1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "ddl1"
        for grind_entry_id in grind_1_derisk_1_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if (
      has_order_tags
      and is_derisk_1
      and not derisk_1_reentry_found
      and (not partial_sell)
      and (grind_2_derisk_1_sub_grind_count < grind_2_derisk_1_max_sub_grinds)
    ):
      if (
        (
          (
            (grind_2_derisk_1_sub_grind_count > 0)
            and grind_2_derisk_1_distance_ratio < grind_2_derisk_1_sub_thresholds[grind_2_derisk_1_sub_grind_count]
          )
          or ((is_derisk or is_derisk_calc) and grind_2_derisk_1_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )
        and ((num_open_grinds == 0) or (slice_profit < -0.02))
        and (
          (last_candle["close"] > (last_candle["close_max_12"] * 0.94))
          and (last_candle["close"] > (last_candle["close_max_24"] * 0.92))
          and (last_candle["close"] > (last_candle["close_max_48"] * 0.90))
          and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.88))
          and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.86))
        )
        and (
          (
            (grind_2_derisk_1_sub_grind_count == 0)
            and (
              is_long_grind_buy
              or (
                (last_candle["rsi_3"] > 16.0)
                and (last_candle["rsi_3_15m"] > 16.0)
                and (last_candle["rsi_3_1h"] > 20.0)
                and (last_candle["rsi_3_4h"] > 20.0)
                and (last_candle["rsi_14"] < 36.0)

                and (last_candle["zlma_50_dec_1h"] == False)
                and (last_candle["zlma_50_dec_4h"] == False)
              )
            )
          )
          or (
            (grind_2_derisk_1_sub_grind_count > 0)
            and (
              is_long_grind_buy
              or (
                (last_candle["rsi_3"] > 12.0)
                and (last_candle["rsi_3_15m"] > 16.0)
                and (last_candle["rsi_3_1h"] > 16.0)
                and (last_candle["rsi_3_4h"] > 16.0)
                and (last_candle["rsi_14"] < 42.0)
              )
            )
          )
        )
      ):
        buy_amount = (
          slice_amount
          * grind_2_derisk_1_stakes[grind_2_derisk_1_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_2_derisk_1_sub_grind_count > 0:
          grind_profit = (exit_rate - grind_2_derisk_1_current_open_rate) / grind_2_derisk_1_current_open_rate
          grind_profit_stake = grind_2_derisk_1_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (dl2) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_2_derisk_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (dl2) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_2_derisk_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "dl2"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_2_derisk_1_sub_grind_count > 0:
      grind_profit = (exit_rate - grind_2_derisk_1_current_open_rate) / grind_2_derisk_1_current_open_rate
      if grind_profit > grind_2_derisk_1_profit_threshold:
        sell_amount = grind_2_derisk_1_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (dl2) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (dl2) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "dl2"
          for grind_entry_id in grind_2_derisk_1_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (grind_2_derisk_1_sub_grind_count > 0)
      and (
        ((exit_rate - grind_2_derisk_1_current_open_rate) / grind_2_derisk_1_current_open_rate)
        < grind_2_derisk_1_stop_grinds
      )
      and (is_derisk or is_derisk_calc)
    ):
      sell_amount = grind_2_derisk_1_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_2_derisk_1_current_open_rate > 0.0:
          grind_profit = (
            ((exit_rate - grind_2_derisk_1_current_open_rate) / grind_2_derisk_1_current_open_rate)
            if grind_2_derisk_1_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (ddl2) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (ddl2) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "ddl2"
        for grind_entry_id in grind_2_derisk_1_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if (not partial_sell) and (grind_1_sub_grind_count < grind_1_max_sub_grinds):
      if (
        (
          ((grind_1_sub_grind_count > 0) and grind_1_distance_ratio < grind_1_sub_thresholds[grind_1_sub_grind_count])
          or ((is_derisk or is_derisk_calc) and grind_1_sub_grind_count == 0)
          or (is_grind_mode and grind_1_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )
        and ((num_open_grinds == 0) or (slice_profit < -0.02))
        and (
          (
            (
              (last_candle["close"] > (last_candle["close_max_12"] * 0.94))
              and (last_candle["close"] > (last_candle["close_max_24"] * 0.92))
              and (last_candle["close"] > (last_candle["close_max_48"] * 0.90))
              and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.88))
              and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.86))
            )
            and (
              (
                (grind_1_sub_grind_count == 0)
                and (
                  is_long_grind_buy
                  or (
                    (last_candle["rsi_3"] > 10.0)
                    and (last_candle["rsi_3_15m"] > 10.0)
                    and (last_candle["rsi_3_1h"] > 10.0)
                    and (last_candle["rsi_3_4h"] > 10.0)
                    and (last_candle["rsi_14"] < 36.0)



                    and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
                  )
                )
              )
              or (
                (grind_1_sub_grind_count > 0)
                and (
                  is_long_grind_buy
                  or (
                    (last_candle["rsi_3"] > 12.0)
                    and (last_candle["rsi_3_15m"] > 12.0)


                    and (last_candle["rsi_14"] < 42.0)
                  )
                )
              )
            )
          )
          or (
            (slice_profit < -0.06)
            and (last_candle["rsi_3"] > 10.0)
            and (last_candle["rsi_3_15m"] > 10.0)
            and (last_candle["rsi_14"] > 28.0)
            and (last_candle["rsi_14"] < 36.0)
            and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
          )
        )
      ):
        buy_amount = (
          slice_amount * grind_1_stakes[grind_1_sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_1_sub_grind_count > 0:
          grind_profit = (exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate
          grind_profit_stake = grind_1_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (gd1) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (gd1) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "gd1"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_1_sub_grind_count > 0:
      grind_profit = (exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate
      if grind_profit > grind_1_profit_threshold:
        sell_amount = grind_1_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (gd1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (gd1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "gd1"
          for grind_entry_id in grind_1_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (
        (grind_1_sub_grind_count > 0)
        and (((exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate) < grind_1_stop_grinds)




        and (is_derisk or is_derisk_calc or is_grind_mode)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 14) or is_backtest)
    ):
      sell_amount = grind_1_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_1_current_open_rate > 0.0:
          grind_profit = (
            ((exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate)
            if grind_1_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (dd1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (dd1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "dd1"
        for grind_entry_id in grind_1_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if has_order_tags and (not partial_sell) and (grind_2_sub_grind_count < grind_2_max_sub_grinds):
      if (
        (
          ((grind_2_sub_grind_count > 0) and grind_2_distance_ratio < grind_2_sub_thresholds[grind_2_sub_grind_count])
          or ((is_derisk or is_derisk_calc) and grind_2_sub_grind_count == 0)
          or (is_grind_mode and grind_2_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )
        and ((num_open_grinds == 0) or (slice_profit < -0.02))
        and (
          (
            (
              (last_candle["close"] > (last_candle["close_max_12"] * 0.94))
              and (last_candle["close"] > (last_candle["close_max_24"] * 0.92))
              and (last_candle["close"] > (last_candle["close_max_48"] * 0.90))
              and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.88))
              and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.86))
            )
            and (
              (
                (grind_2_sub_grind_count == 0)
                and (
                  is_long_grind_buy
                  or (
                    (last_candle["rsi_3"] > 10.0)
                    and (last_candle["rsi_3_15m"] > 10.0)
                    and (last_candle["rsi_3_1h"] > 10.0)
                    and (last_candle["rsi_3_4h"] > 10.0)
                    and (last_candle["rsi_14"] < 36.0)



                    and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
                  )
                )
              )
              or (
                (grind_2_sub_grind_count > 0)
                and (
                  is_long_grind_buy
                  or (
                    (last_candle["rsi_3"] > 12.0)
                    and (last_candle["rsi_3_15m"] > 12.0)


                    and (last_candle["rsi_14"] < 42.0)
                  )
                )
              )
            )
          )
          or (
            (slice_profit < -0.06)
            and (last_candle["rsi_3"] > 10.0)
            and (last_candle["rsi_3_15m"] > 10.0)
            and (last_candle["rsi_14"] > 28.0)
            and (last_candle["rsi_14"] < 36.0)
            and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
          )
        )
      ):
        buy_amount = (
          slice_amount * grind_2_stakes[grind_2_sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_2_sub_grind_count > 0:
          grind_profit = (exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate
          grind_profit_stake = grind_2_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (gd2) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_2_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (gd2) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_2_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "gd2"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_2_sub_grind_count > 0:
      grind_profit = (exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate
      if grind_profit > grind_2_profit_threshold:
        sell_amount = grind_2_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (gd2) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (gd2) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "gd2"
          for grind_entry_id in grind_2_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (
        (grind_2_sub_grind_count > 0)
        and (((exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate) < grind_2_stop_grinds)




        and (is_derisk or is_derisk_calc or is_grind_mode)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 14) or is_backtest)
    ):
      sell_amount = grind_2_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_2_current_open_rate > 0.0:
          grind_profit = (
            ((exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate)
            if grind_2_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (dd2) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (dd2) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "dd2"
        for grind_entry_id in grind_2_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if has_order_tags and (not partial_sell) and (grind_3_sub_grind_count < grind_3_max_sub_grinds):
      if (
        (
          ((grind_3_sub_grind_count > 0) and grind_3_distance_ratio < grind_3_sub_thresholds[grind_3_sub_grind_count])
          or ((is_derisk or is_derisk_calc) and grind_3_sub_grind_count == 0)
          or (is_grind_mode and grind_3_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )
        and ((num_open_grinds == 0) or (slice_profit < -0.02))
        and (
          (
            (
              (last_candle["close"] > (last_candle["close_max_12"] * 0.94))
              and (last_candle["close"] > (last_candle["close_max_24"] * 0.92))
              and (last_candle["close"] > (last_candle["close_max_48"] * 0.90))
              and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.88))
              and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.86))
            )
            and (
              (
                (grind_3_sub_grind_count == 0)
                and (
                  is_long_grind_buy
                  or (
                    (last_candle["rsi_3"] > 10.0)
                    and (last_candle["rsi_3_15m"] > 10.0)
                    and (last_candle["rsi_3_1h"] > 10.0)
                    and (last_candle["rsi_3_4h"] > 10.0)
                    and (last_candle["rsi_14"] < 36.0)



                    and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
                  )
                )
              )
              or (
                (grind_3_sub_grind_count > 0)
                and (
                  is_long_grind_buy
                  or (
                    (last_candle["rsi_3"] > 12.0)
                    and (last_candle["rsi_3_15m"] > 12.0)


                    and (last_candle["rsi_14"] < 42.0)
                  )
                )
              )
            )
          )
          or (
            (slice_profit < -0.06)
            and (last_candle["rsi_3"] > 10.0)
            and (last_candle["rsi_3_15m"] > 10.0)
            and (last_candle["rsi_14"] > 28.0)
            and (last_candle["rsi_14"] < 36.0)
            and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
          )
        )
      ):
        buy_amount = (
          slice_amount * grind_3_stakes[grind_3_sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_3_sub_grind_count > 0:
          grind_profit = (exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate
          grind_profit_stake = grind_3_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (gd3) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_3_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (gd3) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_3_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "gd3"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_3_sub_grind_count > 0:
      grind_profit = (exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate
      if grind_profit > grind_3_profit_threshold:
        sell_amount = grind_3_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (gd3) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (gd3) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "gd3"
          for grind_entry_id in grind_3_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (
        (grind_3_sub_grind_count > 0)
        and (((exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate) < grind_3_stop_grinds)




        and (is_derisk or is_derisk_calc or is_grind_mode)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 14) or is_backtest)
    ):
      sell_amount = grind_3_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_3_current_open_rate > 0.0:
          grind_profit = (
            ((exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate)
            if grind_3_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (dd3) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (dd3) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "dd3"
        for grind_entry_id in grind_3_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if has_order_tags and (not partial_sell) and (grind_4_sub_grind_count < grind_4_max_sub_grinds):
      if (
        (
          ((grind_4_sub_grind_count > 0) and grind_4_distance_ratio < grind_4_sub_thresholds[grind_4_sub_grind_count])
          or ((is_derisk or is_derisk_calc) and grind_4_sub_grind_count == 0)
          or (is_grind_mode and grind_4_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )
        and ((num_open_grinds == 0) or (slice_profit < -0.02))
        and (
          (
            (
              (last_candle["close"] > (last_candle["close_max_12"] * 0.94))
              and (last_candle["close"] > (last_candle["close_max_24"] * 0.92))
              and (last_candle["close"] > (last_candle["close_max_48"] * 0.90))
              and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.88))
              and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.86))
            )
            and (
              (
                (grind_4_sub_grind_count == 0)
                and (
                  is_long_grind_buy
                  or (
                    (last_candle["rsi_3"] > 10.0)
                    and (last_candle["rsi_3_15m"] > 10.0)
                    and (last_candle["rsi_3_1h"] > 10.0)
                    and (last_candle["rsi_3_4h"] > 10.0)
                    and (last_candle["rsi_14"] < 36.0)



                    and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
                  )
                )
              )
              or (
                (grind_4_sub_grind_count > 0)
                and (
                  is_long_grind_buy
                  or (
                    (last_candle["rsi_3"] > 12.0)
                    and (last_candle["rsi_3_15m"] > 12.0)


                    and (last_candle["rsi_14"] < 42.0)
                  )
                )
              )
            )
          )
          or (
            (slice_profit < -0.06)
            and (last_candle["rsi_3"] > 10.0)
            and (last_candle["rsi_3_15m"] > 10.0)
            and (last_candle["rsi_14"] > 28.0)
            and (last_candle["rsi_14"] < 36.0)
            and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
          )
        )
      ):
        buy_amount = (
          slice_amount * grind_4_stakes[grind_4_sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_4_sub_grind_count > 0:
          grind_profit = (exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate
          grind_profit_stake = grind_4_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (gd4) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_4_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (gd4) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_4_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "gd4"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_4_sub_grind_count > 0:
      grind_profit = (exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate
      if grind_profit > grind_4_profit_threshold:
        sell_amount = grind_4_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (gd4) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (gd4) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "gd4"
          for grind_entry_id in grind_4_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (
        (grind_4_sub_grind_count > 0)
        and (((exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate) < grind_4_stop_grinds)




        and (is_derisk or is_derisk_calc or is_grind_mode)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 14) or is_backtest)
    ):
      sell_amount = grind_4_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_4_current_open_rate > 0.0:
          grind_profit = (
            ((exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate)
            if grind_4_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (dd4) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (dd4) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "dd4"
        for grind_entry_id in grind_4_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if has_order_tags and (not partial_sell) and (grind_5_sub_grind_count < grind_5_max_sub_grinds):
      if (
        (
          ((grind_5_sub_grind_count > 0) and grind_5_distance_ratio < grind_5_sub_thresholds[grind_5_sub_grind_count])
          or ((is_derisk or is_derisk_calc) and grind_5_sub_grind_count == 0)
          or (is_grind_mode and grind_5_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )
        and ((num_open_grinds == 0) or (slice_profit < -0.02))
        and (
          (last_candle["close"] > (last_candle["close_max_12"] * 0.94))
          and (last_candle["close"] > (last_candle["close_max_24"] * 0.92))
          and (last_candle["close"] > (last_candle["close_max_48"] * 0.90))
        )
        and (
          (
            (
              (last_candle["close"] > (last_candle["close_max_12"] * 0.94))
              and (last_candle["close"] > (last_candle["close_max_24"] * 0.92))
              and (last_candle["close"] > (last_candle["close_max_48"] * 0.90))
              and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.88))
              and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.86))
            )
            and (
              (
                (grind_5_sub_grind_count == 0)
                and (
                  is_long_grind_buy
                  or (
                    (last_candle["rsi_3"] > 10.0)
                    and (last_candle["rsi_3_15m"] > 10.0)
                    and (last_candle["rsi_3_1h"] > 10.0)
                    and (last_candle["rsi_3_4h"] > 10.0)
                    and (last_candle["rsi_14"] < 36.0)



                    and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
                  )
                )
              )
              or (
                (grind_5_sub_grind_count > 0)
                and (
                  is_long_grind_buy
                  or (
                    (last_candle["rsi_3"] > 12.0)
                    and (last_candle["rsi_3_15m"] > 12.0)


                    and (last_candle["rsi_14"] < 42.0)
                  )
                )
              )
            )
          )
          or (
            (slice_profit < -0.06)
            and (last_candle["rsi_3"] > 10.0)
            and (last_candle["rsi_3_15m"] > 10.0)
            and (last_candle["rsi_14"] > 28.0)
            and (last_candle["rsi_14"] < 36.0)
            and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
          )
        )
      ):
        buy_amount = (
          slice_amount * grind_5_stakes[grind_5_sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_5_sub_grind_count > 0:
          grind_profit = (exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate
          grind_profit_stake = grind_5_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (gd5) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_5_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (gd5) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_5_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "gd5"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_5_sub_grind_count > 0:
      grind_profit = (exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate
      if grind_profit > grind_5_profit_threshold:
        sell_amount = grind_5_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (gd5) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (gd5) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "gd5"
          for grind_entry_id in grind_5_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (
        (grind_5_sub_grind_count > 0)
        and (((exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate) < grind_5_stop_grinds)




        and (is_derisk or is_derisk_calc or is_grind_mode)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 14) or is_backtest)
    ):
      sell_amount = grind_5_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_5_current_open_rate > 0.0:
          grind_profit = (
            ((exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate)
            if grind_5_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (dd5) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (dd5) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "dd5"
        for grind_entry_id in grind_5_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if has_order_tags and (not partial_sell) and (grind_6_sub_grind_count < grind_6_max_sub_grinds):
      if (
        (
          ((grind_6_sub_grind_count > 0) and grind_6_distance_ratio < grind_6_sub_thresholds[grind_6_sub_grind_count])
          or ((is_derisk or is_derisk_calc) and grind_6_sub_grind_count == 0)
          or (is_grind_mode and grind_6_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )

        and (
          (
            (
              (last_candle["close"] > (last_candle["close_max_12"] * 0.94))
              and (last_candle["close"] > (last_candle["close_max_24"] * 0.92))
              and (last_candle["close"] > (last_candle["close_max_48"] * 0.90))
              and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.88))
              and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.86))
            )
            and (
              (
                (grind_6_sub_grind_count == 0)
                and (
                  is_long_grind_buy
                  or (
                    (last_candle["rsi_3"] > 10.0)
                    and (last_candle["rsi_3_15m"] > 10.0)
                    and (last_candle["rsi_3_1h"] > 10.0)
                    and (last_candle["rsi_3_4h"] > 10.0)
                    and (last_candle["rsi_14"] < 36.0)



                    and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
                  )
                )
              )
              or (
                (grind_6_sub_grind_count > 0)
                and (
                  is_long_grind_buy
                  or (
                    (last_candle["rsi_3"] > 12.0)
                    and (last_candle["rsi_3_15m"] > 12.0)


                    and (last_candle["rsi_14"] < 42.0)
                  )
                )
              )
            )
          )
          or (
            (slice_profit < -0.06)
            and (last_candle["rsi_3"] > 10.0)
            and (last_candle["rsi_3_15m"] > 10.0)
            and (last_candle["rsi_14"] > 28.0)
            and (last_candle["rsi_14"] < 42.0)
            and (last_candle["close"] < (last_candle["ema_26"] * 0.994))
          )
        )
      ):
        buy_amount = (
          slice_amount * grind_6_stakes[grind_6_sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_6_sub_grind_count > 0:
          grind_profit = (exit_rate - grind_6_current_open_rate) / grind_6_current_open_rate
          grind_profit_stake = grind_6_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (gd6) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_6_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (gd6) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_6_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "gd6"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_6_sub_grind_count > 0:
      grind_profit = (exit_rate - grind_6_current_open_rate) / grind_6_current_open_rate
      if grind_profit > grind_6_profit_threshold:
        sell_amount = grind_6_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (gd6) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_6_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (gd6) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_6_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "gd6"
          for grind_entry_id in grind_6_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (
        (grind_6_sub_grind_count > 0)
        and (((exit_rate - grind_6_current_open_rate) / grind_6_current_open_rate) < grind_6_stop_grinds)




        and (is_derisk or is_derisk_calc or is_grind_mode)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 14) or is_backtest)
    ):
      sell_amount = grind_6_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_6_current_open_rate > 0.0:
          grind_profit = (
            ((exit_rate - grind_6_current_open_rate) / grind_6_current_open_rate)
            if grind_6_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (dd6) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_6_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (dd6) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_6_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "dd6"
        for grind_entry_id in grind_6_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount

    if (
      is_derisk_1
      and not derisk_1_reentry_found
      and derisk_1_order is not None
      and (
        ((current_rate - derisk_1_order.safe_price) / derisk_1_order.safe_price)
        < (
          self.regular_mode_derisk_1_reentry_futures
          if self.is_futures_mode
          else self.regular_mode_derisk_1_reentry_spot
        )
      )
    ):
      if (
        (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )
        and ((num_open_grinds == 0) or (slice_profit < -0.02))
        and (
          (last_candle["close"] > (last_candle["close_max_12"] * 0.94))
          and (last_candle["close"] > (last_candle["close_max_24"] * 0.92))
          and (last_candle["close"] > (last_candle["close_max_48"] * 0.90))
          and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.88))
          and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.86))
          and (last_candle["close"] > (last_candle["high_max_6_1d"] * 0.84))
          and (last_candle["close"] > (last_candle["high_max_12_1d"] * 0.82))
        )
        and (
          (last_candle["zlma_50_dec_15m"] == False)
          and (last_candle["zlma_50_dec_1h"] == False)
          and (last_candle["zlma_50_dec_4h"] == False)
          and (last_candle["ema_200_dec_48_1h"] == False)
          and (last_candle["ema_200_dec_24_4h"] == False)
          and (last_candle["close"] < last_candle["res_hlevel_4h"])
          and (last_candle["close"] > last_candle["sup_level_4h"])
        )
        and (
          is_long_grind_buy
          or (
            (last_candle["rsi_3"] > 30.0)
            and (last_candle["rsi_3_15m"] > 30.0)
            and (last_candle["rsi_3_1h"] > 30.0)
            and (last_candle["rsi_3_4h"] > 30.0)
            and (last_candle["rsi_14"] < 42.0)
          )
        )
      ):
        buy_amount = derisk_1_order.safe_filled * derisk_1_order.safe_price
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if derisk_1_sub_grind_count > 0:
          grind_profit = (exit_rate - derisk_1_current_open_rate) / derisk_1_current_open_rate
          grind_profit_stake = derisk_1_current_grind_stake_profit
        self.dp.send_msg(
          f"Re-entry (d1) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({derisk_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Re-entry (d1) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({derisk_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "d1"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if (
      has_order_tags

      and derisk_1_reentry_found
      and derisk_1_reentry_order is not None

      and derisk_1_distance_ratio
      < (
        (
          self.regular_mode_derisk_1_reentry_futures
          if self.is_futures_mode
          else self.regular_mode_derisk_1_reentry_spot
        )
        / (trade.leverage if self.is_futures_mode else 1.0)
      )
    ):
      sell_amount = derisk_1_reentry_order.safe_filled * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        self.dp.send_msg(
          f"De-risk (d1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        log.info(
          f"De-risk (d1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        return -ft_sell_amount, "d1"

    if (
      not is_derisk_found
      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 2, 5) or is_backtest)
      and profit_stake
      < (
        slice_amount
        * (
          (self.regular_mode_derisk_futures if self.is_futures_mode else self.regular_mode_derisk_spot)
          if (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 16) or is_backtest)
          else (self.regular_mode_derisk_futures_old if self.is_futures_mode else self.regular_mode_derisk_spot_old)
        )
        / (trade.leverage if self.is_futures_mode else 1.0)
      )
    ):
      sell_amount = trade.amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        self.dp.send_msg(
          f"De-risk [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        log.info(
          f"De-risk [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        return -ft_sell_amount, "d", is_derisk






























































    return None


  def long_grind_buy(self, last_candle: Series, previous_candle: Series, slice_profit: float) -> float:
    if (
      (
        (last_candle["close"] > (last_candle["close_max_12"] * 0.88))
        and (last_candle["close"] > (last_candle["close_max_24"] * 0.82))
        and (last_candle["close"] > (last_candle["close_max_48"] * 0.76))
        and (last_candle["btc_pct_close_max_72_5m"] < 0.03)
        and (last_candle["btc_pct_close_max_24_5m"] < 0.03)
      )
      and (
        (last_candle["enter_long"] == True)
        or (
          (last_candle["rsi_3"] > 10.0)
          and (last_candle["rsi_3_15m"] > 20.0)
          and (last_candle["rsi_3_1h"] > 20.0)
          and (last_candle["rsi_3_4h"] > 20.0)
          and (last_candle["rsi_14"] < 44.0)
          and (last_candle["ha_close"] > last_candle["ha_open"])
          and (last_candle["ema_12"] < (last_candle["ema_26"] * 0.990))
          and (last_candle["cti_20_1h"] < 0.8)
          and (last_candle["rsi_14_1h"] < 80.0)
        )
        or (
          (last_candle["rsi_14"] < 36.0)
          and (last_candle["close"] < (last_candle["sma_16"] * 0.994))
          and (last_candle["rsi_3"] > 16.0)
          and (last_candle["rsi_3_15m"] > 30.0)
          and (last_candle["rsi_3_1h"] > 30.0)
          and (last_candle["rsi_3_4h"] > 30.0)
          and (last_candle["zlma_50_dec_1h"] == False)
        )
        or (
          (last_candle["rsi_14"] < 36.0)
          and (previous_candle["rsi_3"] > 6.0)
          and (last_candle["ema_26"] > last_candle["ema_12"])
          and ((last_candle["ema_26"] - last_candle["ema_12"]) > (last_candle["open"] * 0.010))
          and ((previous_candle["ema_26"] - previous_candle["ema_12"]) > (last_candle["open"] / 100.0))
          and (last_candle["rsi_3_1h"] > 20.0)
          and (last_candle["rsi_3_4h"] > 20.0)
          and (last_candle["cti_20_1h"] < 0.8)
          and (last_candle["rsi_14_1h"] < 80.0)
        )
        or (
          (last_candle["rsi_14"] > 30.0)
          and (last_candle["rsi_14"] < 60.0)
          and (last_candle["hma_70_buy"])
          and (last_candle["close"] < (last_candle["high_max_12_1h"] * 0.90))
          and (last_candle["cti_20_15m"] < 0.5)
          and (last_candle["rsi_14_15m"] < 50.0)
          and (last_candle["cti_20_1h"] < 0.8)
          and (last_candle["rsi_14_1h"] < 80.0)
          and (last_candle["zlma_50_dec_1h"] == False)
          and (last_candle["zlma_50_dec_4h"] == False)
        )
        or (
          (last_candle["rsi_3"] > 12.0)
          and (last_candle["rsi_3_15m"] > 20.0)
          and (last_candle["rsi_3_1h"] > 20.0)
          and (last_candle["rsi_3_4h"] > 20.0)
          and (last_candle["rsi_14"] < 36.0)
          and (last_candle["zlma_50_dec_15m"] == False)
          and (last_candle["zlma_50_dec_1h"] == False)
        )
        or (
          (last_candle["rsi_14"] < 40.0)
          and (last_candle["rsi_14_15m"] < 40.0)
          and (last_candle["rsi_3"] > 6.0)
          and (last_candle["ema_26_15m"] > last_candle["ema_12_15m"])
          and ((last_candle["ema_26_15m"] - last_candle["ema_12_15m"]) > (last_candle["open_15m"] * 0.006))
          and ((previous_candle["ema_26_15m"] - previous_candle["ema_12_15m"]) > (last_candle["open_15m"] / 100.0))
          and (last_candle["rsi_3_15m"] > 10.0)
          and (last_candle["rsi_3_1h"] > 26.0)
          and (last_candle["rsi_3_4h"] > 26.0)
          and (last_candle["cti_20_1h"] < 0.8)
          and (last_candle["rsi_14_1h"] < 80.0)
        )
        or (
          (last_candle["rsi_14"] > 35.0)
          and (last_candle["rsi_3"] > 4.0)
          and (last_candle["rsi_3"] < 46.0)
          and (last_candle["rsi_14"] < previous_candle["rsi_14"])
          and (last_candle["close"] < (last_candle["sma_16"] * 0.982))
          and (last_candle["cti_20"] < -0.6)
          and (last_candle["rsi_3_1h"] > 20.0)
          and (last_candle["rsi_3_4h"] > 20.0)
          and (last_candle["not_downtrend_1d"] == True)
          and (last_candle["zlma_50_dec_1h"] == False)
        )
        or (
          (last_candle["rsi_3"] > 12.0)
          and (last_candle["rsi_3_15m"] > 26.0)
          and (last_candle["rsi_3_1h"] > 26.0)
          and (last_candle["rsi_3_4h"] > 26.0)
          and (last_candle["rsi_14"] < 40.0)
          and (last_candle["ema_12"] < (last_candle["ema_26"] * 0.994))
          and (last_candle["cti_20_1h"] < 0.8)
          and (last_candle["rsi_14_1h"] < 80.0)
          and (last_candle["cti_20_4h"] < 0.8)
          and (last_candle["rsi_14_4h"] < 80.0)
          and (last_candle["ema_200_dec_48_1h"] == False)
        )
        or (
          (last_candle["rsi_14"] < 60.0)
          and (last_candle["hma_55_buy"])
          and (last_candle["rsi_3_1h"] > 4.0)
          and (last_candle["rsi_3_4h"] > 4.0)
          and (last_candle["cti_20_15m"] < 0.8)
          and (last_candle["cti_20_1h"] < 0.8)
          and (last_candle["cti_20_4h"] < 0.8)
          and (last_candle["rsi_14_1h"] < 80.0)
          and (last_candle["close"] < (last_candle["high_max_12_1h"] * 0.90))
          and (last_candle["zlma_50_dec_1h"] == False)
          and (last_candle["zlma_50_dec_4h"] == False)
        )
        or (
          (last_candle["rsi_3"] > 12.0)
          and (last_candle["rsi_3_15m"] > 30.0)
          and (last_candle["rsi_3_1h"] > 30.0)
          and (last_candle["rsi_3_4h"] > 30.0)
          and (last_candle["rsi_14"] < 40.0)
          and (last_candle["cti_20_15m"] < 0.8)
          and (last_candle["rsi_14_15m"] < 70.0)
          and (last_candle["cti_20_1h"] < 0.8)
          and (last_candle["rsi_14_1h"] < 80.0)
          and (last_candle["cti_20_4h"] < 0.8)
          and (last_candle["rsi_14_4h"] < 80.0)
          and (last_candle["r_14_1h"] > -80.0)
          and (last_candle["ema_12"] < (last_candle["ema_26"] * 0.995))
          and (last_candle["zlma_50_dec_1h"] == False)
        )
        or (
          (last_candle["rsi_3"] > 12.0)
          and (last_candle["rsi_3_15m"] > 30.0)
          and (last_candle["rsi_3_1h"] > 30.0)
          and (last_candle["rsi_3_4h"] > 30.0)
          and (last_candle["rsi_14"] < 40.0)
          and (last_candle["rsi_14_1d"] < 70.0)
          and (last_candle["close"] > last_candle["sup_level_1h"])
          and (last_candle["close"] > last_candle["sup_level_4h"])
          and (last_candle["close"] > last_candle["sup_level_1d"])
          and (last_candle["not_downtrend_1h"] == True)
          and (last_candle["not_downtrend_4h"] == True)
          and (last_candle["not_downtrend_1d"] == True)
        )
        or (
          (last_candle["rsi_3"] > 12.0)
          and (last_candle["rsi_3_15m"] > 30.0)
          and (last_candle["rsi_3_1h"] > 30.0)
          and (last_candle["rsi_3_4h"] > 30.0)
          and (last_candle["rsi_14"] < 40.0)
          and (last_candle["ha_close"] > last_candle["ha_open"])
          and (last_candle["close"] < last_candle["res_hlevel_4h"])
          and (last_candle["close"] > last_candle["sup_level_4h"])
          and (last_candle["close"] < last_candle["res_hlevel_1d"])
          and (last_candle["close"] > last_candle["sup_level_1d"])
          and (last_candle["close"] < last_candle["res3_1d"])
          and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.85))
          and (last_candle["hl_pct_change_24_1h"] < 0.35)
        )
        or (
          (last_candle["rsi_3"] > 12.0)
          and (last_candle["rsi_3_15m"] > 30.0)
          and (last_candle["rsi_3_1h"] > 30.0)
          and (last_candle["rsi_3_4h"] > 30.0)
          and (last_candle["rsi_14"] < 46.0)
          and (previous_candle["chandelier_dir"] < -0)
          and (last_candle["chandelier_dir"] > 0)
          and (last_candle["close"] < last_candle["res_hlevel_4h"])
          and (last_candle["close"] > last_candle["sup_level_4h"])
          and (last_candle["close"] < last_candle["res_hlevel_1d"])
          and (last_candle["close"] > last_candle["sup_level_1d"])
        )
      )
    ):
      return True

    return False


  def long_adjust_trade_position_no_derisk(
    self,
    trade: Trade,
    enter_tags,
    current_time: datetime,
    current_rate: float,
    current_profit: float,
    min_stake: Optional[float],
    max_stake: float,
    current_entry_rate: float,
    current_exit_rate: float,
    current_entry_profit: float,
    current_exit_profit: float,
    last_candle: Series,
    previous_candle: Series,
    filled_orders: "Orders",
    filled_entries: "Orders",
    filled_exits: "Orders",
    exit_rate: float,
    slice_amount: float,
    slice_profit_entry: float,
    slice_profit: float,
    profit_ratio: float,
    profit_stake: float,
    profit_init_ratio: float,
    current_stake_amount: float,
    has_order_tags: bool,
    **kwargs,
  ) -> tuple[Optional[float], str, bool]:
    is_backtest = self.dp.runmode.value in ["backtest", "hyperopt"]

    max_rebuy_sub_grinds = 0
    regular_mode_rebuy_stakes = (
      self.regular_mode_rebuy_stakes_futures.copy()
      if self.is_futures_mode
      else self.regular_mode_rebuy_stakes_spot.copy()
    )
    regular_mode_rebuy_sub_thresholds = (
      self.regular_mode_rebuy_thresholds_futures if self.is_futures_mode else self.regular_mode_rebuy_thresholds_spot
    )
    if (slice_amount * regular_mode_rebuy_stakes[0] / trade.leverage) < min_stake:
      multi = min_stake / slice_amount / regular_mode_rebuy_stakes[0]
      for i, _ in enumerate(regular_mode_rebuy_stakes):
        regular_mode_rebuy_stakes[i] *= multi
    max_rebuy_sub_grinds = len(regular_mode_rebuy_stakes)

    max_grind_1_sub_grinds = 0
    regular_mode_grind_1_stakes = (
      self.regular_mode_grind_1_stakes_futures.copy()
      if self.is_futures_mode
      else self.regular_mode_grind_1_stakes_spot.copy()
    )
    regular_mode_grind_1_sub_thresholds = (
      self.regular_mode_grind_1_thresholds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_1_thresholds_spot
    )
    if (slice_amount * regular_mode_grind_1_stakes[0] / trade.leverage) < min_stake:
      multi = min_stake / slice_amount / regular_mode_grind_1_stakes[0]
      for i, _ in enumerate(regular_mode_grind_1_stakes):
        regular_mode_grind_1_stakes[i] *= multi
    max_grind_1_sub_grinds = len(regular_mode_grind_1_stakes)
    regular_mode_grind_1_stop_grinds = (
      self.regular_mode_grind_1_stop_grinds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_1_stop_grinds_spot
    )
    regular_mode_grind_1_profit_threshold = (
      self.regular_mode_grind_1_profit_threshold_futures
      if self.is_futures_mode
      else self.regular_mode_grind_1_profit_threshold_spot
    )

    max_grind_2_sub_grinds = 0
    regular_mode_grind_2_stakes = (
      self.regular_mode_grind_2_stakes_futures.copy()
      if self.is_futures_mode
      else self.regular_mode_grind_2_stakes_spot.copy()
    )
    regular_mode_grind_2_sub_thresholds = (
      self.regular_mode_grind_2_thresholds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_2_thresholds_spot
    )
    if (slice_amount * regular_mode_grind_2_stakes[0] / trade.leverage) < min_stake:
      multi = min_stake / slice_amount / regular_mode_grind_2_stakes[0]
      for i, _ in enumerate(regular_mode_grind_2_stakes):
        regular_mode_grind_2_stakes[i] *= multi
    max_grind_2_sub_grinds = len(regular_mode_grind_2_stakes)
    regular_mode_grind_2_stop_grinds = (
      self.regular_mode_grind_2_stop_grinds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_2_stop_grinds_spot
    )
    regular_mode_grind_2_profit_threshold = (
      self.regular_mode_grind_2_profit_threshold_futures
      if self.is_futures_mode
      else self.regular_mode_grind_2_profit_threshold_spot
    )

    max_grind_3_sub_grinds = 0
    regular_mode_grind_3_stakes = (
      self.regular_mode_grind_3_stakes_futures.copy()
      if self.is_futures_mode
      else self.regular_mode_grind_3_stakes_spot.copy()
    )
    regular_mode_grind_3_sub_thresholds = (
      self.regular_mode_grind_3_thresholds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_3_thresholds_spot
    )
    if (slice_amount * regular_mode_grind_3_stakes[0] / trade.leverage) < min_stake:
      multi = min_stake / slice_amount / regular_mode_grind_3_stakes[0]
      for i, _ in enumerate(regular_mode_grind_3_stakes):
        regular_mode_grind_3_stakes[i] *= multi
    max_grind_3_sub_grinds = len(regular_mode_grind_3_stakes)
    regular_mode_grind_3_stop_grinds = (
      self.regular_mode_grind_3_stop_grinds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_3_stop_grinds_spot
    )
    regular_mode_grind_3_profit_threshold = (
      self.regular_mode_grind_3_profit_threshold_futures
      if self.is_futures_mode
      else self.regular_mode_grind_3_profit_threshold_spot
    )

    max_grind_4_sub_grinds = 0
    regular_mode_grind_4_stakes = (
      self.regular_mode_grind_4_stakes_futures.copy()
      if self.is_futures_mode
      else self.regular_mode_grind_4_stakes_spot.copy()
    )
    regular_mode_grind_4_sub_thresholds = (
      self.regular_mode_grind_4_thresholds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_4_thresholds_spot
    )
    if (slice_amount * regular_mode_grind_4_stakes[0] / trade.leverage) < min_stake:
      multi = min_stake / slice_amount / regular_mode_grind_4_stakes[0]
      for i, _ in enumerate(regular_mode_grind_4_stakes):
        regular_mode_grind_4_stakes[i] *= multi
    max_grind_4_sub_grinds = len(regular_mode_grind_4_stakes)
    regular_mode_grind_4_stop_grinds = (
      self.regular_mode_grind_4_stop_grinds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_4_stop_grinds_spot
    )
    regular_mode_grind_4_profit_threshold = (
      self.regular_mode_grind_4_profit_threshold_futures
      if self.is_futures_mode
      else self.regular_mode_grind_4_profit_threshold_spot
    )

    max_grind_5_sub_grinds = 0
    regular_mode_grind_5_stakes = (
      self.regular_mode_grind_5_stakes_futures.copy()
      if self.is_futures_mode
      else self.regular_mode_grind_5_stakes_spot.copy()
    )
    regular_mode_grind_5_sub_thresholds = (
      self.regular_mode_grind_5_thresholds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_5_thresholds_spot
    )
    if (slice_amount * regular_mode_grind_5_stakes[0] / trade.leverage) < min_stake:
      multi = min_stake / slice_amount / regular_mode_grind_5_stakes[0]
      for i, _ in enumerate(regular_mode_grind_5_stakes):
        regular_mode_grind_5_stakes[i] *= multi
    max_grind_5_sub_grinds = len(regular_mode_grind_5_stakes)
    regular_mode_grind_5_stop_grinds = (
      self.regular_mode_grind_5_stop_grinds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_5_stop_grinds_spot
    )
    regular_mode_grind_5_profit_threshold = (
      self.regular_mode_grind_5_profit_threshold_futures
      if self.is_futures_mode
      else self.regular_mode_grind_5_profit_threshold_spot
    )

    partial_sell = False
    is_derisk = False
    is_derisk_1 = False
    rebuy_sub_grind_count = 0
    rebuy_total_amount = 0.0
    rebuy_total_cost = 0.0
    rebuy_current_open_rate = 0.0
    rebuy_current_grind_stake = 0.0
    rebuy_current_grind_stake_profit = 0.0
    rebuy_is_sell_found = False
    rebuy_found = False
    rebuy_buy_orders = []
    rebuy_distance_ratio = 0.0
    grind_1_sub_grind_count = 0
    grind_1_total_amount = 0.0
    grind_1_total_cost = 0.0
    grind_1_current_open_rate = 0.0
    grind_1_current_grind_stake = 0.0
    grind_1_current_grind_stake_profit = 0.0
    grind_1_is_sell_found = False
    grind_1_found = False
    grind_1_buy_orders = []
    grind_1_distance_ratio = 0.0
    grind_2_sub_grind_count = 0
    grind_2_total_amount = 0.0
    grind_2_total_cost = 0.0
    grind_2_current_open_rate = 0.0
    grind_2_current_grind_stake = 0.0
    grind_2_current_grind_stake_profit = 0.0
    grind_2_is_sell_found = False
    grind_2_found = False
    grind_2_buy_orders = []
    grind_2_distance_ratio = 0.0
    grind_3_sub_grind_count = 0
    grind_3_total_amount = 0.0
    grind_3_total_cost = 0.0
    grind_3_current_open_rate = 0.0
    grind_3_current_grind_stake = 0.0
    grind_3_current_grind_stake_profit = 0.0
    grind_3_is_sell_found = False
    grind_3_found = False
    grind_3_buy_orders = []
    grind_3_distance_ratio = 0.0
    grind_4_sub_grind_count = 0
    grind_4_total_amount = 0.0
    grind_4_total_cost = 0.0
    grind_4_current_open_rate = 0.0
    grind_4_current_grind_stake = 0.0
    grind_4_current_grind_stake_profit = 0.0
    grind_4_is_sell_found = False
    grind_4_found = False
    grind_4_buy_orders = []
    grind_4_distance_ratio = 0.0
    grind_5_sub_grind_count = 0
    grind_5_total_amount = 0.0
    grind_5_total_cost = 0.0
    grind_5_current_open_rate = 0.0
    grind_5_current_grind_stake = 0.0
    grind_5_current_grind_stake_profit = 0.0
    grind_5_is_sell_found = False
    grind_5_found = False
    grind_5_buy_orders = []
    grind_5_distance_ratio = 0.0
    for order in reversed(filled_orders):
      if (order.ft_order_side == "buy") and (order is not filled_orders[0]):
        order_tag = ""
        if has_order_tags:
          if order.ft_order_tag is not None:
            order_tag = order.ft_order_tag
        if not grind_1_is_sell_found and order_tag == "g1":
          grind_1_sub_grind_count += 1
          grind_1_total_amount += order.safe_filled
          grind_1_total_cost += order.safe_filled * order.safe_price
          grind_1_buy_orders.append(order.id)
          if not grind_1_found:
            grind_1_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_1_found = True
        elif not grind_2_is_sell_found and order_tag == "g2":
          grind_2_sub_grind_count += 1
          grind_2_total_amount += order.safe_filled
          grind_2_total_cost += order.safe_filled * order.safe_price
          grind_2_buy_orders.append(order.id)
          if not grind_2_found:
            grind_2_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_2_found = True
        elif not grind_3_is_sell_found and order_tag == "g3":
          grind_3_sub_grind_count += 1
          grind_3_total_amount += order.safe_filled
          grind_3_total_cost += order.safe_filled * order.safe_price
          grind_3_buy_orders.append(order.id)
          if not grind_3_found:
            grind_3_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_3_found = True
        elif not grind_4_is_sell_found and order_tag == "g4":
          grind_4_sub_grind_count += 1
          grind_4_total_amount += order.safe_filled
          grind_4_total_cost += order.safe_filled * order.safe_price
          grind_4_buy_orders.append(order.id)
          if not grind_4_found:
            grind_4_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_4_found = True
        elif not grind_5_is_sell_found and order_tag == "g5":
          grind_5_sub_grind_count += 1
          grind_5_total_amount += order.safe_filled
          grind_5_total_cost += order.safe_filled * order.safe_price
          grind_5_buy_orders.append(order.id)
          if not grind_5_found:
            grind_5_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_5_found = True
        elif not rebuy_is_sell_found and order_tag not in [
          "g1",
          "g2",
          "g3",
          "g4",
          "g5",
          "g6",
          "sg1",
          "sg2",
          "sg3",
          "sg4",
          "sg5",
          "sg6",
          "dl1",
          "dl2",
          "gd1",
          "gd2",
          "gd3",
          "gd4",
          "gd5",
          "gd6",
          "gm0",
          "gmd0",
        ]:
          rebuy_sub_grind_count += 1
          rebuy_total_amount += order.safe_filled
          rebuy_total_cost += order.safe_filled * order.safe_price
          rebuy_buy_orders.append(order.id)
          if not rebuy_found:
            rebuy_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            rebuy_found = True
      elif order.ft_order_side == "sell":
        if (
          order is filled_exits[-1]
          and (order.safe_remaining * exit_rate / (trade.leverage if self.is_futures_mode else 1.0)) > min_stake
        ):
          partial_sell = True
          break
        order_tag = ""
        if has_order_tags:
          if order.ft_order_tag is not None:
            sell_order_tag = order.ft_order_tag
            order_mode = sell_order_tag.split(" ", 1)
            if len(order_mode) > 0:
              order_tag = order_mode[0]
        if order_tag in ["g1", "sg1"]:
          grind_1_is_sell_found = True
        elif order_tag in ["g2", "sg2"]:
          grind_2_is_sell_found = True
        elif order_tag in ["g3", "sg3"]:
          grind_3_is_sell_found = True
        elif order_tag in ["g4", "sg4"]:
          grind_4_is_sell_found = True
        elif order_tag in ["g5", "sg5"]:
          grind_5_is_sell_found = True
        elif order_tag in ["d", "d1", "dd0", "ddl1", "ddl2", "dd1", "dd2", "dd3", "dd4", "dd5", "dd6"]:
          is_derisk = True
          if order_tag in ["d1"]:
            is_derisk_1 = True
          grind_1_is_sell_found = True
          grind_2_is_sell_found = True
          grind_3_is_sell_found = True
          grind_4_is_sell_found = True
          grind_5_is_sell_found = True
          rebuy_is_sell_found = True
        elif order_tag not in [
          "p",
          "g1",
          "g2",
          "g3",
          "g4",
          "g5",
          "g6",
          "sg1",
          "sg2",
          "sg3",
          "sg4",
          "sg5",
          "sg6",
          "dl1",
          "dl2",
          "gd1",
          "gd2",
          "gd3",
          "gd4",
          "gd5",
          "gd6",
          "gm0",
          "gmd0",
        ]:
          rebuy_is_sell_found = True
        if not is_derisk:
          start_amount = filled_orders[0].safe_filled
          current_amount = 0.0
          for order2 in filled_orders:
            if order2.ft_order_side == "buy":
              current_amount += order2.safe_filled
            elif order2.ft_order_side == "sell":
              current_amount -= order2.safe_filled
            if order2 is order:
              if current_amount < (start_amount * 0.95):
                is_derisk = True

        if (
          rebuy_is_sell_found
          and grind_1_is_sell_found
          and grind_2_is_sell_found
          and grind_3_is_sell_found
          and grind_4_is_sell_found
          and grind_5_is_sell_found
        ):
          break

    if is_derisk:
      return None, "", is_derisk
    if not has_order_tags and len(filled_exits) > 0:
      return None, "", is_derisk

    if rebuy_sub_grind_count > 0:
      rebuy_current_open_rate = rebuy_total_cost / rebuy_total_amount
      rebuy_current_grind_stake = rebuy_total_amount * exit_rate * (1 - trade.fee_close)
      rebuy_current_grind_stake_profit = rebuy_current_grind_stake - rebuy_total_cost
    if grind_1_sub_grind_count > 0:
      grind_1_current_open_rate = grind_1_total_cost / grind_1_total_amount
      grind_1_current_grind_stake = grind_1_total_amount * exit_rate * (1 - trade.fee_close)
      grind_1_current_grind_stake_profit = grind_1_current_grind_stake - grind_1_total_cost
    if grind_2_sub_grind_count > 0:
      grind_2_current_open_rate = grind_2_total_cost / grind_2_total_amount
      grind_2_current_grind_stake = grind_2_total_amount * exit_rate * (1 - trade.fee_close)
      grind_2_current_grind_stake_profit = grind_2_current_grind_stake - grind_2_total_cost
    if grind_3_sub_grind_count > 0:
      grind_3_current_open_rate = grind_3_total_cost / grind_3_total_amount
      grind_3_current_grind_stake = grind_3_total_amount * exit_rate * (1 - trade.fee_close)
      grind_3_current_grind_stake_profit = grind_3_current_grind_stake - grind_3_total_cost
    if grind_4_sub_grind_count > 0:
      grind_4_current_open_rate = grind_4_total_cost / grind_4_total_amount
      grind_4_current_grind_stake = grind_4_total_amount * exit_rate * (1 - trade.fee_close)
      grind_4_current_grind_stake_profit = grind_4_current_grind_stake - grind_4_total_cost
    if grind_5_sub_grind_count > 0:
      grind_5_current_open_rate = grind_5_total_cost / grind_5_total_amount
      grind_5_current_grind_stake = grind_5_total_amount * exit_rate * (1 - trade.fee_close)
      grind_5_current_grind_stake_profit = grind_5_current_grind_stake - grind_5_total_cost

    num_open_grinds = (
      grind_1_sub_grind_count
      + grind_2_sub_grind_count
      + grind_3_sub_grind_count
      + grind_4_sub_grind_count
      + grind_5_sub_grind_count
    )

    if partial_sell:
      order = filled_exits[-1]
      sell_amount = order.safe_remaining * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        self.dp.send_msg(
          f"Exit (remaining) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {order.safe_remaining} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        order_tag = "p"
        if has_order_tags:
          if order.ft_order_tag is not None:
            order_tag = order.ft_order_tag
        return -ft_sell_amount, order_tag, is_derisk

    is_long_grind_buy = self.long_grind_buy(last_candle, previous_candle, slice_profit)

    if (not partial_sell) and (not rebuy_is_sell_found) and (rebuy_sub_grind_count < max_rebuy_sub_grinds):
      if (
        (0 <= rebuy_sub_grind_count < max_rebuy_sub_grinds)
        and (slice_profit_entry < regular_mode_rebuy_sub_thresholds[rebuy_sub_grind_count])
        and (
          (rebuy_distance_ratio if (rebuy_sub_grind_count > 0) else profit_init_ratio)
          < (regular_mode_rebuy_sub_thresholds[rebuy_sub_grind_count])
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=12) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.06))
        and (
          (last_candle["close"] > (last_candle["close_max_12"] * 0.94))
          and (last_candle["close"] > (last_candle["close_max_24"] * 0.92))
          and (last_candle["close"] > (last_candle["close_max_48"] * 0.90))
          and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.88))
          and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.86))
          and (last_candle["close"] > (last_candle["high_max_6_1d"] * 0.84))
          and (last_candle["close"] > (last_candle["high_max_12_1d"] * 0.82))
          and (last_candle["btc_pct_close_max_72_5m"] < 0.03)
          and (last_candle["btc_pct_close_max_24_5m"] < 0.03)
        )
        and (
          is_long_grind_buy
          or (
            (last_candle["rsi_3"] > 30.0)
            and (last_candle["rsi_3_15m"] > 30.0)
            and (last_candle["rsi_3_1h"] > 30.0)
            and (last_candle["rsi_3_4h"] > 30.0)
            and (last_candle["rsi_14"] < 36.0)

            and (last_candle["zlma_50_dec_1h"] == False)
            and (last_candle["zlma_50_dec_4h"] == False)
          )
        )
      ):
        buy_amount = (
          slice_amount
          * regular_mode_rebuy_stakes[rebuy_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount > max_stake:
          buy_amount = max_stake
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None, "", is_derisk
        self.dp.send_msg(
          f"Rebuy (r) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        log.info(
          f"Rebuy (r) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        order_tag = "r"
        return buy_amount, order_tag, is_derisk


    if has_order_tags and (not partial_sell) and (grind_1_sub_grind_count < max_grind_1_sub_grinds):
      if (
        (
          (grind_1_distance_ratio if (grind_1_sub_grind_count > 0) else profit_init_ratio)
          < (regular_mode_grind_1_sub_thresholds[grind_1_sub_grind_count])
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )
        and ((num_open_grinds == 0) or (slice_profit < -0.02))
        and (
          (last_candle["close"] > (last_candle["close_max_12"] * 0.92))
          and (last_candle["close"] > (last_candle["close_max_24"] * 0.90))
          and (last_candle["close"] > (last_candle["close_max_48"] * 0.88))
          and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.86))
          and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.84))
          and (last_candle["close"] > (last_candle["high_max_6_1d"] * 0.76))
          and (last_candle["close"] > (last_candle["high_max_12_1d"] * 0.70))
        )
        and (
          is_long_grind_buy
          or (
            (last_candle["rsi_3"] > 16.0)
            and (last_candle["rsi_3_15m"] > 16.0)
            and (last_candle["rsi_3_1h"] > 20.0)
            and (last_candle["rsi_3_4h"] > 20.0)
            and (last_candle["rsi_14"] < 36.0)

            and (last_candle["zlma_50_dec_1h"] == False)
            and (last_candle["zlma_50_dec_4h"] == False)
            and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
          )
        )
      ):
        buy_amount = (
          slice_amount
          * regular_mode_grind_1_stakes[grind_1_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None, "", is_derisk
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_1_sub_grind_count > 0:
          grind_profit = (exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate
          grind_profit_stake = grind_1_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (g1) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (g1) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "g1"
        return buy_amount, order_tag, is_derisk

    if has_order_tags and grind_1_sub_grind_count > 0:
      grind_profit = (exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate
      if grind_profit > regular_mode_grind_1_profit_threshold:
        sell_amount = grind_1_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (g1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (g1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "g1"
          for grind_entry_id in grind_1_buy_orders:
            order_tag += " " + str(grind_entry_id)
          return -ft_sell_amount, order_tag, is_derisk

    if (
      (
        (grind_1_sub_grind_count > 0)
        and self.regular_mode_use_grind_stops
        and (((exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate) < regular_mode_grind_1_stop_grinds)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 16) or is_backtest)
    ):
      sell_amount = grind_1_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_1_current_open_rate > 0.0:
          grind_profit = (
            ((exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate)
            if grind_1_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (sg1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (sg1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "sg1"
        for grind_entry_id in grind_1_buy_orders:
          order_tag += " " + str(grind_entry_id)
        return -ft_sell_amount, order_tag, is_derisk


    if has_order_tags and (not partial_sell) and (grind_2_sub_grind_count < max_grind_2_sub_grinds):
      if (
        (
          (grind_2_distance_ratio if (grind_2_sub_grind_count > 0) else profit_init_ratio)
          < (regular_mode_grind_2_sub_thresholds[grind_2_sub_grind_count])
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )
        and ((num_open_grinds == 0) or (slice_profit < -0.02))
        and (
          (last_candle["close"] > (last_candle["close_max_12"] * 0.92))
          and (last_candle["close"] > (last_candle["close_max_24"] * 0.90))
          and (last_candle["close"] > (last_candle["close_max_48"] * 0.88))
          and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.86))
          and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.84))
          and (last_candle["close"] > (last_candle["high_max_6_1d"] * 0.76))
          and (last_candle["close"] > (last_candle["high_max_12_1d"] * 0.70))
        )
        and (
          is_long_grind_buy
          or (
            (last_candle["rsi_3"] > 16.0)
            and (last_candle["rsi_3_15m"] > 16.0)
            and (last_candle["rsi_3_1h"] > 20.0)
            and (last_candle["rsi_3_4h"] > 20.0)
            and (last_candle["rsi_14"] < 36.0)

            and (last_candle["zlma_50_dec_1h"] == False)
            and (last_candle["zlma_50_dec_4h"] == False)
            and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
          )
        )
      ):
        buy_amount = (
          slice_amount
          * regular_mode_grind_2_stakes[grind_2_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None, "", is_derisk
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_2_sub_grind_count > 0:
          grind_profit = (exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate
          grind_profit_stake = grind_2_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (g2) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_2_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (g2) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_2_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "g2"
        return buy_amount, order_tag, is_derisk

    if has_order_tags and grind_2_sub_grind_count > 0:
      grind_profit = (exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate
      if grind_profit > regular_mode_grind_2_profit_threshold:
        sell_amount = grind_2_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (g2) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (g2) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "g2"
          for grind_entry_id in grind_2_buy_orders:
            order_tag += " " + str(grind_entry_id)
          return -ft_sell_amount, order_tag, is_derisk

    if (
      (
        (grind_2_sub_grind_count > 0)
        and self.regular_mode_use_grind_stops
        and (((exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate) < regular_mode_grind_2_stop_grinds)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 16) or is_backtest)
    ):
      sell_amount = grind_2_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_2_current_open_rate > 0.0:
          grind_profit = (
            ((exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate)
            if grind_2_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (sg2) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (sg2) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "sg2"
        for grind_entry_id in grind_2_buy_orders:
          order_tag += " " + str(grind_entry_id)
        return -ft_sell_amount, order_tag, is_derisk


    if has_order_tags and (not partial_sell) and (grind_3_sub_grind_count < max_grind_3_sub_grinds):
      if (
        (
          (grind_3_distance_ratio if (grind_3_sub_grind_count > 0) else profit_init_ratio)
          < (regular_mode_grind_3_sub_thresholds[grind_3_sub_grind_count])
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )
        and ((num_open_grinds == 0) or (slice_profit < -0.02))
        and (
          (
            (
              (last_candle["close"] > (last_candle["close_max_12"] * 0.92))
              and (last_candle["close"] > (last_candle["close_max_24"] * 0.90))
              and (last_candle["close"] > (last_candle["close_max_48"] * 0.88))
              and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.86))
              and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.84))
              and (last_candle["close"] > (last_candle["high_max_6_1d"] * 0.76))
              and (last_candle["close"] > (last_candle["high_max_12_1d"] * 0.70))
            )
            and (
              is_long_grind_buy
              or (
                (last_candle["rsi_3"] > 12.0)
                and (last_candle["rsi_3_15m"] > 16.0)
                and (last_candle["rsi_3_1h"] > 16.0)
                and (last_candle["rsi_3_4h"] > 16.0)
                and (last_candle["rsi_14"] < 36.0)


                and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
              )
            )
          )
          or (
            (slice_profit < -0.06)
            and (last_candle["rsi_3"] > 10.0)
            and (last_candle["rsi_3_15m"] > 10.0)

            and (last_candle["rsi_14"] < 36.0)
            and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
          )
        )
      ):
        buy_amount = (
          slice_amount
          * regular_mode_grind_3_stakes[grind_3_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None, "", is_derisk
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_3_sub_grind_count > 0:
          grind_profit = (exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate
          grind_profit_stake = grind_3_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (g3) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_3_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (g3) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_3_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "g3"
        return buy_amount, order_tag, is_derisk

    if has_order_tags and grind_3_sub_grind_count > 0:
      grind_profit = (exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate
      if grind_profit > regular_mode_grind_3_profit_threshold:
        sell_amount = grind_3_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (g3) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (g3) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "g3"
          for grind_entry_id in grind_3_buy_orders:
            order_tag += " " + str(grind_entry_id)
          return -ft_sell_amount, order_tag, is_derisk

    if (
      (
        (grind_3_sub_grind_count > 0)
        and self.regular_mode_use_grind_stops
        and (((exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate) < regular_mode_grind_3_stop_grinds)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 16) or is_backtest)
    ):
      sell_amount = grind_3_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_3_current_open_rate > 0.0:
          grind_profit = (
            ((exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate)
            if grind_3_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (sg3) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (sg3) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "sg3"
        for grind_entry_id in grind_3_buy_orders:
          order_tag += " " + str(grind_entry_id)
        return -ft_sell_amount, order_tag, is_derisk


    if has_order_tags and (not partial_sell) and (grind_4_sub_grind_count < max_grind_4_sub_grinds):
      if (
        (
          (grind_4_distance_ratio if (grind_4_sub_grind_count > 0) else profit_init_ratio)
          < (regular_mode_grind_4_sub_thresholds[grind_4_sub_grind_count])
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )
        and ((num_open_grinds == 0) or (slice_profit < -0.02))
        and (
          (
            (
              (last_candle["close"] > (last_candle["close_max_12"] * 0.92))
              and (last_candle["close"] > (last_candle["close_max_24"] * 0.90))
              and (last_candle["close"] > (last_candle["close_max_48"] * 0.88))
              and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.86))
              and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.84))
              and (last_candle["close"] > (last_candle["high_max_6_1d"] * 0.76))
              and (last_candle["close"] > (last_candle["high_max_12_1d"] * 0.70))
            )
            and (
              is_long_grind_buy
              or (
                (last_candle["rsi_3"] > 12.0)
                and (last_candle["rsi_3_15m"] > 16.0)
                and (last_candle["rsi_3_1h"] > 16.0)
                and (last_candle["rsi_3_4h"] > 16.0)
                and (last_candle["rsi_14"] < 36.0)


                and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
              )
            )
          )
          or (
            (slice_profit < -0.06)
            and (last_candle["rsi_3"] > 10.0)
            and (last_candle["rsi_3_15m"] > 10.0)

            and (last_candle["rsi_14"] < 36.0)
            and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
          )
        )
      ):
        buy_amount = (
          slice_amount
          * regular_mode_grind_4_stakes[grind_4_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None, "", is_derisk
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_4_sub_grind_count > 0:
          grind_profit = (exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate
          grind_profit_stake = grind_4_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (g4) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_4_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (g4) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_4_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "g4"
        return buy_amount, order_tag, is_derisk

    if has_order_tags and grind_4_sub_grind_count > 0:
      grind_profit = (exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate
      if grind_profit > regular_mode_grind_4_profit_threshold:
        sell_amount = grind_4_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (g4) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (g4) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "g4"
          for grind_entry_id in grind_4_buy_orders:
            order_tag += " " + str(grind_entry_id)
          return -ft_sell_amount, order_tag, is_derisk

    if (
      (
        (grind_4_sub_grind_count > 0)
        and self.regular_mode_use_grind_stops
        and (((exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate) < regular_mode_grind_4_stop_grinds)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 16) or is_backtest)
    ):
      sell_amount = grind_4_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_4_current_open_rate > 0.0:
          grind_profit = (
            ((exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate)
            if grind_4_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (sg4) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (sg4) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "sg4"
        for grind_entry_id in grind_4_buy_orders:
          order_tag += " " + str(grind_entry_id)
        return -ft_sell_amount, order_tag, is_derisk


    if has_order_tags and (not partial_sell) and (grind_5_sub_grind_count < max_grind_5_sub_grinds):
      if (
        (
          (grind_5_distance_ratio if (grind_5_sub_grind_count > 0) else profit_init_ratio)
          < (regular_mode_grind_5_sub_thresholds[grind_5_sub_grind_count])
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit < -0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit < -0.06)
        )
        and ((num_open_grinds == 0) or (slice_profit < -0.02))
        and (
          (
            (
              (last_candle["close"] > (last_candle["close_max_12"] * 0.92))
              and (last_candle["close"] > (last_candle["close_max_24"] * 0.90))
              and (last_candle["close"] > (last_candle["close_max_48"] * 0.88))
              and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.86))
              and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.84))
              and (last_candle["close"] > (last_candle["high_max_6_1d"] * 0.76))
              and (last_candle["close"] > (last_candle["high_max_12_1d"] * 0.70))
            )
            and (
              is_long_grind_buy
              or (
                (last_candle["rsi_3"] > 12.0)
                and (last_candle["rsi_3_15m"] > 16.0)
                and (last_candle["rsi_3_1h"] > 16.0)
                and (last_candle["rsi_3_4h"] > 16.0)
                and (last_candle["rsi_14"] < 36.0)


                and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
              )
            )
          )
          or (
            (slice_profit < -0.06)
            and (last_candle["rsi_3"] > 10.0)
            and (last_candle["rsi_3_15m"] > 10.0)

            and (last_candle["rsi_14"] < 36.0)
            and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
          )
        )
      ):
        buy_amount = (
          slice_amount
          * regular_mode_grind_5_stakes[grind_5_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None, "", is_derisk
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_5_sub_grind_count > 0:
          grind_profit = (exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate
          grind_profit_stake = grind_5_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (g5) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_5_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (g5) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_5_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "g5"
        return buy_amount, order_tag, is_derisk

    if has_order_tags and grind_5_sub_grind_count > 0:
      grind_profit = (exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate
      if grind_profit > regular_mode_grind_5_profit_threshold:
        sell_amount = grind_5_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (g5) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (g5) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "g5"
          for grind_entry_id in grind_5_buy_orders:
            order_tag += " " + str(grind_entry_id)
          return -ft_sell_amount, order_tag, is_derisk

    if (
      (
        (grind_5_sub_grind_count > 0)
        and self.regular_mode_use_grind_stops
        and (((exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate) < regular_mode_grind_5_stop_grinds)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 16) or is_backtest)
    ):
      sell_amount = grind_5_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_5_current_open_rate > 0.0:
          grind_profit = (
            ((exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate)
            if grind_5_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (sg5) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (sg5) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "sg5"
        for grind_entry_id in grind_5_buy_orders:
          order_tag += " " + str(grind_entry_id)
        return -ft_sell_amount, order_tag, is_derisk

    if profit_stake < (
      slice_amount
      * (
        (self.regular_mode_derisk_futures if self.is_futures_mode else self.regular_mode_derisk_spot)
        if (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 3, 19) or is_backtest)
        else (self.regular_mode_derisk_futures_old if self.is_futures_mode else self.regular_mode_derisk_spot_old)
      )
      / (trade.leverage if self.is_futures_mode else 1.0)
    ):
      sell_amount = trade.amount * exit_rate / trade.leverage - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        self.dp.send_msg(
          f"De-risk [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        log.info(
          f"De-risk [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        return -ft_sell_amount, "d", is_derisk

    if (
      has_order_tags
      and not is_derisk_1
      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 5) or is_backtest)
      and profit_stake
      < (
        slice_amount
        * (
          (self.regular_mode_derisk_1_futures if self.is_futures_mode else self.regular_mode_derisk_1_spot)
          if (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 16) or is_backtest)
          else (
            self.regular_mode_derisk_1_futures_old if self.is_futures_mode else self.regular_mode_derisk_1_spot_old
          )
        )
      )
    ):
      sell_amount = trade.amount * exit_rate / trade.leverage - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        self.dp.send_msg(
          f"De-risk (d1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        log.info(
          f"De-risk (d1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        return -ft_sell_amount, "d1", is_derisk

    return None, "", is_derisk


  def long_rebuy_adjust_trade_position(
    self,
    trade: Trade,
    enter_tags,
    current_time: datetime,
    current_rate: float,
    current_profit: float,
    min_stake: Optional[float],
    max_stake: float,
    current_entry_rate: float,
    current_exit_rate: float,
    current_entry_profit: float,
    current_exit_profit: float,
    **kwargs,
  ) -> Optional[float]:
    df, _ = self.dp.get_analyzed_dataframe(trade.pair, self.timeframe)
    if len(df) < 2:
      return None
    last_candle = df.iloc[-1].squeeze()
    previous_candle = df.iloc[-2].squeeze()

    filled_orders = trade.select_filled_orders()
    filled_entries = trade.select_filled_orders(trade.entry_side)
    filled_exits = trade.select_filled_orders(trade.exit_side)
    count_of_entries = trade.nr_of_successful_entries
    count_of_exits = trade.nr_of_successful_exits

    if count_of_entries == 0:
      return None

    has_order_tags = False
    if hasattr(filled_orders[0], "ft_order_tag"):
      has_order_tags = True

    if count_of_exits > 0:
      return self.long_grind_adjust_trade_position(
        trade,
        enter_tags,
        current_time,
        current_rate,
        current_profit,
        min_stake,
        max_stake,
        current_entry_rate,
        current_exit_rate,
        current_entry_profit,
        current_exit_profit,
      )

    exit_rate = current_rate
    if self.dp.runmode.value in ("live", "dry_run"):
      ticker = self.dp.ticker(trade.pair)
      if ("bid" in ticker) and ("ask" in ticker):
        if trade.is_short:
          if self.config["exit_pricing"]["price_side"] in ["ask", "other"]:
            if ticker["ask"] is not None:
              exit_rate = ticker["ask"]
        else:
          if self.config["exit_pricing"]["price_side"] in ["bid", "other"]:
            if ticker["bid"] is not None:
              exit_rate = ticker["bid"]

    profit_stake, profit_ratio, profit_current_stake_ratio, profit_init_ratio = self.calc_total_profit(
      trade, filled_entries, filled_exits, exit_rate
    )

    slice_amount = filled_entries[0].cost
    slice_profit = (exit_rate - filled_orders[-1].safe_price) / filled_orders[-1].safe_price
    slice_profit_entry = (exit_rate - filled_entries[-1].safe_price) / filled_entries[-1].safe_price
    slice_profit_exit = (
      ((exit_rate - filled_exits[-1].safe_price) / filled_exits[-1].safe_price) if count_of_exits > 0 else 0.0
    )

    current_stake_amount = trade.amount * current_rate

    is_rebuy = False

    rebuy_mode_stakes = self.rebuy_mode_stakes_futures if self.is_futures_mode else self.rebuy_mode_stakes_spot
    max_sub_grinds = len(rebuy_mode_stakes)
    rebuy_mode_sub_thresholds = (
      self.rebuy_mode_thresholds_futures if self.is_futures_mode else self.rebuy_mode_thresholds_spot
    )
    partial_sell = False
    sub_grind_count = 0
    total_amount = 0.0
    total_cost = 0.0
    current_open_rate = 0.0
    current_grind_stake = 0.0
    current_grind_stake_profit = 0.0
    for order in reversed(filled_orders):
      if (order.ft_order_side == "buy") and (order is not filled_orders[0]):
        sub_grind_count += 1
        total_amount += order.safe_filled
        total_cost += order.safe_filled * order.safe_price
      elif order.ft_order_side == "sell":
        if (order.safe_remaining * exit_rate / (trade.leverage if self.is_futures_mode else 1.0)) > min_stake:
          partial_sell = True
        break
    if sub_grind_count > 0:
      current_open_rate = total_cost / total_amount
      current_grind_stake = total_amount * exit_rate * (1 - trade.fee_close)
      current_grind_stake_profit = current_grind_stake - total_cost

    if (not partial_sell) and (sub_grind_count < max_sub_grinds):
      if (
        ((0 <= sub_grind_count < max_sub_grinds) and (slice_profit_entry < rebuy_mode_sub_thresholds[sub_grind_count]))
        and (
          (last_candle["close"] > (last_candle["close_max_12"] * 0.94))
          and (last_candle["close"] > (last_candle["close_max_24"] * 0.92))
          and (last_candle["close"] > (last_candle["close_max_48"] * 0.90))
          and (last_candle["close"] > (last_candle["high_max_24_1h"] * 0.88))
          and (last_candle["close"] > (last_candle["high_max_48_1h"] * 0.86))
          and (last_candle["btc_pct_close_max_72_5m"] < 0.03)
          and (last_candle["btc_pct_close_max_24_5m"] < 0.03)
        )
        and (
          (last_candle["rsi_3"] > 10.0)
          and (last_candle["rsi_3_15m"] > 10.0)
          and (last_candle["rsi_3_1h"] > 10.0)
          and (last_candle["rsi_3_4h"] > 10.0)
          and (last_candle["rsi_14"] < 36.0)
          and (last_candle["close"] < (last_candle["ema_26"] * 0.988))
        )
      ):
        buy_amount = (
          slice_amount * rebuy_mode_stakes[sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount > max_stake:
          buy_amount = max_stake
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        self.dp.send_msg(
          f"Rebuy (r) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        log.info(
          f"Rebuy (r) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        if has_order_tags:
          return buy_amount, "r"
        else:
          return buy_amount

      if profit_stake < (
        slice_amount
        * (self.rebuy_mode_derisk_futures if self.is_futures_mode else self.rebuy_mode_derisk_spot)
        / (trade.leverage if self.is_futures_mode else 1.0)
      ):
        sell_amount = trade.amount * exit_rate / trade.leverage - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          grind_profit = 0.0
          self.dp.send_msg(
            f"Rebuy de-risk (d1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
          )
          log.info(
            f"Rebuy de-risk (d1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
          )
          if has_order_tags:
            return -ft_sell_amount, "d1"
          else:
            return -ft_sell_amount

    return None






























  def short_exit_normal(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    sell = False

    sell, signal_name = self.short_exit_signals(
      self.short_normal_mode_name,
      profit_init_ratio,
      max_profit,
      max_loss,
      last_candle,
      previous_candle_1,
      previous_candle_2,
      previous_candle_3,
      previous_candle_4,
      previous_candle_5,
      trade,
      current_time,
      enter_tags,
    )

    if not sell:
      sell, signal_name = self.short_exit_main(
        self.short_normal_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_williams_r(
        self.short_normal_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_dec(
        self.short_normal_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_stoploss(
        self.short_normal_mode_name,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )


    if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
      previous_rate = self.target_profit_cache.data[pair]["rate"]
      previous_profit = self.target_profit_cache.data[pair]["profit"]
      previous_sell_reason = self.target_profit_cache.data[pair]["sell_reason"]
      previous_time_profit_reached = datetime.fromisoformat(self.target_profit_cache.data[pair]["time_profit_reached"])

      sell_max, signal_name_max = self.exit_profit_target(
        self.short_normal_mode_name,
        pair,
        trade,
        current_time,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        last_candle,
        previous_candle_1,
        previous_rate,
        previous_profit,
        previous_sell_reason,
        previous_time_profit_reached,
        enter_tags,
      )
      if sell_max and signal_name_max is not None:
        return True, f"{signal_name_max}_m"
      if previous_sell_reason in [f"exit_{self.short_normal_mode_name}_stoploss_u_e"]:
        if profit_ratio > (previous_profit + 0.005):
          mark_pair, mark_signal = self.mark_profit_target(
            self.short_normal_mode_name,
            pair,
            True,
            previous_sell_reason,
            trade,
            current_time,
            current_rate,
            profit_ratio,
            last_candle,
            previous_candle_1,
          )
          if mark_pair:
            self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
      elif (profit_init_ratio > (previous_profit + 0.001)) and (
        previous_sell_reason not in [f"exit_{self.short_normal_mode_name}_stoploss_doom"]
      ):

        mark_pair, mark_signal = self.mark_profit_target(
          self.short_normal_mode_name,
          pair,
          True,
          previous_sell_reason,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if sell and signal_name is not None:
      previous_profit = None
      if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
        previous_profit = self.target_profit_cache.data[pair]["profit"]
      if signal_name in [
        f"exit_{self.short_normal_mode_name}_stoploss_doom",
        f"exit_{self.short_normal_mode_name}_stoploss_u_e",
      ]:
        mark_pair, mark_signal = self.mark_profit_target(
          self.short_normal_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
        else:

          return True, f"{signal_name}"
      elif (previous_profit is None) or (previous_profit < profit_init_ratio):
        mark_pair, mark_signal = self.mark_profit_target(
          self.short_normal_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)
        else:

          return True, f"{signal_name}"
    else:
      if profit_init_ratio >= self.profit_max_thresholds[0]:
        previous_profit = None
        if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
          previous_profit = self.target_profit_cache.data[pair]["profit"]
        if (previous_profit is None) or (previous_profit < profit_init_ratio):
          mark_signal = f"exit_profit_{self.short_normal_mode_name}_max"
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if signal_name not in [
      f"exit_profit_{self.short_normal_mode_name}_max",
      f"exit_{self.short_normal_mode_name}_stoploss_doom",
      f"exit_{self.short_normal_mode_name}_stoploss_u_e",
    ]:
      if sell and (signal_name is not None):
        return True, f"{signal_name}"

    return False, None


  def short_exit_pump(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    sell = False

    sell, signal_name = self.short_exit_signals(
      self.short_pump_mode_name,
      profit_init_ratio,
      max_profit,
      max_loss,
      last_candle,
      previous_candle_1,
      previous_candle_2,
      previous_candle_3,
      previous_candle_4,
      previous_candle_5,
      trade,
      current_time,
      enter_tags,
    )

    if not sell:
      sell, signal_name = self.short_exit_main(
        self.short_pump_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_williams_r(
        self.short_pump_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_dec(
        self.short_pump_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_stoploss(
        self.short_pump_mode_name,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )


    if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
      previous_rate = self.target_profit_cache.data[pair]["rate"]
      previous_profit = self.target_profit_cache.data[pair]["profit"]
      previous_sell_reason = self.target_profit_cache.data[pair]["sell_reason"]
      previous_time_profit_reached = datetime.fromisoformat(self.target_profit_cache.data[pair]["time_profit_reached"])

      sell_max, signal_name_max = self.exit_profit_target(
        self.short_pump_mode_name,
        pair,
        trade,
        current_time,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        last_candle,
        previous_candle_1,
        previous_rate,
        previous_profit,
        previous_sell_reason,
        previous_time_profit_reached,
        enter_tags,
      )
      if sell_max and signal_name_max is not None:
        return True, f"{signal_name_max}_m"
      if previous_sell_reason in [f"exit_{self.short_pump_mode_name}_stoploss_u_e"]:
        if profit_ratio > (previous_profit + 0.005):
          mark_pair, mark_signal = self.mark_profit_target(
            self.short_pump_mode_name,
            pair,
            True,
            previous_sell_reason,
            trade,
            current_time,
            current_rate,
            profit_ratio,
            last_candle,
            previous_candle_1,
          )
          if mark_pair:
            self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
      elif (profit_init_ratio > (previous_profit + 0.001)) and (
        previous_sell_reason not in [f"exit_{self.short_pump_mode_name}_stoploss_doom"]
      ):

        mark_pair, mark_signal = self.mark_profit_target(
          self.short_pump_mode_name,
          pair,
          True,
          previous_sell_reason,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if sell and signal_name is not None:
      previous_profit = None
      if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
        previous_profit = self.target_profit_cache.data[pair]["profit"]
      if signal_name in [
        f"exit_{self.short_pump_mode_name}_stoploss_doom",
        f"exit_{self.short_pump_mode_name}_stoploss_u_e",
      ]:
        mark_pair, mark_signal = self.mark_profit_target(
          self.short_pump_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
        else:

          return True, f"{signal_name}"
      elif (previous_profit is None) or (previous_profit < profit_init_ratio):
        mark_pair, mark_signal = self.mark_profit_target(
          self.short_pump_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)
        else:

          return True, f"{signal_name}"
    else:
      if profit_init_ratio >= self.profit_max_thresholds[2]:
        previous_profit = None
        if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
          previous_profit = self.target_profit_cache.data[pair]["profit"]
        if (previous_profit is None) or (previous_profit < profit_init_ratio):
          mark_signal = f"exit_profit_{self.short_pump_mode_name}_max"
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if signal_name not in [
      f"exit_profit_{self.short_pump_mode_name}_max",
      f"exit_{self.short_pump_mode_name}_stoploss_doom",
      f"exit_{self.short_pump_mode_name}_stoploss_u_e",
    ]:
      if sell and (signal_name is not None):
        return True, f"{signal_name}"

    return False, None


  def short_exit_quick(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    sell = False

    sell, signal_name = self.short_exit_signals(
      self.short_quick_mode_name,
      profit_init_ratio,
      max_profit,
      max_loss,
      last_candle,
      previous_candle_1,
      previous_candle_2,
      previous_candle_3,
      previous_candle_4,
      previous_candle_5,
      trade,
      current_time,
      enter_tags,
    )

    if not sell:
      sell, signal_name = self.short_exit_main(
        self.short_quick_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_williams_r(
        self.short_quick_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_dec(
        self.short_quick_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_stoploss(
        self.short_quick_mode_name,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      if (0.09 >= profit_init_ratio > 0.02) and (last_candle["rsi_14"] < 22.0):
        sell, signal_name = True, f"exit_{self.short_quick_mode_name}_q_1"

      if (0.09 >= profit_init_ratio > 0.02) and (last_candle["cti_20"] < -0.95):
        sell, signal_name = True, f"exit_{self.short_quick_mode_name}_q_2"

      if (0.09 >= profit_init_ratio > 0.02) and (last_candle["r_14"] <= -0.99):
        sell, signal_name = True, f"exit_{self.short_quick_mode_name}_q_3"


    if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
      previous_rate = self.target_profit_cache.data[pair]["rate"]
      previous_profit = self.target_profit_cache.data[pair]["profit"]
      previous_sell_reason = self.target_profit_cache.data[pair]["sell_reason"]
      previous_time_profit_reached = datetime.fromisoformat(self.target_profit_cache.data[pair]["time_profit_reached"])

      sell_max, signal_name_max = self.exit_profit_target(
        self.short_quick_mode_name,
        pair,
        trade,
        current_time,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        last_candle,
        previous_candle_1,
        previous_rate,
        previous_profit,
        previous_sell_reason,
        previous_time_profit_reached,
        enter_tags,
      )
      if sell_max and signal_name_max is not None:
        return True, f"{signal_name_max}_m"
      if previous_sell_reason in [f"exit_{self.short_quick_mode_name}_stoploss_u_e"]:
        if profit_ratio > (previous_profit + 0.005):
          mark_pair, mark_signal = self.mark_profit_target(
            self.short_quick_mode_name,
            pair,
            True,
            previous_sell_reason,
            trade,
            current_time,
            current_rate,
            profit_ratio,
            last_candle,
            previous_candle_1,
          )
          if mark_pair:
            self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
      elif (profit_init_ratio > (previous_profit + 0.001)) and (
        previous_sell_reason not in [f"exit_{self.short_quick_mode_name}_stoploss_doom"]
      ):

        mark_pair, mark_signal = self.mark_profit_target(
          self.short_quick_mode_name,
          pair,
          True,
          previous_sell_reason,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if sell and signal_name is not None:
      previous_profit = None
      if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
        previous_profit = self.target_profit_cache.data[pair]["profit"]
      if signal_name in [
        f"exit_{self.short_quick_mode_name}_stoploss_doom",
        f"exit_{self.short_quick_mode_name}_stoploss_u_e",
      ]:
        mark_pair, mark_signal = self.mark_profit_target(
          self.short_quick_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
        else:

          return True, f"{signal_name}"
      elif (previous_profit is None) or (previous_profit < profit_init_ratio):
        mark_pair, mark_signal = self.mark_profit_target(
          self.short_quick_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)
        else:

          return True, f"{signal_name}"
    else:
      if profit_init_ratio >= self.profit_max_thresholds[4]:
        previous_profit = None
        if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
          previous_profit = self.target_profit_cache.data[pair]["profit"]
        if (previous_profit is None) or (previous_profit < profit_init_ratio):
          mark_signal = f"exit_profit_{self.short_quick_mode_name}_max"
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if signal_name not in [
      f"exit_profit_{self.short_quick_mode_name}_max",
      f"exit_{self.short_quick_mode_name}_stoploss_doom",
      f"exit_{self.short_quick_mode_name}_stoploss_u_e",
    ]:
      if sell and (signal_name is not None):
        return True, f"{signal_name}"

    return False, None


  def short_exit_rebuy(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    sell = False

    sell, signal_name = self.short_exit_signals(
      self.short_rebuy_mode_name,
      profit_init_ratio,
      max_profit,
      max_loss,
      last_candle,
      previous_candle_1,
      previous_candle_2,
      previous_candle_3,
      previous_candle_4,
      previous_candle_5,
      trade,
      current_time,
      enter_tags,
    )

    if not sell:
      sell, signal_name = self.short_exit_main(
        self.short_rebuy_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_williams_r(
        self.short_rebuy_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_dec(
        self.short_rebuy_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      if profit_stake < -(
        filled_entries[0].cost
        * (self.stop_threshold_futures_rebuy if self.is_futures_mode else self.stop_threshold_spot_rebuy)
        / (trade.leverage if self.is_futures_mode else 1.0)
      ):
        sell, signal_name = True, f"exit_{self.short_rebuy_mode_name}_stoploss_doom"


    if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
      previous_rate = self.target_profit_cache.data[pair]["rate"]
      previous_profit = self.target_profit_cache.data[pair]["profit"]
      previous_sell_reason = self.target_profit_cache.data[pair]["sell_reason"]
      previous_time_profit_reached = datetime.fromisoformat(self.target_profit_cache.data[pair]["time_profit_reached"])

      sell_max, signal_name_max = self.exit_profit_target(
        self.short_rebuy_mode_name,
        pair,
        trade,
        current_time,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        last_candle,
        previous_candle_1,
        previous_rate,
        previous_profit,
        previous_sell_reason,
        previous_time_profit_reached,
        enter_tags,
      )
      if sell_max and signal_name_max is not None:
        return True, f"{signal_name_max}_m"
      if previous_sell_reason in [f"exit_{self.short_rebuy_mode_name}_stoploss_u_e"]:
        if profit_ratio > (previous_profit + 0.005):
          mark_pair, mark_signal = self.mark_profit_target(
            self.short_rebuy_mode_name,
            pair,
            True,
            previous_sell_reason,
            trade,
            current_time,
            current_rate,
            profit_ratio,
            last_candle,
            previous_candle_1,
          )
          if mark_pair:
            self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
      elif (profit_init_ratio > (previous_profit + 0.001)) and (
        previous_sell_reason not in [f"exit_{self.short_rebuy_mode_name}_stoploss_doom"]
      ):

        mark_pair, mark_signal = self.mark_profit_target(
          self.short_rebuy_mode_name,
          pair,
          True,
          previous_sell_reason,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if sell and signal_name is not None:
      previous_profit = None
      if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
        previous_profit = self.target_profit_cache.data[pair]["profit"]
      if signal_name in [
        f"exit_{self.short_rebuy_mode_name}_stoploss_doom",
        f"exit_{self.short_rebuy_mode_name}_stoploss_u_e",
      ]:
        mark_pair, mark_signal = self.mark_profit_target(
          self.short_rebuy_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
        else:

          return True, f"{signal_name}"
      elif (previous_profit is None) or (previous_profit < profit_init_ratio):
        mark_pair, mark_signal = self.mark_profit_target(
          self.short_rebuy_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)
        else:

          return True, f"{signal_name}"
    else:
      if profit_init_ratio >= self.profit_max_thresholds[6]:
        previous_profit = None
        if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
          previous_profit = self.target_profit_cache.data[pair]["profit"]
        if (previous_profit is None) or (previous_profit < profit_init_ratio):
          mark_signal = f"exit_profit_{self.short_rebuy_mode_name}_max"
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if signal_name not in [f"exit_profit_{self.short_rebuy_mode_name}_max"]:
      if sell and (signal_name is not None):
        return True, f"{signal_name}"

    return False, None


  def short_exit_high_profit(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    sell = False

    sell, signal_name = self.short_exit_signals(
      self.short_high_profit_mode_name,
      profit_init_ratio,
      max_profit,
      max_loss,
      last_candle,
      previous_candle_1,
      previous_candle_2,
      previous_candle_3,
      previous_candle_4,
      previous_candle_5,
      trade,
      current_time,
      enter_tags,
    )

    if not sell:
      sell, signal_name = self.short_exit_main(
        self.short_high_profit_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_williams_r(
        self.short_high_profit_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_stoploss(
        self.short_high_profit_mode_name,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )


    if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
      previous_rate = self.target_profit_cache.data[pair]["rate"]
      previous_profit = self.target_profit_cache.data[pair]["profit"]
      previous_sell_reason = self.target_profit_cache.data[pair]["sell_reason"]
      previous_time_profit_reached = datetime.fromisoformat(self.target_profit_cache.data[pair]["time_profit_reached"])

      sell_max, signal_name_max = self.exit_profit_target(
        self.short_high_profit_mode_name,
        pair,
        trade,
        current_time,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        last_candle,
        previous_candle_1,
        previous_rate,
        previous_profit,
        previous_sell_reason,
        previous_time_profit_reached,
        enter_tags,
      )
      if sell_max and signal_name_max is not None:
        return True, f"{signal_name_max}_m"
      if previous_sell_reason in [f"exit_{self.short_high_profit_mode_name}_stoploss_u_e"]:
        if profit_ratio > (previous_profit + 0.005):
          mark_pair, mark_signal = self.mark_profit_target(
            self.short_high_profit_mode_name,
            pair,
            True,
            previous_sell_reason,
            trade,
            current_time,
            current_rate,
            profit_ratio,
            last_candle,
            previous_candle_1,
          )
          if mark_pair:
            self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
      elif (profit_init_ratio > (previous_profit + 0.001)) and (
        previous_sell_reason not in [f"exit_{self.short_high_profit_mode_name}_stoploss_doom"]
      ):

        mark_pair, mark_signal = self.mark_profit_target(
          self.short_high_profit_mode_name,
          pair,
          True,
          previous_sell_reason,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if sell and signal_name is not None:
      previous_profit = None
      if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
        previous_profit = self.target_profit_cache.data[pair]["profit"]
      if signal_name in [
        f"exit_{self.short_high_profit_mode_name}_stoploss_doom",
        f"exit_{self.short_high_profit_mode_name}_stoploss_u_e",
      ]:
        mark_pair, mark_signal = self.mark_profit_target(
          self.short_high_profit_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
        else:

          return True, f"{signal_name}"
      elif (previous_profit is None) or (previous_profit < profit_init_ratio):
        mark_pair, mark_signal = self.mark_profit_target(
          self.short_high_profit_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)
        else:

          return True, f"{signal_name}"
    else:
      if profit_init_ratio >= self.profit_max_thresholds[8]:
        previous_profit = None
        if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
          previous_profit = self.target_profit_cache.data[pair]["profit"]
        if (previous_profit is None) or (previous_profit < profit_init_ratio):
          mark_signal = f"exit_profit_{self.short_high_profit_mode_name}_max"
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if signal_name not in [
      f"exit_profit_{self.short_high_profit_mode_name}_max",
      f"exit_{self.short_high_profit_mode_name}_stoploss_doom",
      f"exit_{self.short_high_profit_mode_name}_stoploss_u_e",
    ]:
      if sell and (signal_name is not None):
        return True, f"{signal_name}"

    return False, None


  def short_exit_rapid(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    sell = False

    sell, signal_name = self.short_exit_signals(
      self.short_rapid_mode_name,
      profit_init_ratio,
      max_profit,
      max_loss,
      last_candle,
      previous_candle_1,
      previous_candle_2,
      previous_candle_3,
      previous_candle_4,
      previous_candle_5,
      trade,
      current_time,
      enter_tags,
    )

    if not sell:
      sell, signal_name = self.short_exit_main(
        self.short_rapid_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_williams_r(
        self.short_rapid_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_dec(
        self.short_rapid_mode_name,
        profit_init_ratio,
        max_profit,
        max_loss,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      sell, signal_name = self.short_exit_stoploss(
        self.short_rapid_mode_name,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        max_profit,
        max_loss,
        filled_entries,
        filled_exits,
        last_candle,
        previous_candle_1,
        previous_candle_2,
        previous_candle_3,
        previous_candle_4,
        previous_candle_5,
        trade,
        current_time,
        enter_tags,
      )

    if not sell:
      if (0.09 >= profit_init_ratio > 0.01) and (last_candle["rsi_14"] < 22.0):
        sell, signal_name = True, f"exit_{self.short_rapid_mode_name}_rpd_1"

      if (0.09 >= profit_init_ratio > 0.01) and (last_candle["cti_20"] < -0.95):
        sell, signal_name = True, f"exit_{self.short_rapid_mode_name}_rpd_2"

      if (0.09 >= profit_init_ratio > 0.01) and (last_candle["r_14"] <= -0.99):
        sell, signal_name = True, f"exit_{self.short_rapid_mode_name}_rpd_3"

      if profit_stake < -(
        filled_entries[0].cost
        * (self.stop_threshold_futures_rapid if self.is_futures_mode else self.stop_threshold_spot_rapid)
        / (trade.leverage if self.is_futures_mode else 1.0)
      ):
        sell, signal_name = True, f"exit_{self.short_rapid_mode_name}_stoploss_doom"


    if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
      previous_rate = self.target_profit_cache.data[pair]["rate"]
      previous_profit = self.target_profit_cache.data[pair]["profit"]
      previous_sell_reason = self.target_profit_cache.data[pair]["sell_reason"]
      previous_time_profit_reached = datetime.fromisoformat(self.target_profit_cache.data[pair]["time_profit_reached"])

      sell_max, signal_name_max = self.exit_profit_target(
        self.short_rapid_mode_name,
        pair,
        trade,
        current_time,
        current_rate,
        profit_stake,
        profit_ratio,
        profit_current_stake_ratio,
        profit_init_ratio,
        last_candle,
        previous_candle_1,
        previous_rate,
        previous_profit,
        previous_sell_reason,
        previous_time_profit_reached,
        enter_tags,
      )
      if sell_max and signal_name_max is not None:
        return True, f"{signal_name_max}_m"
      if previous_sell_reason in [f"exit_{self.short_rapid_mode_name}_stoploss_u_e"]:
        if profit_ratio > (previous_profit + 0.005):
          mark_pair, mark_signal = self.mark_profit_target(
            self.short_rapid_mode_name,
            pair,
            True,
            previous_sell_reason,
            trade,
            current_time,
            current_rate,
            profit_ratio,
            last_candle,
            previous_candle_1,
          )
          if mark_pair:
            self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
      elif (profit_init_ratio > (previous_profit + 0.001)) and (
        previous_sell_reason not in [f"exit_{self.short_rapid_mode_name}_stoploss_doom"]
      ):

        mark_pair, mark_signal = self.mark_profit_target(
          self.short_rapid_mode_name,
          pair,
          True,
          previous_sell_reason,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if sell and signal_name is not None:
      previous_profit = None
      if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
        previous_profit = self.target_profit_cache.data[pair]["profit"]
      if signal_name in [
        f"exit_{self.short_rapid_mode_name}_stoploss_doom",
        f"exit_{self.short_rapid_mode_name}_stoploss_u_e",
      ]:
        mark_pair, mark_signal = self.mark_profit_target(
          self.short_rapid_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_ratio, current_time)
        else:

          return True, f"{signal_name}"
      elif (previous_profit is None) or (previous_profit < profit_init_ratio):
        mark_pair, mark_signal = self.mark_profit_target(
          self.short_rapid_mode_name,
          pair,
          sell,
          signal_name,
          trade,
          current_time,
          current_rate,
          profit_init_ratio,
          last_candle,
          previous_candle_1,
        )
        if mark_pair:
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)
        else:

          return True, f"{signal_name}"
    else:
      if profit_init_ratio >= 0.01:
        previous_profit = None
        if self.target_profit_cache is not None and pair in self.target_profit_cache.data:
          previous_profit = self.target_profit_cache.data[pair]["profit"]
        if (previous_profit is None) or (previous_profit < profit_init_ratio):
          mark_signal = f"exit_profit_{self.short_rapid_mode_name}_max"
          self._set_profit_target(pair, mark_signal, current_rate, profit_init_ratio, current_time)

    if signal_name not in [f"exit_profit_{self.short_rapid_mode_name}_max"]:
      if sell and (signal_name is not None):
        return True, f"{signal_name}"

    return False, None


  def short_exit_grind(
    self,
    pair: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    enter_tags,
  ) -> tuple:
    if len(filled_exits) > 30 and profit_init_ratio > 1.0:
      return True, f"exit_{self.short_grind_mode_name}_g"
    return False, None


  def short_exit_signals(
    self,
    mode_name: str,
    current_profit: float,
    max_profit: float,
    max_loss: float,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    buy_tag,
  ) -> tuple:

    if (
      (last_candle["rsi_14"] < 30.0)
      and (last_candle["close"] < last_candle["bb20_2_low"])
      and (previous_candle_1["close"] < previous_candle_1["bb20_2_low"])
      and (previous_candle_2["close"] < previous_candle_2["bb20_2_low"])
      and (previous_candle_3["close"] < previous_candle_3["bb20_2_low"])
      and (previous_candle_4["close"] < previous_candle_4["bb20_2_low"])
    ):
      if last_candle["close"] < last_candle["ema_200"]:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_1_1_1"
      else:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_1_2_1"

    elif (
      (last_candle["rsi_14"] < 20.0)
      and (last_candle["close"] < last_candle["bb20_2_low"])
      and (previous_candle_1["close"] < previous_candle_1["bb20_2_low"])
      and (previous_candle_2["close"] < previous_candle_2["bb20_2_low"])
    ):
      if last_candle["close"] < last_candle["ema_200"]:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_2_1_1"
      else:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_2_2_1"

    elif last_candle["rsi_14"] < 15.0:
      if last_candle["close"] < last_candle["ema_200"]:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_3_1_1"
      else:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_3_2_1"

    elif (last_candle["rsi_14"] < 20.0) and (last_candle["rsi_14_1h"] < 22.0):
      if last_candle["close"] < last_candle["ema_200"]:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_4_1_1"
      else:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_4_2_1"

    elif (
      (last_candle["close"] > last_candle["ema_200"])
      and (last_candle["close"] < last_candle["ema_50"])
      and (last_candle["rsi_14"] < 21.0)
    ):
      if current_profit > 0.01:
        return True, f"exit_{mode_name}_6_1"

    elif (last_candle["rsi_14_1h"] < 21.0) and (last_candle["crossed_above_ema_12_26"]):
      if last_candle["close"] < last_candle["ema_200"]:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_7_1_1"
      else:
        if current_profit > 0.01:
          return True, f"exit_{mode_name}_7_2_1"









    return False, None


  def short_exit_main(
    self,
    mode_name: str,
    current_profit: float,
    max_profit: float,
    max_loss: float,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    buy_tag,
  ) -> tuple:
    if last_candle["close"] < last_candle["sma_200"]:
      if 0.02 > current_profit >= 0.01:
        if last_candle["rsi_14"] > 72.0:
          return True, f"exit_{mode_name}_o_1"
      elif 0.03 > current_profit >= 0.02:
        if last_candle["rsi_14"] > 70.0:
          return True, f"exit_{mode_name}_o_2"
      elif 0.04 > current_profit >= 0.03:
        if last_candle["rsi_14"] > 68.0:
          return True, f"exit_{mode_name}_o_3"
      elif 0.05 > current_profit >= 0.04:
        if last_candle["rsi_14"] > 66.0:
          return True, f"exit_{mode_name}_o_4"
      elif 0.06 > current_profit >= 0.05:
        if last_candle["rsi_14"] > 64.0:
          return True, f"exit_{mode_name}_o_5"
      elif 0.07 > current_profit >= 0.06:
        if last_candle["rsi_14"] > 62.0:
          return True, f"exit_{mode_name}_o_6"
      elif 0.08 > current_profit >= 0.07:
        if last_candle["rsi_14"] > 60.0:
          return True, f"exit_{mode_name}_o_7"
      elif 0.09 > current_profit >= 0.08:
        if last_candle["rsi_14"] > 58.0:
          return True, f"exit_{mode_name}_o_8"
      elif 0.1 > current_profit >= 0.09:
        if last_candle["rsi_14"] > 56.0:
          return True, f"exit_{mode_name}_o_9"
      elif 0.12 > current_profit >= 0.1:
        if last_candle["rsi_14"] > 54.0:
          return True, f"exit_{mode_name}_o_10"
      elif 0.2 > current_profit >= 0.12:
        if last_candle["rsi_14"] > 56.0:
          return True, f"exit_{mode_name}_o_11"
      elif current_profit >= 0.2:
        if last_candle["rsi_14"] > 58.0:
          return True, f"exit_{mode_name}_o_12"
    elif last_candle["close"] > last_candle["sma_200"]:
      if 0.02 > current_profit >= 0.01:
        if last_candle["rsi_14"] > 70.0:
          return True, f"exit_{mode_name}_u_1"
      elif 0.03 > current_profit >= 0.02:
        if last_candle["rsi_14"] > 68.0:
          return True, f"exit_{mode_name}_u_2"
      elif 0.04 > current_profit >= 0.03:
        if last_candle["rsi_14"] > 66.0:
          return True, f"exit_{mode_name}_u_3"
      elif 0.05 > current_profit >= 0.04:
        if last_candle["rsi_14"] > 64.0:
          return True, f"exit_{mode_name}_u_4"
      elif 0.06 > current_profit >= 0.05:
        if last_candle["rsi_14"] > 62.0:
          return True, f"exit_{mode_name}_u_5"
      elif 0.07 > current_profit >= 0.06:
        if last_candle["rsi_14"] > 60.0:
          return True, f"exit_{mode_name}_u_6"
      elif 0.08 > current_profit >= 0.07:
        if last_candle["rsi_14"] > 58.0:
          return True, f"exit_{mode_name}_u_7"
      elif 0.09 > current_profit >= 0.08:
        if last_candle["rsi_14"] > 56.0:
          return True, f"exit_{mode_name}_u_8"
      elif 0.1 > current_profit >= 0.09:
        if last_candle["rsi_14"] > 54.0:
          return True, f"exit_{mode_name}_u_9"
      elif 0.12 > current_profit >= 0.1:
        if last_candle["rsi_14"] > 52.0:
          return True, f"exit_{mode_name}_u_10"
      elif 0.2 > current_profit >= 0.12:
        if last_candle["rsi_14"] > 54.0:
          return True, f"exit_{mode_name}_u_11"
      elif current_profit >= 0.2:
        if last_candle["rsi_14"] > 56.0:
          return True, f"exit_{mode_name}_u_12"

    return False, None


  def short_exit_williams_r(
    self,
    mode_name: str,
    current_profit: float,
    max_profit: float,
    max_loss: float,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    buy_tag,
  ) -> tuple:
    if 0.02 > current_profit >= 0.01:
      if last_candle["r_480"] < -99.8:
        return True, f"exit_{mode_name}_w_1_1"
      elif (last_candle["r_14"] <= -99.0) and (last_candle["rsi_14"] < 22.0):
        return True, f"exit_{mode_name}_w_1_2"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["rsi_14"] > 54.0):
        return True, f"exit_{mode_name}_w_1_3"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["rsi_14"] < 26.0) and (last_candle["r_480_1h"] < -75.0):
        return True, f"exit_{mode_name}_w_1_4"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["cti_20"] < -0.95):
        return True, f"exit_{mode_name}_w_1_5"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] < 30.0)
        and (last_candle["r_480_1h"] < -90.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 40.0)
        and (last_candle["cti_20_1d"] < -0.80)
      ):
        return True, f"exit_{mode_name}_w_1_6"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_4h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 50.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_1_7"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["rsi_14_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.02)
      ):
        return True, f"exit_{mode_name}_w_1_8"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 24.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["cti_20_1d"] <= -0.70)
      ):
        return True, f"exit_{mode_name}_w_1_9"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_1_10"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["rsi_14_1h"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["r_480_1h"] < -70.0)
      ):
        return True, f"exit_{mode_name}_w_1_11"
      elif (
        (last_candle["r_14"] <= -76.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_1_12"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["rsi_14_4h"] <= 35.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_1_13"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1h"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
      ):
        return True, f"exit_{mode_name}_w_1_14"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["change_pct_1h"] > 0.00)
      ):
        return True, f"exit_{mode_name}_w_1_15"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["bot_wick_pct_1d"] > 0.16)
        and (last_candle["close"] > (last_candle["low_min_24_1h"] * 1.20))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_1_16"
      elif (
        (last_candle["r_14"] <= -84.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["r_480_1h"] < -70.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["bot_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_1_17"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_1_18"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 20.0)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_1_19"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 35.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_1_20"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 35.0)
        and (last_candle["cti_20_dec_3_1h"] == False)
        and (last_candle["r_480_4h"] > -50.0)
      ):
        return True, f"exit_{mode_name}_w_1_21"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_1_22"
    elif 0.03 > current_profit >= 0.02:
      if last_candle["r_480"] < -99.7:
        return True, f"exit_{mode_name}_w_2_1"
      elif (last_candle["r_14"] <= -99.0) and (last_candle["rsi_14"] < 23.0):
        return True, f"exit_{mode_name}_w_2_2"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["rsi_14"] > 52.0):
        return True, f"exit_{mode_name}_w_2_3"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["rsi_14"] < 27.0) and (last_candle["r_480_1h"] < -75.0):
        return True, f"exit_{mode_name}_w_2_4"
      elif (last_candle["r_14"] <= -97.0) and (last_candle["cti_20"] < -0.95):
        return True, f"exit_{mode_name}_w_2_5"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -80.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 40.0)
        and (last_candle["cti_20_1d"] < -0.80)
      ):
        return True, f"exit_{mode_name}_w_2_6"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_4h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 50.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_2_7"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["rsi_14_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.02)
      ):
        return True, f"exit_{mode_name}_w_2_8"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["rsi_14_15m"] <= 32.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["cti_20_1d"] <= -0.70)
      ):
        return True, f"exit_{mode_name}_w_2_9"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_2_10"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["rsi_14_1h"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["r_480_1h"] < -70.0)
      ):
        return True, f"exit_{mode_name}_w_2_11"
      elif (
        (last_candle["r_14"] <= -76.0)
        and (last_candle["rsi_14"] <= 28.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_2_12"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["rsi_14_4h"] <= 35.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_2_13"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1h"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
      ):
        return True, f"exit_{mode_name}_w_2_14"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["change_pct_1h"] > 0.00)
      ):
        return True, f"exit_{mode_name}_w_2_15"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["bot_wick_pct_1d"] > 0.16)
        and (last_candle["close"] > (last_candle["low_min_24_1h"] * 1.20))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_2_16"
      elif (
        (last_candle["r_14"] <= -82.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["r_480_1h"] < -70.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["bot_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_2_17"
      elif (
        (last_candle["r_14"] <= -95.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_2_18"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 20.0)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_2_19"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 35.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_2_20"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["rsi_14_15m"] <= 35.0)
        and (last_candle["cti_20_dec_3_1h"] == False)
        and (last_candle["r_480_4h"] > -50.0)
      ):
        return True, f"exit_{mode_name}_w_2_21"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_2_22"
    elif 0.04 > current_profit >= 0.03:
      if last_candle["r_480"] < -99.6:
        return True, f"exit_{mode_name}_w_3_1"
      elif (last_candle["r_14"] <= -99.0) and (last_candle["rsi_14"] < 24.0):
        return True, f"exit_{mode_name}_w_3_2"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["rsi_14"] > 50.0):
        return True, f"exit_{mode_name}_w_3_3"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["rsi_14"] < 28.0) and (last_candle["r_480_1h"] < -75.0):
        return True, f"exit_{mode_name}_w_3_4"
      elif (last_candle["r_14"] <= -96.0) and (last_candle["cti_20"] < -0.95):
        return True, f"exit_{mode_name}_w_3_5"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -80.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 40.0)
        and (last_candle["cti_20_1d"] < -0.80)
      ):
        return True, f"exit_{mode_name}_w_3_6"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_4h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 50.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_3_7"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["rsi_14_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.02)
      ):
        return True, f"exit_{mode_name}_w_3_8"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 28.0)
        and (last_candle["rsi_14_15m"] <= 34.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["cti_20_1d"] <= -0.70)
      ):
        return True, f"exit_{mode_name}_w_3_9"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_3_10"
      elif (
        (last_candle["r_14"] <= -97.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["rsi_14_1h"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["r_480_1h"] < -70.0)
      ):
        return True, f"exit_{mode_name}_w_3_11"
      elif (
        (last_candle["r_14"] <= -76.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_3_12"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["rsi_14_4h"] <= 35.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_3_13"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1h"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
      ):
        return True, f"exit_{mode_name}_w_3_14"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["change_pct_1h"] > 0.00)
      ):
        return True, f"exit_{mode_name}_w_3_15"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["bot_wick_pct_1d"] > 0.16)
        and (last_candle["close"] > (last_candle["low_min_24_1h"] * 1.20))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_3_16"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["r_480_1h"] < -70.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["bot_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_3_17"
      elif (
        (last_candle["r_14"] <= -65.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_3_18"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 20.0)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_3_19"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 35.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_3_20"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 34.0)
        and (last_candle["rsi_14_15m"] <= 35.0)
        and (last_candle["cti_20_dec_3_1h"] == False)
        and (last_candle["r_480_4h"] > -50.0)
      ):
        return True, f"exit_{mode_name}_w_3_21"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_3_22"
    elif 0.05 > current_profit >= 0.04:
      if last_candle["r_480"] < -99.5:
        return True, f"exit_{mode_name}_w_4_1"
      elif (last_candle["r_14"] <= -99.0) and (last_candle["rsi_14"] < 25.0):
        return True, f"exit_{mode_name}_w_4_2"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["rsi_14"] > 48.0):
        return True, f"exit_{mode_name}_w_4_3"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["rsi_14"] < 29.0) and (last_candle["r_480_1h"] < -75.0):
        return True, f"exit_{mode_name}_w_4_4"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["cti_20"] < -0.95):
        return True, f"exit_{mode_name}_w_4_5"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -80.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 40.0)
        and (last_candle["cti_20_1d"] < -0.80)
      ):
        return True, f"exit_{mode_name}_w_4_6"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_4h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 50.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_4_7"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["rsi_14_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.02)
      ):
        return True, f"exit_{mode_name}_w_4_8"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 36.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["cti_20_1d"] <= -0.70)
      ):
        return True, f"exit_{mode_name}_w_4_9"
      elif (
        (last_candle["r_14"] <= -86.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_4_10"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["rsi_14_1h"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["r_480_1h"] < -70.0)
      ):
        return True, f"exit_{mode_name}_w_4_11"
      elif (
        (last_candle["r_14"] <= -76.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_4_12"
      elif (
        (last_candle["r_14"] <= -86.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["rsi_14_4h"] <= 35.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_4_13"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1h"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
      ):
        return True, f"exit_{mode_name}_w_4_14"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["change_pct_1h"] > 0.00)
      ):
        return True, f"exit_{mode_name}_w_4_15"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["bot_wick_pct_1d"] > 0.16)
        and (last_candle["close"] > (last_candle["low_min_24_1h"] * 1.20))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_4_16"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["r_480_1h"] < -70.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["bot_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_4_17"
      elif (
        (last_candle["r_14"] <= -60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_4_18"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 20.0)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_4_19"
      elif (
        (last_candle["r_14"] <= -86.0)
        and (last_candle["rsi_14"] <= 35.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_4_20"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 36.0)
        and (last_candle["rsi_14_15m"] <= 35.0)
        and (last_candle["cti_20_dec_3_1h"] == False)
        and (last_candle["r_480_4h"] > -50.0)
      ):
        return True, f"exit_{mode_name}_w_4_21"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_4_22"
    elif 0.06 > current_profit >= 0.05:
      if last_candle["r_480"] < -99.4:
        return True, f"exit_{mode_name}_w_5_1"
      elif (last_candle["r_14"] <= -99.0) and (last_candle["rsi_14"] < 26.0):
        return True, f"exit_{mode_name}_w_5_2"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["rsi_14"] > 46.0):
        return True, f"exit_{mode_name}_w_5_3"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["rsi_14"] < 30.0) and (last_candle["r_480_1h"] < -75.0):
        return True, f"exit_{mode_name}_w_5_4"
      elif (last_candle["r_14"] <= -94.0) and (last_candle["cti_20"] < -0.95):
        return True, f"exit_{mode_name}_w_5_5"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -80.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 40.0)
        and (last_candle["cti_20_1d"] < -0.80)
      ):
        return True, f"exit_{mode_name}_w_5_6"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_4h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 50.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_5_7"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["rsi_14_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.02)
      ):
        return True, f"exit_{mode_name}_w_5_8"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["rsi_14_15m"] <= 38.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["cti_20_1d"] <= -0.70)
      ):
        return True, f"exit_{mode_name}_w_5_9"
      elif (
        (last_candle["r_14"] <= -85.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_5_10"
      elif (
        (last_candle["r_14"] <= -95.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["rsi_14_1h"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["r_480_1h"] < -70.0)
      ):
        return True, f"exit_{mode_name}_w_5_11"
      elif (
        (last_candle["r_14"] <= -76.0)
        and (last_candle["rsi_14"] <= 34.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_5_12"
      elif (
        (last_candle["r_14"] <= -84.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["rsi_14_4h"] <= 35.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_5_13"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1h"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
      ):
        return True, f"exit_{mode_name}_w_5_14"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["change_pct_1h"] > 0.00)
      ):
        return True, f"exit_{mode_name}_w_5_15"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["bot_wick_pct_1d"] > 0.16)
        and (last_candle["close"] > (last_candle["low_min_24_1h"] * 1.20))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_5_16"
      elif (
        (last_candle["r_14"] <= -76.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["r_480_1h"] < -70.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["bot_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_5_17"
      elif (
        (last_candle["r_14"] <= -55.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_5_18"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 20.0)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_5_19"
      elif (
        (last_candle["r_14"] <= -84.0)
        and (last_candle["rsi_14"] <= 35.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_5_20"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 38.0)
        and (last_candle["rsi_14_15m"] <= 35.0)
        and (last_candle["cti_20_dec_3_1h"] == False)
        and (last_candle["r_480_4h"] > -50.0)
      ):
        return True, f"exit_{mode_name}_w_5_21"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_5_22"
    elif 0.07 > current_profit >= 0.06:
      if last_candle["r_480"] < -99.3:
        return True, f"exit_{mode_name}_w_6_1"
      elif (last_candle["r_14"] <= -99.0) and (last_candle["rsi_14"] < 25.0):
        return True, f"exit_{mode_name}_w_6_2"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["rsi_14"] > 48.0):
        return True, f"exit_{mode_name}_w_6_3"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["rsi_14"] < 29.0) and (last_candle["r_480_1h"] < -75.0):
        return True, f"exit_{mode_name}_w_6_4"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["cti_20"] < -0.95):
        return True, f"exit_{mode_name}_w_6_5"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -80.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 40.0)
        and (last_candle["cti_20_1d"] < -0.80)
      ):
        return True, f"exit_{mode_name}_w_6_6"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_4h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 50.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_6_7"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["rsi_14_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.02)
      ):
        return True, f"exit_{mode_name}_w_6_8"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 36.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["cti_20_1d"] <= -0.70)
      ):
        return True, f"exit_{mode_name}_w_6_9"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_6_10"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["rsi_14_1h"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["r_480_1h"] < -70.0)
      ):
        return True, f"exit_{mode_name}_w_6_11"
      elif (
        (last_candle["r_14"] <= -86.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_6_12"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["rsi_14_4h"] <= 35.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_6_13"
      elif (
        (last_candle["r_14"] <= -82.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1h"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
      ):
        return True, f"exit_{mode_name}_w_6_14"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["change_pct_1h"] > 0.00)
      ):
        return True, f"exit_{mode_name}_w_6_15"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["bot_wick_pct_1d"] > 0.16)
        and (last_candle["close"] > (last_candle["low_min_24_1h"] * 1.20))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_6_16"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["r_480_1h"] < -70.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["bot_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_6_17"
      elif (
        (last_candle["r_14"] <= -65.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_6_18"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 20.0)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_6_19"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 35.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_6_20"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 36.0)
        and (last_candle["rsi_14_15m"] <= 35.0)
        and (last_candle["cti_20_dec_3_1h"] == False)
        and (last_candle["r_480_4h"] > -50.0)
      ):
        return True, f"exit_{mode_name}_w_6_21"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_6_22"
    elif 0.08 > current_profit >= 0.07:
      if last_candle["r_480"] < -99.5:
        return True, f"exit_{mode_name}_w_7_1"
      elif (last_candle["r_14"] <= -99.0) and (last_candle["rsi_14"] < 24.0):
        return True, f"exit_{mode_name}_w_7_2"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["rsi_14"] > 50.0):
        return True, f"exit_{mode_name}_w_7_3"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["rsi_14"] < 28.0) and (last_candle["r_480_1h"] < -75.0):
        return True, f"exit_{mode_name}_w_7_4"
      elif (last_candle["r_14"] <= -96.0) and (last_candle["cti_20"] < -0.95):
        return True, f"exit_{mode_name}_w_7_5"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -85.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 40.0)
        and (last_candle["cti_20_1d"] < -0.80)
      ):
        return True, f"exit_{mode_name}_w_7_6"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_4h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 50.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_7_7"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["rsi_14_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.02)
      ):
        return True, f"exit_{mode_name}_w_7_8"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 28.0)
        and (last_candle["rsi_14_15m"] <= 34.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["cti_20_1d"] <= -0.70)
      ):
        return True, f"exit_{mode_name}_w_7_9"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 36.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_7_10"
      elif (
        (last_candle["r_14"] <= -97.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["rsi_14_1h"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["r_480_1h"] < -70.0)
      ):
        return True, f"exit_{mode_name}_w_7_11"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_7_12"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["rsi_14_4h"] <= 35.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_7_13"
      elif (
        (last_candle["r_14"] <= -84.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1h"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
      ):
        return True, f"exit_{mode_name}_w_7_14"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["change_pct_1h"] > 0.00)
      ):
        return True, f"exit_{mode_name}_w_7_15"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["bot_wick_pct_1d"] > 0.16)
        and (last_candle["close"] > (last_candle["low_min_24_1h"] * 1.20))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_7_16"
      elif (
        (last_candle["r_14"] <= -84.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["r_480_1h"] < -70.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["bot_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_7_17"
      elif (
        (last_candle["r_14"] <= -75.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_7_18"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 20.0)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_7_19"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 35.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_7_20"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 34.0)
        and (last_candle["rsi_14_15m"] <= 35.0)
        and (last_candle["cti_20_dec_3_1h"] == False)
        and (last_candle["r_480_4h"] > -50.0)
      ):
        return True, f"exit_{mode_name}_w_7_21"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_7_22"
    elif 0.09 > current_profit >= 0.08:
      if last_candle["r_480"] < -99.6:
        return True, f"exit_{mode_name}_w_8_1"
      elif (last_candle["r_14"] <= -99.0) and (last_candle["rsi_14"] < 23.0):
        return True, f"exit_{mode_name}_w_8_2"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["rsi_14"] > 52.0):
        return True, f"exit_{mode_name}_w_8_3"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["rsi_14"] < 27.0) and (last_candle["r_480_1h"] < -75.0):
        return True, f"exit_{mode_name}_w_8_4"
      elif (last_candle["r_14"] <= -97.0) and (last_candle["cti_20"] < -0.95):
        return True, f"exit_{mode_name}_w_8_5"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -85.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 40.0)
        and (last_candle["cti_20_1d"] < -0.80)
      ):
        return True, f"exit_{mode_name}_w_8_6"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_4h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 50.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_8_7"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["rsi_14_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.02)
      ):
        return True, f"exit_{mode_name}_w_8_8"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["rsi_14_15m"] <= 32.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["cti_20_1d"] <= -0.70)
      ):
        return True, f"exit_{mode_name}_w_8_9"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 34.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_8_10"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["rsi_14_1h"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["r_480_1h"] < -70.0)
      ):
        return True, f"exit_{mode_name}_w_8_11"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 28.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_8_12"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["rsi_14_4h"] <= 35.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_8_13"
      elif (
        (last_candle["r_14"] <= -86.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1h"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
      ):
        return True, f"exit_{mode_name}_w_8_14"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["change_pct_1h"] > 0.00)
      ):
        return True, f"exit_{mode_name}_w_8_15"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["bot_wick_pct_1d"] > 0.16)
        and (last_candle["close"] > (last_candle["low_min_24_1h"] * 1.20))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_8_16"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["r_480_1h"] < -70.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["bot_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_8_17"
      elif (
        (last_candle["r_14"] <= -85.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_8_18"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 20.0)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_8_19"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 35.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_8_20"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["rsi_14_15m"] <= 35.0)
        and (last_candle["cti_20_dec_3_1h"] == False)
        and (last_candle["r_480_4h"] > -50.0)
      ):
        return True, f"exit_{mode_name}_w_8_21"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_8_22"
    elif 0.1 > current_profit >= 0.09:
      if last_candle["r_480"] < -99.7:
        return True, f"exit_{mode_name}_w_9_1"
      elif (last_candle["r_14"] <= -99.0) and (last_candle["rsi_14"] < 22.0):
        return True, f"exit_{mode_name}_w_9_2"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["rsi_14"] > 54.0):
        return True, f"exit_{mode_name}_w_9_3"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["rsi_14"] < 26.0) and (last_candle["r_480_1h"] < -75.0):
        return True, f"exit_{mode_name}_w_9_4"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["cti_20"] < -0.95):
        return True, f"exit_{mode_name}_w_9_5"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -85.0)
        and (last_candle["r_480_4h"] < -90.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 40.0)
        and (last_candle["cti_20_1d"] < -0.80)
      ):
        return True, f"exit_{mode_name}_w_9_6"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] < 40.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_4h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 50.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_9_7"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["rsi_14_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.02)
      ):
        return True, f"exit_{mode_name}_w_9_8"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 24.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["cti_20_1d"] <= -0.70)
      ):
        return True, f"exit_{mode_name}_w_9_9"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_9_10"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["rsi_14_1h"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["r_480_1h"] < -70.0)
      ):
        return True, f"exit_{mode_name}_w_9_11"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_9_12"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["rsi_14_4h"] <= 35.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_9_13"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1h"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
      ):
        return True, f"exit_{mode_name}_w_9_14"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["change_pct_1h"] > 0.00)
      ):
        return True, f"exit_{mode_name}_w_9_15"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["bot_wick_pct_1d"] > 0.16)
        and (last_candle["close"] > (last_candle["low_min_24_1h"] * 1.20))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_9_16"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["r_480_1h"] < -70.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["bot_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_9_17"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_9_18"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 20.0)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_9_19"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 35.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_9_20"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 35.0)
        and (last_candle["cti_20_dec_3_1h"] == False)
        and (last_candle["r_480_4h"] > -50.0)
      ):
        return True, f"exit_{mode_name}_w_9_21"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_9_22"
    elif 0.12 > current_profit >= 0.1:
      if last_candle["r_480"] > -99.8:
        return True, f"exit_{mode_name}_w_10_1"
      elif (last_candle["r_14"] <= -99.0) and (last_candle["rsi_14"] > 79.0):
        return True, f"exit_{mode_name}_w_10_2"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["rsi_14"] < 44.0):
        return True, f"exit_{mode_name}_w_10_3"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["rsi_14"] < 25.0) and (last_candle["r_480_1h"] < -75.0):
        return True, f"exit_{mode_name}_w_10_4"
      elif (last_candle["r_14"] <= -99.0) and (last_candle["cti_20"] < -0.95):
        return True, f"exit_{mode_name}_w_10_5"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] < 35.0)
        and (last_candle["r_480_1h"] < -90.0)
        and (last_candle["r_480_4h"] < -95.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 40.0)
        and (last_candle["cti_20_1d"] < -0.80)
      ):
        return True, f"exit_{mode_name}_w_10_6"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] < 35.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_4h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 50.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_10_7"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 25.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["rsi_14_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.02)
      ):
        return True, f"exit_{mode_name}_w_10_8"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 22.0)
        and (last_candle["rsi_14_15m"] <= 28.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["cti_20_1d"] <= -0.70)
      ):
        return True, f"exit_{mode_name}_w_10_9"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_10_10"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 28.0)
        and (last_candle["rsi_14_15m"] <= 28.0)
        and (last_candle["rsi_14_1h"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["r_480_1h"] < -70.0)
      ):
        return True, f"exit_{mode_name}_w_10_11"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 24.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_10_12"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["rsi_14_4h"] <= 35.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_10_13"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1h"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
      ):
        return True, f"exit_{mode_name}_w_10_14"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["change_pct_1h"] > 0.00)
      ):
        return True, f"exit_{mode_name}_w_10_15"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["bot_wick_pct_1d"] > 0.16)
        and (last_candle["close"] > (last_candle["low_min_24_1h"] * 1.20))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_10_16"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["r_480_1h"] < -70.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["bot_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_10_17"
      elif (
        (last_candle["r_14"] <= -95.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_10_18"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 20.0)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_10_19"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 35.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_10_20"
      elif (
        (last_candle["r_14"] >= -99.0)
        and (last_candle["rsi_14"] <= 28.0)
        and (last_candle["rsi_14_15m"] <= 35.0)
        and (last_candle["cti_20_dec_3_1h"] == False)
        and (last_candle["r_480_4h"] > -50.0)
      ):
        return True, f"exit_{mode_name}_w_10_21"
      elif (
        (last_candle["r_14"] >= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_10_22"
    elif 0.2 > current_profit >= 0.12:
      if last_candle["r_480"] < -99.6:
        return True, f"exit_{mode_name}_w_11_1"
      elif (last_candle["r_14"] <= -99.0) and (last_candle["rsi_14"] < 20.0):
        return True, f"exit_{mode_name}_w_11_2"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["rsi_14"] > 58.0):
        return True, f"exit_{mode_name}_w_11_3"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["rsi_14"] < 24.0) and (last_candle["r_480_1h"] < -75.0):
        return True, f"exit_{mode_name}_w_11_4"
      elif (last_candle["r_14"] <= -99.5) and (last_candle["cti_20"] < -0.95):
        return True, f"exit_{mode_name}_w_11_5"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] < 30.0)
        and (last_candle["r_480_1h"] < -90.0)
        and (last_candle["r_480_4h"] < -95.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 40.0)
        and (last_candle["cti_20_1d"] < -0.80)
      ):
        return True, f"exit_{mode_name}_w_11_6"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] < 30.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_4h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 50.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_11_7"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 24.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["rsi_14_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.02)
      ):
        return True, f"exit_{mode_name}_w_11_8"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 22.0)
        and (last_candle["rsi_14_15m"] <= 26.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["cti_20_1d"] <= -0.70)
      ):
        return True, f"exit_{mode_name}_w_11_9"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 28.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_11_10"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["rsi_14_15m"] <= 26.0)
        and (last_candle["rsi_14_1h"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["r_480_1h"] < -70.0)
      ):
        return True, f"exit_{mode_name}_w_11_11"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 22.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_11_12"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["rsi_14_4h"] <= 35.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_11_13"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1h"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
      ):
        return True, f"exit_{mode_name}_w_11_14"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["change_pct_1h"] > 0.00)
      ):
        return True, f"exit_{mode_name}_w_11_15"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["bot_wick_pct_1d"] > 0.16)
        and (last_candle["close"] > (last_candle["low_min_24_1h"] * 1.20))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_11_16"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["r_480_1h"] < -70.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["bot_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_11_17"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_11_18"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 35.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 20.0)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_11_19"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_11_20"
      elif (
        (last_candle["r_14"] >= -99.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["cti_20_dec_3_1h"] == False)
        and (last_candle["r_480_4h"] > -50.0)
      ):
        return True, f"exit_{mode_name}_w_11_21"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 25.0)
        and (last_candle["rsi_14_15m"] <= 35.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_11_22"
    elif current_profit >= 0.2:
      if last_candle["r_480"] < -99.8:
        return True, f"exit_{mode_name}_w_12_1"
      elif (last_candle["r_14"] <= -99.0) and (last_candle["rsi_14"] < 19.0):
        return True, f"exit_{mode_name}_w_12_2"
      elif (last_candle["r_14"] <= -98.0) and (last_candle["rsi_14"] > 60.0):
        return True, f"exit_{mode_name}_w_12_3"
      elif (last_candle["r_14"] <= -95.0) and (last_candle["rsi_14"] < 23.0) and (last_candle["r_480_1h"] < -75.0):
        return True, f"exit_{mode_name}_w_12_4"
      elif (last_candle["r_14"] <= -0.99) and (last_candle["cti_20"] < -0.95):
        return True, f"exit_{mode_name}_w_12_5"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] < 25.0)
        and (last_candle["r_480_1h"] < -95.0)
        and (last_candle["r_480_4h"] < -95.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 40.0)
        and (last_candle["cti_20_1d"] < -0.80)
      ):
        return True, f"exit_{mode_name}_w_12_6"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] < 25.0)
        and (last_candle["r_480_1h"] < -75.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["rsi_14_1h"] < 50.0)
        and (last_candle["rsi_14_4h"] < 50.0)
        and (last_candle["rsi_14_1d"] < 50.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_12_7"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 22.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["rsi_14_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.02)
      ):
        return True, f"exit_{mode_name}_w_12_8"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 20.0)
        and (last_candle["rsi_14_15m"] <= 24.0)
        and (last_candle["cti_20_4h"] >= 0.50)
        and (last_candle["cti_20_1d"] <= -0.70)
      ):
        return True, f"exit_{mode_name}_w_12_9"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_12_10"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 24.0)
        and (last_candle["rsi_14_15m"] <= 24.0)
        and (last_candle["rsi_14_1h"] <= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["r_480_1h"] < -70.0)
      ):
        return True, f"exit_{mode_name}_w_12_11"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 20.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["change_pct_1d"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_12_12"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["rsi_14_4h"] <= 35.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.01)
      ):
        return True, f"exit_{mode_name}_w_12_13"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_1h"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
      ):
        return True, f"exit_{mode_name}_w_12_14"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_4h"] <= 25.0)
        and (last_candle["r_480_4h"] < -85.0)
        and (last_candle["change_pct_1h"] > 0.00)
      ):
        return True, f"exit_{mode_name}_w_12_15"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["bot_wick_pct_1d"] > 0.16)
        and (last_candle["close"] > (last_candle["low_min_24_1h"] * 1.20))
        and (last_candle["hl_pct_change_6_1d"] > 0.75)
      ):
        return True, f"exit_{mode_name}_w_12_16"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["r_480_1h"] < -70.0)
        and (last_candle["r_480_4h"] < -75.0)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["bot_wick_pct_1d"] > 0.08)
      ):
        return True, f"exit_{mode_name}_w_12_17"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.90)
      ):
        return True, f"exit_{mode_name}_w_12_18"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 20.0)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["hl_pct_change_48_1h"] > 0.50)
      ):
        return True, f"exit_{mode_name}_w_12_19"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 30.0)
        and (last_candle["r_480_4h"] < -70.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_12_20"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 20.0)
        and (last_candle["rsi_14_15m"] <= 25.0)
        and (last_candle["cti_20_dec_3_1h"] == False)
        and (last_candle["r_480_4h"] > -50.0)
      ):
        return True, f"exit_{mode_name}_w_12_21"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 20.0)
        and (last_candle["rsi_14_15m"] <= 25.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_w_12_22"

    return False, None


  def short_exit_dec(
    self,
    mode_name: str,
    current_profit: float,
    max_profit: float,
    max_loss: float,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    buy_tag,
  ) -> tuple:
    if 0.02 > current_profit >= 0.01:
      if (
        (last_candle["r_14"] < -90.0)
        and (last_candle["rsi_14"] < 34.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_1_1"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_1_2"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["cti_20_1d"] <= -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_4h"] >= 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_3"
      elif (
        (last_candle["r_14"] <= -60.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1h"] > 0.03)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_4"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["change_pct_1d"] > 0.02)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_5"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_6"
      elif (
        (last_candle["rsi_3"] <= 40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_24"] == False)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_7"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_8"
      elif (
        (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_3_1h"] >= 90.0)
        and (last_candle["rsi_3_4h"] >= 94.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_9"
      elif (
        (last_candle["rsi_14"] >= 54.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_1h"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_1_10"
      elif (
        (last_candle["r_14"] <= -95.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1d"] > 0.10)
        and (last_candle["not_downtrend_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_11"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_12"
      elif (
        (last_candle["r_14"] <= -85.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["change_pct_4h"] > 0.03)
        and (last_candle["bot_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_13"
      elif (
        (last_candle["rsi_14"] >= 54.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_3_1d"] >= 94.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_14"
      elif (
        (last_candle["r_14"] <= -95.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_15"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] <= -0.70)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_16"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_1_17"
      elif (
        (last_candle["r_14"] <= -40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_18"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_max_6_1d"] <= 15.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.06)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_1_19"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_max_6_1d"] <= 30.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_20"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["close"] > (last_candle["low_min_48_1h"] * 1.25))
      ):
        return True, f"exit_{mode_name}_d_1_21"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.01)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.30))
      ):
        return True, f"exit_{mode_name}_d_1_22"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_1_23"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["change_pct_1h"] > 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_1_24"
      elif (
        (last_candle["r_14"] <= -75.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["rsi_3_1d"] >= 74.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_1_25"
    elif 0.03 > current_profit >= 0.02:
      if (
        (last_candle["r_14"] < -84.0)
        and (last_candle["rsi_14"] < 44.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_2_1"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_2_2"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["cti_20_1d"] <= -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_4h"] >= 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_3"
      elif (
        (last_candle["r_14"] <= -60.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1h"] > 0.03)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_4"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["change_pct_1d"] > 0.02)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_5"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_3"] <= 30.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_6"
      elif (
        (last_candle["rsi_3"] <= 40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_24"] == False)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_7"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 34.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_8"
      elif (
        (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_3_1h"] >= 90.0)
        and (last_candle["rsi_3_4h"] >= 94.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_9"
      elif (
        (last_candle["rsi_14"] <= 48.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_1h"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_2_10"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1d"] > 0.10)
        and (last_candle["not_downtrend_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_11"
      elif (
        (last_candle["r_14"] <= -91.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_12"
      elif (
        (last_candle["r_14"] <= -85.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["change_pct_4h"] > 0.03)
        and (last_candle["bot_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_13"
      elif (
        (last_candle["rsi_14"] <= 48.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_3_1d"] >= 94.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_14"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_15"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] <= -0.70)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_16"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_2_17"
      elif (
        (last_candle["r_14"] <= -40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_18"
      elif (
        (last_candle["r_14"] <= -60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_max_6_1d"] <= 15.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.06)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_2_19"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_max_6_1d"] <= 30.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_2_20"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["close"] > (last_candle["low_min_48_1h"] * 1.25))
      ):
        return True, f"exit_{mode_name}_d_2_21"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.01)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.30))
      ):
        return True, f"exit_{mode_name}_d_2_22"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_2_23"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["change_pct_1h"] > 0.02)
        and (last_candle["not_downtrend_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_24"
      elif (
        (last_candle["r_14"] <= -75.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["rsi_3_1d"] >= 74.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_2_25"
    elif 0.04 > current_profit >= 0.03:
      if (
        (last_candle["r_14"] < -84.0)
        and (last_candle["rsi_14"] < 46.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_3_1"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_3_2"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["cti_20_1d"] <= -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_4h"] >= 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_3"
      elif (
        (last_candle["r_14"] <= -60.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1h"] > 0.03)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_4"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["change_pct_1d"] > 0.02)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_5"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_3"] <= 30.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_6"
      elif (
        (last_candle["rsi_3"] <= 40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_24"] == False)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_7"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 36.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_8"
      elif (
        (last_candle["rsi_14"] >= 58.0)
        and (last_candle["rsi_3_1h"] >= 90.0)
        and (last_candle["rsi_3_4h"] >= 94.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_9"
      elif (
        (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_1h"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_3_10"
      elif (
        (last_candle["r_14"] <= -93.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1d"] > 0.10)
        and (last_candle["not_downtrend_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_11"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_12"
      elif (
        (last_candle["r_14"] <= -85.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["change_pct_4h"] > 0.03)
        and (last_candle["bot_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_13"
      elif (
        (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_3_1d"] >= 94.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_14"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] <= 34.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_15"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] <= -0.70)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_16"
      elif (
        (last_candle["r_14"] <= -84.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_3_17"
      elif (
        (last_candle["r_14"] <= -40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_18"
      elif (
        (last_candle["r_14"] <= -60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_max_6_1d"] <= 15.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.06)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_3_19"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_max_6_1d"] <= 30.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_20"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["close"] > (last_candle["low_min_48_1h"] * 1.25))
      ):
        return True, f"exit_{mode_name}_d_3_21"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.01)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.30))
      ):
        return True, f"exit_{mode_name}_d_3_22"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_3_23"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_3_1h"] >= 20.0)
        and (last_candle["change_pct_1h"] > 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_3_24"
      elif (
        (last_candle["r_14"] <= -75.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["rsi_3_1d"] >= 74.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_3_25"
    elif 0.05 > current_profit >= 0.04:
      if (
        (last_candle["r_14"] < -84.0)
        and (last_candle["rsi_14"] < 48.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_4_1"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_4_2"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["cti_20_1d"] <= -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_4h"] >= 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_3"
      elif (
        (last_candle["r_14"] <= -60.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1h"] > 0.03)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_4"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["change_pct_1d"] > 0.02)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_5"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_3"] <= 30.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_6"
      elif (
        (last_candle["rsi_3"] <= 40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_24"] == False)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_7"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 38.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_8"
      elif (
        (last_candle["rsi_14"] <= 44.0)
        and (last_candle["rsi_3_1h"] >= 90.0)
        and (last_candle["rsi_3_4h"] >= 94.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_9"
      elif (
        (last_candle["rsi_14"] >= 48.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_1h"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_4_10"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1d"] > 0.10)
        and (last_candle["not_downtrend_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_11"
      elif (
        (last_candle["r_14"] >= -11.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_12"
      elif (
        (last_candle["r_14"] <= -85.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["change_pct_4h"] > 0.03)
        and (last_candle["bot_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_13"
      elif (
        (last_candle["rsi_14"] >= 48.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_3_1d"] >= 94.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_14"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 36.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_15"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] <= -0.70)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_16"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_4_17"
      elif (
        (last_candle["r_14"] <= -40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_18"
      elif (
        (last_candle["r_14"] <= -60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_max_6_1d"] <= 15.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.06)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_4_19"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_max_6_1d"] <= 30.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_20"
      elif (
        (last_candle["r_14"] <= -86.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["close"] > (last_candle["low_min_48_1h"] * 1.25))
      ):
        return True, f"exit_{mode_name}_d_4_21"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.01)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.30))
      ):
        return True, f"exit_{mode_name}_d_4_22"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_4_23"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["change_pct_1h"] > 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_4_24"
      elif (
        (last_candle["r_14"] <= -75.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["rsi_3_1d"] >= 74.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_4_25"
    elif 0.06 > current_profit >= 0.05:
      if (
        (last_candle["r_14"] < -84.0)
        and (last_candle["rsi_14"] < 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_5_1"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_5_2"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["cti_20_1d"] <= -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_4h"] >= 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_3"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1h"] > 0.03)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_4"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["change_pct_1d"] > 0.02)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_5"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_3"] <= 30.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_6"
      elif (
        (last_candle["rsi_3"] <= 40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_24"] == False)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_7"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_8"
      elif (
        (last_candle["rsi_14"] >= 54.0)
        and (last_candle["rsi_3_1h"] >= 90.0)
        and (last_candle["rsi_3_4h"] >= 94.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_9"
      elif (
        (last_candle["rsi_14"] >= 46.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_1h"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_5_10"
      elif (
        (last_candle["r_14"] <= -91.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1d"] > 0.10)
        and (last_candle["not_downtrend_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_11"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_12"
      elif (
        (last_candle["r_14"] <= -85.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["change_pct_4h"] > 0.03)
        and (last_candle["bot_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_13"
      elif (
        (last_candle["rsi_14"] >= 46.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_3_1d"] >= 94.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_14"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_15"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] <= -0.70)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_16"
      elif (
        (last_candle["r_14"] <= -76.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_5_17"
      elif (
        (last_candle["r_14"] <= -40.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_18"
      elif (
        (last_candle["r_14"] <= -60.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_max_6_1d"] <= 15.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.06)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_5_19"
      elif (
        (last_candle["r_14"] <= -70.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_max_6_1d"] <= 30.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_20"
      elif (
        (last_candle["r_14"] <= -84.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["close"] > (last_candle["low_min_48_1h"] * 1.25))
      ):
        return True, f"exit_{mode_name}_d_5_21"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.01)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.30))
      ):
        return True, f"exit_{mode_name}_d_5_22"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_5_23"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["change_pct_1h"] > 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_5_24"
      elif (
        (last_candle["r_14"] <= -75.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["rsi_3_1d"] >= 74.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_5_25"
    elif 0.07 > current_profit >= 0.06:
      if (
        (last_candle["r_14"] < -84.0)
        and (last_candle["rsi_14"] < 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_6_1"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_6_2"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["cti_20_1d"] <= -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_4h"] >= 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_3"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1h"] > 0.03)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_4"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_3"] <= 6.0)
        and (last_candle["rsi_14"] <= 36.0)
        and (last_candle["change_pct_1d"] > 0.02)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_5"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_6"
      elif (
        (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_24"] == False)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_7"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 38.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_8"
      elif (
        (last_candle["rsi_14"] <= 44.0)
        and (last_candle["rsi_3_1h"] >= 90.0)
        and (last_candle["rsi_3_4h"] >= 94.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_9"
      elif (
        (last_candle["rsi_14"] >= 48.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_1h"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_6_10"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1d"] > 0.10)
        and (last_candle["not_downtrend_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_11"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 34.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_12"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["change_pct_4h"] > 0.03)
        and (last_candle["bot_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_13"
      elif (
        (last_candle["rsi_14"] >= 48.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_3_1d"] >= 94.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_14"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] <= 36.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_15"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 34.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] <= -0.70)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_16"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_6_17"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_18"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_max_6_1d"] <= 15.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.06)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_6_19"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_max_6_1d"] <= 30.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_20"
      elif (
        (last_candle["r_14"] <= -86.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["close"] > (last_candle["low_min_48_1h"] * 1.25))
      ):
        return True, f"exit_{mode_name}_d_6_21"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.01)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.30))
      ):
        return True, f"exit_{mode_name}_d_6_22"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_6_23"
      elif (
        (last_candle["r_14"] <= -84.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["change_pct_1h"] > 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_6_24"
      elif (
        (last_candle["r_14"] <= -80.0)
        and (last_candle["rsi_14"] <= 38.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["rsi_3_1d"] >= 26.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_6_25"
    elif 0.08 > current_profit >= 0.07:
      if (
        (last_candle["r_14"] < -84.0)
        and (last_candle["rsi_14"] < 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_7_1"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_7_2"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["cti_20_1d"] <= -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_4h"] >= 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_3"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1h"] > 0.03)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_4"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_3"] <= 4.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["change_pct_1d"] > 0.02)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_5"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_6"
      elif (
        (last_candle["rsi_3"] <= 5.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_24"] == False)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_7"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 36.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_8"
      elif (
        (last_candle["rsi_14"] <= 42.0)
        and (last_candle["rsi_3_1h"] >= 90.0)
        and (last_candle["rsi_3_4h"] >= 94.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_9"
      elif (
        (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_1h"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_7_10"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1d"] > 0.10)
        and (last_candle["not_downtrend_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_11"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_12"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["change_pct_4h"] > 0.03)
        and (last_candle["bot_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_13"
      elif (
        (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_3_1d"] >= 94.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_14"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 34.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_15"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] <= -0.70)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_16"
      elif (
        (last_candle["r_14"] <= -86.0)
        and (last_candle["rsi_14"] <= 38.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_7_17"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_18"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_max_6_1d"] <= 15.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.06)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_7_19"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_max_6_1d"] <= 30.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_20"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["close"] > (last_candle["low_min_48_1h"] * 1.25))
      ):
        return True, f"exit_{mode_name}_d_7_21"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.01)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.30))
      ):
        return True, f"exit_{mode_name}_d_7_22"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_7_23"
      elif (
        (last_candle["r_14"] <= -86.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["change_pct_1h"] > 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_7_24"
      elif (
        (last_candle["r_14"] <= -85.0)
        and (last_candle["rsi_14"] <= 36.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["rsi_3_1d"] >= 74.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_7_25"
    elif 0.09 > current_profit >= 0.08:
      if (
        (last_candle["r_14"] < -84.0)
        and (last_candle["rsi_14"] < 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_8_1"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_8_2"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["cti_20_1d"] <= -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_4h"] >= 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_3"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1h"] > 0.03)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_4"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 2.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["change_pct_1d"] > 0.02)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_5"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_3"] <= 5.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_6"
      elif (
        (last_candle["rsi_3"] <= 2.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_24"] == False)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_7"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 34.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_8"
      elif (
        (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_3_1h"] >= 90.0)
        and (last_candle["rsi_3_4h"] >= 94.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_9"
      elif (
        (last_candle["rsi_14"] <= 48.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_1h"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_8_10"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 36.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1d"] > 0.10)
        and (last_candle["not_downtrend_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_11"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_12"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["change_pct_4h"] > 0.03)
        and (last_candle["bot_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_13"
      elif (
        (last_candle["rsi_14"] <= 48.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_3_1d"] >= 94.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_14"
      elif (
        (last_candle["r_14"] <= -95.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_15"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] <= -0.70)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_16"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 36.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_8_17"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_18"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_max_6_1d"] <= 15.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.06)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_8_19"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_max_6_1d"] <= 30.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_20"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["close"] > (last_candle["low_min_48_1h"] * 1.25))
      ):
        return True, f"exit_{mode_name}_d_8_21"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.01)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.30))
      ):
        return True, f"exit_{mode_name}_d_8_22"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_8_23"
      elif (
        (last_candle["r_14"] <= -88.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["change_pct_1h"] > 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_8_24"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_14"] <= 34.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["rsi_3_1d"] >= 74.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_8_25"
    elif 0.1 > current_profit >= 0.09:
      if (
        (last_candle["r_14"] < -84.0)
        and (last_candle["rsi_14"] < 48.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_9_1"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_9_2"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["cti_20_1d"] <= -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_4h"] >= 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_3"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1h"] > 0.03)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_4"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 2.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["change_pct_1d"] > 0.02)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_5"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 2.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.70)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_6"
      elif (
        (last_candle["rsi_3"] >= 99.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_24"] == False)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_7"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_8"
      elif (
        (last_candle["rsi_14"] <= 38.0)
        and (last_candle["rsi_3_1h"] >= 90.0)
        and (last_candle["rsi_3_4h"] >= 94.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_9"
      elif (
        (last_candle["rsi_14"] >= 54.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_1h"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_9_10"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 36.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1d"] > 0.10)
        and (last_candle["not_downtrend_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_11"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_12"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["change_pct_4h"] > 0.03)
        and (last_candle["bot_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_13"
      elif (
        (last_candle["rsi_14"] >= 54.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_3_1d"] >= 94.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_14"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_15"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 28.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] <= -0.70)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_16"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 34.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_9_17"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_18"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_max_6_1d"] <= 15.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.06)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_9_19"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_max_6_1d"] <= 30.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_20"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["close"] > (last_candle["low_min_48_1h"] * 1.25))
      ):
        return True, f"exit_{mode_name}_d_9_21"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.01)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.30))
      ):
        return True, f"exit_{mode_name}_d_9_22"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_9_23"
      elif (
        (last_candle["r_14"] <= -90.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["change_pct_1h"] > 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_9_24"
      elif (
        (last_candle["r_14"] <= -92.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["rsi_3_1d"] >= 74.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_9_25"
    elif 0.12 > current_profit >= 0.1:
      if (
        (last_candle["r_14"] < -84.0)
        and (last_candle["rsi_14"] < 46.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_10_1"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_10_2"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["cti_20_1d"] <= -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_4h"] >= 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_3"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1h"] > 0.03)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_4"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 2.0)
        and (last_candle["rsi_14"] <= 24.0)
        and (last_candle["change_pct_1d"] > 0.02)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_5"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 2.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_6"
      elif (
        (last_candle["rsi_3"] >= 99.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_24"] == False)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_7"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_8"
      elif (
        (last_candle["rsi_14"] >= 64.0)
        and (last_candle["rsi_3_1h"] >= 90.0)
        and (last_candle["rsi_3_4h"] >= 94.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_9"
      elif (
        (last_candle["rsi_14"] <= 44.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_1h"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_10_10"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1d"] > 0.10)
        and (last_candle["not_downtrend_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_11"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 24.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_12"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["change_pct_4h"] > 0.03)
        and (last_candle["bot_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_13"
      elif (
        (last_candle["rsi_14"] <= 44.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_3_1d"] >= 94.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_14"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 28.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_15"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] <= -0.70)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_16"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 32.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_10_17"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_18"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_max_6_1d"] <= 15.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.06)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_10_19"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_max_6_1d"] <= 30.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_20"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["close"] > (last_candle["low_min_48_1h"] * 1.25))
      ):
        return True, f"exit_{mode_name}_d_10_21"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.01)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.30))
      ):
        return True, f"exit_{mode_name}_d_10_22"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_10_23"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["change_pct_1h"] > 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_10_24"
      elif (
        (last_candle["r_14"] <= -94.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["rsi_3_1d"] >= 74.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_10_25"
    elif 0.2 > current_profit >= 0.12:
      if (
        (last_candle["r_14"] < -84.0)
        and (last_candle["rsi_14"] < 44.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_11_1"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_11_2"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["cti_20_1d"] <= -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_4h"] >= 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_3"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_1h"] > 0.03)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_4"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 2.0)
        and (last_candle["rsi_14"] <= 22.0)
        and (last_candle["change_pct_1d"] > 0.02)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_5"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 2.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_6"
      elif (
        (last_candle["rsi_3"] <= 1.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_24"] == False)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_7"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_8"
      elif (
        (last_candle["rsi_14"] >= 66.0)
        and (last_candle["rsi_3_1h"] >= 90.0)
        and (last_candle["rsi_3_4h"] >= 94.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_9"
      elif (
        (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_1h"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_11_10"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1d"] > 0.10)
        and (last_candle["not_downtrend_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_11"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 22.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_12"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["change_pct_4h"] > 0.03)
        and (last_candle["bot_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_13"
      elif (
        (last_candle["rsi_14"] <= 42.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_3_1d"] >= 94.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_14"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_15"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 24.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] <= -0.70)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_16"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_11_17"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_18"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_max_6_1d"] <= 15.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.06)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_11_19"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_max_6_1d"] <= 30.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_20"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["close"] > (last_candle["low_min_48_1h"] * 1.25))
      ):
        return True, f"exit_{mode_name}_d_11_21"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.01)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.30))
      ):
        return True, f"exit_{mode_name}_d_11_22"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_11_23"
      elif (
        (last_candle["r_14"] <= -96.0)
        and (last_candle["rsi_3"] <= 20.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["change_pct_1h"] > 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_11_24"
      elif (
        (last_candle["r_14"] <= -98.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["rsi_3_1d"] >= 74.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_11_25"
    elif current_profit >= 0.2:
      if (
        (last_candle["r_14"] < -90.0)
        and (last_candle["rsi_14"] < 34.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_12_1"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["ema_200_dec_4_1d"] == False)
        and (last_candle["change_pct_4h"] > 0.03)
      ):
        return True, f"exit_{mode_name}_d_12_2"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_14_4h"] <= 50.0)
        and (last_candle["cti_20_1d"] <= -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_4h"] >= 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_3"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 10.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] > -25.0)
        and (last_candle["cti_20_1d"] < -0.50)
        and (last_candle["rsi_14_1d"] <= 30.0)
        and (last_candle["change_pct_1h"] > 0.03)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_4"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 2.0)
        and (last_candle["rsi_14"] <= 20.0)
        and (last_candle["change_pct_1d"] > 0.02)
        and (last_candle["change_pct_4h"] > 0.02)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_5"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 2.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.80)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_6"
      elif (
        (last_candle["rsi_3"] >= 99.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_4h"] <= -0.50)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_24"] == False)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_7"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 24.0)
        and (last_candle["r_480_4h"] > -10.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_8"
      elif (
        (last_candle["rsi_14"] >= 70.0)
        and (last_candle["rsi_3_1h"] >= 90.0)
        and (last_candle["rsi_3_4h"] >= 94.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_9"
      elif (
        (last_candle["rsi_14"] >= 64.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["rsi_14_4h"] <= 40.0)
        and (last_candle["change_pct_1h"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_12_10"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["change_pct_1d"] > 0.10)
        and (last_candle["not_downtrend_4h"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_11"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 20.0)
        and (last_candle["cti_20_1d"] <= -0.80)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_12"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 26.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["change_pct_4h"] > 0.03)
        and (last_candle["bot_wick_pct_4h"] > 0.03)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_13"
      elif (
        (last_candle["rsi_14"] >= 60.0)
        and (last_candle["rsi_14_15m"] >= 60.0)
        and (last_candle["rsi_3_1d"] >= 94.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_14"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 24.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["r_480_4h"] <= -70.0)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_15"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 22.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["cti_20_1d"] <= -0.70)
        and (last_candle["change_pct_1d"] > 0.04)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_16"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 28.0)
        and (last_candle["rsi_14_15m"] <= 40.0)
        and (last_candle["rsi_14_1h"] <= 40.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["r_480_4h"] > -30.0)
        and (last_candle["change_pct_1d"] > 0.04)
      ):
        return True, f"exit_{mode_name}_d_12_17"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_24_15m"] == False)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_24_4h"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_18"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_max_6_1d"] <= 15.0)
        and (last_candle["change_pct_1h"] > 0.01)
        and (last_candle["change_pct_4h"] > 0.06)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_12_19"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_max_6_1d"] <= 30.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_20"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 50.0)
        and (last_candle["cti_20_dec_3_1d"] == False)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["close"] > (last_candle["low_min_48_1h"] * 1.25))
      ):
        return True, f"exit_{mode_name}_d_12_21"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 40.0)
        and (last_candle["rsi_14_max_6_4h"] <= 30.0)
        and (last_candle["change_pct_4h"] > 0.01)
        and (last_candle["not_downtrend_4h"] == True)
        and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.30))
      ):
        return True, f"exit_{mode_name}_d_12_22"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 30.0)
        and (last_candle["not_downtrend_1h"] == True)
        and (last_candle["not_downtrend_1d"] == True)
        and (last_candle["close"] > (last_candle["low_min_6_1d"] * 1.20))
      ):
        return True, f"exit_{mode_name}_d_12_23"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_3"] <= 5.0)
        and (last_candle["rsi_14"] >= 50.0)
        and (last_candle["rsi_14_15m"] >= 50.0)
        and (last_candle["rsi_3_1h"] >= 80.0)
        and (last_candle["change_pct_1h"] > 0.02)
        and (last_candle["not_downtrend_1h"] == True)
      ):
        return True, f"exit_{mode_name}_d_12_24"
      elif (
        (last_candle["r_14"] <= -99.0)
        and (last_candle["rsi_14"] <= 22.0)
        and (last_candle["rsi_14_15m"] <= 50.0)
        and (last_candle["rsi_14_1d"] <= 50.0)
        and (last_candle["rsi_3_1d"] >= 74.0)
        and (last_candle["ema_200_dec_48_1h"] == False)
        and (last_candle["ema_200_dec_4_1d"] == False)
      ):
        return True, f"exit_{mode_name}_d_12_25"

    return False, None


  def short_exit_stoploss(
    self,
    mode_name: str,
    current_rate: float,
    profit_stake: float,
    profit_ratio: float,
    profit_current_stake_ratio: float,
    profit_init_ratio: float,
    max_profit: float,
    max_loss: float,
    filled_entries,
    filled_exits,
    last_candle,
    previous_candle_1,
    previous_candle_2,
    previous_candle_3,
    previous_candle_4,
    previous_candle_5,
    trade: "Trade",
    current_time: "datetime",
    buy_tag,
  ) -> tuple:
    is_backtest = self.dp.runmode.value in ["backtest", "hyperopt"]

    if (
      self.is_futures_mode is False
      and profit_stake
      < -(filled_entries[0].cost * self.stop_threshold / (trade.leverage if self.is_futures_mode else 1.0))

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2023, 6, 13) or is_backtest)
    ):
      return True, f"exit_{mode_name}_stoploss"

    if (
      self.is_futures_mode is True
      and profit_stake
      < -(filled_entries[0].cost * self.stop_threshold_futures / (trade.leverage if self.is_futures_mode else 1.0))

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2023, 10, 17) or is_backtest)
    ):
      return True, f"exit_{mode_name}_stoploss"

    return False, None
















  def short_grind_adjust_trade_position(
    self,
    trade: Trade,
    enter_tags,
    current_time: datetime,
    current_rate: float,
    current_profit: float,
    min_stake: Optional[float],
    max_stake: float,
    current_entry_rate: float,
    current_exit_rate: float,
    current_entry_profit: float,
    current_exit_profit: float,
    **kwargs,
  ):
    is_backtest = self.dp.runmode.value in ["backtest", "hyperopt"]
    df, _ = self.dp.get_analyzed_dataframe(trade.pair, self.timeframe)
    if len(df) < 2:
      return None
    last_candle = df.iloc[-1].squeeze()
    previous_candle = df.iloc[-2].squeeze()

    filled_orders = trade.select_filled_orders()
    filled_entries = trade.select_filled_orders(trade.entry_side)
    filled_exits = trade.select_filled_orders(trade.exit_side)
    count_of_entries = trade.nr_of_successful_entries
    count_of_exits = trade.nr_of_successful_exits

    if count_of_entries == 0:
      return None

    if len(filled_orders) < 1:
      return None
    has_order_tags = False
    if hasattr(filled_orders[0], "ft_order_tag"):
      has_order_tags = True

    exit_rate = current_rate
    if self.dp.runmode.value in ("live", "dry_run"):
      ticker = self.dp.ticker(trade.pair)
      if ("bid" in ticker) and ("ask" in ticker):
        if trade.is_short:
          if self.config["exit_pricing"]["price_side"] in ["ask", "other"]:
            if ticker["ask"] is not None:
              exit_rate = ticker["ask"]
        else:
          if self.config["exit_pricing"]["price_side"] in ["bid", "other"]:
            if ticker["bid"] is not None:
              exit_rate = ticker["bid"]

    profit_stake, profit_ratio, profit_current_stake_ratio, profit_init_ratio = self.calc_total_profit(
      trade, filled_entries, filled_exits, exit_rate
    )

    slice_amount = filled_entries[0].cost
    slice_profit = (exit_rate - filled_orders[-1].safe_price) / filled_orders[-1].safe_price
    slice_profit_entry = (exit_rate - filled_entries[-1].safe_price) / filled_entries[-1].safe_price
    slice_profit_exit = (
      ((exit_rate - filled_exits[-1].safe_price) / filled_exits[-1].safe_price) if count_of_exits > 0 else 0.0
    )

    current_stake_amount = trade.amount * current_rate
    is_derisk = trade.amount < (filled_entries[0].safe_filled * 0.95)
    is_derisk_calc = False
    is_rebuy_mode = all(c in self.short_rebuy_mode_tags for c in enter_tags) or (
      any(c in self.short_rebuy_mode_tags for c in enter_tags)
      and all(c in (self.short_rebuy_mode_tags + self.short_grind_mode_tags) for c in enter_tags)
    )
    is_grind_mode = all(c in self.short_grind_mode_tags for c in enter_tags)

    if is_rebuy_mode:
      slice_amount /= self.rebuy_mode_stake_multiplier

    elif is_grind_mode:
      slice_amount /= (
        self.grind_mode_stake_multiplier_futures[0]
        if self.is_futures_mode
        else self.grind_mode_stake_multiplier_spot[0]
      )
    elif not is_derisk and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 2, 5) or is_backtest):
      rebuy_stake, order_tag, is_derisk_calc = self.short_adjust_trade_position_no_derisk(
        trade,
        enter_tags,
        current_time,
        current_rate,
        current_profit,
        min_stake,
        max_stake,
        current_entry_rate,
        current_exit_rate,
        current_entry_profit,
        current_exit_profit,
        last_candle,
        previous_candle,
        filled_orders,
        filled_entries,
        filled_exits,
        exit_rate,
        slice_amount,
        slice_profit_entry,
        slice_profit,
        profit_ratio,
        profit_stake,
        profit_init_ratio,
        current_stake_amount,
        has_order_tags,
      )
      if rebuy_stake is not None:
        if has_order_tags:
          return rebuy_stake, order_tag
        else:
          return rebuy_stake
      elif count_of_exits == 0:
        return None
      elif not is_derisk_calc:
        return None

    if not is_rebuy_mode and not is_grind_mode:

      if trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 2, 5) or is_backtest:
        slice_amount /= (
          self.regular_mode_stake_multiplier_futures[0]
          if self.is_futures_mode
          else self.regular_mode_stake_multiplier_spot[0]
        )

    grind_derisk = self.grind_derisk_futures if self.is_futures_mode else self.grind_derisk_spot

    grind_1_max_sub_grinds = 0
    grind_1_stakes = self.grind_1_stakes_futures.copy() if self.is_futures_mode else self.grind_1_stakes_spot.copy()
    grind_1_sub_thresholds = (
      self.grind_1_sub_thresholds_futures if self.is_futures_mode else self.grind_1_sub_thresholds_spot
    )
    if (slice_amount * grind_1_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = slice_amount / min_stake
      for i, _ in enumerate(grind_1_stakes):
        grind_1_stakes[i] *= multi
    grind_1_max_sub_grinds = len(grind_1_stakes)
    grind_1_stop_grinds = self.grind_1_stop_grinds_futures if self.is_futures_mode else self.grind_1_stop_grinds_spot
    grind_1_profit_threshold = (
      self.grind_1_profit_threshold_futures if self.is_futures_mode else self.grind_1_profit_threshold_spot
    )

    grind_2_max_sub_grinds = 0
    grind_2_stakes = self.grind_2_stakes_futures.copy() if self.is_futures_mode else self.grind_2_stakes_spot.copy()
    grind_2_sub_thresholds = (
      self.grind_2_sub_thresholds_futures if self.is_futures_mode else self.grind_2_sub_thresholds_spot
    )
    if (slice_amount * grind_2_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = slice_amount / min_stake
      for i, _ in enumerate(grind_2_stakes):
        grind_2_stakes[i] *= multi
    grind_2_max_sub_grinds = len(grind_2_stakes)
    grind_2_stop_grinds = self.grind_2_stop_grinds_futures if self.is_futures_mode else self.grind_2_stop_grinds_spot
    grind_2_profit_threshold = (
      self.grind_2_profit_threshold_futures if self.is_futures_mode else self.grind_2_profit_threshold_spot
    )

    grind_3_max_sub_grinds = 0
    grind_3_stakes = self.grind_3_stakes_futures.copy() if self.is_futures_mode else self.grind_3_stakes_spot.copy()
    grind_3_sub_thresholds = (
      self.grind_3_sub_thresholds_futures if self.is_futures_mode else self.grind_3_sub_thresholds_spot
    )
    if (slice_amount * grind_3_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = slice_amount / min_stake
      for i, _ in enumerate(grind_3_stakes):
        grind_3_stakes[i] *= multi
    grind_3_max_sub_grinds = len(grind_3_stakes)
    grind_3_stop_grinds = self.grind_3_stop_grinds_futures if self.is_futures_mode else self.grind_3_stop_grinds_spot
    grind_3_profit_threshold = (
      self.grind_3_profit_threshold_futures if self.is_futures_mode else self.grind_3_profit_threshold_spot
    )

    grind_4_max_sub_grinds = 0
    grind_4_stakes = self.grind_4_stakes_futures.copy() if self.is_futures_mode else self.grind_4_stakes_spot.copy()
    grind_4_sub_thresholds = (
      self.grind_4_sub_thresholds_futures if self.is_futures_mode else self.grind_4_sub_thresholds_spot
    )
    if (slice_amount * grind_4_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = slice_amount / min_stake
      for i, _ in enumerate(grind_4_stakes):
        grind_4_stakes[i] *= multi
    grind_4_max_sub_grinds = len(grind_4_stakes)
    grind_4_stop_grinds = self.grind_4_stop_grinds_futures if self.is_futures_mode else self.grind_4_stop_grinds_spot
    grind_4_profit_threshold = (
      self.grind_4_profit_threshold_futures if self.is_futures_mode else self.grind_4_profit_threshold_spot
    )

    grind_5_max_sub_grinds = 0
    grind_5_stakes = self.grind_5_stakes_futures.copy() if self.is_futures_mode else self.grind_5_stakes_spot.copy()
    grind_5_sub_thresholds = (
      self.grind_5_sub_thresholds_futures if self.is_futures_mode else self.grind_5_sub_thresholds_spot
    )
    if (slice_amount * grind_5_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = slice_amount / min_stake
      for i, _ in enumerate(grind_5_stakes):
        grind_5_stakes[i] *= multi
    grind_5_max_sub_grinds = len(grind_5_stakes)
    grind_5_stop_grinds = self.grind_5_stop_grinds_futures if self.is_futures_mode else self.grind_5_stop_grinds_spot
    grind_5_profit_threshold = (
      self.grind_5_profit_threshold_futures if self.is_futures_mode else self.grind_5_profit_threshold_spot
    )

    grind_6_max_sub_grinds = 0
    grind_6_stakes = self.grind_6_stakes_futures.copy() if self.is_futures_mode else self.grind_6_stakes_spot.copy()
    grind_6_sub_thresholds = (
      self.grind_6_sub_thresholds_futures if self.is_futures_mode else self.grind_6_sub_thresholds_spot
    )
    if (slice_amount * grind_6_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = slice_amount / min_stake
      for i, _ in enumerate(grind_6_stakes):
        grind_6_stakes[i] *= multi
    grind_6_max_sub_grinds = len(grind_6_stakes)
    grind_6_stop_grinds = self.grind_6_stop_grinds_futures if self.is_futures_mode else self.grind_6_stop_grinds_spot
    grind_6_profit_threshold = (
      self.grind_6_profit_threshold_futures if self.is_futures_mode else self.grind_6_profit_threshold_spot
    )

    grind_1_derisk_1_max_sub_grinds = 0
    grind_1_derisk_1_stakes = (
      self.grind_1_derisk_1_stakes_futures.copy() if self.is_futures_mode else self.grind_1_derisk_1_stakes_spot.copy()
    )
    grind_1_derisk_1_sub_thresholds = (
      self.grind_1_derisk_1_sub_thresholds_futures
      if self.is_futures_mode
      else self.grind_1_derisk_1_sub_thresholds_spot
    )
    if (slice_amount * grind_1_derisk_1_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = slice_amount / min_stake
      for i, _ in enumerate(grind_1_derisk_1_stakes):
        grind_1_derisk_1_stakes[i] *= multi
    grind_1_derisk_1_max_sub_grinds = len(grind_1_derisk_1_stakes)
    grind_1_derisk_1_stop_grinds = (
      self.grind_1_derisk_1_stop_grinds_futures if self.is_futures_mode else self.grind_1_derisk_1_stop_grinds_spot
    )
    grind_1_derisk_1_profit_threshold = (
      self.grind_1_derisk_1_profit_threshold_futures
      if self.is_futures_mode
      else self.grind_1_derisk_1_profit_threshold_spot
    )

    grind_2_derisk_1_max_sub_grinds = 0
    grind_2_derisk_1_stakes = (
      self.grind_2_derisk_1_stakes_futures.copy() if self.is_futures_mode else self.grind_2_derisk_1_stakes_spot.copy()
    )
    grind_2_derisk_1_sub_thresholds = (
      self.grind_2_derisk_1_sub_thresholds_futures
      if self.is_futures_mode
      else self.grind_2_derisk_1_sub_thresholds_spot
    )
    if (slice_amount * grind_2_derisk_1_stakes[0] / (trade.leverage if self.is_futures_mode else 1.0)) < min_stake:
      multi = slice_amount / min_stake
      for i, _ in enumerate(grind_2_derisk_1_stakes):
        grind_2_derisk_1_stakes[i] *= multi
    grind_2_derisk_1_max_sub_grinds = len(grind_2_derisk_1_stakes)
    grind_2_derisk_1_stop_grinds = (
      self.grind_2_derisk_1_stop_grinds_futures if self.is_futures_mode else self.grind_2_derisk_1_stop_grinds_spot
    )
    grind_2_derisk_1_profit_threshold = (
      self.grind_2_derisk_1_profit_threshold_futures
      if self.is_futures_mode
      else self.grind_2_derisk_1_profit_threshold_spot
    )

    partial_sell = False
    is_derisk_found = False  # d de-risk
    is_derisk_1 = False
    is_derisk_1_found = False  # d1 de-risk exit
    derisk_1_order = None
    derisk_1_reentry_order = None
    derisk_1_sub_grind_count = 0
    derisk_1_total_amount = 0.0
    derisk_1_total_cost = 0.0
    derisk_1_current_open_rate = 0.0
    derisk_1_current_grind_stake = 0.0
    derisk_1_current_grind_stake_profit = 0.0
    derisk_1_is_sell_found = False
    derisk_1_reentry_found = False
    derisk_1_buy_orders = []
    derisk_1_distance_ratio = 0.0
    grind_1_sub_grind_count = 0
    grind_1_total_amount = 0.0
    grind_1_total_cost = 0.0
    grind_1_current_open_rate = 0.0
    grind_1_current_grind_stake = 0.0
    grind_1_current_grind_stake_profit = 0.0
    grind_1_is_sell_found = False
    grind_1_found = False
    grind_1_buy_orders = []
    grind_1_distance_ratio = 0.0
    grind_2_sub_grind_count = 0
    grind_2_total_amount = 0.0
    grind_2_total_cost = 0.0
    grind_2_current_open_rate = 0.0
    grind_2_current_grind_stake = 0.0
    grind_2_current_grind_stake_profit = 0.0
    grind_2_is_sell_found = False
    grind_2_found = False
    grind_2_buy_orders = []
    grind_2_distance_ratio = 0.0
    grind_3_sub_grind_count = 0
    grind_3_total_amount = 0.0
    grind_3_total_cost = 0.0
    grind_3_current_open_rate = 0.0
    grind_3_current_grind_stake = 0.0
    grind_3_current_grind_stake_profit = 0.0
    grind_3_is_sell_found = False
    grind_3_found = False
    grind_3_buy_orders = []
    grind_3_distance_ratio = 0.0
    grind_4_sub_grind_count = 0
    grind_4_total_amount = 0.0
    grind_4_total_cost = 0.0
    grind_4_current_open_rate = 0.0
    grind_4_current_grind_stake = 0.0
    grind_4_current_grind_stake_profit = 0.0
    grind_4_is_sell_found = False
    grind_4_found = False
    grind_4_buy_orders = []
    grind_4_distance_ratio = 0.0
    grind_5_sub_grind_count = 0
    grind_5_total_amount = 0.0
    grind_5_total_cost = 0.0
    grind_5_current_open_rate = 0.0
    grind_5_current_grind_stake = 0.0
    grind_5_current_grind_stake_profit = 0.0
    grind_5_is_sell_found = False
    grind_5_found = False
    grind_5_buy_orders = []
    grind_5_distance_ratio = 0.0
    grind_6_sub_grind_count = 0
    grind_6_total_amount = 0.0
    grind_6_total_cost = 0.0
    grind_6_current_open_rate = 0.0
    grind_6_current_grind_stake = 0.0
    grind_6_current_grind_stake_profit = 0.0
    grind_6_is_sell_found = False
    grind_6_found = False
    grind_6_buy_orders = []
    grind_6_distance_ratio = 0.0
    grind_1_derisk_1_sub_grind_count = 0
    grind_1_derisk_1_total_amount = 0.0
    grind_1_derisk_1_total_cost = 0.0
    grind_1_derisk_1_current_open_rate = 0.0
    grind_1_derisk_1_current_grind_stake = 0.0
    grind_1_derisk_1_current_grind_stake_profit = 0.0
    grind_1_derisk_1_is_sell_found = False
    grind_1_derisk_1_found = False
    grind_1_derisk_1_buy_orders = []
    grind_1_derisk_1_distance_ratio = 0.0
    grind_2_derisk_1_sub_grind_count = 0
    grind_2_derisk_1_total_amount = 0.0
    grind_2_derisk_1_total_cost = 0.0
    grind_2_derisk_1_current_open_rate = 0.0
    grind_2_derisk_1_current_grind_stake = 0.0
    grind_2_derisk_1_current_grind_stake_profit = 0.0
    grind_2_derisk_1_is_sell_found = False
    grind_2_derisk_1_found = False
    grind_2_derisk_1_buy_orders = []
    grind_2_derisk_1_distance_ratio = 0.0
    for order in reversed(filled_orders):
      if (order.ft_order_side == "sell") and (order is not filled_orders[0]):
        order_tag = ""
        if has_order_tags:
          if order.ft_order_tag is not None:
            order_tag = order.ft_order_tag
        if not is_derisk_1 and order_tag == "d1":
          derisk_1_sub_grind_count += 1
          derisk_1_total_amount += order.safe_filled
          derisk_1_total_cost += order.safe_filled * order.safe_price
          derisk_1_buy_orders.append(order.id)
          if not derisk_1_reentry_found and not is_derisk_1:
            derisk_1_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            derisk_1_reentry_found = True
            derisk_1_reentry_order = order
        elif not grind_1_derisk_1_is_sell_found and order_tag == "dl1":
          grind_1_derisk_1_sub_grind_count += 1
          grind_1_derisk_1_total_amount += order.safe_filled
          grind_1_derisk_1_total_cost += order.safe_filled * order.safe_price
          grind_1_derisk_1_buy_orders.append(order.id)
          if not grind_1_derisk_1_found:
            grind_1_derisk_1_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_1_derisk_1_found = True
        elif not grind_2_derisk_1_is_sell_found and order_tag == "dl2":
          grind_2_derisk_1_sub_grind_count += 1
          grind_2_derisk_1_total_amount += order.safe_filled
          grind_2_derisk_1_total_cost += order.safe_filled * order.safe_price
          grind_2_derisk_1_buy_orders.append(order.id)
          if not grind_2_derisk_1_found:
            grind_2_derisk_1_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_2_derisk_1_found = True
        elif not grind_6_is_sell_found and order_tag == "gd6":
          grind_6_sub_grind_count += 1
          grind_6_total_amount += order.safe_filled
          grind_6_total_cost += order.safe_filled * order.safe_price
          grind_6_buy_orders.append(order.id)
          if not grind_6_found:
            grind_6_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_6_found = True
        elif not grind_5_is_sell_found and order_tag == "gd5":
          grind_5_sub_grind_count += 1
          grind_5_total_amount += order.safe_filled
          grind_5_total_cost += order.safe_filled * order.safe_price
          grind_5_buy_orders.append(order.id)
          if not grind_5_found:
            grind_5_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_5_found = True
        elif not grind_4_is_sell_found and order_tag == "gd4":
          grind_4_sub_grind_count += 1
          grind_4_total_amount += order.safe_filled
          grind_4_total_cost += order.safe_filled * order.safe_price
          grind_4_buy_orders.append(order.id)
          if not grind_4_found:
            grind_4_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_4_found = True
        elif not grind_3_is_sell_found and order_tag == "gd3":
          grind_3_sub_grind_count += 1
          grind_3_total_amount += order.safe_filled
          grind_3_total_cost += order.safe_filled * order.safe_price
          grind_3_buy_orders.append(order.id)
          if not grind_3_found:
            grind_3_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_3_found = True
        elif not grind_2_is_sell_found and order_tag == "gd2":
          grind_2_sub_grind_count += 1
          grind_2_total_amount += order.safe_filled
          grind_2_total_cost += order.safe_filled * order.safe_price
          grind_2_buy_orders.append(order.id)
          if not grind_2_found:
            grind_2_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_2_found = True
        elif not grind_1_is_sell_found and order_tag not in [
          "r",
          "d1",
          "dl1",
          "dl2",
          "g1",
          "g2",
          "g3",
          "g4",
          "g5",
          "g6",
          "gd2",
          "gd3",
          "gd4",
          "gd5",
          "gd6",
          "gm0",
          "gmd0",
        ]:
          grind_1_sub_grind_count += 1
          grind_1_total_amount += order.safe_filled
          grind_1_total_cost += order.safe_filled * order.safe_price
          grind_1_buy_orders.append(order.id)
          if not grind_1_found:
            grind_1_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_1_found = True
      elif order.ft_order_side == "buy":
        if (
          order is filled_exits[-1]
          and (order.safe_remaining * exit_rate / (trade.leverage if self.is_futures_mode else 1.0)) > min_stake
        ):
          partial_sell = True
          break
        order_tag = ""
        if has_order_tags:
          if order.ft_order_tag is not None:
            sell_order_tag = order.ft_order_tag
            order_mode = sell_order_tag.split(" ", 1)
            if len(order_mode) > 0:
              order_tag = order_mode[0]
        if order_tag in ["dl1", "ddl1"]:
          grind_1_derisk_1_is_sell_found = True
        elif order_tag in ["dl2", "ddl2"]:
          grind_2_derisk_1_is_sell_found = True
        elif order_tag in ["gd6", "dd6"]:
          grind_6_is_sell_found = True
        elif order_tag in ["gd5", "dd5"]:
          grind_5_is_sell_found = True
        if order_tag in ["gd4", "dd4"]:
          grind_4_is_sell_found = True
        elif order_tag in ["gd3", "dd3"]:
          grind_3_is_sell_found = True
        elif order_tag in ["gd2", "dd2"]:
          grind_2_is_sell_found = True
        elif order_tag in ["d1"]:
          if not is_derisk_1_found:
            is_derisk_1_found = True
            is_derisk_1 = True
            derisk_1_order = order
        elif order_tag in ["p", "r", "d", "dd0", "partial_exit", ""]:
          if order_tag in ["d"]:
            is_derisk_found = True
            is_derisk = True
          grind_1_is_sell_found = True
          grind_2_is_sell_found = True
          grind_3_is_sell_found = True
          grind_4_is_sell_found = True
          grind_5_is_sell_found = True
          grind_1_derisk_1_is_sell_found = True
          grind_2_derisk_1_is_sell_found = True
        elif order_tag not in [
          "dl1",
          "ddl1",
          "dl2",
          "ddl2",
          "g1",
          "g2",
          "g3",
          "g4",
          "g5",
          "g6",
          "gd2",
          "gd3",
          "gd4",
          "gd5",
          "gd6",
          "dd2",
          "dd3",
          "dd4",
          "dd5",
          "dd6",
          "gm0",
          "gmd0",
        ]:
          grind_1_is_sell_found = True

    if derisk_1_sub_grind_count > 0:
      derisk_1_current_open_rate = derisk_1_total_cost / derisk_1_total_amount
      derisk_1_current_grind_stake = derisk_1_total_amount * exit_rate * (1 + trade.fee_close)
      derisk_1_current_grind_stake_profit = derisk_1_total_cost - derisk_1_current_grind_stake
    if grind_1_sub_grind_count > 0:
      grind_1_current_open_rate = grind_1_total_cost / grind_1_total_amount
      grind_1_current_grind_stake = grind_1_total_amount * exit_rate * (1 + trade.fee_close)
      grind_1_current_grind_stake_profit = grind_1_total_cost - grind_1_current_grind_stake
    if grind_2_sub_grind_count > 0:
      grind_2_current_open_rate = grind_2_total_cost / grind_2_total_amount
      grind_2_current_grind_stake = grind_2_total_amount * exit_rate * (1 + trade.fee_close)
      grind_2_current_grind_stake_profit = grind_2_total_cost - grind_2_current_grind_stake
    if grind_3_sub_grind_count > 0:
      grind_3_current_open_rate = grind_3_total_cost / grind_3_total_amount
      grind_3_current_grind_stake = grind_3_total_amount * exit_rate * (1 + trade.fee_close)
      grind_3_current_grind_stake_profit = grind_3_total_cost - grind_3_current_grind_stake
    if grind_4_sub_grind_count > 0:
      grind_4_current_open_rate = grind_4_total_cost / grind_4_total_amount
      grind_4_current_grind_stake = grind_4_total_amount * exit_rate * (1 + trade.fee_close)
      grind_4_current_grind_stake_profit = grind_4_total_cost - grind_4_current_grind_stake
    if grind_5_sub_grind_count > 0:
      grind_5_current_open_rate = grind_5_total_cost / grind_5_total_amount
      grind_5_current_grind_stake = grind_5_total_amount * exit_rate * (1 + trade.fee_close)
      grind_5_current_grind_stake_profit = grind_5_total_cost - grind_5_current_grind_stake
    if grind_6_sub_grind_count > 0:
      grind_6_current_open_rate = grind_6_total_cost / grind_6_total_amount
      grind_6_current_grind_stake = grind_6_total_amount * exit_rate * (1 + trade.fee_close)
      grind_6_current_grind_stake_profit = grind_6_total_cost - grind_6_current_grind_stake
    if grind_1_derisk_1_sub_grind_count > 0:
      grind_1_derisk_1_current_open_rate = grind_1_derisk_1_total_cost / grind_1_derisk_1_total_amount
      grind_1_derisk_1_current_grind_stake = grind_1_derisk_1_total_amount * exit_rate * (1 + trade.fee_close)
      grind_1_derisk_1_current_grind_stake_profit = grind_1_derisk_1_total_cost - grind_1_derisk_1_current_grind_stake
    if grind_2_derisk_1_sub_grind_count > 0:
      grind_2_derisk_1_current_open_rate = grind_2_derisk_1_total_cost / grind_2_derisk_1_total_amount
      grind_2_derisk_1_current_grind_stake = grind_2_derisk_1_total_amount * exit_rate * (1 + trade.fee_close)
      grind_2_derisk_1_current_grind_stake_profit = grind_2_derisk_1_total_cost - grind_2_derisk_1_current_grind_stake

    num_open_grinds = (
      grind_1_sub_grind_count
      + grind_2_sub_grind_count
      + grind_3_sub_grind_count
      + grind_4_sub_grind_count
      + grind_5_sub_grind_count
      + grind_6_sub_grind_count
      + grind_1_derisk_1_sub_grind_count
      + grind_2_derisk_1_sub_grind_count
    )

    if partial_sell:
      order = filled_exits[-1]
      sell_amount = order.safe_remaining * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        self.dp.send_msg(
          f"Exit (remaining) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {order.safe_remaining} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        order_tag = "p"
        if has_order_tags:
          if order.ft_order_tag is not None:
            order_tag = order.ft_order_tag
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount

    if is_grind_mode and (
      (filled_entries[0].safe_filled * (trade.stake_amount / trade.amount) - (min_stake * 1.5)) > min_stake
    ):
      is_first_entry_exit_found = False
      for order in filled_orders:
        if order.ft_order_side == "sell":
          order_tag = ""
          if has_order_tags:
            if order.ft_order_tag is not None:
              sell_order_tag = order.ft_order_tag
              order_mode = sell_order_tag.split(" ", 1)
              if len(order_mode) > 0:
                order_tag = order_mode[0]
          else:

            is_first_entry_exit_found = True
          if order_tag in ["gm0", "gmd0"]:
            is_first_entry_exit_found = True
            break
      if not is_first_entry_exit_found:
        first_entry = filled_entries[0]
        first_entry_distance_ratio = -(exit_rate - first_entry.safe_price) / first_entry.safe_price

        if first_entry_distance_ratio > (
          self.grind_mode_first_entry_profit_threshold_spot
          if self.is_futures_mode
          else self.grind_mode_first_entry_profit_threshold_spot
        ):
          sell_amount = first_entry.safe_filled * exit_rate / trade.leverage
          if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
            sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
          ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
          if sell_amount > min_stake and ft_sell_amount > min_stake:
            grind_profit = -(exit_rate - first_entry.safe_price) / first_entry.safe_price
            coin_amount = sell_amount / exit_rate
            self.dp.send_msg(
              f"Grinding exit (gm0) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {coin_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
            )
            log.info(
              f"Grinding exit (gm0) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {coin_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
            )
            order_tag = "gm0"
            for grind_entry_id in grind_1_buy_orders:
              order_tag += " " + str(grind_entry_id)
            if has_order_tags:
              return -ft_sell_amount, order_tag
            else:
              return -ft_sell_amount

        if first_entry_distance_ratio < (
          self.grind_mode_first_entry_stop_threshold_spot
          if self.is_futures_mode
          else self.grind_mode_first_entry_stop_threshold_spot
        ):
          sell_amount = first_entry.safe_filled * exit_rate / trade.leverage
          if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
            sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
          ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
          if sell_amount > min_stake and ft_sell_amount > min_stake:
            grind_profit = -(exit_rate - first_entry.safe_price) / first_entry.safe_price
            coin_amount = sell_amount / exit_rate
            self.dp.send_msg(
              f"Grinding de-risk (gmd0) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {coin_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
            )
            log.info(
              f"Grinding de-risk (gmd0) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {coin_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
            )
            order_tag = "gmd0"
            for grind_entry_id in grind_1_buy_orders:
              order_tag += " " + str(grind_entry_id)
            if has_order_tags:
              return -ft_sell_amount, order_tag
            else:
              return -ft_sell_amount

    is_short_grind_buy = self.short_grind_buy(last_candle, previous_candle, slice_profit)


    if (
      has_order_tags
      and is_derisk_1
      and not derisk_1_reentry_found
      and (not partial_sell)
      and (grind_1_derisk_1_sub_grind_count < grind_1_derisk_1_max_sub_grinds)
    ):
      if (
        (
          (
            (grind_1_derisk_1_sub_grind_count > 0)
            and -grind_1_derisk_1_distance_ratio < grind_1_derisk_1_sub_thresholds[grind_1_derisk_1_sub_grind_count]
          )
          or ((is_derisk or is_derisk_calc) and grind_1_derisk_1_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (last_candle["close"] < (last_candle["close_min_12"] * 1.06))
          and (last_candle["close"] < (last_candle["close_min_24"] * 1.08))
          and (last_candle["close"] < (last_candle["close_min_48"] * 1.10))
          and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.12))
          and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.14))
          and (last_candle["close"] < (last_candle["low_min_6_1d"] * 1.16))
          and (last_candle["close"] < (last_candle["low_min_12_1d"] * 1.18))
        )
        and (
          (last_candle["zlma_50_dec_15m"] == True)
          and (last_candle["zlma_50_dec_1h"] == True)
          and (last_candle["zlma_50_dec_4h"] == True)
          and (last_candle["ema_200_dec_48_1h"] == True)
          and (last_candle["ema_200_dec_24_4h"] == True)
        )
        and (
          is_short_grind_buy
          or (
            (last_candle["rsi_3"] < 84.0)
            and (last_candle["rsi_3_15m"] < 84.0)
            and (last_candle["rsi_3_1h"] < 70.0)
            and (last_candle["rsi_3_4h"] < 70.0)
            and (last_candle["rsi_14"] > 64.0)
            and (last_candle["close"] < last_candle["res_hlevel_4h"])
            and (last_candle["close"] > last_candle["sup_level_4h"])
          )
        )
      ):
        buy_amount = (
          slice_amount
          * grind_1_derisk_1_stakes[grind_1_derisk_1_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_1_derisk_1_sub_grind_count > 0:
          grind_profit = -(exit_rate - grind_1_derisk_1_current_open_rate) / grind_1_derisk_1_current_open_rate
          grind_profit_stake = grind_1_derisk_1_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (dl1) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_1_derisk_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (dl1) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_1_derisk_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "dl1"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_1_derisk_1_sub_grind_count > 0:
      grind_profit = -(exit_rate - grind_1_derisk_1_current_open_rate) / grind_1_derisk_1_current_open_rate
      if grind_profit > grind_1_derisk_1_profit_threshold:
        sell_amount = grind_1_derisk_1_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (dl1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (dl1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "dl1"
          for grind_entry_id in grind_1_derisk_1_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (-grind_1_derisk_1_distance_ratio < grind_1_derisk_1_stop_grinds) and (is_derisk or is_derisk_calc):
      sell_amount = grind_1_derisk_1_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_1_derisk_1_current_open_rate > 0.0:
          grind_profit = (
            -((exit_rate - grind_1_derisk_1_current_open_rate) / grind_1_derisk_1_current_open_rate)
            if grind_1_derisk_1_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (ddl1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (ddl1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "ddl1"
        for grind_entry_id in grind_1_derisk_1_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if (
      has_order_tags
      and is_derisk_1
      and not derisk_1_reentry_found
      and (not partial_sell)
      and (grind_2_derisk_1_sub_grind_count < grind_2_derisk_1_max_sub_grinds)
    ):
      if (
        (
          (
            (grind_2_derisk_1_sub_grind_count > 0)
            and -grind_2_derisk_1_distance_ratio < grind_2_derisk_1_sub_thresholds[grind_2_derisk_1_sub_grind_count]
          )
          or ((is_derisk or is_derisk_calc) and grind_2_derisk_1_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (last_candle["close"] < (last_candle["close_min_12"] * 1.06))
          and (last_candle["close"] < (last_candle["close_min_24"] * 1.08))
          and (last_candle["close"] < (last_candle["close_min_48"] * 1.10))
          and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.12))
          and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.14))
        )
        and (
          (
            (grind_2_derisk_1_sub_grind_count == 0)
            and (
              is_short_grind_buy
              or (
                (last_candle["rsi_3"] < 84.0)
                and (last_candle["rsi_3_15m"] < 84.0)
                and (last_candle["rsi_3_1h"] < 80.0)
                and (last_candle["rsi_3_4h"] < 80.0)
                and (last_candle["rsi_14"] > 64.0)

                and (last_candle["zlma_50_dec_1h"] == True)
                and (last_candle["zlma_50_dec_4h"] == True)
              )
            )
          )
          or (
            (grind_2_derisk_1_sub_grind_count > 0)
            and (
              is_short_grind_buy
              or (
                (last_candle["rsi_3"] < 88.0)
                and (last_candle["rsi_3_15m"] < 84.0)
                and (last_candle["rsi_3_1h"] < 84.0)
                and (last_candle["rsi_3_4h"] < 84.0)
                and (last_candle["rsi_14"] > 54.0)
              )
            )
          )
        )
      ):
        buy_amount = (
          slice_amount
          * grind_2_derisk_1_stakes[grind_2_derisk_1_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_2_derisk_1_sub_grind_count > 0:
          grind_profit = -(exit_rate - grind_2_derisk_1_current_open_rate) / grind_2_derisk_1_current_open_rate
          grind_profit_stake = grind_2_derisk_1_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (dl2) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_2_derisk_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (dl2) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_2_derisk_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "dl2"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_2_derisk_1_sub_grind_count > 0:
      grind_profit = -(exit_rate - grind_2_derisk_1_current_open_rate) / grind_2_derisk_1_current_open_rate
      if grind_profit > grind_2_derisk_1_profit_threshold:
        sell_amount = grind_2_derisk_1_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (dl2) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (dl2) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "dl2"
          for grind_entry_id in grind_2_derisk_1_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (grind_2_derisk_1_sub_grind_count > 0)
      and (
        (-(exit_rate - grind_2_derisk_1_current_open_rate) / grind_2_derisk_1_current_open_rate)
        < grind_2_derisk_1_stop_grinds
      )
      and (is_derisk or is_derisk_calc)
    ):
      sell_amount = grind_2_derisk_1_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_2_derisk_1_current_open_rate > 0.0:
          grind_profit = (
            (-(exit_rate - grind_2_derisk_1_current_open_rate) / grind_2_derisk_1_current_open_rate)
            if grind_2_derisk_1_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (ddl2) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (ddl2) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_derisk_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "ddl2"
        for grind_entry_id in grind_2_derisk_1_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if (not partial_sell) and (grind_1_sub_grind_count < grind_1_max_sub_grinds):
      if (
        (
          ((grind_1_sub_grind_count > 0) and -grind_1_distance_ratio < grind_1_sub_thresholds[grind_1_sub_grind_count])
          or ((is_derisk or is_derisk_calc) and grind_1_sub_grind_count == 0)
          or (is_grind_mode and grind_1_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (
            (
              (last_candle["close"] < (last_candle["close_min_12"] * 1.06))
              and (last_candle["close"] < (last_candle["close_min_24"] * 1.08))
              and (last_candle["close"] < (last_candle["close_min_48"] * 1.10))
              and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.12))
              and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.14))
            )
            and (
              (
                (grind_1_sub_grind_count == 0)
                and (
                  is_short_grind_buy
                  or (
                    (last_candle["rsi_3"] < 90.0)
                    and (last_candle["rsi_3_15m"] < 90.0)
                    and (last_candle["rsi_3_1h"] < 90.0)
                    and (last_candle["rsi_3_4h"] < 90.0)
                    and (last_candle["rsi_14"] > 64.0)



                    and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
                  )
                )
              )
              or (
                (grind_1_sub_grind_count > 0)
                and (
                  is_short_grind_buy
                  or (
                    (last_candle["rsi_3"] < 88.0)
                    and (last_candle["rsi_3_15m"] < 88.0)


                    and (last_candle["rsi_14"] > 58.0)
                  )
                )
              )
            )
          )
          or (
            (slice_profit < 0.06)
            and (last_candle["rsi_3"] < 90.0)
            and (last_candle["rsi_3_15m"] < 90.0)
            and (last_candle["rsi_14"] < 72.0)
            and (last_candle["rsi_14"] > 64.0)
            and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
          )
        )
      ):
        buy_amount = (
          slice_amount * grind_1_stakes[grind_1_sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_1_sub_grind_count > 0:
          grind_profit = -(exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate
          grind_profit_stake = grind_1_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (gd1) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (gd1) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "gd1"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_1_sub_grind_count > 0:
      grind_profit = -(exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate
      if grind_profit > grind_1_profit_threshold:
        sell_amount = grind_1_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (gd1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (gd1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "gd1"
          for grind_entry_id in grind_1_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (
        (grind_1_sub_grind_count > 0)
        and ((-(exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate) < grind_1_stop_grinds)




        and (is_derisk or is_derisk_calc or is_grind_mode)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 14) or is_backtest)
    ):
      sell_amount = grind_1_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_1_current_open_rate > 0.0:
          grind_profit = (
            (-(exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate)
            if grind_1_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (dd1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (dd1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "dd1"
        for grind_entry_id in grind_1_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if has_order_tags and (not partial_sell) and (grind_2_sub_grind_count < grind_2_max_sub_grinds):
      if (
        (
          ((grind_2_sub_grind_count > 0) and -grind_2_distance_ratio < grind_2_sub_thresholds[grind_2_sub_grind_count])
          or ((is_derisk or is_derisk_calc) and grind_2_sub_grind_count == 0)
          or (is_grind_mode and grind_2_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (
            (
              (last_candle["close"] < (last_candle["close_min_12"] * 1.06))
              and (last_candle["close"] < (last_candle["close_min_24"] * 1.08))
              and (last_candle["close"] < (last_candle["close_min_48"] * 1.10))
              and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.12))
              and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.14))
            )
            and (
              (
                (grind_2_sub_grind_count == 0)
                and (
                  is_short_grind_buy
                  or (
                    (last_candle["rsi_3"] < 90.0)
                    and (last_candle["rsi_3_15m"] < 90.0)
                    and (last_candle["rsi_3_1h"] < 90.0)
                    and (last_candle["rsi_3_4h"] < 90.0)
                    and (last_candle["rsi_14"] > 64.0)



                    and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
                  )
                )
              )
              or (
                (grind_2_sub_grind_count > 0)
                and (
                  is_short_grind_buy
                  or (
                    (last_candle["rsi_3"] < 88.0)
                    and (last_candle["rsi_3_15m"] < 88.0)


                    and (last_candle["rsi_14"] > 58.0)
                  )
                )
              )
            )
          )
          or (
            (slice_profit < 0.06)
            and (last_candle["rsi_3"] < 90.0)
            and (last_candle["rsi_3_15m"] < 90.0)
            and (last_candle["rsi_14"] < 72.0)
            and (last_candle["rsi_14"] > 64.0)
            and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
          )
        )
      ):
        buy_amount = (
          slice_amount * grind_2_stakes[grind_2_sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_2_sub_grind_count > 0:
          grind_profit = -(exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate
          grind_profit_stake = grind_2_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (gd2) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_2_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (gd2) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_2_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "gd2"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_2_sub_grind_count > 0:
      grind_profit = -(exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate
      if grind_profit > grind_2_profit_threshold:
        sell_amount = grind_2_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (gd2) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (gd2) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "gd2"
          for grind_entry_id in grind_2_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (
        (grind_2_sub_grind_count > 0)
        and ((-(exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate) < grind_2_stop_grinds)




        and (is_derisk or is_derisk_calc or is_grind_mode)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 14) or is_backtest)
    ):
      sell_amount = grind_2_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_2_current_open_rate > 0.0:
          grind_profit = (
            (-(exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate)
            if grind_2_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (dd2) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (dd2) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "dd2"
        for grind_entry_id in grind_2_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if has_order_tags and (not partial_sell) and (grind_3_sub_grind_count < grind_3_max_sub_grinds):
      if (
        (
          ((grind_3_sub_grind_count > 0) and -grind_3_distance_ratio < grind_3_sub_thresholds[grind_3_sub_grind_count])
          or ((is_derisk or is_derisk_calc) and grind_3_sub_grind_count == 0)
          or (is_grind_mode and grind_3_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (
            (
              (last_candle["close"] < (last_candle["close_min_12"] * 1.06))
              and (last_candle["close"] < (last_candle["close_min_24"] * 1.08))
              and (last_candle["close"] < (last_candle["close_min_48"] * 1.10))
              and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.12))
              and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.14))
            )
            and (
              (
                (grind_3_sub_grind_count == 0)
                and (
                  is_short_grind_buy
                  or (
                    (last_candle["rsi_3"] < 90.0)
                    and (last_candle["rsi_3_15m"] < 90.0)
                    and (last_candle["rsi_3_1h"] < 90.0)
                    and (last_candle["rsi_3_4h"] < 90.0)
                    and (last_candle["rsi_14"] > 64.0)



                    and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
                  )
                )
              )
              or (
                (grind_3_sub_grind_count > 0)
                and (
                  is_short_grind_buy
                  or (
                    (last_candle["rsi_3"] < 88.0)
                    and (last_candle["rsi_3_15m"] < 88.0)


                    and (last_candle["rsi_14"] > 58.0)
                  )
                )
              )
            )
          )
          or (
            (slice_profit < 0.06)
            and (last_candle["rsi_3"] < 90.0)
            and (last_candle["rsi_3_15m"] < 90.0)
            and (last_candle["rsi_14"] < 72.0)
            and (last_candle["rsi_14"] > 64.0)
            and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
          )
        )
      ):
        buy_amount = (
          slice_amount * grind_3_stakes[grind_3_sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_3_sub_grind_count > 0:
          grind_profit = -(exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate
          grind_profit_stake = grind_3_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (gd3) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_3_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (gd3) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_3_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "gd3"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_3_sub_grind_count > 0:
      grind_profit = -(exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate
      if grind_profit > grind_3_profit_threshold:
        sell_amount = grind_3_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (gd3) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (gd3) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "gd3"
          for grind_entry_id in grind_3_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (
        (grind_3_sub_grind_count > 0)
        and ((-(exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate) < grind_3_stop_grinds)




        and (is_derisk or is_derisk_calc or is_grind_mode)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 14) or is_backtest)
    ):
      sell_amount = grind_3_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_3_current_open_rate > 0.0:
          grind_profit = (
            (-(exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate)
            if grind_3_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (dd3) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (dd3) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "dd3"
        for grind_entry_id in grind_3_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if has_order_tags and (not partial_sell) and (grind_4_sub_grind_count < grind_4_max_sub_grinds):
      if (
        (
          ((grind_4_sub_grind_count > 0) and -grind_4_distance_ratio < grind_4_sub_thresholds[grind_4_sub_grind_count])
          or ((is_derisk or is_derisk_calc) and grind_4_sub_grind_count == 0)
          or (is_grind_mode and grind_4_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (
            (
              (last_candle["close"] < (last_candle["close_min_12"] * 1.06))
              and (last_candle["close"] < (last_candle["close_min_24"] * 1.08))
              and (last_candle["close"] < (last_candle["close_min_48"] * 1.10))
              and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.12))
              and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.14))
            )
            and (
              (
                (grind_4_sub_grind_count == 0)
                and (
                  is_short_grind_buy
                  or (
                    (last_candle["rsi_3"] < 90.0)
                    and (last_candle["rsi_3_15m"] < 90.0)
                    and (last_candle["rsi_3_1h"] < 90.0)
                    and (last_candle["rsi_3_4h"] < 90.0)
                    and (last_candle["rsi_14"] > 64.0)



                    and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
                  )
                )
              )
              or (
                (grind_4_sub_grind_count > 0)
                and (
                  is_short_grind_buy
                  or (
                    (last_candle["rsi_3"] < 88.0)
                    and (last_candle["rsi_3_15m"] < 88.0)


                    and (last_candle["rsi_14"] > 58.0)
                  )
                )
              )
            )
          )
          or (
            (slice_profit < 0.06)
            and (last_candle["rsi_3"] < 90.0)
            and (last_candle["rsi_3_15m"] < 90.0)
            and (last_candle["rsi_14"] < 72.0)
            and (last_candle["rsi_14"] > 64.0)
            and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
          )
        )
      ):
        buy_amount = (
          slice_amount * grind_4_stakes[grind_4_sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_4_sub_grind_count > 0:
          grind_profit = -(exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate
          grind_profit_stake = grind_4_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (gd4) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_4_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (gd4) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_4_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "gd4"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_4_sub_grind_count > 0:
      grind_profit = -(exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate
      if grind_profit > grind_4_profit_threshold:
        sell_amount = grind_4_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (gd4) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (gd4) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "gd4"
          for grind_entry_id in grind_4_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (
        (grind_4_sub_grind_count > 0)
        and ((-(exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate) < grind_4_stop_grinds)




        and (is_derisk or is_derisk_calc or is_grind_mode)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 14) or is_backtest)
    ):
      sell_amount = grind_4_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_4_current_open_rate > 0.0:
          grind_profit = (
            (-(exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate)
            if grind_4_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (dd4) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (dd4) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "dd4"
        for grind_entry_id in grind_4_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if has_order_tags and (not partial_sell) and (grind_5_sub_grind_count < grind_5_max_sub_grinds):
      if (
        (
          ((grind_5_sub_grind_count > 0) and -grind_5_distance_ratio < grind_5_sub_thresholds[grind_5_sub_grind_count])
          or ((is_derisk or is_derisk_calc) and grind_5_sub_grind_count == 0)
          or (is_grind_mode and grind_5_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (
            (
              (last_candle["close"] < (last_candle["close_min_12"] * 1.06))
              and (last_candle["close"] < (last_candle["close_min_24"] * 1.08))
              and (last_candle["close"] < (last_candle["close_min_48"] * 1.10))
              and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.12))
              and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.14))
            )
            and (
              (
                (grind_5_sub_grind_count == 0)
                and (
                  is_short_grind_buy
                  or (
                    (last_candle["rsi_3"] < 90.0)
                    and (last_candle["rsi_3_15m"] < 90.0)
                    and (last_candle["rsi_3_1h"] < 90.0)
                    and (last_candle["rsi_3_4h"] < 90.0)
                    and (last_candle["rsi_14"] > 64.0)



                    and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
                  )
                )
              )
              or (
                (grind_5_sub_grind_count > 0)
                and (
                  is_short_grind_buy
                  or (
                    (last_candle["rsi_3"] < 88.0)
                    and (last_candle["rsi_3_15m"] < 88.0)


                    and (last_candle["rsi_14"] > 58.0)
                  )
                )
              )
            )
          )
          or (
            (slice_profit < 0.06)
            and (last_candle["rsi_3"] < 90.0)
            and (last_candle["rsi_3_15m"] < 90.0)
            and (last_candle["rsi_14"] < 72.0)
            and (last_candle["rsi_14"] > 64.0)
            and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
          )
        )
      ):
        buy_amount = (
          slice_amount * grind_5_stakes[grind_5_sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_5_sub_grind_count > 0:
          grind_profit = -(exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate
          grind_profit_stake = grind_5_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (gd5) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_5_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (gd5) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_5_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "gd5"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_5_sub_grind_count > 0:
      grind_profit = -(exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate
      if grind_profit > grind_5_profit_threshold:
        sell_amount = grind_5_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (gd5) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (gd5) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "gd5"
          for grind_entry_id in grind_5_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (
        (grind_5_sub_grind_count > 0)
        and ((-(exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate) < grind_5_stop_grinds)




        and (is_derisk or is_derisk_calc or is_grind_mode)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 14) or is_backtest)
    ):
      sell_amount = grind_5_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_5_current_open_rate > 0.0:
          grind_profit = (
            (-(exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate)
            if grind_5_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (dd5) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (dd5) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "dd5"
        for grind_entry_id in grind_5_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount


    if has_order_tags and (not partial_sell) and (grind_6_sub_grind_count < grind_6_max_sub_grinds):
      if (
        (
          ((grind_6_sub_grind_count > 0) and -grind_6_distance_ratio < grind_6_sub_thresholds[grind_6_sub_grind_count])
          or ((is_derisk or is_derisk_calc) and grind_6_sub_grind_count == 0)
          or (is_grind_mode and grind_6_sub_grind_count == 0)
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (
            (
              (last_candle["close"] < (last_candle["close_min_12"] * 1.06))
              and (last_candle["close"] < (last_candle["close_min_24"] * 1.08))
              and (last_candle["close"] < (last_candle["close_min_48"] * 1.10))
              and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.12))
              and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.14))
            )
            and (
              (
                (grind_6_sub_grind_count == 0)
                and (
                  is_short_grind_buy
                  or (
                    (last_candle["rsi_3"] < 90.0)
                    and (last_candle["rsi_3_15m"] < 90.0)
                    and (last_candle["rsi_3_1h"] < 90.0)
                    and (last_candle["rsi_3_4h"] < 90.0)
                    and (last_candle["rsi_14"] > 64.0)



                    and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
                  )
                )
              )
              or (
                (grind_6_sub_grind_count > 0)
                and (
                  is_short_grind_buy
                  or (
                    (last_candle["rsi_3"] < 88.0)
                    and (last_candle["rsi_3_15m"] < 88.0)


                    and (last_candle["rsi_14"] > 58.0)
                  )
                )
              )
            )
          )
          or (
            (slice_profit < 0.06)
            and (last_candle["rsi_3"] < 90.0)
            and (last_candle["rsi_3_15m"] < 90.0)
            and (last_candle["rsi_14"] < 72.0)
            and (last_candle["rsi_14"] > 58.0)
            and (last_candle["close"] > (last_candle["ema_26"] * 1.006))
          )
        )
      ):
        buy_amount = (
          slice_amount * grind_6_stakes[grind_6_sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_6_sub_grind_count > 0:
          grind_profit = -(exit_rate - grind_6_current_open_rate) / grind_6_current_open_rate
          grind_profit_stake = grind_6_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (gd6) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_6_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (gd6) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_6_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "gd6"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if grind_6_sub_grind_count > 0:
      grind_profit = -(exit_rate - grind_6_current_open_rate) / grind_6_current_open_rate
      if grind_profit > grind_6_profit_threshold:
        sell_amount = grind_6_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (gd6) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_6_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (gd6) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_6_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "gd6"
          for grind_entry_id in grind_6_buy_orders:
            order_tag += " " + str(grind_entry_id)
          if has_order_tags:
            return -ft_sell_amount, order_tag
          else:
            return -ft_sell_amount

    if (
      (
        (grind_6_sub_grind_count > 0)
        and ((-(exit_rate - grind_6_current_open_rate) / grind_6_current_open_rate) < grind_6_stop_grinds)




        and (is_derisk or is_derisk_calc or is_grind_mode)
      )

      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 14) or is_backtest)
    ):
      sell_amount = grind_6_total_amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        if grind_6_current_open_rate > 0.0:
          grind_profit = (
            (-(exit_rate - grind_6_current_open_rate) / grind_6_current_open_rate)
            if grind_6_is_sell_found
            else profit_ratio
          )
        self.dp.send_msg(
          f"Grinding stop exit (dd6) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_6_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        log.info(
          f"Grinding stop exit (dd6) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_6_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}%"
        )
        order_tag = "dd6"
        for grind_entry_id in grind_6_buy_orders:
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount

    if (
      is_derisk_1
      and not derisk_1_reentry_found
      and derisk_1_order is not None
      and (
        (-(current_rate - derisk_1_order.safe_price) / derisk_1_order.safe_price)
        < (
          self.regular_mode_derisk_1_reentry_futures
          if self.is_futures_mode
          else self.regular_mode_derisk_1_reentry_spot
        )
      )
    ):
      if (
        (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (last_candle["close"] < (last_candle["close_min_12"] * 1.06))
          and (last_candle["close"] < (last_candle["close_min_24"] * 1.08))
          and (last_candle["close"] < (last_candle["close_min_48"] * 1.10))
          and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.12))
          and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.14))
          and (last_candle["close"] < (last_candle["low_min_6_1d"] * 1.16))
          and (last_candle["close"] < (last_candle["low_min_12_1d"] * 1.18))
        )
        and (
          (last_candle["zlma_50_dec_15m"] == True)
          and (last_candle["zlma_50_dec_1h"] == True)
          and (last_candle["zlma_50_dec_4h"] == True)
          and (last_candle["ema_200_dec_48_1h"] == True)
          and (last_candle["ema_200_dec_24_4h"] == True)
          and (last_candle["close"] < last_candle["res_hlevel_4h"])
          and (last_candle["close"] > last_candle["sup_level_4h"])
        )
        and (
          is_short_grind_buy
          or (
            (last_candle["rsi_3"] < 70.0)
            and (last_candle["rsi_3_15m"] < 70.0)
            and (last_candle["rsi_3_1h"] < 70.0)
            and (last_candle["rsi_3_4h"] < 70.0)
            and (last_candle["rsi_14"] > 58.0)
          )
        )
      ):
        buy_amount = derisk_1_order.safe_filled * derisk_1_order.safe_price
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if derisk_1_sub_grind_count > 0:
          grind_profit = -(exit_rate - derisk_1_current_open_rate) / derisk_1_current_open_rate
          grind_profit_stake = derisk_1_current_grind_stake_profit
        self.dp.send_msg(
          f"Re-entry (d1) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({derisk_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Re-entry (d1) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({derisk_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "d1"
        if has_order_tags:
          return buy_amount, order_tag
        else:
          return buy_amount

    if (
      has_order_tags

      and derisk_1_reentry_found
      and derisk_1_reentry_order is not None

      and -derisk_1_distance_ratio
      < (
        (self.regular_mode_derisk_1_futures if self.is_futures_mode else self.regular_mode_derisk_1_spot)
        / (trade.leverage if self.is_futures_mode else 1.0)
      )
    ):
      sell_amount = derisk_1_reentry_order.safe_filled * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        self.dp.send_msg(
          f"De-risk (d1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        log.info(
          f"De-risk (d1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        return -ft_sell_amount, "d1"

    if (
      not is_derisk_found
      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 2, 5) or is_backtest)
      and profit_stake
      < (
        slice_amount
        * (
          (self.regular_mode_derisk_futures if self.is_futures_mode else self.regular_mode_derisk_spot)
          if (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 5) or is_backtest)
          else (self.regular_mode_derisk_futures_old if self.is_futures_mode else self.regular_mode_derisk_spot_old)
        )
        / (trade.leverage if self.is_futures_mode else 1.0)
      )
    ):
      sell_amount = trade.amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        self.dp.send_msg(
          f"De-risk [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        log.info(
          f"De-risk [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        return -ft_sell_amount, "d", is_derisk

    if (
      (
        (profit_stake < (slice_amount * grind_derisk / (trade.leverage if self.is_futures_mode else 1.0)))
        and (
          (
            (trade.amount * exit_rate / (trade.leverage if self.is_futures_mode else 1.0))
            - (
              (
                derisk_1_total_amount
                + grind_1_derisk_1_total_amount
                + grind_2_derisk_1_total_amount
                + grind_1_total_amount
                + grind_2_total_amount
                + grind_3_total_amount
                + grind_4_total_amount
                + grind_5_total_amount
                + grind_6_total_amount
              )
              * exit_rate
              / (trade.leverage if self.is_futures_mode else 1.0)
            )
          )
          > (min_stake * 3.0)
        )

        and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2023, 12, 19) or is_backtest)
      )

      and (
        (trade.open_date_utc.replace(tzinfo=None) >= datetime(2023, 8, 28) or is_backtest)
        or (filled_entries[-1].order_date_utc.replace(tzinfo=None) >= datetime(2023, 8, 28) or is_backtest)
      )
    ):
      sell_amount = trade.amount * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        self.dp.send_msg(
          f"De-risk (dd0) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        log.info(
          f"De-risk (dd0) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        order_tag = "dd0"
        for grind_entry_id in (
          grind_1_buy_orders
          + grind_2_buy_orders
          + grind_3_buy_orders
          + grind_4_buy_orders
          + grind_5_buy_orders
          + grind_6_buy_orders
          + grind_1_derisk_1_buy_orders
          + grind_2_derisk_1_buy_orders
        ):
          order_tag += " " + str(grind_entry_id)
        if has_order_tags:
          return -ft_sell_amount, order_tag
        else:
          return -ft_sell_amount

    return None


  def short_grind_buy(self, last_candle: Series, previous_candle: Series, slice_profit: float) -> float:
    if (
      (
        (last_candle["close"] < (last_candle["close_min_12"] * 1.12))
        and (last_candle["close"] < (last_candle["close_min_24"] * 1.18))
        and (last_candle["close"] < (last_candle["close_min_48"] * 1.24))
        and (last_candle["btc_pct_close_min_72_5m"] < -0.03)
        and (last_candle["btc_pct_close_min_24_5m"] < -0.03)
      )
      and (
        (last_candle["enter_short"] == True)
        or (
          (last_candle["rsi_3"] < 90.0)
          and (last_candle["rsi_3_15m"] < 80.0)
          and (last_candle["rsi_3_1h"] < 80.0)
          and (last_candle["rsi_3_4h"] < 80.0)
          and (last_candle["rsi_14"] > 56.0)
          and (last_candle["ha_close"] < last_candle["ha_open"])
          and (last_candle["ema_12"] > (last_candle["ema_26"] * 1.010))
          and (last_candle["cti_20_1h"] > -0.80)
          and (last_candle["rsi_14_1h"] < 80.0)
        )
        or (
          (last_candle["rsi_14"] > 64.0)
          and (last_candle["close"] > (last_candle["sma_16"] * 1.006))
          and (last_candle["rsi_3"] < 86.0)
          and (last_candle["rsi_3_15m"] < 70.0)
          and (last_candle["rsi_3_1h"] < 70.0)
          and (last_candle["rsi_3_4h"] < 70.0)
          and (last_candle["zlma_50_dec_1h"] == True)
        )
        or (
          (last_candle["rsi_14"] > 64.0)
          and (previous_candle["rsi_3"] < 94.0)
          and (last_candle["ema_26"] < last_candle["ema_12"])
          and ((last_candle["ema_26"] - last_candle["ema_12"]) > (last_candle["open"] * 0.010))
          and ((previous_candle["ema_26"] - previous_candle["ema_12"]) > (last_candle["open"] / 100.0))
          and (last_candle["rsi_3_1h"] < 80.0)
          and (last_candle["rsi_3_4h"] < 80.0)
          and (last_candle["cti_20_1h"] > -0.80)
          and (last_candle["rsi_14_1h"] > 20.0)
        )
        or (
          (last_candle["rsi_14"] < 70.0)
          and (last_candle["rsi_14"] > 40.0)
          and (last_candle["hma_70_buy"] == False)
          and (last_candle["close"] < (last_candle["low_min_12_1h"] * 0.90))
          and (last_candle["cti_20_15m"] > -0.50)
          and (last_candle["rsi_14_15m"] > 50.0)
          and (last_candle["cti_20_1h"] > -0.80)
          and (last_candle["rsi_14_1h"] > 20.0)
          and (last_candle["zlma_50_dec_1h"] == True)
          and (last_candle["zlma_50_dec_4h"] == True)
        )
        or (
          (last_candle["rsi_3"] < 88.0)
          and (last_candle["rsi_3_15m"] < 80.0)
          and (last_candle["rsi_3_1h"] < 80.0)
          and (last_candle["rsi_3_4h"] < 80.0)
          and (last_candle["rsi_14"] > 64.0)
          and (last_candle["zlma_50_dec_15m"] == True)
          and (last_candle["zlma_50_dec_1h"] == True)
        )
        or (
          (last_candle["rsi_14"] > 60.0)
          and (last_candle["rsi_14_15m"] > 60.0)
          and (last_candle["rsi_3"] < 94.0)
          and (last_candle["ema_26_15m"] < last_candle["ema_12_15m"])
          and ((last_candle["ema_12_15m"] - last_candle["ema_26_15m"]) > (last_candle["open_15m"] * 0.006))
          and ((previous_candle["ema_12_15m"] - previous_candle["ema_26_15m"]) > (last_candle["open_15m"] / 100.0))
          and (last_candle["rsi_3_15m"] < 90.0)
          and (last_candle["rsi_3_1h"] < 74.0)
          and (last_candle["rsi_3_4h"] < 74.0)
          and (last_candle["cti_20_1h"] > -0.80)
          and (last_candle["rsi_14_1h"] > 20.0)
        )
        or (
          (last_candle["rsi_14"] < 65.0)
          and (last_candle["rsi_3"] < 96.0)
          and (last_candle["rsi_3"] > 54.0)
          and (last_candle["rsi_14"] > previous_candle["rsi_14"])
          and (last_candle["close"] > (last_candle["sma_16"] * 1.018))
          and (last_candle["cti_20"] > 0.60)
          and (last_candle["rsi_3_1h"] < 80.0)
          and (last_candle["rsi_3_4h"] < 80.0)
          and (last_candle["not_downtrend_1d"] == False)
          and (last_candle["zlma_50_dec_1h"] == True)
        )
        or (
          (last_candle["rsi_3"] < 88.0)
          and (last_candle["rsi_3_15m"] < 74.0)
          and (last_candle["rsi_3_1h"] < 74.0)
          and (last_candle["rsi_3_4h"] < 74.0)
          and (last_candle["rsi_14"] > 60.0)
          and (last_candle["ema_12"] > (last_candle["ema_26"] * 1.006))
          and (last_candle["cti_20_1h"] > -0.80)
          and (last_candle["rsi_14_1h"] > 20.0)
          and (last_candle["cti_20_4h"] > -0.80)
          and (last_candle["rsi_14_4h"] > 20.0)
          and (last_candle["ema_200_dec_48_1h"] == True)
        )
        or (
          (last_candle["rsi_14"] > 40.0)
          and (last_candle["hma_55_buy"] == False)
          and (last_candle["rsi_3_1h"] < 96.0)
          and (last_candle["rsi_3_4h"] < 96.0)
          and (last_candle["cti_20_15m"] > -0.80)
          and (last_candle["cti_20_1h"] > -0.80)
          and (last_candle["cti_20_4h"] > -0.80)
          and (last_candle["rsi_14_1h"] > 20.0)
          and (last_candle["close"] > (last_candle["low_min_12_1h"] * 1.10))
          and (last_candle["zlma_50_dec_1h"] == True)
          and (last_candle["zlma_50_dec_4h"] == True)
        )
        or (
          (last_candle["rsi_3"] < 88.0)
          and (last_candle["rsi_3_15m"] < 70.0)
          and (last_candle["rsi_3_1h"] < 70.0)
          and (last_candle["rsi_3_4h"] < 70.0)
          and (last_candle["rsi_14"] > 60.0)
          and (last_candle["cti_20_15m"] > -0.80)
          and (last_candle["rsi_14_15m"] > 30.0)
          and (last_candle["cti_20_1h"] > -0.80)
          and (last_candle["rsi_14_1h"] > 20.0)
          and (last_candle["cti_20_4h"] > -0.80)
          and (last_candle["rsi_14_4h"] > 20.0)
          and (last_candle["r_14_1h"] < -20.0)
          and (last_candle["ema_12"] > (last_candle["ema_26"] * 1.005))
          and (last_candle["zlma_50_dec_1h"] == True)
        )
        or (
          (last_candle["rsi_3"] < 88.0)
          and (last_candle["rsi_3_15m"] < 70.0)
          and (last_candle["rsi_3_1h"] < 70.0)
          and (last_candle["rsi_3_4h"] < 70.0)
          and (last_candle["rsi_14"] > 60.0)
          and (last_candle["rsi_14_1d"] > 30.0)
          and (last_candle["close"] > last_candle["sup_level_1h"])
          and (last_candle["close"] > last_candle["sup_level_4h"])
          and (last_candle["close"] > last_candle["sup_level_1d"])
          and (last_candle["not_downtrend_1h"] == False)
          and (last_candle["not_downtrend_4h"] == False)
          and (last_candle["not_downtrend_1d"] == False)
        )
        or (
          (last_candle["rsi_3"] < 88.0)
          and (last_candle["rsi_3_15m"] < 70.0)
          and (last_candle["rsi_3_1h"] < 70.0)
          and (last_candle["rsi_3_4h"] < 70.0)
          and (last_candle["rsi_14"] > 60.0)
          and (last_candle["ha_close"] < last_candle["ha_open"])
          and (last_candle["close"] < last_candle["res_hlevel_4h"])
          and (last_candle["close"] > last_candle["sup_level_4h"])
          and (last_candle["close"] < last_candle["res_hlevel_1d"])
          and (last_candle["close"] > last_candle["sup_level_1d"])
          and (last_candle["close"] > last_candle["sup3_1d"])
          and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.15))
          and (last_candle["hl_pct_change_24_1h"] > -0.35)
        )
        or (
          (last_candle["rsi_3"] < 88.0)
          and (last_candle["rsi_3_15m"] < 70.0)
          and (last_candle["rsi_3_1h"] < 70.0)
          and (last_candle["rsi_3_4h"] < 70.0)
          and (last_candle["rsi_14"] > 54.0)
          and (previous_candle["chandelier_dir"] > 0)
          and (last_candle["chandelier_dir"] < -0)
          and (last_candle["close"] < last_candle["res_hlevel_4h"])
          and (last_candle["close"] > last_candle["sup_level_4h"])
          and (last_candle["close"] < last_candle["res_hlevel_1d"])
          and (last_candle["close"] > last_candle["sup_level_1d"])
        )
      )
    ):
      return True

    return False


  def short_adjust_trade_position_no_derisk(
    self,
    trade: Trade,
    enter_tags,
    current_time: datetime,
    current_rate: float,
    current_profit: float,
    min_stake: Optional[float],
    max_stake: float,
    current_entry_rate: float,
    current_exit_rate: float,
    current_entry_profit: float,
    current_exit_profit: float,
    last_candle: Series,
    previous_candle: Series,
    filled_orders: "Orders",
    filled_entries: "Orders",
    filled_exits: "Orders",
    exit_rate: float,
    slice_amount: float,
    slice_profit_entry: float,
    slice_profit: float,
    profit_ratio: float,
    profit_stake: float,
    profit_init_ratio: float,
    current_stake_amount: float,
    has_order_tags: bool,
    **kwargs,
  ) -> tuple[Optional[float], str, bool]:
    is_backtest = self.dp.runmode.value in ["backtest", "hyperopt"]

    max_rebuy_sub_grinds = 0
    regular_mode_rebuy_stakes = (
      self.regular_mode_rebuy_stakes_futures.copy()
      if self.is_futures_mode
      else self.regular_mode_rebuy_stakes_spot.copy()
    )
    regular_mode_rebuy_sub_thresholds = (
      self.regular_mode_rebuy_thresholds_futures if self.is_futures_mode else self.regular_mode_rebuy_thresholds_spot
    )
    if (slice_amount * regular_mode_rebuy_stakes[0] / trade.leverage) < min_stake:
      multi = min_stake / slice_amount / regular_mode_rebuy_stakes[0]
      for i, _ in enumerate(regular_mode_rebuy_stakes):
        regular_mode_rebuy_stakes[i] *= multi
    max_rebuy_sub_grinds = len(regular_mode_rebuy_stakes)

    max_grind_1_sub_grinds = 0
    regular_mode_grind_1_stakes = (
      self.regular_mode_grind_1_stakes_futures.copy()
      if self.is_futures_mode
      else self.regular_mode_grind_1_stakes_spot.copy()
    )
    regular_mode_grind_1_sub_thresholds = (
      self.regular_mode_grind_1_thresholds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_1_thresholds_spot
    )
    if (slice_amount * regular_mode_grind_1_stakes[0] / trade.leverage) < min_stake:
      multi = min_stake / slice_amount / regular_mode_grind_1_stakes[0]
      for i, _ in enumerate(regular_mode_grind_1_stakes):
        regular_mode_grind_1_stakes[i] *= multi
    max_grind_1_sub_grinds = len(regular_mode_grind_1_stakes)
    regular_mode_grind_1_stop_grinds = (
      self.regular_mode_grind_1_stop_grinds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_1_stop_grinds_spot
    )
    regular_mode_grind_1_profit_threshold = (
      self.regular_mode_grind_1_profit_threshold_futures
      if self.is_futures_mode
      else self.regular_mode_grind_1_profit_threshold_spot
    )

    max_grind_2_sub_grinds = 0
    regular_mode_grind_2_stakes = (
      self.regular_mode_grind_2_stakes_futures.copy()
      if self.is_futures_mode
      else self.regular_mode_grind_2_stakes_spot.copy()
    )
    regular_mode_grind_2_sub_thresholds = (
      self.regular_mode_grind_2_thresholds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_2_thresholds_spot
    )
    if (slice_amount * regular_mode_grind_2_stakes[0] / trade.leverage) < min_stake:
      multi = min_stake / slice_amount / regular_mode_grind_2_stakes[0]
      for i, _ in enumerate(regular_mode_grind_2_stakes):
        regular_mode_grind_2_stakes[i] *= multi
    max_grind_2_sub_grinds = len(regular_mode_grind_2_stakes)
    regular_mode_grind_2_stop_grinds = (
      self.regular_mode_grind_2_stop_grinds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_2_stop_grinds_spot
    )
    regular_mode_grind_2_profit_threshold = (
      self.regular_mode_grind_2_profit_threshold_futures
      if self.is_futures_mode
      else self.regular_mode_grind_2_profit_threshold_spot
    )

    max_grind_3_sub_grinds = 0
    regular_mode_grind_3_stakes = (
      self.regular_mode_grind_3_stakes_futures.copy()
      if self.is_futures_mode
      else self.regular_mode_grind_3_stakes_spot.copy()
    )
    regular_mode_grind_3_sub_thresholds = (
      self.regular_mode_grind_3_thresholds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_3_thresholds_spot
    )
    if (slice_amount * regular_mode_grind_3_stakes[0] / trade.leverage) < min_stake:
      multi = min_stake / slice_amount / regular_mode_grind_3_stakes[0]
      for i, _ in enumerate(regular_mode_grind_3_stakes):
        regular_mode_grind_3_stakes[i] *= multi
    max_grind_3_sub_grinds = len(regular_mode_grind_3_stakes)
    regular_mode_grind_3_stop_grinds = (
      self.regular_mode_grind_3_stop_grinds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_3_stop_grinds_spot
    )
    regular_mode_grind_3_profit_threshold = (
      self.regular_mode_grind_3_profit_threshold_futures
      if self.is_futures_mode
      else self.regular_mode_grind_3_profit_threshold_spot
    )

    max_grind_4_sub_grinds = 0
    regular_mode_grind_4_stakes = (
      self.regular_mode_grind_4_stakes_futures.copy()
      if self.is_futures_mode
      else self.regular_mode_grind_4_stakes_spot.copy()
    )
    regular_mode_grind_4_sub_thresholds = (
      self.regular_mode_grind_4_thresholds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_4_thresholds_spot
    )
    if (slice_amount * regular_mode_grind_4_stakes[0] / trade.leverage) < min_stake:
      multi = min_stake / slice_amount / regular_mode_grind_4_stakes[0]
      for i, _ in enumerate(regular_mode_grind_4_stakes):
        regular_mode_grind_4_stakes[i] *= multi
    max_grind_4_sub_grinds = len(regular_mode_grind_4_stakes)
    regular_mode_grind_4_stop_grinds = (
      self.regular_mode_grind_4_stop_grinds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_4_stop_grinds_spot
    )
    regular_mode_grind_4_profit_threshold = (
      self.regular_mode_grind_4_profit_threshold_futures
      if self.is_futures_mode
      else self.regular_mode_grind_4_profit_threshold_spot
    )

    max_grind_5_sub_grinds = 0
    regular_mode_grind_5_stakes = (
      self.regular_mode_grind_5_stakes_futures.copy()
      if self.is_futures_mode
      else self.regular_mode_grind_5_stakes_spot.copy()
    )
    regular_mode_grind_5_sub_thresholds = (
      self.regular_mode_grind_5_thresholds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_5_thresholds_spot
    )
    if (slice_amount * regular_mode_grind_5_stakes[0] / trade.leverage) < min_stake:
      multi = min_stake / slice_amount / regular_mode_grind_5_stakes[0]
      for i, _ in enumerate(regular_mode_grind_5_stakes):
        regular_mode_grind_5_stakes[i] *= multi
    max_grind_5_sub_grinds = len(regular_mode_grind_5_stakes)
    regular_mode_grind_5_stop_grinds = (
      self.regular_mode_grind_5_stop_grinds_futures
      if self.is_futures_mode
      else self.regular_mode_grind_5_stop_grinds_spot
    )
    regular_mode_grind_5_profit_threshold = (
      self.regular_mode_grind_5_profit_threshold_futures
      if self.is_futures_mode
      else self.regular_mode_grind_5_profit_threshold_spot
    )

    partial_sell = False
    is_derisk = False
    is_derisk_1 = False
    rebuy_sub_grind_count = 0
    rebuy_total_amount = 0.0
    rebuy_total_cost = 0.0
    rebuy_current_open_rate = 0.0
    rebuy_current_grind_stake = 0.0
    rebuy_current_grind_stake_profit = 0.0
    rebuy_is_sell_found = False
    rebuy_found = False
    rebuy_buy_orders = []
    rebuy_distance_ratio = 0.0
    grind_1_sub_grind_count = 0
    grind_1_total_amount = 0.0
    grind_1_total_cost = 0.0
    grind_1_current_open_rate = 0.0
    grind_1_current_grind_stake = 0.0
    grind_1_current_grind_stake_profit = 0.0
    grind_1_is_sell_found = False
    grind_1_found = False
    grind_1_buy_orders = []
    grind_1_distance_ratio = 0.0
    grind_2_sub_grind_count = 0
    grind_2_total_amount = 0.0
    grind_2_total_cost = 0.0
    grind_2_current_open_rate = 0.0
    grind_2_current_grind_stake = 0.0
    grind_2_current_grind_stake_profit = 0.0
    grind_2_is_sell_found = False
    grind_2_found = False
    grind_2_buy_orders = []
    grind_2_distance_ratio = 0.0
    grind_3_sub_grind_count = 0
    grind_3_total_amount = 0.0
    grind_3_total_cost = 0.0
    grind_3_current_open_rate = 0.0
    grind_3_current_grind_stake = 0.0
    grind_3_current_grind_stake_profit = 0.0
    grind_3_is_sell_found = False
    grind_3_found = False
    grind_3_buy_orders = []
    grind_3_distance_ratio = 0.0
    grind_4_sub_grind_count = 0
    grind_4_total_amount = 0.0
    grind_4_total_cost = 0.0
    grind_4_current_open_rate = 0.0
    grind_4_current_grind_stake = 0.0
    grind_4_current_grind_stake_profit = 0.0
    grind_4_is_sell_found = False
    grind_4_found = False
    grind_4_buy_orders = []
    grind_4_distance_ratio = 0.0
    grind_5_sub_grind_count = 0
    grind_5_total_amount = 0.0
    grind_5_total_cost = 0.0
    grind_5_current_open_rate = 0.0
    grind_5_current_grind_stake = 0.0
    grind_5_current_grind_stake_profit = 0.0
    grind_5_is_sell_found = False
    grind_5_found = False
    grind_5_buy_orders = []
    grind_5_distance_ratio = 0.0
    for order in reversed(filled_orders):
      if (order.ft_order_side == "sell") and (order is not filled_orders[0]):
        order_tag = ""
        if has_order_tags:
          if order.ft_order_tag is not None:
            order_tag = order.ft_order_tag
        if not grind_1_is_sell_found and order_tag == "g1":
          grind_1_sub_grind_count += 1
          grind_1_total_amount += order.safe_filled
          grind_1_total_cost += order.safe_filled * order.safe_price
          grind_1_buy_orders.append(order.id)
          if not grind_1_found:
            grind_1_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_1_found = True
        elif not grind_2_is_sell_found and order_tag == "g2":
          grind_2_sub_grind_count += 1
          grind_2_total_amount += order.safe_filled
          grind_2_total_cost += order.safe_filled * order.safe_price
          grind_2_buy_orders.append(order.id)
          if not grind_2_found:
            grind_2_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_2_found = True
        elif not grind_3_is_sell_found and order_tag == "g3":
          grind_3_sub_grind_count += 1
          grind_3_total_amount += order.safe_filled
          grind_3_total_cost += order.safe_filled * order.safe_price
          grind_3_buy_orders.append(order.id)
          if not grind_3_found:
            grind_3_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_3_found = True
        elif not grind_4_is_sell_found and order_tag == "g4":
          grind_4_sub_grind_count += 1
          grind_4_total_amount += order.safe_filled
          grind_4_total_cost += order.safe_filled * order.safe_price
          grind_4_buy_orders.append(order.id)
          if not grind_4_found:
            grind_4_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_4_found = True
        elif not grind_5_is_sell_found and order_tag == "g5":
          grind_5_sub_grind_count += 1
          grind_5_total_amount += order.safe_filled
          grind_5_total_cost += order.safe_filled * order.safe_price
          grind_5_buy_orders.append(order.id)
          if not grind_5_found:
            grind_5_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            grind_5_found = True
        elif not rebuy_is_sell_found and order_tag not in [
          "g1",
          "g2",
          "g3",
          "g4",
          "g5",
          "g6",
          "dl1",
          "dl2",
          "gd1",
          "gd2",
          "gd3",
          "gd4",
          "gd5",
          "gd6",
          "gm0",
          "gmd0",
        ]:
          rebuy_sub_grind_count += 1
          rebuy_total_amount += order.safe_filled
          rebuy_total_cost += order.safe_filled * order.safe_price
          rebuy_buy_orders.append(order.id)
          if not rebuy_found:
            rebuy_distance_ratio = (exit_rate - order.safe_price) / order.safe_price
            rebuy_found = True
      elif order.ft_order_side == "buy":
        if (
          order is filled_exits[-1]
          and (order.safe_remaining * exit_rate / (trade.leverage if self.is_futures_mode else 1.0)) > min_stake
        ):
          partial_sell = True
          break
        order_tag = ""
        if has_order_tags:
          if order.ft_order_tag is not None:
            sell_order_tag = order.ft_order_tag
            order_mode = sell_order_tag.split(" ", 1)
            if len(order_mode) > 0:
              order_tag = order_mode[0]
        if order_tag == "g1":
          grind_1_is_sell_found = True
        elif order_tag == "g2":
          grind_2_is_sell_found = True
        elif order_tag == "g3":
          grind_3_is_sell_found = True
        elif order_tag == "g4":
          grind_4_is_sell_found = True
        elif order_tag == "g5":
          grind_5_is_sell_found = True
        elif order_tag in ["d", "d1", "dd0", "ddl1", "ddl2", "dd1", "dd2", "dd3", "dd4", "dd5", "dd6"]:
          is_derisk = True
          if order_tag in ["d1"]:
            is_derisk_1 = True
          grind_1_is_sell_found = True
          grind_2_is_sell_found = True
          grind_3_is_sell_found = True
          grind_4_is_sell_found = True
          grind_5_is_sell_found = True
          rebuy_is_sell_found = True
        elif order_tag not in [
          "p",
          "g1",
          "g2",
          "g3",
          "g4",
          "g5",
          "g6",
          "dl1",
          "dl2",
          "gd1",
          "gd2",
          "gd3",
          "gd4",
          "gd5",
          "gd6",
          "gm0",
          "gmd0",
        ]:
          rebuy_is_sell_found = True
        if not is_derisk:
          start_amount = filled_orders[0].safe_filled
          current_amount = 0.0
          for order2 in filled_orders:
            if order2.ft_order_side == "sell":
              current_amount += order2.safe_filled
            elif order2.ft_order_side == "buy":
              current_amount -= order2.safe_filled
            if order2 is order:
              if current_amount < (start_amount * 0.95):
                is_derisk = True

        if (
          rebuy_is_sell_found
          and grind_1_is_sell_found
          and grind_2_is_sell_found
          and grind_3_is_sell_found
          and grind_4_is_sell_found
          and grind_5_is_sell_found
        ):
          break

    if is_derisk:
      return None, "", is_derisk
    if not has_order_tags and len(filled_exits) > 0:
      return None, "", is_derisk

    if rebuy_sub_grind_count > 0:
      rebuy_current_open_rate = rebuy_total_cost / rebuy_total_amount
      rebuy_current_grind_stake = rebuy_total_amount * exit_rate * (1 + trade.fee_close)
      rebuy_current_grind_stake_profit = rebuy_total_cost - rebuy_current_grind_stake
    if grind_1_sub_grind_count > 0:
      grind_1_current_open_rate = grind_1_total_cost / grind_1_total_amount
      grind_1_current_grind_stake = grind_1_total_amount * exit_rate * (1 + trade.fee_close)
      grind_1_current_grind_stake_profit = grind_1_total_cost - grind_1_current_grind_stake
    if grind_2_sub_grind_count > 0:
      grind_2_current_open_rate = grind_2_total_cost / grind_2_total_amount
      grind_2_current_grind_stake = grind_2_total_amount * exit_rate * (1 + trade.fee_close)
      grind_2_current_grind_stake_profit = grind_2_total_cost - grind_2_current_grind_stake
    if grind_3_sub_grind_count > 0:
      grind_3_current_open_rate = grind_3_total_cost / grind_3_total_amount
      grind_3_current_grind_stake = grind_3_total_amount * exit_rate * (1 + trade.fee_close)
      grind_3_current_grind_stake_profit = grind_3_total_cost - grind_3_current_grind_stake
    if grind_4_sub_grind_count > 0:
      grind_4_current_open_rate = grind_4_total_cost / grind_4_total_amount
      grind_4_current_grind_stake = grind_4_total_amount * exit_rate * (1 + trade.fee_close)
      grind_4_current_grind_stake_profit = grind_4_total_cost - grind_4_current_grind_stake
    if grind_5_sub_grind_count > 0:
      grind_5_current_open_rate = grind_5_total_cost / grind_5_total_amount
      grind_5_current_grind_stake = grind_5_total_amount * exit_rate * (1 + trade.fee_close)
      grind_5_current_grind_stake_profit = grind_5_total_cost - grind_5_current_grind_stake

    num_open_grinds = (
      grind_1_sub_grind_count
      + grind_2_sub_grind_count
      + grind_3_sub_grind_count
      + grind_4_sub_grind_count
      + grind_5_sub_grind_count
    )

    if partial_sell:
      order = filled_exits[-1]
      sell_amount = order.safe_remaining * exit_rate / trade.leverage
      if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
        sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        self.dp.send_msg(
          f"Exit (remaining) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {order.safe_remaining} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        order_tag = "p"
        if has_order_tags:
          if order.ft_order_tag is not None:
            order_tag = order.ft_order_tag
        return -ft_sell_amount, order_tag, is_derisk

    is_short_grind_buy = self.short_grind_buy(last_candle, previous_candle, slice_profit)

    if (not partial_sell) and (not rebuy_is_sell_found) and (rebuy_sub_grind_count < max_rebuy_sub_grinds):
      if (
        (0 <= rebuy_sub_grind_count < max_rebuy_sub_grinds)
        and (slice_profit_entry < regular_mode_rebuy_sub_thresholds[rebuy_sub_grind_count])
        and (
          (-rebuy_distance_ratio if (rebuy_sub_grind_count > 0) else profit_init_ratio)
          < (regular_mode_rebuy_sub_thresholds[rebuy_sub_grind_count])
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=12) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.06))
        and (
          (last_candle["close"] < (last_candle["close_min_12"] * 1.06))
          and (last_candle["close"] < (last_candle["close_min_24"] * 1.08))
          and (last_candle["close"] < (last_candle["close_min_48"] * 1.10))
          and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.12))
          and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.14))
          and (last_candle["close"] < (last_candle["low_min_6_1d"] * 1.16))
          and (last_candle["close"] < (last_candle["low_min_12_1d"] * 1.18))
          and (last_candle["btc_pct_close_min_72_5m"] > -0.03)
          and (last_candle["btc_pct_close_min_24_5m"] > -0.03)
        )
        and (
          is_short_grind_buy
          or (
            (last_candle["rsi_3"] < 70.0)
            and (last_candle["rsi_3_15m"] < 70.0)
            and (last_candle["rsi_3_1h"] < 70.0)
            and (last_candle["rsi_3_4h"] < 70.0)
            and (last_candle["rsi_14"] > 64.0)
            and (last_candle["zlma_50_dec_1h"] == True)
            and (last_candle["zlma_50_dec_4h"] == True)
          )
        )
      ):
        buy_amount = (
          slice_amount
          * regular_mode_rebuy_stakes[rebuy_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount > max_stake:
          buy_amount = max_stake
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None, "", is_derisk
        self.dp.send_msg(
          f"Rebuy (r) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        log.info(
          f"Rebuy (r) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        order_tag = "r"
        return buy_amount, order_tag, is_derisk


    if has_order_tags and (not partial_sell) and (grind_1_sub_grind_count < max_grind_1_sub_grinds):
      if (
        (
          (-grind_1_distance_ratio if (grind_1_sub_grind_count > 0) else profit_init_ratio)
          < (regular_mode_grind_1_sub_thresholds[grind_1_sub_grind_count])
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (last_candle["close"] < (last_candle["close_min_12"] * 1.08))
          and (last_candle["close"] < (last_candle["close_min_24"] * 1.10))
          and (last_candle["close"] < (last_candle["close_min_48"] * 1.12))
          and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.14))
          and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.16))
          and (last_candle["close"] < (last_candle["low_min_6_1d"] * 1.24))
          and (last_candle["close"] < (last_candle["low_min_12_1d"] * 1.30))
        )
        and (
          is_short_grind_buy
          or (
            (last_candle["rsi_3"] < 84.0)
            and (last_candle["rsi_3_15m"] < 84.0)
            and (last_candle["rsi_3_1h"] < 80.0)
            and (last_candle["rsi_3_4h"] < 80.0)
            and (last_candle["rsi_14"] > 64.0)
            and (last_candle["zlma_50_dec_1h"] == True)
            and (last_candle["zlma_50_dec_4h"] == True)
            and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
          )
        )
      ):
        buy_amount = (
          slice_amount
          * regular_mode_grind_1_stakes[grind_1_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None, "", is_derisk
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_1_sub_grind_count > 0:
          grind_profit = -(exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate
          grind_profit_stake = grind_1_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (g1) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (g1) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_1_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "g1"
        return buy_amount, order_tag, is_derisk

    if has_order_tags and grind_1_sub_grind_count > 0:
      grind_profit = -(exit_rate - grind_1_current_open_rate) / grind_1_current_open_rate
      if grind_profit > regular_mode_grind_1_profit_threshold:
        sell_amount = grind_1_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (g1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (g1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_1_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "g1"
          for grind_entry_id in grind_1_buy_orders:
            order_tag += " " + str(grind_entry_id)
          return -ft_sell_amount, order_tag, is_derisk


    if has_order_tags and (not partial_sell) and (grind_2_sub_grind_count < max_grind_2_sub_grinds):
      if (
        (
          (-grind_2_distance_ratio if (grind_2_sub_grind_count > 0) else profit_init_ratio)
          < (regular_mode_grind_2_sub_thresholds[grind_2_sub_grind_count])
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (last_candle["close"] < (last_candle["close_min_12"] * 1.08))
          and (last_candle["close"] < (last_candle["close_min_24"] * 1.10))
          and (last_candle["close"] < (last_candle["close_min_48"] * 1.12))
          and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.14))
          and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.16))
          and (last_candle["close"] < (last_candle["low_min_6_1d"] * 1.24))
          and (last_candle["close"] < (last_candle["low_min_12_1d"] * 1.30))
        )
        and (
          is_short_grind_buy
          or (
            (last_candle["rsi_3"] < 84.0)
            and (last_candle["rsi_3_15m"] < 84.0)
            and (last_candle["rsi_3_1h"] < 80.0)
            and (last_candle["rsi_3_4h"] < 80.0)
            and (last_candle["rsi_14"] > 64.0)
            and (last_candle["zlma_50_dec_1h"] == True)
            and (last_candle["zlma_50_dec_4h"] == True)
            and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
          )
        )
      ):
        buy_amount = (
          slice_amount
          * regular_mode_grind_2_stakes[grind_2_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None, "", is_derisk
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_2_sub_grind_count > 0:
          grind_profit = -(exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate
          grind_profit_stake = grind_2_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (g2) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_2_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (g2) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_2_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "g2"
        return buy_amount, order_tag, is_derisk

    if has_order_tags and grind_2_sub_grind_count > 0:
      grind_profit = -(exit_rate - grind_2_current_open_rate) / grind_2_current_open_rate
      if grind_profit > regular_mode_grind_2_profit_threshold:
        sell_amount = grind_2_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (g2) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (g2) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_2_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "g2"
          for grind_entry_id in grind_2_buy_orders:
            order_tag += " " + str(grind_entry_id)
          return -ft_sell_amount, order_tag, is_derisk


    if has_order_tags and (not partial_sell) and (grind_3_sub_grind_count < max_grind_3_sub_grinds):
      if (
        (
          (-grind_3_distance_ratio if (grind_3_sub_grind_count > 0) else profit_init_ratio)
          < (regular_mode_grind_3_sub_thresholds[grind_3_sub_grind_count])
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (
            (
              (last_candle["close"] < (last_candle["close_min_12"] * 1.08))
              and (last_candle["close"] < (last_candle["close_min_24"] * 1.10))
              and (last_candle["close"] < (last_candle["close_min_48"] * 1.12))
              and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.14))
              and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.16))
              and (last_candle["close"] < (last_candle["low_min_6_1d"] * 1.24))
              and (last_candle["close"] < (last_candle["low_min_12_1d"] * 1.30))
            )
            and (
              is_short_grind_buy
              or (
                (last_candle["rsi_3"] < 88.0)
                and (last_candle["rsi_3_15m"] < 84.0)
                and (last_candle["rsi_3_1h"] < 84.0)
                and (last_candle["rsi_3_4h"] < 84.0)
                and (last_candle["rsi_14"] > 64.0)


                and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
              )
            )
          )
          or (
            (slice_profit > 0.06)
            and (last_candle["rsi_3"] < 90.0)
            and (last_candle["rsi_3_15m"] < 90.0)

            and (last_candle["rsi_14"] > 64.0)
            and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
          )
        )
      ):
        buy_amount = (
          slice_amount
          * regular_mode_grind_3_stakes[grind_3_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None, "", is_derisk
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_3_sub_grind_count > 0:
          grind_profit = -(exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate
          grind_profit_stake = grind_3_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (g3) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_3_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (g3) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_3_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "g3"
        return buy_amount, order_tag, is_derisk

    if has_order_tags and grind_3_sub_grind_count > 0:
      grind_profit = -(exit_rate - grind_3_current_open_rate) / grind_3_current_open_rate
      if grind_profit > regular_mode_grind_3_profit_threshold:
        sell_amount = grind_3_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (g3) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (g3) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_3_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "g3"
          for grind_entry_id in grind_3_buy_orders:
            order_tag += " " + str(grind_entry_id)
          return -ft_sell_amount, order_tag, is_derisk


    if has_order_tags and (not partial_sell) and (grind_4_sub_grind_count < max_grind_4_sub_grinds):
      if (
        (
          (-grind_4_distance_ratio if (grind_4_sub_grind_count > 0) else profit_init_ratio)
          < (regular_mode_grind_4_sub_thresholds[grind_4_sub_grind_count])
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (
            (
              (last_candle["close"] < (last_candle["close_min_12"] * 1.08))
              and (last_candle["close"] < (last_candle["close_min_24"] * 1.10))
              and (last_candle["close"] < (last_candle["close_min_48"] * 1.12))
              and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.14))
              and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.16))
              and (last_candle["close"] < (last_candle["low_min_6_1d"] * 1.24))
              and (last_candle["close"] < (last_candle["low_min_12_1d"] * 1.30))
            )
            and (
              is_short_grind_buy
              or (
                (last_candle["rsi_3"] < 88.0)
                and (last_candle["rsi_3_15m"] < 84.0)
                and (last_candle["rsi_3_1h"] < 84.0)
                and (last_candle["rsi_3_4h"] < 84.0)
                and (last_candle["rsi_14"] > 64.0)


                and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
              )
            )
          )
          or (
            (slice_profit > 0.06)
            and (last_candle["rsi_3"] < 90.0)
            and (last_candle["rsi_3_15m"] < 90.0)

            and (last_candle["rsi_14"] > 64.0)
            and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
          )
        )
      ):
        buy_amount = (
          slice_amount
          * regular_mode_grind_4_stakes[grind_4_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None, "", is_derisk
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_4_sub_grind_count > 0:
          grind_profit = -(exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate
          grind_profit_stake = grind_4_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (g4) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_4_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (g4) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_4_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "g4"
        return buy_amount, order_tag, is_derisk

    if has_order_tags and grind_4_sub_grind_count > 0:
      grind_profit = -(exit_rate - grind_4_current_open_rate) / grind_4_current_open_rate
      if grind_profit > regular_mode_grind_4_profit_threshold:
        sell_amount = grind_4_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (g4) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (g4) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_4_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "g4"
          for grind_entry_id in grind_4_buy_orders:
            order_tag += " " + str(grind_entry_id)
          return -ft_sell_amount, order_tag, is_derisk


    if has_order_tags and (not partial_sell) and (grind_5_sub_grind_count < max_grind_5_sub_grinds):
      if (
        (
          (-grind_5_distance_ratio if (grind_5_sub_grind_count > 0) else profit_init_ratio)
          < (regular_mode_grind_5_sub_thresholds[grind_5_sub_grind_count])
        )
        and (current_time - timedelta(minutes=10) > filled_entries[-1].order_filled_utc)
        and ((current_time - timedelta(hours=2) > filled_orders[-1].order_filled_utc) or (slice_profit > 0.02))
        and (
          (num_open_grinds == 0)
          or (current_time - timedelta(hours=6) > filled_orders[-1].order_filled_utc)
          or (slice_profit > 0.06)
        )

        and (
          (
            (
              (last_candle["close"] < (last_candle["close_min_12"] * 1.08))
              and (last_candle["close"] < (last_candle["close_min_24"] * 1.10))
              and (last_candle["close"] < (last_candle["close_min_48"] * 1.12))
              and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.14))
              and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.16))
              and (last_candle["close"] < (last_candle["low_min_6_1d"] * 1.24))
              and (last_candle["close"] < (last_candle["low_min_12_1d"] * 1.30))
            )
            and (
              is_short_grind_buy
              or (
                (last_candle["rsi_3"] < 88.0)
                and (last_candle["rsi_3_15m"] < 84.0)
                and (last_candle["rsi_3_1h"] < 84.0)
                and (last_candle["rsi_3_4h"] < 84.0)
                and (last_candle["rsi_14"] > 64.0)


                and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
              )
            )
          )
          or (
            (slice_profit > 0.06)
            and (last_candle["rsi_3"] < 90.0)
            and (last_candle["rsi_3_15m"] < 90.0)

            and (last_candle["rsi_14"] > 64.0)
            and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
          )
        )
      ):
        buy_amount = (
          slice_amount
          * regular_mode_grind_5_stakes[grind_5_sub_grind_count]
          / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        if buy_amount > max_stake:
          return None, "", is_derisk
        grind_profit = 0.0
        grind_profit_stake = 0.0
        if grind_5_sub_grind_count > 0:
          grind_profit = -(exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate
          grind_profit_stake = grind_5_current_grind_stake_profit
        self.dp.send_msg(
          f"Grinding entry (g5) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_5_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        log.info(
          f"Grinding entry (g5) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_5_current_grind_stake_profit} {self.config['stake_currency']})"
        )
        order_tag = "g5"
        return buy_amount, order_tag, is_derisk

    if has_order_tags and grind_5_sub_grind_count > 0:
      grind_profit = -(exit_rate - grind_5_current_open_rate) / grind_5_current_open_rate
      if grind_profit > regular_mode_grind_5_profit_threshold:
        sell_amount = grind_5_total_amount * exit_rate / trade.leverage
        if ((current_stake_amount / trade.leverage) - sell_amount) < (min_stake * 1.55):
          sell_amount = (trade.amount * exit_rate / trade.leverage) - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Grinding exit (g5) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          log.info(
            f"Grinding exit (g5) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Coin amount: {grind_5_total_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}% | Grind profit: {(grind_profit * 100.0):.2f}% ({grind_profit * sell_amount * trade.leverage} {self.config['stake_currency']})"
          )
          order_tag = "g5"
          for grind_entry_id in grind_5_buy_orders:
            order_tag += " " + str(grind_entry_id)
          return -ft_sell_amount, order_tag, is_derisk

    if profit_stake < (
      slice_amount
      * (
        (self.regular_mode_derisk_futures if self.is_futures_mode else self.regular_mode_derisk_spot)
        if (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 3, 19) or is_backtest)
        else (self.regular_mode_derisk_futures_old if self.is_futures_mode else self.regular_mode_derisk_spot_old)
      )
      / (trade.leverage if self.is_futures_mode else 1.0)
    ):
      sell_amount = trade.amount * exit_rate / trade.leverage - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        self.dp.send_msg(
          f"De-risk [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        log.info(
          f"De-risk [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        return -ft_sell_amount, "d", is_derisk

    if (
      has_order_tags
      and not is_derisk_1
      and (trade.open_date_utc.replace(tzinfo=None) >= datetime(2024, 4, 5) or is_backtest)
      and profit_stake
      < (
        slice_amount
        * (self.regular_mode_derisk_1_futures if self.is_futures_mode else self.regular_mode_derisk_1_spot)
        / (trade.leverage if self.is_futures_mode else 1.0)
      )
    ):
      sell_amount = trade.amount * exit_rate / trade.leverage - (min_stake * 1.55)
      ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
      if sell_amount > min_stake and ft_sell_amount > min_stake:
        grind_profit = 0.0
        self.dp.send_msg(
          f"De-risk (d1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        log.info(
          f"De-risk (d1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        return -ft_sell_amount, "d1", is_derisk

    return None, "", is_derisk


  def short_rebuy_adjust_trade_position(
    self,
    trade: Trade,
    enter_tags,
    current_time: datetime,
    current_rate: float,
    current_profit: float,
    min_stake: Optional[float],
    max_stake: float,
    current_entry_rate: float,
    current_exit_rate: float,
    current_entry_profit: float,
    current_exit_profit: float,
    **kwargs,
  ) -> Optional[float]:
    df, _ = self.dp.get_analyzed_dataframe(trade.pair, self.timeframe)
    if len(df) < 2:
      return None
    last_candle = df.iloc[-1].squeeze()
    previous_candle = df.iloc[-2].squeeze()

    filled_orders = trade.select_filled_orders()
    filled_entries = trade.select_filled_orders(trade.entry_side)
    filled_exits = trade.select_filled_orders(trade.exit_side)
    count_of_entries = trade.nr_of_successful_entries
    count_of_exits = trade.nr_of_successful_exits

    if count_of_entries == 0:
      return None

    has_order_tags = False
    if hasattr(filled_orders[0], "ft_order_tag"):
      has_order_tags = True

    if count_of_exits > 0:
      return self.short_grind_adjust_trade_position(
        trade,
        enter_tags,
        current_time,
        current_rate,
        current_profit,
        min_stake,
        max_stake,
        current_entry_rate,
        current_exit_rate,
        current_entry_profit,
        current_exit_profit,
      )

    exit_rate = current_rate
    if self.dp.runmode.value in ("live", "dry_run"):
      ticker = self.dp.ticker(trade.pair)
      if ("bid" in ticker) and ("ask" in ticker):
        if trade.is_short:
          if self.config["exit_pricing"]["price_side"] in ["ask", "other"]:
            if ticker["ask"] is not None:
              exit_rate = ticker["ask"]
        else:
          if self.config["exit_pricing"]["price_side"] in ["bid", "other"]:
            if ticker["bid"] is not None:
              exit_rate = ticker["bid"]

    profit_stake, profit_ratio, profit_current_stake_ratio, profit_init_ratio = self.calc_total_profit(
      trade, filled_entries, filled_exits, exit_rate
    )

    slice_amount = filled_entries[0].cost
    slice_profit = (exit_rate - filled_orders[-1].safe_price) / filled_orders[-1].safe_price
    slice_profit_entry = (exit_rate - filled_entries[-1].safe_price) / filled_entries[-1].safe_price
    slice_profit_exit = (
      ((exit_rate - filled_exits[-1].safe_price) / filled_exits[-1].safe_price) if count_of_exits > 0 else 0.0
    )

    current_stake_amount = trade.amount * current_rate

    is_rebuy = False

    rebuy_mode_stakes = self.rebuy_mode_stakes_futures if self.is_futures_mode else self.rebuy_mode_stakes_spot
    max_sub_grinds = len(rebuy_mode_stakes)
    rebuy_mode_sub_thresholds = (
      self.rebuy_mode_thresholds_futures if self.is_futures_mode else self.rebuy_mode_thresholds_spot
    )
    partial_sell = False
    sub_grind_count = 0
    total_amount = 0.0
    total_cost = 0.0
    current_open_rate = 0.0
    current_grind_stake = 0.0
    current_grind_stake_profit = 0.0
    for order in reversed(filled_orders):
      if (order.ft_order_side == "buy") and (order is not filled_orders[0]):
        sub_grind_count += 1
        total_amount += order.safe_filled
        total_cost += order.safe_filled * order.safe_price
      elif order.ft_order_side == "sell":
        if (order.safe_remaining * exit_rate / (trade.leverage if self.is_futures_mode else 1.0)) > min_stake:
          partial_sell = True
        break
    if sub_grind_count > 0:
      current_open_rate = total_cost / total_amount
      current_grind_stake = total_amount * exit_rate * (1 - trade.fee_close)
      current_grind_stake_profit = current_grind_stake - total_cost

    if (not partial_sell) and (sub_grind_count < max_sub_grinds):
      if (
        ((0 <= sub_grind_count < max_sub_grinds) and (slice_profit_entry < rebuy_mode_sub_thresholds[sub_grind_count]))
        and (
          (last_candle["close"] < (last_candle["close_min_12"] * 1.06))
          and (last_candle["close"] < (last_candle["close_min_24"] * 1.08))
          and (last_candle["close"] < (last_candle["close_min_48"] * 1.10))
          and (last_candle["close"] < (last_candle["low_min_24_1h"] * 1.12))
          and (last_candle["close"] < (last_candle["low_min_48_1h"] * 1.14))
          and (last_candle["btc_pct_close_min_72_5m"] > 0.03)
          and (last_candle["btc_pct_close_min_24_5m"] > 0.03)
        )
        and (
          (last_candle["rsi_3"] < 90.0)
          and (last_candle["rsi_3_15m"] < 90.0)
          and (last_candle["rsi_3_1h"] < 90.0)
          and (last_candle["rsi_3_4h"] < 90.0)
          and (last_candle["rsi_14"] > 64.0)
          and (last_candle["close"] > (last_candle["ema_26"] * 1.012))
        )
      ):
        buy_amount = (
          slice_amount * rebuy_mode_stakes[sub_grind_count] / (trade.leverage if self.is_futures_mode else 1.0)
        )
        if buy_amount > max_stake:
          buy_amount = max_stake
        if buy_amount < (min_stake * 1.5):
          buy_amount = min_stake * 1.5
        self.dp.send_msg(
          f"Rebuy (r) [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        log.info(
          f"Rebuy (r) [{current_time}] [{trade.pair}] | Rate: {current_rate} | Stake amount: {buy_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
        )
        if has_order_tags:
          return buy_amount, "r"
        else:
          return buy_amount

      if profit_stake < (
        slice_amount
        * (self.rebuy_mode_derisk_futures if self.is_futures_mode else self.rebuy_mode_derisk_spot)
        / (trade.leverage if self.is_futures_mode else 1.0)
      ):
        sell_amount = trade.amount * exit_rate / trade.leverage - (min_stake * 1.55)
        ft_sell_amount = sell_amount * trade.leverage * (trade.stake_amount / trade.amount) / exit_rate
        if sell_amount > min_stake and ft_sell_amount > min_stake:
          self.dp.send_msg(
            f"Rebuy de-risk (d1) [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
          )
          log.info(
            f"Rebuy de-risk (d1) [{current_time}] [{trade.pair}] | Rate: {exit_rate} | Stake amount: {sell_amount} | Profit (stake): {profit_stake} | Profit: {(profit_ratio * 100.0):.2f}%"
          )
          if has_order_tags:
            return -ft_sell_amount, "d1"
          else:
            return -ft_sell_amount

    return None




  @property
  def protections(self):
      return [
          {
              "method": "StoplossGuard",
              "lookback_period_candles": 12,
              "trade_limit": 3,
              "stop_duration_candles": 6,
              "required_profit": 0.0,
              "only_per_pair": False,
              "only_per_side": False
          }
      ]





def is_support(row_data) -> bool:
  conditions = []
  for row in range(len(row_data) - 1):
    if row < len(row_data) // 2:
      conditions.append(row_data[row] > row_data[row + 1])
    else:
      conditions.append(row_data[row] < row_data[row + 1])
  result = reduce(lambda x, y: x & y, conditions)
  return result


def is_resistance(row_data) -> bool:
  conditions = []
  for row in range(len(row_data) - 1):
    if row < len(row_data) // 2:
      conditions.append(row_data[row] < row_data[row + 1])
    else:
      conditions.append(row_data[row] > row_data[row + 1])
  result = reduce(lambda x, y: x & y, conditions)
  return result


def ewo(df, ema1_length=5, ema2_length=35):
  ema1 = ta.EMA(df, timeperiod=ema1_length)
  ema2 = ta.EMA(df, timeperiod=ema2_length)
  emadiff = (ema1 - ema2) / df["close"] * 100.0
  return emadiff


def pivot_points(df: DataFrame, mode="fibonacci") -> Series:
  if mode == "simple":
    hlc3_pivot = (df["high"] + df["low"] + df["close"]).shift(1) / 3
    res1 = hlc3_pivot * 2 - df["low"].shift(1)
    sup1 = hlc3_pivot * 2 - df["high"].shift(1)
    res2 = hlc3_pivot + (df["high"] - df["low"]).shift()
    sup2 = hlc3_pivot - (df["high"] - df["low"]).shift()
    res3 = hlc3_pivot * 2 + (df["high"] - 2 * df["low"]).shift()
    sup3 = hlc3_pivot * 2 - (2 * df["high"] - df["low"]).shift()
    return hlc3_pivot, res1, res2, res3, sup1, sup2, sup3
  elif mode == "fibonacci":
    hlc3_pivot = (df["high"] + df["low"] + df["close"]).shift(1) / 3
    hl_range = (df["high"] - df["low"]).shift(1)
    res1 = hlc3_pivot + 0.382 * hl_range
    sup1 = hlc3_pivot - 0.382 * hl_range
    res2 = hlc3_pivot + 0.618 * hl_range
    sup2 = hlc3_pivot - 0.618 * hl_range
    res3 = hlc3_pivot + 1 * hl_range
    sup3 = hlc3_pivot - 1 * hl_range
    return hlc3_pivot, res1, res2, res3, sup1, sup2, sup3
  elif mode == "DeMark":
    demark_pivot_lt = df["low"] * 2 + df["high"] + df["close"]
    demark_pivot_eq = df["close"] * 2 + df["low"] + df["high"]
    demark_pivot_gt = df["high"] * 2 + df["low"] + df["close"]
    demark_pivot = np.where(
      (df["close"] < df["open"]),
      demark_pivot_lt,
      np.where((df["close"] > df["open"]), demark_pivot_gt, demark_pivot_eq),
    )
    dm_pivot = demark_pivot / 4
    dm_res = demark_pivot / 2 - df["low"]
    dm_sup = demark_pivot / 2 - df["high"]
    return dm_pivot, dm_res, dm_sup


def heikin_ashi(df, smooth_inputs=False, smooth_outputs=False, length=10):
  df = df[["open", "close", "high", "low"]].copy().fillna(0)
  if smooth_inputs:
    df["open_s"] = ta.EMA(df["open"], timeframe=length)
    df["high_s"] = ta.EMA(df["high"], timeframe=length)
    df["low_s"] = ta.EMA(df["low"], timeframe=length)
    df["close_s"] = ta.EMA(df["close"], timeframe=length)

    open_ha = (df["open_s"].shift(1) + df["close_s"].shift(1)) / 2
    high_ha = df.loc[:, ["high_s", "open_s", "close_s"]].max(axis=1)
    low_ha = df.loc[:, ["low_s", "open_s", "close_s"]].min(axis=1)
    close_ha = (df["open_s"] + df["high_s"] + df["low_s"] + df["close_s"]) / 4
  else:
    open_ha = (df["open"].shift(1) + df["close"].shift(1)) / 2
    high_ha = df.loc[:, ["high", "open", "close"]].max(axis=1)
    low_ha = df.loc[:, ["low", "open", "close"]].min(axis=1)
    close_ha = (df["open"] + df["high"] + df["low"] + df["close"]) / 4

  open_ha = open_ha.fillna(0)
  high_ha = high_ha.fillna(0)
  low_ha = low_ha.fillna(0)
  close_ha = close_ha.fillna(0)

  if smooth_outputs:
    open_sha = ta.EMA(open_ha, timeframe=length)
    high_sha = ta.EMA(high_ha, timeframe=length)
    low_sha = ta.EMA(low_ha, timeframe=length)
    close_sha = ta.EMA(close_ha, timeframe=length)

    return open_sha, close_sha, low_sha
  else:
    return open_ha, close_ha, low_ha


def range_percent_change(self, df: DataFrame, method, length: int) -> float:
  """
  Rolling Percentage Change Maximum across interval.

  :param df: DataFrame The original OHLC df
  :param method: High to Low / Open to Close
  :param length: int The length to look back
  """
  if method == "HL":
    return (df["high"].rolling(length).max() - df["low"].rolling(length).min()) / df["low"].rolling(length).min()
  elif method == "OC":
    return (df["open"].rolling(length).max() - df["close"].rolling(length).min()) / df["close"].rolling(length).min()
  else:
    raise ValueError(f"Method {method} not defined!")


def top_percent_change(self, df: DataFrame, length: int) -> float:
  """
  Percentage change of the current close from the range maximum Open price

  :param df: DataFrame The original OHLC df
  :param length: int The length to look back
  """
  if length == 0:
    return (df["open"] - df["close"]) / df["close"]
  else:
    return (df["open"].rolling(length).max() - df["close"]) / df["close"]

def resample(indicator):

    return indicator

def two_bands_check(dataframe):
    check = (dataframe[resample("low")] < dataframe[resample("kc_lowerband")]) & (dataframe[resample("high")] > dataframe[resample("kc_upperband")])
    return ~check

def ema_cross_check(dataframe):
    dataframe["ema20_50_cross"] = qtpylib.crossed_below(dataframe[resample("ema20")], dataframe[resample("ema50")])
    dataframe["ema20_200_cross"] = qtpylib.crossed_below(dataframe[resample("ema20")], dataframe[resample("ema200")])
    dataframe["ema50_200_cross"] = qtpylib.crossed_below(dataframe[resample("ema50")], dataframe[resample("ema200")])
    return ~(dataframe["ema20_50_cross"] | dataframe["ema20_200_cross"] | dataframe["ema50_200_cross"])

def green_candle(dataframe):
    return dataframe[resample("open")] < dataframe[resample("close")]

def keltner_middleband_check(dataframe):
    return (dataframe[resample("low")] < dataframe[resample("kc_middleband")]) & (dataframe[resample("high")] > dataframe[resample("kc_middleband")])

def keltner_lowerband_check(dataframe):
    return (dataframe[resample("low")] < dataframe[resample("kc_lowerband")]) & (dataframe[resample("high")] > dataframe[resample("kc_lowerband")])

def bollinger_lowerband_check(dataframe):
    return (dataframe[resample("low")] < dataframe[resample("bollinger_lowerband")]) & (dataframe[resample("high")] > dataframe[resample("bollinger_lowerband")])

def bollinger_keltner_check(dataframe):
    return (dataframe[resample("bollinger_lowerband")] < dataframe[resample("kc_lowerband")]) & (dataframe[resample("bollinger_upperband")] > dataframe[resample("kc_upperband")])

def ema_check(dataframe):
    check = (dataframe[resample("ema9")] < dataframe[resample("ema20")]) & (dataframe[resample("ema20")] < dataframe[resample("ema50")]) & (dataframe[resample("ema50")] < dataframe[resample("ema200")])
    return ~check

def initialize_divergences_lists(dataframe: DataFrame):
    dataframe["total_bullish_divergences"] = np.empty(len(dataframe["close"])) * np.nan
    dataframe["total_bullish_divergences_count"] = np.empty(len(dataframe["close"])) * np.nan
    dataframe["total_bullish_divergences_count"] = [0 if x != x else x for x in dataframe["total_bullish_divergences_count"]]
    dataframe["total_bullish_divergences_names"] = np.empty(len(dataframe["close"])) * np.nan
    dataframe["total_bullish_divergences_names"] = ["" if x != x else x for x in dataframe["total_bullish_divergences_names"]]
    dataframe["total_bearish_divergences"] = np.empty(len(dataframe["close"])) * np.nan
    dataframe["total_bearish_divergences_count"] = np.empty(len(dataframe["close"])) * np.nan
    dataframe["total_bearish_divergences_count"] = [0 if x != x else x for x in dataframe["total_bearish_divergences_count"]]
    dataframe["total_bearish_divergences_names"] = np.empty(len(dataframe["close"])) * np.nan
    dataframe["total_bearish_divergences_names"] = ["" if x != x else x for x in dataframe["total_bearish_divergences_names"]]

def add_divergences(dataframe: DataFrame, indicator: str):
    bearish_divergences, bearish_lines, bullish_divergences, bullish_lines = divergence_finder_dataframe(dataframe, indicator)
    dataframe["bearish_divergence_" + indicator + "_occurence"] = bearish_divergences


    dataframe["bullish_divergence_" + indicator + "_occurence"] = bullish_divergences



def divergence_finder_dataframe(dataframe: DataFrame, indicator_source: str) -> Tuple[pd.Series, pd.Series]:
    bearish_lines = [np.empty(len(dataframe["close"])) * np.nan]
    bearish_divergences = np.empty(len(dataframe["close"])) * np.nan
    bullish_lines = [np.empty(len(dataframe["close"])) * np.nan]
    bullish_divergences = np.empty(len(dataframe["close"])) * np.nan
    low_iterator = []
    high_iterator = []
    for index, row in enumerate(dataframe.itertuples(index=True, name="Pandas")):
        if np.isnan(row.pivot_lows):
            low_iterator.append(0 if len(low_iterator) == 0 else low_iterator[-1])
        else:
            low_iterator.append(index)
        if np.isnan(row.pivot_highs):
            high_iterator.append(0 if len(high_iterator) == 0 else high_iterator[-1])
        else:
            high_iterator.append(index)
    for index, row in enumerate(dataframe.itertuples(index=True, name="Pandas")):
        bearish_occurence = bearish_divergence_finder(dataframe, dataframe[indicator_source], high_iterator, index)
        if bearish_occurence != None:
            prev_pivot, current_pivot = bearish_occurence




            bearish_prev_pivot = dataframe.loc[prev_pivot, "close"]
            bearish_current_pivot = dataframe.loc[current_pivot, "close"]
            bearish_ind_prev_pivot = dataframe.loc[prev_pivot, indicator_source]
            bearish_ind_current_pivot = dataframe.loc[current_pivot, indicator_source]
            length = current_pivot - prev_pivot
            bearish_lines_index = 0
            can_exist = True
            while True:
                can_draw = True
                if bearish_lines_index <= len(bearish_lines):
                    bearish_lines.append(np.empty(len(dataframe["close"])) * np.nan)
                actual_bearish_lines = bearish_lines[bearish_lines_index]
                for i in range(length + 1):
                    point = bearish_prev_pivot + (bearish_current_pivot - bearish_prev_pivot) * i / length
                    indicator_point = bearish_ind_prev_pivot + (bearish_ind_current_pivot - bearish_ind_prev_pivot) * i / length
                    if i != 0 and i != length:

                        if point <= dataframe.loc[prev_pivot + i, "close"] or indicator_point <= dataframe.loc[prev_pivot + i, indicator_source]:
                            can_exist = False
                    if not np.isnan(actual_bearish_lines[prev_pivot + i]):
                        can_draw = False
                if not can_exist:
                    break
                if can_draw:
                    for i in range(length + 1):
                        actual_bearish_lines[prev_pivot + i] = bearish_prev_pivot + (bearish_current_pivot - bearish_prev_pivot) * i / length
                    break
                bearish_lines_index = bearish_lines_index + 1
            if can_exist:
                bearish_divergences[index] = row.close

                dataframe.loc[index, "total_bearish_divergences"] = row.close
                if index > 30:


                    dataframe.loc[index - 30, "total_bearish_divergences_count"] = dataframe.loc[index - 30, "total_bearish_divergences_count"] + 1
                    dataframe.loc[index - 30, "total_bearish_divergences_names"] = dataframe.loc[index - 30, "total_bearish_divergences_names"] + indicator_source.upper() + "<br>"
        bullish_occurence = bullish_divergence_finder(dataframe, dataframe[indicator_source], low_iterator, index)
        if bullish_occurence != None:
            prev_pivot, current_pivot = bullish_occurence




            bullish_prev_pivot = dataframe.loc[prev_pivot, "close"]
            bullish_current_pivot = dataframe.loc[current_pivot, "close"]
            bullish_ind_prev_pivot = dataframe.loc[prev_pivot, indicator_source]
            bullish_ind_current_pivot = dataframe.loc[current_pivot, indicator_source]
            length = current_pivot - prev_pivot
            bullish_lines_index = 0
            can_exist = True
            while True:
                can_draw = True
                if bullish_lines_index <= len(bullish_lines):
                    bullish_lines.append(np.empty(len(dataframe["close"])) * np.nan)
                actual_bullish_lines = bullish_lines[bullish_lines_index]
                for i in range(length + 1):
                    point = bullish_prev_pivot + (bullish_current_pivot - bullish_prev_pivot) * i / length
                    indicator_point = bullish_ind_prev_pivot + (bullish_ind_current_pivot - bullish_ind_prev_pivot) * i / length
                    if i != 0 and i != length:

                        if point >= dataframe.loc[prev_pivot + i, "close"] or indicator_point >= dataframe.loc[prev_pivot + i, indicator_source]:
                            can_exist = False
                    if not np.isnan(actual_bullish_lines[prev_pivot + i]):
                        can_draw = False
                if not can_exist:
                    break
                if can_draw:
                    for i in range(length + 1):
                        actual_bullish_lines[prev_pivot + i] = bullish_prev_pivot + (bullish_current_pivot - bullish_prev_pivot) * i / length
                    break
                bullish_lines_index = bullish_lines_index + 1
            if can_exist:
                bullish_divergences[index] = row.close

                dataframe.loc[index, "total_bullish_divergences"] = row.close
                if index > 30:


                    dataframe.loc[index - 30, "total_bullish_divergences_count"] = dataframe.loc[index - 30, "total_bullish_divergences_count"] + 1
                    dataframe.loc[index - 30, "total_bullish_divergences_names"] = dataframe.loc[index - 30, "total_bullish_divergences_names"] + indicator_source.upper() + "<br>"
    return (bearish_divergences, bearish_lines, bullish_divergences, bullish_lines)

def bearish_divergence_finder(dataframe, indicator, high_iterator, index):
    if high_iterator[index] == index:
        current_pivot = high_iterator[index]
        occurences = list(dict.fromkeys(high_iterator))
        current_index = occurences.index(high_iterator[index])
        for i in range(current_index - 1, current_index - 6, -1):
            prev_pivot = occurences[i]
            if np.isnan(prev_pivot):
                return

            if dataframe.loc[current_pivot, "pivot_highs"] < dataframe.loc[prev_pivot, "pivot_highs"] and indicator[current_pivot] > indicator[prev_pivot] or (dataframe.loc[current_pivot, "pivot_highs"] > dataframe.loc[prev_pivot, "pivot_highs"] and indicator[current_pivot] < indicator[prev_pivot]):
                return (prev_pivot, current_pivot)
    return None

def bullish_divergence_finder(dataframe, indicator, low_iterator, index):
    if low_iterator[index] == index:
        current_pivot = low_iterator[index]
        occurences = list(dict.fromkeys(low_iterator))
        current_index = occurences.index(low_iterator[index])
        for i in range(current_index - 1, current_index - 6, -1):
            prev_pivot = occurences[i]
            if np.isnan(prev_pivot):
                return

            if dataframe.loc[current_pivot, "pivot_lows"] < dataframe.loc[prev_pivot, "pivot_lows"] and indicator[current_pivot] > indicator[prev_pivot] or (dataframe.loc[current_pivot, "pivot_lows"] > dataframe.loc[prev_pivot, "pivot_lows"] and indicator[current_pivot] < indicator[prev_pivot]):
                return (prev_pivot, current_pivot)
    return None

class PivotSource(Enum):
    HighLow = 0
    Close = 1

def pivot_points_S(dataframe: DataFrame, window: int=5, pivot_source: PivotSource=PivotSource.Close) -> DataFrame:
    high_source = None
    low_source = None
    if pivot_source == PivotSource.Close:
        high_source = "close"
        low_source = "close"
    elif pivot_source == PivotSource.HighLow:
        high_source = "high"
        low_source = "low"
    pivot_points_lows = np.empty(len(dataframe["close"])) * np.nan
    pivot_points_highs = np.empty(len(dataframe["close"])) * np.nan
    last_values = deque()

    for index, row in enumerate(dataframe.itertuples(index=True, name="Pandas")):
        last_values.append(row)
        if len(last_values) >= window * 2 + 1:
            current_value = last_values[window]
            is_greater = True
            is_less = True
            for window_index in range(0, window):
                left = last_values[window_index]
                right = last_values[2 * window - window_index]
                local_is_greater, local_is_less = check_if_pivot_is_greater_or_less(current_value, high_source, low_source, left, right)
                is_greater &= local_is_greater
                is_less &= local_is_less
            if is_greater:
                pivot_points_highs[index - window] = getattr(current_value, high_source)
            if is_less:
                pivot_points_lows[index - window] = getattr(current_value, low_source)
            last_values.popleft()

    if len(last_values) >= window + 2:
        current_value = last_values[-2]
        is_greater = True
        is_less = True
        for window_index in range(0, window):
            left = last_values[-2 - window_index - 1]
            right = last_values[-1]
            local_is_greater, local_is_less = check_if_pivot_is_greater_or_less(current_value, high_source, low_source, left, right)
            is_greater &= local_is_greater
            is_less &= local_is_less
        if is_greater:
            pivot_points_highs[index - 1] = getattr(current_value, high_source)
        if is_less:
            pivot_points_lows[index - 1] = getattr(current_value, low_source)
    return pd.DataFrame(index=dataframe.index, data={"pivot_lows": pivot_points_lows, "pivot_highs": pivot_points_highs})

def check_if_pivot_is_greater_or_less(current_value, high_source: str, low_source: str, left, right) -> Tuple[bool, bool]:
    is_greater = True
    is_less = True
    if getattr(current_value, high_source) < getattr(left, high_source) or getattr(current_value, high_source) < getattr(right, high_source):
        is_greater = False
    if getattr(current_value, low_source) > getattr(left, low_source) or getattr(current_value, low_source) > getattr(right, low_source):
        is_less = False
    return (is_greater, is_less)

def emaKeltner(df):
    keltner = {}
    atr = qtpylib.atr(df, window=10)
    ema20 = ta.EMA(df, timeperiod=20)
    keltner["upper"] = ema20 + atr
    keltner["mid"] = ema20
    keltner["lower"] = ema20 - atr
    return keltner
  
def chaikin_money_flow(df, n=20, fillna=False) -> Series:
  """Chaikin Money Flow (CMF)
  It measures the amount of Money Flow Volume over a specific period.
  http://stockcharts.com/school/doku.php?id=chart_school:technical_indicators:chaikin_money_flow_cmf
  Args:
      df(pandas.Dataframe): df containing ohlcv
      n(int): n period.
      fillna(bool): if True, fill nan values.
  Returns:
      pandas.Series: New feature generated.
  """
  mfv = ((df["close"] - df["low"]) - (df["high"] - df["close"])) / (df["high"] - df["low"])
  mfv = mfv.fillna(0.0)  # float division by zero
  mfv *= df["volume"]
  cmf = mfv.rolling(n, min_periods=0).sum() / df["volume"].rolling(n, min_periods=0).sum()
  if fillna:
    cmf = cmf.replace([np.inf, -np.inf], np.nan).fillna(0)
  return Series(cmf, name="cmf")





class Cache:
  def __init__(self, path):
    self.path = path
    self.data = {}
    self._mtime = None
    self._previous_data = {}
    try:
      self.load()
    except FileNotFoundError:
      pass

  @staticmethod
  def rapidjson_load_kwargs():
    return {"number_mode": rapidjson.NM_NATIVE, "parse_mode": rapidjson.PM_COMMENTS | rapidjson.PM_TRAILING_COMMAS}

  @staticmethod
  def rapidjson_dump_kwargs():
    return {"number_mode": rapidjson.NM_NATIVE}

  def load(self):
    if not self._mtime or self.path.stat().st_mtime_ns != self._mtime:
      self._load()

  def save(self):
    if self.data != self._previous_data:
      self._save()

  def process_loaded_data(self, data):
    return data

  def _load(self):

    with self.path.open("r") as rfh:
      try:
        data = rapidjson.load(rfh, **self.rapidjson_load_kwargs())
      except rapidjson.JSONDecodeError as exc:
        log.error("Failed to load JSON from %s: %s", self.path, exc)
      else:
        self.data = self.process_loaded_data(data)
        self._previous_data = copy.deepcopy(self.data)
        self._mtime = self.path.stat().st_mtime_ns

  def _save(self):

    rapidjson.dump(self.data, self.path.open("w"), **self.rapidjson_dump_kwargs())
    self._mtime = self.path.stat().st_mtime
    self._previous_data = copy.deepcopy(self.data)


class HoldsCache(Cache):
  @staticmethod
  def rapidjson_load_kwargs():
    return {
      "number_mode": rapidjson.NM_NATIVE,
      "parse_mode": rapidjson.PM_COMMENTS | rapidjson.PM_TRAILING_COMMAS,
      "object_hook": HoldsCache._object_hook,
    }

  @staticmethod
  def rapidjson_dump_kwargs():
    return {
      "number_mode": rapidjson.NM_NATIVE,
      "mapping_mode": rapidjson.MM_COERCE_KEYS_TO_STRINGS,
    }

  def save(self):
    raise RuntimeError("The holds cache does not allow programatical save")

  def process_loaded_data(self, data):
    trade_ids = data.get("trade_ids")
    trade_pairs = data.get("trade_pairs")

    if not trade_ids and not trade_pairs:
      return data

    open_trades = {}
    for trade in Trade.get_trades_proxy(is_open=True):
      open_trades[trade.id] = open_trades[trade.pair] = trade

    r_trade_ids = {}
    if trade_ids:
      if isinstance(trade_ids, dict):

        for trade_id, profit_ratio in trade_ids.items():
          if not isinstance(trade_id, int):
            log.error("The trade_id(%s) defined under 'trade_ids' in %s is not an integer", trade_id, self.path)
            continue
          if not isinstance(profit_ratio, float):
            log.error(
              "The 'profit_ratio' config value(%s) for trade_id %s in %s is not a float",
              profit_ratio,
              trade_id,
              self.path,
            )
          if trade_id in open_trades:
            formatted_profit_ratio = f"{profit_ratio * 100}%"
            log.warning(
              "The trade %s is configured to HOLD until the profit ratio of %s is met",
              open_trades[trade_id],
              formatted_profit_ratio,
            )
            r_trade_ids[trade_id] = profit_ratio
          else:
            log.warning(
              "The trade_id(%s) is no longer open. Please remove it from 'trade_ids' in %s",
              trade_id,
              self.path,
            )
      else:

        profit_ratio = data.get("profit_ratio")
        if profit_ratio:
          if not isinstance(profit_ratio, float):
            log.error("The 'profit_ratio' config value(%s) in %s is not a float", profit_ratio, self.path)
        else:
          profit_ratio = 0.005
        formatted_profit_ratio = f"{profit_ratio * 100}%"
        for trade_id in trade_ids:
          if not isinstance(trade_id, int):
            log.error("The trade_id(%s) defined under 'trade_ids' in %s is not an integer", trade_id, self.path)
            continue
          if trade_id in open_trades:
            log.warning(
              "The trade %s is configured to HOLD until the profit ratio of %s is met",
              open_trades[trade_id],
              formatted_profit_ratio,
            )
            r_trade_ids[trade_id] = profit_ratio
          else:
            log.warning(
              "The trade_id(%s) is no longer open. Please remove it from 'trade_ids' in %s",
              trade_id,
              self.path,
            )

    r_trade_pairs = {}
    if trade_pairs:
      for trade_pair, profit_ratio in trade_pairs.items():
        if not isinstance(trade_pair, str):
          log.error("The trade_pair(%s) defined under 'trade_pairs' in %s is not a string", trade_pair, self.path)
          continue
        if "/" not in trade_pair:
          log.error(
            "The trade_pair(%s) defined under 'trade_pairs' in %s does not look like "
            "a valid '<TOKEN_NAME>/<STAKE_CURRENCY>' formatted pair.",
            trade_pair,
            self.path,
          )
          continue
        if not isinstance(profit_ratio, float):
          log.error(
            "The 'profit_ratio' config value(%s) for trade_pair %s in %s is not a float",
            profit_ratio,
            trade_pair,
            self.path,
          )
        formatted_profit_ratio = f"{profit_ratio * 100}%"
        if trade_pair in open_trades:
          log.warning(
            "The trade %s is configured to HOLD until the profit ratio of %s is met",
            open_trades[trade_pair],
            formatted_profit_ratio,
          )
        else:
          log.warning(
            "The trade pair %s is configured to HOLD until the profit ratio of %s is met",
            trade_pair,
            formatted_profit_ratio,
          )
        r_trade_pairs[trade_pair] = profit_ratio

    r_data = {}
    if r_trade_ids:
      r_data["trade_ids"] = r_trade_ids
    if r_trade_pairs:
      r_data["trade_pairs"] = r_trade_pairs
    return r_data

  @staticmethod
  def _object_hook(data):
    _data = {}
    for key, value in data.items():
      try:
        key = int(key)
      except ValueError:
        pass
      _data[key] = value
    return _data
